package seu;

import java.io.IOException;

import rescuecore2.components.ComponentLauncher;
import rescuecore2.components.ComponentConnectionException;
import rescuecore2.components.TCPComponentLauncher;
import rescuecore2.connection.ConnectionException;
import rescuecore2.registry.Registry;
import rescuecore2.misc.CommandLineOptions;
import rescuecore2.config.Config;
import rescuecore2.config.ConfigException;
import rescuecore2.Constants;
import rescuecore2.log.Logger;

import rescuecore2.standard.entities.StandardEntityFactory;
import rescuecore2.standard.entities.StandardPropertyFactory;
import rescuecore2.standard.messages.StandardMessageFactory;
import seu.agent.*;


/**
   Launcher for sample agents. This will launch as many instances of each of the sample agents as possible, all using one connction.
 */
public final class Main {
    private static final int maxAgent = 30;
    private static final int maxCentre = 5;


    private Main() {}

    /**
       Launch 'em!
       @param args The following arguments are understood: -p <port>, -h <hostname>, -fb <fire brigades>, -pf <police forces>, -at <ambulance teams>
    */
    public static void main(String[] args) {
    	Logger.setLogContext("sample");
        try {
            Registry.SYSTEM_REGISTRY.registerEntityFactory(StandardEntityFactory.INSTANCE);
            Registry.SYSTEM_REGISTRY.registerMessageFactory(StandardMessageFactory.INSTANCE);
            Registry.SYSTEM_REGISTRY.registerPropertyFactory(StandardPropertyFactory.INSTANCE);
            Config config = new Config();
            args = CommandLineOptions.processArgs(args, config);
            int port = config.getIntValue(Constants.KERNEL_PORT_NUMBER_KEY, Constants.DEFAULT_KERNEL_PORT_NUMBER);
//            String host = null;        
            String defaultHost = config.getValue(Constants.KERNEL_HOST_NAME_KEY, Constants.DEFAULT_KERNEL_HOST_NAME);
            int fb = -1;
            int fs = -1;
            int pf = -1;
            int po = -1;
            int at = -1;
            int ac = -1;
//            // CHECKSTYLE:OFF:ModifiedControlVariable
//            for (int i = 0; i < args.length; ++i) {
//                if (args[i].equals(FIRE_BRIGADE_FLAG)) {
                    fb = (args[2].equals("-")) ? maxAgent
                            : Integer.parseInt(args[2]);
//                }
//                else if (args[i].equals(FIRE_STATION_FLAG)) {
                    at = (args[3].equals("-")) ? maxAgent
                            : Integer.parseInt(args[3]);
//                }
//                else if (args[i].equals(POLICE_FORCE_FLAG)) {
                    pf = (args[4].equals("-")) ? maxAgent
                            : Integer.parseInt(args[4]);
//                }
//                else if (args[i].equals(POLICE_OFFICE_FLAG)) {
                    fs = (args[5].equals("-")) ? maxCentre
                            : Integer.parseInt(args[5]);
//                }
//                else if (args[i].equals(AMBULANCE_TEAM_FLAG)) {
                    ac = (args[6].equals("-")) ? maxCentre
                            : Integer.parseInt(args[6]);
//                }
//                else if (args[i].equals(AMBULANCE_CENTRE_FLAG)) {
                    po = (args[7].equals("-")) ? maxCentre
                            : Integer.parseInt(args[7]);
                    
                   String host = (args[0].equals("-")) ? defaultHost : args[0];
//                   String host = "10.168.140.23";
//                }
//                else {
//                    Logger.warn("Unrecognised option: " + args[i]);
//                }
//                System.out.println(fb+" "+fs+" "+pf+" "+po+" "+at+" "+ac+" ");
//            }
            // CHECKSTYLE:ON:ModifiedControlVariable
            ComponentLauncher launcher = new TCPComponentLauncher(host, port, config);
            connect(launcher, fb, fs, pf, po, at, ac, config);
        }
        catch (IOException e) {
            Logger.error("Error connecting agents", e);
        }
        catch (ConfigException e) {
            Logger.error("Configuration error", e);
        }
        catch (ConnectionException e) {
            Logger.error("Error connecting agents", e);
        }
        catch (InterruptedException e) {
            Logger.error("Error connecting agents", e);
        }
    }

    private static void connect(ComponentLauncher launcher, int fb, int fs, int pf, int po, int at, int ac, Config config) throws InterruptedException, ConnectionException {
        int i = 0;
        
        try {
            while (fb-- > 0) {
//                Logger.info("Connecting fire brigade " + (i++) + "...");
                launcher.connect(new FireBrigadeAgent());
                System.out.println("Connecting fire brigade " + (i++) + "... success");
                Logger.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            Logger.info("failed: " + e.getMessage());
        }
        
        try {
            while (fs-- > 0) {
//                Logger.info("Connecting FireStation " + (i++) + "...");
                launcher.connect(new FireStationAgent());
                System.out.println("Connecting FireStation " + (i++) + "... success");
                Logger.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            Logger.info("failed: " + e.getMessage());
        }
        
        try {
            while (pf-- > 0) {
//                Logger.info("Connecting police force " + (i++) + "...");
                launcher.connect(new PoliceForceAgent());
                System.out.println("Connecting police force " + (i++) + "... success");
                Logger.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            Logger.info("failed: " + e.getMessage());
        }
        
        try {
            while (po-- > 0) {
//                Logger.info("Connecting PoliceOffice Agent " + (i++) + "...");
                launcher.connect(new PoliceOfficeAgent());
                System.out.println("Connecting PoliceOffice " + (i++) + "... success");
                Logger.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            Logger.info("failed: " + e.getMessage());
        }
      
        try {
            while (at-- > 0) {
//                Logger.info("Connecting ambulance team " + (i++) + "...");
                launcher.connect(new AmbulanceTeamAgent());
                System.out.println("Connecting ambulance team " + (i++) + "... success");
                Logger.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            Logger.info("failed: " + e.getMessage());
        }
        
        try {
            while (ac-- > 0) {
//                Logger.info("Connecting AmbulanceCentre agent " + (i++) + "...");
                launcher.connect(new AmbulanceCentreAgent());
                System.out.println("Connecting AmbulanceCentre " + (i++) + "... success");
                Logger.info("success");
            }
        }
        catch (ComponentConnectionException e) {
            Logger.info("failed: " + e.getMessage());
        }
        
        
        
    }
}
