//SEU2010

package seu.agent;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;

import rescuecore2.log.Logger;
import rescuecore2.messages.Command;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.messages.AgentMessageManager;
import seu.messages.SayMessageManager;
import seu.states.AmbulanceTeamState;
import seu.states.CivilianState;
import seu.states.FireBrigadeState;
import seu.states.PoliceForceState;
import seu.utilities.DistanceSorter;

/**
 * A sample ambulance team agent.
 */
public class AmbulanceTeamAgent extends PlatoonAgent<AmbulanceTeam> {

	private AmbulanceTeamState myState;

	private Collection<StandardEntity> unexploredBuildings;

	protected List<Civilian> iSelectList = new ArrayList<Civilian>();

	protected List<Civilian> visiableBuriedList = new ArrayList<Civilian>();
	protected List<Civilian> needRescueCivilian = new ArrayList<Civilian>();

	protected int lockIndex = 0;
	// protected ArrayList<Integer> channels;

	public Civilian myTarget;
	public int assTime = 0;
	public int assgin = 0;

	@Override
	protected void postConnect() {
		super.postConnect();
		model.indexClass(StandardEntityURN.CIVILIAN,
				StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.POLICE_FORCE,
				StandardEntityURN.AMBULANCE_TEAM, StandardEntityURN.REFUGE,
				StandardEntityURN.BUILDING);
		unexploredBuildings = model
				.getEntitiesOfType(StandardEntityURN.BUILDING);
	}

	@Override
	protected void think(int time, ChangeSet changed, Collection<Command> heard) {

		prepareForAct(time, changed, heard);

		act(time);

		// prepareForAct(time, changed, heard);
		// channelWork(time, changed, heard);

	}

	public void prepareForAct(int time, ChangeSet changed,
			Collection<Command> heard) {
		if (time < agentIgnoreTime)
			return;

		if (time == agentIgnoreTime) {
			try {
				initialChannels(time);
			} catch (Exception e) {
				System.out.println(me() + " initialChannel failed");
				e.printStackTrace();
			}
		}
		// System.out.println(getTime()+" at "+me()+" 1");
		// update world message
		try {
			updateModel(heard);
		} catch (Exception e) {
			System.out.println(me() + "updateModel failed");
			e.printStackTrace();
		}
		// System.out.println(getTime()+" at "+me()+" 2");
		super.prepareForAct(time, changed);
		// System.out.println(getTime()+" at "+me()+" 3");
		modelState = model.getState();
		myState = me().getState();

		checkAround(changed);

		checkMyVictim();
		// System.out.println(getTime()+" at "+me()+" 4");
		// TODO
		reportUnloadedVictim();
		// System.out.println(getTime()+" at "+me()+" 5");
		initialTarget();
		// System.out.println(getTime()+" at "+me()+" 6");
		// System.out.println("Time = " + time + " I = " +me()
		// + " Buildings = " + model.getState().buildings.size()
		// + " Roads = " + model.getState().roads.size() + "");

		try {
			if (agentMessageManager != null) {
				agentMessageManager.sendMessages();
				if (agentMessageManager.getMessageBytes() != null) {
					// System.out.println("at"+agentMessageManager.getMessageBytes().length);
					sendSpeak(time, speakChannel,
							agentMessageManager.getMessageBytes());
					// System.out.println("at send =  "+(agentMessageManager.getMessageBytes()));
				}
			}
		} catch (Exception e) {
			System.out.println(me() + " sendMessage failed");
			e.printStackTrace();
		}
		// System.out.println(getTime()+" at "+me()+" 7");
	}

	public void act(int time) {
		if (time < agentIgnoreTime)
			return;

		super.act();
		// System.out.println(getTime()+" at "+me()+" 8");
		// System.out.println("Time = "+time+
		// " stateCiv = "+modelState.civilians.size()
		// + " modelCiv = " +model.getEntitiesOfType(StandardEntityURN.CIVILIAN)
		// .size()
		// + " stateHum = " +modelState.humans.size());
		// rescueIfLocked();
		if (action)
			return;

		goToRefuge();
		if (action)
			return;
		// System.out.println(getTime()+" at "+me()+" 9");
		loadHere();
		if (action)
			return;
		// System.out.println(getTime()+" at "+me()+" 10");
		rescueBuriedAgents();
		if (action)
			return;
		// System.out.println(getTime()+" at "+me()+" 11");
		rescueCivs();
		// rescueTargetNoCommand();
		// System.out.println(me()+" 1");
		if (action)
			return;

		// rescueSomeOne();
		//
		// // rescueTarget();
		// if (action)
		// return;
		// System.out.println(getTime()+" at "+me()+" 12");

		// System.out.println(getTime()+" at "+me()+" 13");
		// randomSearch(time);

		atSearch();
		System.out.println(me() + " 2");
		if (action)
			return;

		searchCivilians(true);
		System.out.println(me() + " 3");
		if (action)
			return;
		// System.out.println(getTime()+" at "+me()+" 14");
		// rescueDeadToRefuge();
		// System.out.println("rest");
		restAtRefuge();
		if (action)
			return;
	}

	protected void rescueSomeOne() {
		for (Human next : getTargets()) {
			if (next.getPosition().equals(location().getID())) {
				// Targets in the same place might need rescueing or loading
				if ((next instanceof Civilian) && next.getBuriedness() == 0
						&& !(location() instanceof Refuge)) {
					// Load
					// Logger.info("Loading " + next);
					load(next);
					return;
				}
				if (next.getBuriedness() > 0) {
					// Rescue
					// Logger.info("Rescueing " + next);
					rescue(next);
					return;
				}
			} else {
				// Try to move to the target
				List<EntityID> path = search.breadthFirstSearch(me()
						.getPosition(), next.getPosition());
				if (path != null) {
					// Logger.info("Moving to target");
					move(path);
					return;
				}
			}
		}
	}

	protected void initialChannels(int time) {
		if (useSpeak) {
			if (model.getState().channelCount > 3) {
				sendSubscribe(time, 3);
				speakChannel = 1;
			} else if (model.getState().channelCount > 2) {
				sendSubscribe(time, 1);
				speakChannel = 1;
			} else if (model.getState().channelCount > 1) {
				sendSubscribe(time, 1);
				speakChannel = 1;
			}
			agentMessageManager = new AgentMessageManager(model, me().getID());
		} else {
			sayMessageManager = new SayMessageManager(model, me().getID());
		}
		// System.out.println(me() +" areas = "+modelState.areas.size()+" "
		// +" Buildings = "+modelState.buildings.size()+" "
		// +" Roads = "+modelState.roads.size()+" "
		// +model.getEntitiesOfType(StandardEntityURN.BUILDING).size()
		// +model.getEntitiesOfType(StandardEntityURN.ROAD).size()
		// +model.getEntitiesOfType(StandardEntityURN.BUILDING,StandardEntityURN.ROAD).size());
	}

	private void reportUnloadedVictim() {
		if (m_unloadedHumanoid != null) {

			if (agentMessageManager != null)
				if (getRefuges().isEmpty())
					return;
			agentMessageManager.reportVictim(m_unloadedHumanoid, getRefuges()
					.get(0));
			// reset
			m_unloadedHumanoid = null;
		}
	}

	protected void initialTarget() {
		myState.currentTarget = myState.getCenterAssignedTarget();
		myState.nextTarget = myState.getCenterSecondAssignedTarget();
		if (myState.currentTarget != null) {
			System.out.println(me() + " : " + getTime() + " 1st "
					+ myState.currentTarget + " 2nd " + myState.nextTarget);
		}
	}

	// ============================================================================================//

	// go to refuge
	private void goToRefuge() {
		if (isMyVictimLoaded()) {
			// System.out.println("on at");
			modelState.savedHumanoids.add(getMyLoadedVictim());
			m_unloadedHumanoid = getMyLoadedVictim();

			if (me().getPosition(model) instanceof Refuge) {
				unload();
				// action = true;
			} else {
				if (getRefuges().isEmpty()) {
					if (location() instanceof Road) {
						unload();
					} else {
						randomSearch(getTime());
					}
					// action = true;
				} else {
					moveToRefuge();
					// action = true;
				}
			}
		}

		if (me().getDamage() > 40
				|| (me().getHP() < 5000 && me().getDamage() > 4)) {
			restAtRefuge();
			action = true;
		}

	}

	protected void checkMyVictim() {
		if (myLoadedVictim instanceof Civilian) {
			if (myLoadedVictim.getPosition(model) != me()) {
				myLoadedVictim = null;
			}
		}
	}

	protected boolean isMyVictimLoaded() {
		return myLoadedVictim != null;
	}

	public Human getMyLoadedVictim() {
		return myLoadedVictim;
	}

	// =======================================================================================//
	// load civilian
	private void loadHere() {
		StandardEntity mpos = location();

		ArrayList<Civilian> civilianHereNeedLoad = (NEED_LOAD_C)
				.extract(modelState.updatedCivilians);
		for (Civilian cv : civilianHereNeedLoad)
			if (cv.getPosition(model) != mpos)
				civilianHereNeedLoad.remove(cv);
		if (!civilianHereNeedLoad.isEmpty()) {
			ArrayList<AmbulanceTeam> ambulanceHere = new ArrayList<AmbulanceTeam>();
			ambulanceHere.clear();
			for (Iterator it = modelState.ambulanceTeams.iterator(); it
					.hasNext();) {
				AmbulanceTeam at = (AmbulanceTeam) it.next();
				if (location() == at.getPosition(model)) {
					ambulanceHere.add(at);
				}
			}
			Collections.sort(ambulanceHere, Util.idComparator);
			int index = ambulanceHere.indexOf(me());
			if (index < civilianHereNeedLoad.size()) {
				myState.mySelectedTarget = civilianHereNeedLoad.get(index);

				load(myState.mySelectedTarget);
				// action = true;
			}
		}
	}

	// =======================================================================================//
	// rescue Agents
	private Human myAgentTarget = null;
	private int m_notReachableTime = -30;

	protected void rescueBuriedAgents() {

		if (getTime() < m_notReachableTime + 10) {
			return;
		}

		if (myAgentTarget != null) {
			if (location() == myAgentTarget.getPosition(model)) {// at the same
																	// position
				if (myAgentTarget.getState().isVisible(model)) {// I can see it

					if (checkNeedRescue(model, myAgentTarget)) {// need rescue
						rescue(myAgentTarget);
						return;
						// action = true;
					} else if (checkNeedLoad(model, myAgentTarget)) {// need
																		// load
						load(myAgentTarget);
						// action = true;
						model.getState().savedHumanoids.add(myAgentTarget);
						return;
						// rescueTarget();
					} else {// need neither rescue nor load
						model.getState().savedHumanoids.add(myAgentTarget);
					}
				} else {// I can't see it althought I was on the same place as
						// the target
					model.getState().savedHumanoids.add(myAgentTarget);
				}
			} else {
				move(myAgentTarget.getPosition(model));
				return;
			}
		}

		myAgentTarget = null;

		int minDist = Integer.MAX_VALUE;

		for (Iterator it = modelState.buriedAgents.iterator(); it.hasNext();) {
			Human target = (Human) it.next();
			if (target.getPosition(model) == null)
				continue;
			if (!checkNeedRescue(model, myAgentTarget)
					|| model.getState().savedHumanoids.contains(target))
				continue;

			if (target.getPosition(model) instanceof Building) {
				Building b = (Building) target.getPosition(model);
				if ((b.getState().isBurning() && target.getBuriedness() > 30)
						|| b.getState().isBurned())
					continue;
			}

			int dist = 0;

			try {
				List<EntityID> path = search.breadthFirstSearch(location()
						.getID(), target.getPosition());
				for (EntityID id : path) {
					StandardEntity next = model.getEntity(id);
					if (next instanceof Road) {
						dist += ((Road) next).getState().getLength(model);
					}
				}
				// dist = (int) Router.getASGS(location(),
				// Collections.singleton(target.getPosition(model)),
				// SIMPLE_COST_FUNC, LINE_DIST_HFUNC).cost / 2000;
			} catch (Exception e) {
				dist = (int) (Math.random() * 500);
			}

			if (dist < minDist) {
				minDist = dist;
				myAgentTarget = target;
			}
		}

		if (minDist >= Integer.MAX_VALUE - 1000) {
			m_notReachableTime = getTime();
			return;
		}

		if (myAgentTarget == null || myAgentTarget.getPosition(model) == null)
			return;

		if (location() == myAgentTarget.getPosition(model)) {// at the same
																// position
			if (myAgentTarget.getState().isVisible(model)) {// I can see it

				if (checkNeedRescue(model, myAgentTarget)) {// need rescue
					rescue(myAgentTarget);
					return;
					// action = true;
				} else if (checkNeedLoad(model, myAgentTarget)) {// need load
					load(myAgentTarget);
					// action = true;
					model.getState().savedHumanoids.add(myAgentTarget);
					// rescueTarget();
					return;
				} else {// need neither rescue nor load
					model.getState().savedHumanoids.add(myAgentTarget);
				}
			} else {// I can't see it althought I was on the same place as the
					// target
				model.getState().savedHumanoids.add(myAgentTarget);
			}
		} else {
			move(myAgentTarget.getPosition(model));
			return;
		}
	}

	// =======================================================================================//
	// rescue targets
	private void rescueTarget() {
		checkLastTarget();

		if (myState.currentTarget == null
				|| myState.currentTarget.getPosition(model) == null
				|| modelState.savedHumanoids.contains(myState.currentTarget)) {
			return;
		}

		myState.lastTarget = myState.currentTarget; // used for report, when
													// unloaded

		if (location() == myState.currentTarget.getPosition(model)) {// at the
																		// same
																		// position
			if (myState.currentTarget.getState().isVisible(model)) {// I can see
																	// it
				if (checkNeedRescue(model, myState.currentTarget)) {// need
																	// rescue
					rescue(myState.currentTarget);
					return;
					// action = true;

				} else if (checkNeedLoad(model, myState.currentTarget)) {// need
																			// load

					model.getState().savedHumanoids.add(myState.currentTarget);

					load(myState.currentTarget);
					return;
					// action = true;

					// rescueTarget();
				} else {// need neither rescue nor load

					model.getState().savedHumanoids.add(myState.currentTarget);

					rescueTarget();
				}
			} else {// I can't see it althought I was on the same place as the
					// target

				myState.currentTarget.getState().setProperty(model, getTime(),
						myState.currentTarget.getPositionProperty(),
						getRefuges().get(0).getID().getValue());
				model.getState().savedHumanoids.add(myState.currentTarget);

				m_unloadedHumanoid = myState.currentTarget;
			}
		} else {
			move(myState.currentTarget.getPosition(model));
			return;
			// action = true;
		}
	}

	protected void checkLastTarget() {
		if (model.getState().savedHumanoids.contains(myState.currentTarget)) {
			myState.currentTarget = myState.nextTarget;
			myState.nextTarget = null;
		}

		if (myState.currentTarget == null) {
			myState.currentTarget = myState.nextTarget;
			myState.nextTarget = null;
		}

		if (myState.currentTarget != null
				&& myState.getStuckCycles(getTime()) > 1) {
			myState.currentTarget = myState.nextTarget;
		}

	}

	// ========================================================================================//
	// AIS************************AIS******************************AIS//
	protected void rescueTargetNoCommand() {
		updateLastConcentration();
		updateRescueList();
		updateAllConcentration();

		targetEvaluate();

		rescueBySelf();
	}

	protected void updateRescueList() {
		iSelectList.clear();

		List<Civilian> tmp = new ArrayList<Civilian>();

		// for (Iterator it = modelState.buriedAgents.iterator(); it.hasNext();)
		// {
		// Human target = (Human) it.next();
		// if (target.getPosition(model) == null)
		// continue;
		//
		// if ((target.getPosition(model) instanceof Building)) {
		// Building bld = (Building) target.getPosition(model);
		// if(target instanceof AmbulanceTeam){
		// AmbulanceTeam at = (AmbulanceTeam) target;
		// if ((at.getState().needRescue(model) ||
		// at.getState().needLoad(model))
		// && !bld.getState().isBurned()&& (!bld.getState().isBurning()
		// || target.getBuriedness() < 81)) {
		//
		// if (!REACHABLE_C.eval(target.getPosition(model)))
		// continue;
		//
		// iSelectList.add(target);
		// }
		// }
		// else if(target instanceof FireBrigade){
		// FireBrigade fb = (FireBrigade) target;
		// if ((fb.getState().needRescue(model) ||
		// fb.getState().needLoad(model))
		// && !bld.getState().isBurned()&& (!bld.getState().isBurning()
		// || target.getBuriedness() < 81)) {
		//
		// if (!REACHABLE_C.eval(target.getPosition(model)))
		// continue;
		//
		// iSelectList.add(target);
		// }
		// }
		// else if(target instanceof PoliceForce){
		// PoliceForce pf = (PoliceForce) target;
		// if ((pf.getState().needRescue(model) ||
		// pf.getState().needLoad(model))
		// && !bld.getState().isBurned()&& (!bld.getState().isBurning()
		// || target.getBuriedness() < 81)) {
		//
		// if (!REACHABLE_C.eval(target.getPosition(model)))
		// continue;
		//
		// iSelectList.add(target);
		// }
		// }
		// }
		// }
		for (Iterator it = modelState.civilians.iterator(); it.hasNext();) {
			Civilian target = (Civilian) it.next();
			CivilianState cst = target.getState();
			if (target.getPosition(model) == null)
				continue;

			if (modelState.savedHumanoids.contains(target))
				continue;

			target.getState().updateCivilianDeadTime(model);

			if (!REACHABLE_C.eval(target.getPosition(model)))
				continue;

			if ((cst.needRescue(model) || cst.needLoad(model))
					&& cst.isRescueAble(model)) {
				if ((target.getPosition(model) instanceof Building)) {
					Building bld = (Building) target.getPosition(model);

					if (smallIndexLoader.contains(me())
							&& target.getBuriedness() == 0)
						iSelectList.add(target);

					if (!bld.getState().isBurned()
							&& (!bld.getState().isBurning() || target
									.getBuriedness() < 31)) {
						if (!smallIndexLoader.contains(me()))
							iSelectList.add(target);
						else
							tmp.add(target);
					}
				}
			}

		}

		if (iSelectList.isEmpty())
			iSelectList.addAll(tmp);

		iSelectList.removeAll(modelState.savedHumanoids);
	}

	private void rescueBySelf() {

		int humIndex = 0;
		while (humIndex < iSelectList.size()) {
			Civilian civ = iSelectList.get(humIndex);
			if (civ == myState.getLastTarget() && hasCheckLast) {
				humIndex++;
				continue;
			}
			if (needChangeToLast(civ)) {
				civ = myState.getLastTarget();
			}
			if (civ == myState.getLastTarget() && myState.getStuckCycles() >= 2
					&& !(location() instanceof Building)) {
				hasCheckLast = true;
				continue;
			}

			myState.mySelectedTarget = civ;
			if (REACHABLE_C.eval(civ))
				rescueTarget_noChannel();
			// else if (!SeuConstants.IS_COMPETETION_MODE)
			// {
			// world.logFile.writeString(human + "  pos000  = "
			// + human.getMotionlessPosition() + "  time = "
			// + getTime(), 12, "noReachable");
			// }

			// no action be performed
			if (myState.mySelectedTarget == myState.getLastTarget())
				hasCheckLast = true;
			else
				humIndex++;

			// if(humIndex > 10) break;//to reduce the calculation
		}
		// try
		// {
		//
		// } catch (ActionCommandException ace)
		// {
		// me().setLastTarget(me().mySelectedTarget);
		// hasCheckLast = false;
		// throw new ActionCommandException();
		// }
		me().getState().setLastTarget(null);
		hasCheckLast = true;
	}

	private final List<AmbulanceTeam> smallIndexLoader = new ArrayList<AmbulanceTeam>();

	protected void initialSmallIndexLoader() {
		for (AmbulanceTeam sa : modelState.ambulanceTeams) {
			if (modelState.ambulanceTeams.indexOf(sa) < 2)
				smallIndexLoader.add(sa);
		}
	}

	private AmbulanceTeam getNearestSAFromLoadList(Civilian tar) {
		if (tar == null)
			return null;
		int minDist = Integer.MAX_VALUE;
		AmbulanceTeam nearest = null;
		for (AmbulanceTeam sa : smallIndexLoader) {
			if (sa.getPosition(model) == null)
				continue;
			int dist = model.getDistance(sa, tar);
			if (dist < minDist) {
				nearest = sa;
				minDist = model.getDistance(sa, tar);
			}
		}
		return nearest;
	}

	private int getNumATRescueing(Human myTarget) {
		int n = 0;
		for (AmbulanceTeam sa : modelState.ambulanceTeams) {
			if (sa.getPosition(model) == myTarget.getPosition(model))
				n++;
		}
		return n;
	}

	private boolean assignedSmallBuriedness() {
		// Assign two AT to the small getBuriedness target.
		if (location() == myState.mySelectedTarget.getPosition(model))
			return true;

		if (getNumATRescueing(myState.mySelectedTarget) > 0)
			return false;

		if (smallIndexLoader.size() < 2)
			return true;

		if (getNearestSAFromLoadList(myState.mySelectedTarget) == me()
				&& smallIndexLoader.contains(me()))
			return true;
		if (smallIndexLoader.contains(me())
				&& getNearestSAFromLoadList(myState.mySelectedTarget) != null
				&& getNearestSAFromLoadList(myState.mySelectedTarget)
						.getState().isBusyNow(model))
			return true;

		return false;
	}

	private void rescueTarget_noChannel() {
		if (myState.mySelectedTarget == null)
			return;

		// if(world.ambulanceTeams.indexOf(me()) == 0)
		// System.out.println(" civNum "+world.civilians.size()+"   list= "+iSelectList.size()+" in "+getTime());

		// if (!SeuConstants.IS_COMPETETION_MODE)
		// {
		// world.logFile.writeString(me().mySelectedTarget + "  bury = "
		// + me().mySelectedTarget.getBuriedness() + "  at = "
		// + getTime(), 12, "ATar");
		// }

		if (myState.mySelectedTarget.getBuriedness() < 5) {
			// getBuriedness is small, which means that it's unnecessary
			// to assign many at to rescue.
			if (!assignedSmallBuriedness())
				return;
		}

		if (location() == myState.mySelectedTarget.getPosition(model)) {// at
																		// the
																		// same
																		// getRealPosition
			if (myState.mySelectedTarget.getState().isVisible(model)) {// I can
																		// see
																		// it

				if (myState.mySelectedTarget.getState().needRescue(model)) {// need
																			// rescue
					// System.out.println(" Rescue "+me().mySelectedTarget+" in "+getTime());
					rescue(myState.mySelectedTarget);
					// action = true;
				} else if (myState.mySelectedTarget.getState().needLoad(model)) {// need
																					// load

					model.getState().savedHumanoids
							.add(myState.mySelectedTarget);

					load(myState.mySelectedTarget);

				} else {// need neither rescue nor load
					// System.out.println("no needRescue and needLoad is "
					// + me().mySelectedTarget + " in " + getTime());

					model.getState().savedHumanoids
							.add(myState.mySelectedTarget);
				}
			} else {// I can't see it althought I was on the same place as the
					// target
				myState.mySelectedTarget.getState().setProperty(model,
						getTime(),
						myState.mySelectedTarget.getPositionProperty(),
						new int[] { getRefuges().get(0).getID().getValue() });
				model.getState().savedHumanoids.add(myState.mySelectedTarget);

			}
		} else {// on the different getRealPosition
			// move to my target's getRealPosition.

			move(myState.mySelectedTarget.getPosition(model));
		}
	}

	private boolean hasCheckLast = false;

	private boolean needChangeToLast(Civilian checkTarget) {
		if (hasCheckLast || myState.getLastTarget() == null)
			return false;
		if (checkTarget == myState.getLastTarget())
			return false;
		if (!iSelectList.contains(myState.getLastTarget())) {
			hasCheckLast = true;
			return false;
		}
		if (AGENT_C.eval(checkTarget) && !AGENT_C.eval(myState.getLastTarget()))
			return false;

		float checkCon = checkTarget.getState().concentration();
		float lastCon = myState.getLastTarget().getState().concentration();
		if (checkCon > 1.35 * lastCon)
			return false;
		else
			return true;
	}

	public StandardEntity getNearestObject(List<StandardEntity> objects) {
		if (objects == null || objects.isEmpty()) {
			System.err.println("getNearestObject : objects is empty");
			return null;
		}

		return modelState.getNearestObjectInRange(me(), objects);
	}

	public float getUpdatedStimValue(Civilian civ) {
		CivilianState cst = civ.getState();
		float stm = 0.8f * cst.lastStimValue();

		stm += getUpdatedStimPlus(civ);
		cst.setLastStimValue(stm);
		return stm;
	}

	protected float w_disMe = 0.45f;// 0.19f;//dis to me
	protected float w_disRef = 0.28f;// 0.11f;//dis to ref
	protected float w_disBB = 0.5f;// 0.5f;//dis to burning building
	protected float w_atNum = 0.32f;// 0.60f;//at around num
	protected float w_civNum = 0.12f;// 0.19f;//civ around num

	// =================================
	protected float hp_impact = 1.8f;// 1.8f;
	protected float dmg_impact = 0.8f;// 0.2f;
	protected float brdness_impact = 0.4f;// 0.4f;

	protected final float natural_death_rate = 0.005f;

	protected float getUpdatedStimPlus(Civilian civ) {
		float stm = 0, s = 0;

		// dist to me
		s = affinityMe(civ) * w_disMe;//
		stm += s;

		// dist to refuge
		s = affinityRefuge(civ) * w_disRef;//
		stm += s;
		// dist to burning building
		s = affinityBurningBld(civ) * w_disBB;//
		stm += s;

		// ambulance team nums

		s = affinityATNum(civ) * w_atNum;//
		stm += s;

		s = affinityNumAround(civ) * w_civNum;//
		stm += s;

		// natural death rate
		stm -= natural_death_rate;

		// antibodies affection
		s = getAbsAffection(civ);

		stm += s;

		return stm * civ.getState().lastConcentration();

	}

	protected float affinityRefuge(Civilian civ) {
		List<StandardEntity> refuges = new ArrayList<StandardEntity>(
				modelState.refuges);
		if (refuges.isEmpty())
			return 0;
		float aveLength = (modelState.mapHeight + modelState.mapWidth) / 2 / 1000;
		float mapRate = 1;
		if (aveLength > 350)
			mapRate = aveLength / 350;
		float v = 1
				- (model.getDistance(civ, getNearestObject(refuges)) / 1000f)
				/ 50f / mapRate;

		if (v < -1f)
			v = -1f;
		return v;
	}

	protected float affinityBurningBld(Civilian civ) {
		if (model.getState().burningBuildings.size() == 0)
			return 0;
		float affinity = 1f;
		float aveLength = (modelState.mapHeight + modelState.mapWidth) / 2 / 1000;
		float mapRate = 1;
		if (aveLength > 350)
			mapRate = aveLength / 350;

		List<StandardEntity> burnings = new ArrayList<StandardEntity>(
				model.getState().burningBuildings);
		float getDistance = model.getDistance(civ, getNearestObject(burnings)) / 1000f;
		float saveDis = 10f;
		if (getDistance < saveDis)
			affinity = 0.5f + 0.5f * (getDistance / saveDis);
		else
			affinity = 1 - (getDistance - saveDis) / 50f / mapRate;

		if (affinity < -1)
			affinity = -1;
		return affinity;
	}

	protected float affinityATNum(Human hum) {
		if (!(hum.getPosition(model) instanceof Building))
			return 0;
		int n = 0;
		for (AmbulanceTeam sa : modelState.ambulanceTeams) {
			if (sa.getPosition(model) == hum.getPosition(model)) {
				n++;
			}
		}

		float affinity = 1 - n / 2f;
		if (affinity < -1)
			affinity = -1f;
		return affinity;
	}

	protected float affinityMe(Civilian civ) {// dist to me
		float aveLength = (modelState.mapHeight + modelState.mapWidth) / 2 / 1000;
		float mapRate = 1;
		if (aveLength > 350)
			mapRate = aveLength / 350;
		float v = 1 - 0.01f * (model.getDistance(me(), civ) / 1000f / mapRate);
		if (v < -1f)
			v = -1f;

		return v;
	}

	protected float affinityNumAround(Civilian hum) {
		int n = 0;
		for (Civilian civ : modelState.civilians) {
			if (civ.getPosition(model) == null)
				continue;
			if (civ.getHP() <= 100)
				continue;
			if (model.getState().savedHumanoids.contains(civ))
				continue;
			if (civ.getID() == hum.getID())
				continue;
			if (model.getDistance(civ, hum) * 0.001 < 60)
				n++;
		}
		if (n >= 5)
			return 1;
		else
			return -1 + 0.4f * n;
	}

	protected float getAbsAffection(Civilian civ) {
		float v = 0;
		for (Civilian h : iSelectList) {
			v += (getAbsAffinity(civ, h) - getAbsAffinity(h, civ))
					* h.getState().lastConcentration();
		}
		return v;
	}

	protected float getAbsAffinity(Civilian me, Civilian tar) {
		if (me.getID() == tar.getID())
			return 1;
		float v = 0, s = 0;

		int hpDelta = 0;
		int dmgDelta = 0;

		hpDelta = tar.getHP() - me.getHP();
		dmgDelta = tar.getDamage() - me.getDamage();

		s = hp_impact * getHpFuncValue(hpDelta);
		v += s;

		s = dmg_impact * getDamageFuncValue(dmgDelta);
		v += s;

		s = brdness_impact
				* getBuriednessFuncValue(tar.getBuriedness()
						- me.getBuriedness());
		v += s;

		return v;
	}

	protected float getHpFuncValue(int hpDelta) {
		float k2 = 1 / 10000f;
		float k1 = -0.77f * k2;
		float v = 0;
		if (hpDelta > 0) {
			v = 1 + hpDelta * k1;
		} else {
			v = 1 + hpDelta * k2;
		}
		return Math.max(v, 0);
	}

	protected float getDamageFuncValue(int damageDelta) {
		float k1 = -1 / 250f;
		float k2 = -0.77f * k1;
		float v = 0;
		if (damageDelta > 0) {
			v = 1 + damageDelta * k1;
		} else {
			v = 1 + damageDelta * k2;
		}
		return Math.max(v, 0);
	}

	protected float getBuriednessFuncValue(int getBuriednessDelta) {
		float k2 = 1 / 60f;
		float k1 = -0.77f * k2;
		float v = 0;
		if (getBuriednessDelta > 0) {
			v = 1 + getBuriednessDelta * k1;
		} else {
			v = 1 + getBuriednessDelta * k2;
		}
		return Math.max(v, 0);
	}

	// Comparator distanceToMeComparator = new Comparator() {
	// public int compare(Object obj1, Object obj2) {
	// Human civ1 = (Human) obj1;
	// Human civ2 = (Human) obj2;
	// if (model.getDistance(me(), civ1) > model.getDistance(me(), civ2))
	// return 1;
	// else if (model.getDistance(me(), civ1) == model.getDistance(me(),
	// civ2))
	// return 0;
	// else
	// return -1;
	// }
	// };

	Comparator concentrationComparator = new Comparator() {
		@Override
		public int compare(Object obj1, Object obj2) {
			Civilian civ1 = (Civilian) obj1;
			Civilian civ2 = (Civilian) obj2;
			if (civ1.getState().concentration() < civ2.getState()
					.concentration())
				return 1;
			else if (civ1.getState().concentration() == civ2.getState()
					.concentration())
				return 0;
			else
				return -1;
		}
	};

	Comparator getBuriednessComparator = new Comparator() {
		@Override
		public int compare(Object obj1, Object obj2) {
			Human civ1 = (Human) obj1;
			Human civ2 = (Human) obj2;
			if (civ1.getBuriedness() < civ2.getBuriedness())
				return 1;
			else if (civ1.getBuriedness() == civ2.getBuriedness())
				return 0;
			else
				return -1;
		}
	};

	private int m_updateConcentrationTime = 0;

	protected void updateAllConcentration() {
		m_updateConcentrationTime = getTime();
		for (Civilian h : iSelectList) {
			updateConcentration(h);
		}
	}

	protected void updateConcentration(Civilian civ) {
		float v, s = getUpdatedStimValue(civ);
		v = (float) (1 / (1 + Math.exp(0.5 - s)));

		if (MOVING_AGENT_C.eval(civ)) {
			v += 1.0;
			if (v > 1.5)
				v = 1.5f;
		}

		civ.getState().setConcentration(v);

	}

	protected void updateLastConcentration() {
		if (m_updateConcentrationTime != getTime()) {
			for (Civilian civ : iSelectList) {
				civ.getState().setLastConcentration(
						civ.getState().concentration());
			}
		}
	}

	protected void targetEvaluate() {
		Collections.sort(iSelectList, concentrationComparator);
	}

	// end
	// ****************************************************************************************//
	// ========================================================================================//
	// random walk
	protected void randomSearch(int time) {
		// List<EntityID> path = search.breadthFirstSearch(location(),
		// unexploredBuildings);
		// if (path != null) {
		// Logger.info("Searching buildings");
		// sendMove(time, path);
		// action = true;
		// return;
		// }
		// Logger.info("Moving randomly");
		sendMove(time, randomWalk());
		// action = true;
	}

	// =========================
	protected boolean checkNeedLoad(StandardWorldModel model, Human target) {
		if (target instanceof Civilian) {
			Civilian civ = (Civilian) target;
			CivilianState cst = civ.getState();
			return cst.needLoad(model);
		} else {
			return false;
		}
	}

	protected boolean checkNeedRescue(StandardWorldModel model, Human target) {
		if (target instanceof Civilian) {
			Civilian civ = (Civilian) target;
			CivilianState cst = civ.getState();
			return cst.needRescue(model);
		} else if (target instanceof AmbulanceTeam) {
			AmbulanceTeam at = (AmbulanceTeam) target;
			AmbulanceTeamState atst = at.getState();
			return atst.needRescue(model);
		} else if (target instanceof FireBrigade) {
			FireBrigade fb = (FireBrigade) target;
			FireBrigadeState fbst = fb.getState();
			return fbst.needRescue(model);
		} else if (target instanceof PoliceForce) {
			PoliceForce pf = (PoliceForce) target;
			PoliceForceState pfst = pf.getState();
			return pfst.needRescue(model);
		} else {
			return false;
		}
	}

	// ======================================================================================//

	// =======================================================================================//

	protected void channelWork(int time, ChangeSet changed,
			Collection<Command> heard) {
		// if (time ==
		// config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY))
		// {
		// // Subscribe to channel 1
		// sendSubscribe(time, 1);
		// }
		// for (Command next : heard) {
		// if(next instanceof AKSpeak){
		// System.out.println("I : "+next.getAgentID()+"get the message "+new
		// String(((AKSpeak)next).getContent()));
		// }
		// Logger.debug("Heard " + next);
		// }

		updateUnexploredBuildings(changed);
		// Am I transporting a civilian to a refuge?
		if (someoneOnBoard()) {
			// Am I at a refuge?
			if (location() instanceof Refuge) {
				// Unload!
				Logger.info("Unloading");
				sendUnload(time);
				return;
			} else {
				// Move to a refuge
				List<EntityID> path = search.breadthFirstSearch(location()
						.getID(), getRefugesID());
				if (path != null) {
					Logger.info("Moving to refuge");
					sendMove(time, path);
					return;
				}
				// What do I do now? Might as well carry on and see if we can
				// dig someone else out.
				Logger.debug("Failed to plan path to refuge");
			}
		}
		// Go through targets (sorted by distance) and check for things we can
		// do
		for (Human next : getTargets()) {
			if (next.getPosition().equals(location().getID())) {
				// Targets in the same place might need rescueing or loading
				if ((next instanceof Civilian) && next.getBuriedness() == 0
						&& !(location() instanceof Refuge)) {
					// Load
					Logger.info("Loading " + next);
					sendLoad(time, next.getID());
					return;
				}
				if (next.getBuriedness() > 0) {
					// Rescue
					Logger.info("Rescueing " + next);
					sendRescue(time, next.getID());
					return;
				}
			} else {
				// Try to move to the target
				List<EntityID> path = search.breadthFirstSearch(location()
						.getID(), next.getPosition());
				if (path != null) {
					Logger.info("Moving to target");
					sendMove(time, path);
					return;
				}
			}
		}
		// Nothing to do
		// List<EntityID> path = search.breadthFirstSearch(location(),
		// unexploredBuildings);
		// if (path != null) {
		// Logger.info("Searching buildings");
		// sendMove(time, path);
		// return;
		// }
		Logger.info("Moving randomly");
		sendMove(time, randomWalk());

	}

	// ===============================================================================================//
	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.AMBULANCE_TEAM);
	}

	private boolean someoneOnBoard() {
		for (StandardEntity next : model
				.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
			if (((Human) next).getPosition().equals(getID())) {
				Logger.debug(next + " is on board");
				return true;
			}
		}
		return false;
	}

	private List<Human> getTargets() {
		List<Human> targets = new ArrayList<Human>();
		for (StandardEntity next : model.getEntitiesOfType(
				StandardEntityURN.CIVILIAN, StandardEntityURN.FIRE_BRIGADE,
				StandardEntityURN.POLICE_FORCE,
				StandardEntityURN.AMBULANCE_TEAM)) {
			Human h = (Human) next;
			if (h == me()) {
				continue;
			}
			if (h.isHPDefined() && h.isBuriednessDefined()
					&& h.isDamageDefined() && h.isPositionDefined()
					&& h.getHP() > 0
					&& (h.getBuriedness() > 0 || h.getDamage() > 0)) {
				targets.add(h);
			}
		}
		Collections.sort(targets, new DistanceSorter(location(), model));
		return targets;
	}

	private void updateUnexploredBuildings(ChangeSet changed) {
		for (EntityID next : changed.getChangedEntities()) {
			StandardEntity e = model.getEntity(next);
			if (e != null) {
				unexploredBuildings.remove(e);
			}
		}
	}

	@Override
	public String toString() {
		return "ambulance team";
	}

	public void rescueCivs() {
		System.out.println(getTime() + "");
		if (myTarget == null) {
			evaluateCivilian();
		}
		if (myTarget == null)
			return;

		if (!isReachable(myTarget.getPosition(model).getID())) {
			assgin++;
			if (assgin > 4)
				assgin = 0;
			System.out.println(getTime() + " assgin again " + assgin);
			evaluateCivilian();
		}
		// if(!(location() == myTarget.getPosition(model)) && (getTime() -
		// assTime) > 5){
		// assgin++;
		// if(assgin > 4) assgin = 0;
		// // System.out.println("assgin again");
		// evaluateCivilian();
		// }
		rescueMyTarget();
	}

	public void evaluateCivilian() {
		List<StandardEntity> temp = new ArrayList<StandardEntity>(
				model.getEntitiesOfType(StandardEntityURN.CIVILIAN));
		List<Civilian> needRescueCivilian = new ArrayList<Civilian>();
		System.out.println(" needRescueCivilian = " + needRescueCivilian);
		for (StandardEntity next : temp) {
			if (next instanceof Civilian) {
				Civilian civ = (Civilian) next;
				if (isAbleToRescue(civ)) {
					needRescueCivilian.add(civ);
				}
			}
		}
		if (needRescueCivilian.isEmpty()) {
			return;
		} else {
			for (Civilian next : needRescueCivilian) {
				setRescueValue(next);
			}
			Collections.sort(needRescueCivilian, valueComparator);
			for (Civilian next : needRescueCivilian) {
				// System.out.println(getTime()+" "+next+" "+next.getState().getRescueValue());
			}

			assTime = getTime();
			if (assgin < needRescueCivilian.size()) {
				myTarget = needRescueCivilian.get(assgin);
			} else {
				myTarget = needRescueCivilian.get(0);
			}

			// System.out.println(assTime +" : " +assgin);

		}

	}

	public void rescueMyTarget() {

		if (myTarget == null || myTarget.getPosition(model) == null
				|| modelState.savedHumanoids.contains(myTarget)) {
			return;// if(!(location() == myTarget.getPosition(model)) &&
					// (getTime() - assTime) > 5){
			// assgin++;
			// if(assgin > 4) assgin = 0;
			// // System.out.println("assgin again");
			// evaluateCivilian();
			// }
		}

		if (location() == myTarget.getPosition(model)) {// at the
														// same
														// position
			if (myTarget.getState().isVisible(model)) {// I can see
														// it
				if (checkNeedRescue(model, myTarget)) {// need
														// rescue
					rescue(myTarget);
					assgin = 0;
					return;
					// action = true;

				} else if (checkNeedLoad(model, myTarget)) {// need
															// load

					model.getState().savedHumanoids.add(myTarget);

					load(myTarget);
					return;
					// action = true;

					// rescueTarget();
				} else {// need neither rescue nor load

					model.getState().savedHumanoids.add(myTarget);

					rescueTarget();
				}
			} else {// I can't see it althought I was on the same place as the
					// target

				myTarget.getState().setProperty(model, getTime(),
						myTarget.getPositionProperty(),
						getRefuges().get(0).getID().getValue());
				model.getState().savedHumanoids.add(myTarget);

				m_unloadedHumanoid = myTarget;
			}
		} else {
			move(myTarget.getPosition(model));
			return;
			// action = true;
		}
	}

	public void setRescueValue(Civilian civ) {
		if (civ.isBuriednessDefined() && civ.isHPDefined()
				&& civ.isPositionDefined()) {
			civ.getState().rescueValue = civ.getBuriedness() * 125
					+ civ.getHP() + model.getDistance(me(), civ) / 30;
		}
	}

	public boolean isAbleToRescue(Civilian civ) {
		if (!civ.isBuriednessDefined())
			return false;
		if (!civ.isDamageDefined())
			return false;
		if (!civ.isHPDefined())
			return false;
		if (!civ.isPositionDefined())
			return false;
		if (isReachable(civ.getID()))
			return false;
		if (civ.getPosition(model) instanceof Refuge
				|| !(civ.getPosition(model) instanceof Building)) {
			return false;
		}
		if (civ.getHP() == 10000 && civ.getDamage() == 0
				&& civ.getBuriedness() == 0) {
			return false;
		}

		if (civ.getDamage() > 199) {
			return false;
		} else if (civ.getDamage() > 199 && civ.getHP() < 5000
				&& civ.getBuriedness() > 5) {
			return false;
		} else if (civ.getDamage() > 199 && civ.getHP() < 2000) {
			return false;
		} else {
			return true;
		}
	}

	Comparator valueComparator = new Comparator() {
		@Override
		public int compare(Object obj1, Object obj2) {
			Civilian civ1 = (Civilian) obj1;
			Civilian civ2 = (Civilian) obj2;
			if (civ1.getState().getRescueValue() > civ2.getState()
					.getRescueValue())
				return 1;
			else if (civ1.getState().getRescueValue() == civ2.getState()
					.getRescueValue())
				return 0;
			else
				return -1;
		}
	};

	protected void atSearch() {
		if (unVisitedBuildings.isEmpty())
			return;
		Collections.sort(unVisitedBuildings, distanceToMeComparator);
		for (Building next : unVisitedBuildings) {
			if (model.getDistance(location(), next) < Constants.SEE_DISTANCE)
				continue;
			move(next);
			return;
		}
	}

	@Override
	public void checkAround(ChangeSet changed) {
		for (EntityID next : changed.getChangedEntities()) {
			// update Roads
			if (model.getEntity(next) instanceof Civilian) {
				Civilian civ = (Civilian) model.getEntity(next);
				if (!needRescueCivilian.contains(civ)) {
					needRescueCivilian.add(civ);
				}
			}
		}
	}

}
