//SEU2010

package seu.agent;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import rescuecore2.messages.Command;
import rescuecore2.standard.components.StandardAgent;
import rescuecore2.standard.entities.*;
import rescuecore2.standard.messages.AKSpeak;
import seu.messages.CenterMessageManager;
import seu.states.RoadState;
import seu.utilities.AgentUtil;
import seu.utilities.Condition;

/**
   A sample centre agent.
 */
public abstract class CentreAgent<E extends StandardEntity> extends StandardAgent<E> {
    
    public CentreAgent() {
    }
    
    protected void postConnect() {
        super.postConnect();
    }
    
    public int speakChannel = -1;
    
    public String toString() {
        return "centre";
    }

	public void prepareForAct(Collection<Command> heard) {
		if(getTime() < agentIgnoreTime)
			return;
		
		try{
			updateModel(heard);
		}catch (Exception e) {
			System.out.println(me()+"updateModel failed");
			e.printStackTrace();
		}

		modelState = model.getState();
		
		try{
			report();
		}catch (Exception e) {
			System.out.println("report failed");
			e.printStackTrace();
		}	
	}

	public void act(){

	}

	private void updateModel(Collection<Command> heard)
	{
    	//TODO
    	for (Command next : heard) {
        	if(next instanceof AKSpeak){
        		AKSpeak message = (AKSpeak) next;
//        		System.out.println(getTime()+" "+message.getContent().length);
        		StandardEntity sender = model.getEntity(message.getAgentID());
//        		System.out.println(me()+" : Sender = "+sender+" Message = "+new String(message.getContent()));
        		if(sender != me()){
//        			hasHeard = true;
        			if (centerMessageManager != null)
        				centerMessageManager.importMessage(message.getContent(), getTime(), sender);
        		}
        	}
    	}
	}
	
	protected void report(){

		// === Reporting Fiery Buildings :
		for(Building b : modelState.updatedBuildings){
			if(model.getDistance(me(), b) > 30000) continue;
			if(!b.getState().isUnburned()){
//				if(virtualCenterMessageManager != null)
//					virtualCenterMessageManager.reportFire(b);
//				else 
				if(centerMessageManager != null) centerMessageManager.reportFire(b);
//				if(b.getState().tempChanged){
//					if(virtualCenterMessageManager != null)
//						virtualCenterMessageManager.reportTemperature(b);
//					else if(centerMessageManager != null)
//						centerMessageManager.reportTemperature(b);
//				}
				b.getState().report.update(getTime());
			}
		}


		// === Reporting Civilians
		for(Civilian civ : modelState.foundVictimCivilianIC){
//			if(virtualCenterMessageManager != null)
//				virtualCenterMessageManager.reportFoundCivilian(civ);//must be victim
//			else 
			if(centerMessageManager != null)
				centerMessageManager.reportFoundCivilian(civ);//must be victim
			civ.getState().report.update(getTime());
		}
		ArrayList<Civilian> victims = new ArrayList<Civilian>(modelState.updatedCivilians);
		victims.removeAll(modelState.foundVictimCivilianIC);
		for(Civilian civ : victims){
			if(civ.getPosition(model) instanceof Refuge)
				continue;
			
//			if(virtualCenterMessageManager != null)
//				virtualCenterMessageManager.reportVictim(civ);
//			else 
			if(centerMessageManager != null)
				centerMessageManager.reportVictim(civ);
			
			civ.getState().report.update(getTime());
		}
		
		
		// === Reporting Roads (Clears and UnClears and Blocks)
		for (Iterator i = modelState.updatedRoads.iterator(); i.hasNext();) {
			Road road = (Road) i.next();
			if(road.getState().getCondition() == RoadState.CONDITION_PASSABLE){
				if(road.getState().reportClear.getCount() > 1)
					continue;
				road.getState().reportClear.update(getTime());
			}
			
			if(road.getState().getCondition() == RoadState.CONDITION_BLOCKED){
				if(road.getState().reportUnClear.getCount() > 1)
					continue;
				road.getState().reportUnClear.update(getTime());
			}
			
//			if(virtualCenterMessageManager != null)
//				virtualCenterMessageManager.reportRoad(road, road.getCondition());
//			else 
			if(centerMessageManager != null)
				centerMessageManager.reportRoad(road, road.getState().getCondition());
			
		}
		

		// === reporting empty building
		
		List visibleBuildingsIC = IN_VISIBLE_DIST.extract(modelState.updatedBuildings);
		
		for (Iterator iterator = visibleBuildingsIC.iterator();iterator.hasNext();) {
			Building bldg = (Building) iterator.next();
			if (bldg.getState().emptyBuildingReport.getCount() > 1	
					|| bldg.getState().isReportedAsEmpty())
				continue;
			if(!modelState.emptyBuildings.contains(bldg))
				continue;
			
//			if(virtualCenterMessageManager != null)
//				virtualCenterMessageManager.reportEmptyBuilding(building);
//			else 
			if(centerMessageManager != null)
				centerMessageManager.reportEmptyBuilding(bldg);

			bldg.getState().emptyBuildingReport.update(getTime());
			bldg.getState().report.update(getTime());
		}
	}
	
	public CenterMessageManager centerMessageManager = null;
//	public VirtualCenterMessageManager virtualCenterMessageManager = null;
//	public int wait=0;
	
	//TODO
	final Condition IN_VISIBLE_DIST = new Condition()
	{
		final int VISIBLE_DIST = seu.utilities.SimulatorsConstants.VISIBLE_DISTANCE;

		@Override
		public boolean eval(Object obj)
		{
			if (obj instanceof StandardEntity)
				return AgentUtil.distance(me(), (StandardEntity) obj) < VISIBLE_DIST;
			else
				throw new Error("Invalid Usage.");
		}
	};
	
}
