//SEU2010

package seu.agent;

import static rescuecore2.misc.Handy.objectsToIDs;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import rescuecore2.log.Logger;
import rescuecore2.messages.Command;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.entities.ExtraProperty;
import seu.messages.AgentMessageManager;
import seu.messages.SayMessageManager;
import seu.states.FireBrigadeState;
import seu.utilities.Condition;
import seu.utilities.DistanceSorter;
import seu.utilities.FireSite;

/**
 * SEU's FireBrigade Editted by Knight 2010 4:44:22
 */
public class FireBrigadeAgent extends PlatoonAgent<FireBrigade> {
	/*
	 * private static final String MAX_WATER_KEY = "fire.tank.maximum"; private
	 * static final String MAX_DISTANCE_KEY = "fire.extinguish.max-distance";
	 * private static final String MAX_POWER_KEY = "fire.extinguish.max-sum";
	 * 
	 * private int maxWater; private int maxDistance; private int maxPower;
	 */

	private FireBrigadeState myState;
	
	public ArrayList<FireSite> fireSites = new ArrayList<FireSite>();

	private FireSite targetSite;

	private Building targetBuilding;

	private Building lastTargetBuilding;

	private FireSite lastTargetSite;

	protected ArrayList<Integer> channels;

	@Override
	public String toString() {
		return "SEU's FireBrigade";
	}

//	@Override
//	protected FireBrigade me() {
//		return (FireBrigade) modelState.me;
//	}

	@Override
	protected void postConnect() {

		super.postConnect();

		model.indexClass(StandardEntityURN.BUILDING, StandardEntityURN.REFUGE);

//		model.getState().initializeNearBuildings();

		model.getState().setBuildingConnections();

		/*
		 * maxWater = config.getIntValue(MAX_WATER_KEY); maxDistance =
		 * config.getIntValue(MAX_DISTANCE_KEY); maxPower =
		 * config.getIntValue(MAX_POWER_KEY);
		 * Logger.info("Sample fire brigade connected: max extinguish distance = "
		 * + maxDistance + ", max power = " + maxPower + ", max tank = " +
		 * maxWater);
		 */
	}

	// Prepare for Act
	// 2010 10:08
	public void prepareForAct(int time, ChangeSet changed, Collection<Command> heard) {

		if(time < agentIgnoreTime) return;

		if (time == agentIgnoreTime) {
			try{
				initialChannels(time);
			}catch (Exception e) {
				System.out.println("initialChannel failed");
				e.printStackTrace();
			}
        }
    	
    	//update world message

		try{
			updateModel(heard);
		}catch (Exception e) {
			System.out.println(me()+"updateModel failed");
			e.printStackTrace();
		}

//    	System.out.println(getTime()+" fb "+me()+" 1");
    	super.prepareForAct(time, changed);
//    	System.out.println(getTime()+" fb "+me()+" 2");
    	modelState = model.getState();
    	myState = me().getState();
    	
		initialEvironment();
//		System.out.println(getTime()+" fb "+me()+" 3");
		updateFireSites();
//		System.out.println(getTime()+" fb "+me()+" 4");
		
		
		
		try{
			if (agentMessageManager != null) {
				agentMessageManager.sendMessages();
				if (agentMessageManager.getMessageBytes() != null) {
//					System.out.println("at"+agentMessageManager.getMessageBytes().length);
					sendSpeak(time, speakChannel, agentMessageManager.getMessageBytes());
					// System.out.println("at send =  "+(agentMessageManager.getMessageBytes()));
				}
			}
		}catch (Exception e) {
			System.out.println(me()+" sendMessage failed");
			e.printStackTrace();
		}	
//      System.out.println(getTime()+" fb "+me()+" 5");
	}

	private void updateFireSites() {
		// 1, add fire or construct some new fire sites
		for (Building b : modelState.burningBuildings) {
			if (b.getFieryness() >= 5 && b.getFieryness() <= 7)
				continue;
			boolean isNewSite = true;
			ArrayList<Integer> siteHaveThisFire = new ArrayList<Integer>();
			for (int i = 0; i < fireSites.size(); i++) {
				FireSite fs = fireSites.get(i);
				if (fs.addFire(b) != -1) {
					siteHaveThisFire.add(new Integer(i));
					isNewSite = false;
				}
			}
			if (isNewSite) {
				FireSite newSite = new FireSite(b, model);
				fireSites.add(newSite);
				continue;
			}
			// 2, merge any overlapped sites
			if (siteHaveThisFire.size() > 1) {
				FireSite mergeSite = fireSites.get(siteHaveThisFire.get(0)
						.intValue());
				for (int i = 1; i < siteHaveThisFire.size(); i++) {
					mergeSite.mergeWith(fireSites.get(siteHaveThisFire.get(i)
							.intValue()));
				}
				ArrayList<FireSite> needToRemove = new ArrayList<FireSite>();
				for (int i = 0; i < siteHaveThisFire.size(); i++)
					needToRemove.add(fireSites.get(siteHaveThisFire.get(i)
							.intValue()));
				fireSites.removeAll(needToRemove);
				fireSites.add(mergeSite);
			}
		}
		// 3, modify the dangerous fires and set value for sites
		for (FireSite fs : fireSites) {
			fs.updateDangerousFires();
			fs.updateSite(model);
			calculateVictimsAroundSite(fs);
		}

	}

	public void calculateVictimsAroundSite(FireSite site) {
		int victimAround = 0;
		int count = 0;
		for (int i = 0; i < 4; i++) {
			count = getNumberVictimInDirection(i, site);
			victimAround += count;
		}
		site.setNumberVictimsAround(victimAround);
	}

	// TODO Here should to add civilian state
	//
	public int getNumberVictimInDirection(int direction, FireSite site) {

		ArrayList<Human> temp = new ArrayList<Human>();

		for (Civilian civ : modelState.civilians) {
			if (civ.getPosition(model) == null)
				continue;
			if (!(civ.getPosition(model) instanceof Building))
				continue;
			if (civ.getState().needLoad(model)
					|| civ.getState().needRescue(model))
				temp.add(civ);
		}

		ArrayList positions = new ArrayList();
		for(Human hm:temp)
			positions.add(hm.getPosition(model));
		int count = 0;
		int mainDepthSearch = 2000;
		int x_DepthSearch = Math.abs((site.m_maxX - site.m_minX) / 3);
		int y_DepthSearch = Math.abs((site.m_maxY - site.m_minY) / 3);

		switch (direction) {
		case 0:
			Condition northCondition = Y_PRP.gte(site.m_maxY - y_DepthSearch)
					.and(Y_PRP.lte(site.m_maxY + mainDepthSearch));
			count = Util.getConditionCount(positions, northCondition);
			break;
		case 1:
			Condition eastCondition = X_PRP.gte(site.m_maxX - x_DepthSearch)
					.and(X_PRP.lte(site.m_maxX + mainDepthSearch));
			count = Util.getConditionCount(positions, eastCondition);
			break;
		case 2:
			Condition southCondition = Y_PRP.gte(site.m_minY - mainDepthSearch)
					.and(Y_PRP.lte(site.m_minY + y_DepthSearch));
			count = Util.getConditionCount(positions, southCondition);
			break;
		case 3:
			Condition westCondition = X_PRP.gte(site.m_minX - mainDepthSearch)
					.and(X_PRP.lte(site.m_minX + x_DepthSearch));
			count = Util.getConditionCount(positions, westCondition);
			break;

		}
		return count;
	}

	private final ExtraProperty X_PRP = new ExtraProperty() {
		@Override
		public Object eval(Object obj) {
			return new Integer(((Area) obj).getX());
		}
	};

	private final ExtraProperty Y_PRP = new ExtraProperty() {
		@Override
		public Object eval(Object obj) {
			return new Integer(((Area) obj).getY());
		}
	};

	private void initialEvironment() {
		// lastTargetBuilding = targetBuilding;
		lastTargetSite = targetSite;
		targetSite = null;
		targetBuilding = null;
	}

	@Override
	protected void think(int time, ChangeSet changed, Collection<Command> heard) {

		prepareForAct(time, changed, heard);
//
		act(time);
		
//		if (time == config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY)) {
//    		initialChannels(time);
//        }
		
//		if (time == config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY)) {
//            // Subscribe to channel 1
//            sendSubscribe(time, 2);
//        }
//        for (Command next : heard) {
//        	if(next instanceof AKSpeak){   		
//        		AKSpeak message = (AKSpeak) next;
//        		StandardEntity sender = model.getEntity(message.getAgentID());
//        		//System.out.println("Sender = "+sender+" Message = "+new String(message.getContent()));
//        		if (agentMessageManager != null)
//    				agentMessageManager.importMessage(message.getContent(), getTime(), sender);
//            Logger.debug("Heard " + next);
//        	}
//        }
//        super.prepareForAct();
//        if (agentMessageManager != null){
//			agentMessageManager.sendMessages();
//			sendSpeak(time, 2 ,agentMessageManager.getMessageBytes());
//		}

        
//        FireBrigade me = me();
//        // Are we currently filling with water?
//        if (me.isWaterDefined() && me.getWater() < extinguishWater && location() instanceof Refuge) {
//            Logger.info("Filling with water at " + location());
//            sendRest(time);
//            return;
//        }
//        // Are we out of water?
//        if (me.isWaterDefined() && me.getWater() == 0) {
//            // Head for a refuge
//            List<EntityID> path = search.breadthFirstSearch(me().getPosition(), getRefugesID());
//            if (path != null) {
//                Logger.info("Moving to refuge");
//                sendMove(time, path);
//                return;
//            }
//            else {
//                Logger.debug("Couldn't plan a path to a refuge.");
//                path = randomWalk();
//                Logger.info("Moving randomly");
//                sendMove(time, path);
//                return;
//            }
//        }
//        // Find all buildings that are on fire
//        Collection<Building> all = getBurningBuildings();
//        // Can we extinguish any right now?
//        for (Building next : all) {
//            if (model.getDistance(me(), next) <= extinguishDistance) {
//                Logger.info("Extinguishing " + next);
//                sendExtinguish(time, next.getID(), extinguishPower);
////                sendSpeak(time, 1, ("Extinguishing " + next).getBytes());
//                return;
//            }
//        }
//        // Plan a path to a fire
//        for (Building next : all) {
//            List<EntityID> path = planPathToFire(next);
//            if (path != null) {
//                Logger.info("Moving to target");
//                sendMove(time, path);
//                return;
//            }
//        }
//        List<EntityID> path = null;
//        Logger.debug("Couldn't plan a path to a fire.");
//        path = randomWalk();
//        Logger.info("Moving randomly");
//        sendMove(time, path);
	}

	private void act(int time) {
		if(time < agentIgnoreTime) return;
		// TODO need for test
//		moreWaterEveryCycle();
		
		super.act();
//		System.out.println(getTime()+" fb "+me()+" 6");
//		System.out.println(getTime()+" 1");
		if (action)
			return;
		// 
		moveToRefugeIfDamagedOrTankIsEmpty(time);
//		System.out.println(getTime()+" fb "+me()+" 8");
		if (action)
			return;
//		System.out.println(getTime()+" 2");
		// 
		// areWeFillingWaterNow(time, me);
		//
		// // Are we out of water?
		// if (me.isWaterDefined() && me.getWater() == 0) {
		// // Head for a refuge
		// List<EntityID> path = search.breadthFirstSearch(location(),
		// getRefuges());
		// if (path != null) {
		// Logger.info("Moving to refuge");
		// sendMove(time, path);
		// return;
		// } else {
		// Logger.debug("Couldn't plan a path to a refuge.");
		// path = randomWalk();
		// Logger.info("Moving randomly");
		// sendMove(time, path);
		// return;
		// }
		// }

		// Knight add a
		selectFireSites();
//		System.out.println(getTime()+" fb "+me()+" 9");
//		System.out.println(getTime()+" e1");
		// Find all buildings that are on fire
		// Collection<Building> all = getBurningBuildings();

		// Can we extinguish any right now?
		if(targetSite != null){
//			System.out.println(getTime()+" e2");
//			System.out.println(getTime()+" "+targetSite.toString());
			for (Building next : targetSite.burningBuildings()) {
//				System.out.println(getTime()+" "+next);
				if(!(location() instanceof Refuge)){
					if (model.getDistance(me(), next) <= extinguishDistance) {
//						Logger.info("Extinguishing " + next);
						extinguish(next.getID());
//						sendSpeak(time, 1, ("Extinguishing " + next.getID()).getBytes());
						return;
					}
				}
				else if(location() instanceof Refuge){
					if (model.getDistance(me(), next) <= Constants.SEE_DISTANCE) {
//						Logger.info("Extinguishing " + next);
						extinguish(next.getID());
//						sendSpeak(time, 1, ("Extinguishing " + next.getID()).getBytes());
						return;
					}
				}
				
			}
			// Plan a path to a fire
			for (Building next : targetSite.burningBuildings()) {
//				System.out.println(getTime()+" "+next);
				List<EntityID> path = planPathToFire(next);
				if (path != null) {
					Logger.info("Moving to target");
					move(path);
//					System.out.println(getTime()+" move "+path);
					return;
				}
			}
			
		}
		
//		fbSearch();
//		if (action)return;
//		System.out.println(getTime()+" fb "+me()+" 9");
//		System.out.println(getTime()+" 4");
		searchCivilians(true);
//		System.out.println(getTime()+" fb "+me()+" 10");
//		List<EntityID> path = null;
//		Logger.debug("Couldn't plan a path to a fire.");
//		path = randomWalk();
//		Logger.info("Moving Randomly");
//		sendMove(time, path);
//		System.out.println(getTime()+" 5");

		if (action)
			return;
//		System.out.println(getTime()+" 6");
		
	}

	HashMap<FireSite, Float> groupValueMap = new HashMap<FireSite, Float>();

	private void selectFireSites() {

		sortFireSite();
		for (FireSite site : fireSites) {

			if (site.burningBuildings().isEmpty())
				continue;

			targetSite = site;
			

		}
	}

	private void sortFireSite() {

		for (FireSite fs : fireSites) {
			if (fs.burningBuildings().isEmpty()) {
				groupValueMap.put(fs, Float.MIN_VALUE);
				continue;
			}
			int selfDistanceValue = 30 * 30;

			int distance = fs.distance(me().getX(), me().getY()) / 1000;

			if (distance > 30) {
				selfDistanceValue = (int) Math.pow(distance, 2);
			}

			float value = fs.getBurningArea() / 10 + selfDistanceValue;

			float victim = fs.getNumberVictimsAround() / 50f
					/ (fs.burningBuildings().size() + 1f);

			value += victim;

			if (lastTargetSite == fs) {
				value *= 1.06f;
			}

			groupValueMap.put(fs, value);

		}

		Collections.sort(fireSites, new HashValueComaparator<FireSite>(
				groupValueMap));
	}

	public class HashValueComaparator<T> implements Comparator<T> {
		Map<T, Float> hashValue;

		public HashValueComaparator(Map<T, Float> valueMap) {
			hashValue = valueMap;
		}

		public int compare(T t1, T t2) {
			if (hashValue.get(t2) > hashValue.get(t1)) {
				return 1;
			} else if (hashValue.get(t1) > hashValue.get(t2)) {
				return -1;
			} else {
				return 0;
			}

		}

	}

	//
	protected void initialChannels(int time){
		if(useSpeak){
    		if(model.getState().channelCount > 3){
        		sendSubscribe(time, 3);
        		speakChannel = 2;
        	}
        	else if(model.getState().channelCount > 2){
        		sendSubscribe(time, 1);
        		speakChannel = 2;
        	}
        	else if(model.getState().channelCount > 1){
        		sendSubscribe(time, 1);
        		speakChannel = -1;
        	}
    		agentMessageManager = new AgentMessageManager(model, me().getID());
    	}
		else {
			sayMessageManager = new SayMessageManager(model, me().getID());
		}
//    	System.out.println(me() +" areas = "+modelState.areas.size()+" "
//    			+" Buildings = "+modelState.buildings.size()+" "
//    			+" Roads = "+modelState.roads.size()+" "
//    			+model.getEntitiesOfType(StandardEntityURN.BUILDING).size()
//    			+model.getEntitiesOfType(StandardEntityURN.ROAD).size()
//    			+model.getEntitiesOfType(StandardEntityURN.BUILDING,StandardEntityURN.ROAD).size());
    }
	
	

	/**
	 * @param time
	 * @param me
	 * @author Knight Guan
	 */
	// private void areWeFillingWaterNow(int time, FireBrigade me) {
	// // Are we currently filling with water?
	// if (me.isWaterDefined() && me.getWater() < extinguishWater
	// && location() instanceof Refuge) {
	// Logger.info("Filling with water at " + location());
	// sendRest(time);
	// return;
	// }
	// }

	// more extinguish
	private void moreWaterEveryCycle() {
		if (lastTargetBuilding != null) {
			if (lastTargetBuilding.getState().getFieryness() >= 5
					&& lastTargetBuilding.getState().getFieryness() <= 7
					&& lastTargetBuilding.getState().moreExtinguishCount < 1
					&& lastTargetBuilding.getState().moreExtinguishCount >= 0
					&& model.getDistance(me(), lastTargetBuilding) < extinguishDistance
					&& hasBurnedDownAround(lastTargetBuilding)) {

				// System.err.println("FireBrigade:" + me().id + " Time:"
				// + getTime() + "extraExtinguish:" + targetBuilding);

				lastTargetBuilding.getState().moreExtinguishCount += 1;
				sendExtinguish(getTime(), lastTargetBuilding.getID(),
						extinguishPower);

			} else {
				lastTargetBuilding.getState().moreExtinguishCount = 0;
				lastTargetBuilding = null;
			}
		}

	}

	private boolean hasBurnedDownAround(Building buil) {
		// 
		boolean flag = false;
		for (Building b : buil.getState().connectedBuildings) {
			if (b.getFieryness() == 8) {
				flag = true;
				break;
			}
		}
		return flag;
	}

	protected final ExtraProperty distancePrp = new ExtraProperty() {
		@Override
		public Object eval(Object obj) {
			if (obj instanceof StandardEntity)
				return new Integer(model
						.getDistance(me(), (StandardEntity) obj));
			return null;
		}
	};

	private void moveToRefugeIfDamagedOrTankIsEmpty(int time) {

		int requiredWater = 10000;

		if (targetSite != null) {
			Building nearestRefuge = (Building) distancePrp
					.min(modelState.refuges);
			if (model.getDistance(me(), nearestRefuge) <= 140000) {
				if (targetSite.getTotalArea() <= 9050)
					requiredWater = 5000;
				else
					requiredWater = 15000;
			}
		}
		int fbCount = 0;
		if (me().getPosition(model) instanceof Refuge) {
			for (int i = 0; i < modelState.fireBrigades.size(); i++) {
				FireBrigade fb = modelState.fireBrigades.get(i);
				if (fb.getPosition(model) == me().getPosition(model)
						&& fb.getState().isVisible(model))
					fbCount++;
			}
			requiredWater = 16000 - fbCount * 1000;
			requiredWater = Util.max(requiredWater, 10000);
			requiredWater = Util.min(requiredWater, extinguishWater);
			if (me().getWater() < requiredWater) {
				rest();
			}
		}
		if ((me().getDamage() > 70 && me().getHP() <= 4000)
				|| me().getWater() == 0) {
			restAtRefuge(time);
		}
	}

	private void restAtRefuge(int time) {
		// Refugee
		List<EntityID> path = search.breadthFirstSearch(location().getID(),
				getRefugesID());
		if (path != null) {
			Logger.info("Moving to refuge");
			sendMove(time, path);
			return;
		} else {
			Logger.debug("Couldn't plan a path to a refuge.");
			path = randomWalk();
			Logger.info("Moving Randomly");
			sendMove(time, path);
			return;
		}

	}

	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
	}

	private List<Building> getBurningBuildings() {
		Collection<StandardEntity> e = model
				.getEntitiesOfType(StandardEntityURN.BUILDING);
		List<Building> result = new ArrayList<Building>();
		for (StandardEntity next : e) {
			if (next instanceof Building) {
				Building b = (Building) next;
				if (b.isOnFire()) {
					result.add(b);
				}
			}
		}
		// Sort by distance
		Collections.sort(result, new DistanceSorter(location(), model));
		return result;
	}

	private List<EntityID> planPathToFire(Building target) {
		// Try to get to anything within maxDistance of the target
		Collection<EntityID> targets = target.getNeighbours();
		for(EntityID next : targets){
			if(model.getEntity(next) instanceof Building){
				Building b = (Building)model.getEntity(next);
				if(b.getState().isBurning()){
					targets.remove(b);
				}
			}
		}
		if (targets.isEmpty()) {
			if(location() instanceof Refuge){
				targets.addAll(objectsToIDs(model.getObjectsInRange(target,
						Constants.SEE_DISTANCE)));
			}
			else{
				targets.addAll(objectsToIDs(model.getObjectsInRange(target,
						extinguishDistance)));
			}
//			System.out.println("plantofire is empty");
			
		}
		
		if (targets.isEmpty()) {
			return null;
		}	
//		List<EntityID> targetID = new ArrayList<EntityID>();
//		for(StandardEntity  se : targets){
//			targetID.add(se.getID());
//		}

		return search.breadthFirstSearch(location().getID(), targets);

	}
	
	protected void fbSearch(){
		if(unVisitedBuildings.isEmpty()) return;
		Collections.sort(unVisitedBuildings, distanceToMeComparator);
		for(Building next : unVisitedBuildings){
			if(model.getDistance(location(), next) < Constants.SEE_DISTANCE) continue;
			move(next);
			return;
		}
	}
}
