//SEU2010

package seu.agent;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.messages.Command;
import rescuecore2.standard.components.StandardAgent;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityFactory;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.messages.AKSay;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.messages.AgentMessageManager;
import seu.messages.SayMessageManager;
import seu.utilities.Condition;
import seu.utilities.PathMaker;
import seu.utilities.Partition.RadialPartition;

/**
 * Abstract base class for sample agents.
 * 
 * @param <E>
 *            The subclass of StandardEntity this agent wants to control.
 */
public abstract class PlatoonAgent<E extends StandardEntity> extends
		StandardAgent<E> implements seu.utilities.ConstantConditions
{
	private static final int RANDOM_WALK_LENGTH = 50;

	protected AgentMessageManager agentMessageManager = null;
	public SayMessageManager sayMessageManager = null;

	protected List<RadialPartition> searchPartitions;
	protected RadialPartition basePartition = null;
	protected RadialPartition myPartition;

	public List<Human> myExploreTeams = new ArrayList<Human>();
	public List<EntityID> movePath = new ArrayList<EntityID>();
	// protected int worldTime;
	/**
	 * The search algorithm.
	 */
	protected PathMaker search;
	public int speakChannel = -1;
	/**
	 * Cache of building IDs.
	 */
	protected List<EntityID> buildingIDs;

	/**
	 * Cache of road IDs.
	 */
	protected List<EntityID> roadIDs;

	/**
	 * Cache of refuge IDs.
	 */
	protected List<EntityID> refugeIDs;

	private Map<EntityID, Set<EntityID>> neighbours;

	private Area currentPosition;

	private Area lastPosition1;
	private Area lastPosition2;
	private Area lastPosition3;

	boolean isClear = false;
	boolean isRescue = false;
	boolean isExtinguish = false;
	boolean trymove = false;
	boolean moveAgain = false;

	boolean action = false;

	protected Human m_unloadedHumanoid = null;
	protected Human myLoadedVictim = null;

	protected boolean submittedAction = false;
	protected boolean submittedMoveAct = false;

	protected final Map<Building, Integer> fullBuildingsTimeMap = new HashMap<Building, Integer>();
	protected final Collection fullBuildings = fullBuildingsTimeMap.keySet();

	protected Human giveWayRequirer = null;
	private final int m_prevTimeRoamed = -1;
	protected boolean roamed = false;

	protected final HashSet<Building> visibleBuildingsIC = new HashSet<Building>();

	@Override
	protected void postConnect()
	{
		super.postConnect();

		search = new PathMaker(model);

		for (StandardEntity next : model)
		{
			if (next instanceof Building)
			{
				unVisitedBuildings.add((Building) next);
			}
			if (next instanceof Road)
			{
				unKnownRoads.add((Road) next);
			}
			if (next instanceof Refuge)
			{
				unKnownRefuges.add((Refuge) next);
			}
		}

		if (me() instanceof AmbulanceTeam)
		{
			myExploreTeams.addAll(model.getState().ambulanceTeams);
		} else if (me() instanceof FireBrigade)
		{
			myExploreTeams.addAll(model.getState().fireBrigades);
		} else if (me() instanceof PoliceForce)
		{
			myExploreTeams.addAll(model.getState().policeForces);
		}

		Collections.sort(myExploreTeams, Util.idComparator);
		// searchPartitions = new
		// ArrayList<RadialPartition>(model.getState().radialParitions);
	}

	public void prepareForAct(int time, ChangeSet changed)
	{

		if (getTime() < agentIgnoreTime)
			return;

		// if (getTime() > 9 && !isRadioChannelCollapsed && !hasHeard)
		// {
		// System.out
		// .println(me()
		// +
		// " Warning ............. \n *******    Communication collapsed!    ***");
		// isRadioChannelCollapsed = true;
		// }
		// try{
		// updateMyGraph(model);
		// }catch (Exception e) {
		// System.out.println(me()+" updateMyGraph failed");
		// e.printStackTrace();
		// }
		//		

		try
		{
			if (me() instanceof AmbulanceTeam)
			{
				AmbulanceTeam at = (AmbulanceTeam) me();
				submittedMoveAct = at.getState().submittedMoveAct();
				submittedAction = at.getState().submittedAction();
			} else if (me() instanceof FireBrigade)
			{
				FireBrigade fb = (FireBrigade) me();
				submittedMoveAct = fb.getState().submittedMoveAct();
				submittedAction = fb.getState().submittedAction();
			} else if (me() instanceof PoliceForce)
			{
				PoliceForce pf = (PoliceForce) me();
				submittedMoveAct = pf.getState().submittedMoveAct();
				submittedAction = pf.getState().submittedAction();
			}
		} catch (Exception e)
		{
			System.out.println(me() + " submit failed");
			e.printStackTrace();
		}

		checkAround(changed);

		checkAgentCurrentPosition();

		checkFullBuildings();

		updateBuildingSet();

		// updatePrevRoads();

		updateForCivilianSearch();

		// updateModel(heard);

		// stepWorldCG();

		if (agentMessageManager != null)
		{
			try
			{
				report(agentMessageManager);
			} catch (Exception e)
			{
				System.out.println("report failed");
				e.printStackTrace();
			}
		}

		doSomethingIfWereStuckByHumanoids();

		try
		{
			if (!useSpeak)
			{
				HashSet addtionMsgs = new HashSet(agentMessageManager
						.getMessagesToCenters());
				addtionMsgs.addAll(agentMessageManager.getMessagesToAgents());
				if (sayMessageManager != null)
				{
					sayMessageManager.sendAdditionalMessages(addtionMsgs);
					if (sayMessageManager.getMessageBytes() != null)
					{
						sendSay(time, sayMessageManager.getMessageBytes());
					}
				}

			} else
			{
				HashSet addtionMsgs = new HashSet(agentMessageManager
						.getMessagesToAgents());
				if (sayMessageManager != null)
				{
					sayMessageManager.sendAdditionalMessages(addtionMsgs);
					if (sayMessageManager.getMessageBytes() != null)
					{
						sendSay(time, sayMessageManager.getMessageBytes());
					}
				}
			}
		} catch (Exception e)
		{
			System.out.println(me() + " sayMessage failed");
			e.printStackTrace();
		}

	}

	protected void act()
	{

		if (getTime() < agentIgnoreTime)
			return;
		action = false;

		pfClearInBuilding();
		// if(me() instanceof PoliceForce){
		// System.out.println(me()+" "+getTime()+" 1");
		// }
		if (action)
			return;

		tryMove();
		// if(me() instanceof PoliceForce){
		// System.out.println(me()+" "+getTime()+" 2");
		// }
		if (action)
			return;

		isClear = false;
		isRescue = false;
		isExtinguish = false;
		// exitIfIsBurning();
		// if(action)return;
		restIfIsDamaging();
		// if(me() instanceof FireBrigade){
		// System.out.println(getTime()+" fb "+me()+" 6");
		// }
		// if(me() instanceof PoliceForce){
		// System.out.println(me()+" "+getTime()+" 3");
		// }
		if (action)
			return;

		if (giveWayRequirer != null)
		{

			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString(me()+"  give way  in: "+getTime(), 10,
			// "action");

			// moveToAvailableBldgsIfRequire();
		}

		if ((me() instanceof FireBrigade))
		{
			FireBrigade fb = (FireBrigade) me();
			if (fb.getState().getWereLockedByBlockade(getTime()))
			{
				extinguishNearFire();
				if (action)
					return;
				searchNearBuilding();
				if (action)
					return;
				// tryToGetOutOfBlockades();
				// if(action)return;
			}
		}
		if ((me() instanceof AmbulanceTeam))
		{
			AmbulanceTeam at = (AmbulanceTeam) me();
			if (at.getState().getWereLockedByBlockade(getTime()))
			{
				searchNearBuilding();
				if (action)
					return;
				// tryToGetOutOfBlockades();
				// if(action)return;
			}
		}
		// if(me() instanceof FireBrigade){
		// System.out.println(getTime()+" fb "+me()+" 7");
		// }

		// if ((me() instanceof FireBrigade)) {
		// FireBrigade fb = (FireBrigade) me();
		// if (fb.getState().getStuckCycles() >= 2) {
		// if(!roamed || getTime()-m_prevTimeRoamed <= 1)
		// roam();
		// else roamed = false;
		// }
		// }
		// else if ((me() instanceof AmbulanceTeam)) {
		// AmbulanceTeam at = (AmbulanceTeam) me();
		// if (at.getState().getStuckCycles() >= 2) {
		// if(!roamed || getTime()-m_prevTimeRoamed <= 1)
		// roam();
		// else roamed = false;
		// }
		// }
		// else if ((me() instanceof PoliceForce)) {
		// PoliceForce pf = (PoliceForce) me();
		// if (pf.getState().getStuckCycles() >= 2) {
		// if(!roamed || getTime()-m_prevTimeRoamed <= 1)
		// roam();
		// else roamed = false;
		// }
		// }

	}

	protected void pfClearInBuilding()
	{
		if (me() instanceof PoliceForce && location() instanceof Building)
		{
			for (StandardEntity next : search.findNeighbours(location()))
			{
				if (next instanceof Area)
				{
					Area ar = (Area) next;
					if (!ar.isBlockadesDefined())
						return;
					if (ar.getBlockades().isEmpty())
						return;
					if (model.getDistance(location(), ar) > Constants.SEE_DISTANCE)
					{
						// System.out.println("not see");
						return;
					}
					clear(ar.getBlockades().get(0));
				}
			}
		}
	}

	protected void tryMove()
	{

		if (location() instanceof Refuge && me() instanceof FireBrigade)
			return;
		if (me() instanceof PoliceForce && isClear)
			return;
		if (me() instanceof AmbulanceTeam && isRescue)
			return;
		if (me() instanceof FireBrigade && isExtinguish)
			return;

		if (moveAgain)
		{
			move(randomWalk());

			if (trymove)
			{
				trymove = false;
			} else
			{
				moveAgain = false;
			}
		}

		// if(me() instanceof PoliceForce && location() instanceof Building){
		// trymove = true;
		// moveAgain = true;
		// List<EntityID> temp = new ArrayList<EntityID>();
		// temp.add(currentPosition.getID());
		// move(temp, currentPosition.getX(), currentPosition.getY());
		// }

		if (lastPosition2 != null && lastPosition1 != null
				&& currentPosition != null)
		{
			if ((lastPosition2 == lastPosition1)
					&& (lastPosition2 == currentPosition))
			{
				trymove = true;
				moveAgain = true;
				List<EntityID> temp = new ArrayList<EntityID>();
				temp.add(currentPosition.getID());
				move(temp, currentPosition.getX(), currentPosition.getY());
			}
		}
	}

	/**
	 * Construct a random walk starting from this agent's current location.
	 * Buildings will only be entered at the end of the walk.
	 * 
	 * @return A random walk.
	 */
	protected List<EntityID> randomWalk()
	{
		List<EntityID> result = new ArrayList<EntityID>(RANDOM_WALK_LENGTH);
		Set<StandardEntity> seen = new HashSet<StandardEntity>();
		StandardEntity current = location();
		for (int i = 0; i < RANDOM_WALK_LENGTH; ++i)
		{
			result.add(current.getID());
			seen.add(current);
			List<StandardEntity> neighbours = new ArrayList<StandardEntity>(
					search.findNeighbours(current));
			List<StandardEntity> safeNeighbours = new ArrayList<StandardEntity>();
			List<StandardEntity> fireNeighbours = new ArrayList<StandardEntity>();
			List<StandardEntity> visitedNeighbours = new ArrayList<StandardEntity>();
			List<Road> checkedRoads = new ArrayList<Road>();
			// List<Road> needClearRoads = new ArrayList<Road>();
			for (StandardEntity it : neighbours)
			{
				if (it == lastPosition1)
				{
					// System.out.println(" delete "+it);
					continue;
				}
				if (it instanceof Road)
				{
					Road r = (Road) it;
					if (me() instanceof PoliceForce)
					{
						if (modelState.checkedRoads.contains(r))
						{
							if (r.isBlockadesDefined()
									& r.getBlockades().isEmpty())
							{
								checkedRoads.add(r);
							}
						} else
						{
							safeNeighbours.add(r);
						}
					} else
					{
						safeNeighbours.add(it);
					}
				} else if ((it instanceof Building))
				{
					// if(me() instanceof PoliceForce)continue;

					Building b = (Building) it;
					if (b.getState().isBurning() || b instanceof Refuge)
					{
						fireNeighbours.add(b);
						// System.out.println(b+" is on fire in walk");
						continue;
					} else if (modelState.visitedBuildings.contains(b))
					{
						visitedNeighbours.add(b);
						// System.out.println(b+" is visited in walk");
						continue;
					}
					safeNeighbours.add(it);
				}
				// else{
				// System.out.println(it+" not Area in walk");
				// }
			}

			if (safeNeighbours.isEmpty())
			{
				if (me() instanceof PoliceForce)
				{
					// System.out.println("1"+neighbours);
					// System.out.println("2"+safeNeighbours);
					if (!checkedRoads.isEmpty() || !visitedNeighbours.isEmpty())
					{
						safeNeighbours.addAll(checkedRoads);
						safeNeighbours.addAll(visitedNeighbours);
					} else
					{
						safeNeighbours.add(lastPosition1);
						safeNeighbours.addAll(fireNeighbours);
					}
					// System.out.println("3"+safeNeighbours);
				} else
				{
					if (!visitedNeighbours.isEmpty())
					{
						safeNeighbours.addAll(visitedNeighbours);
					} else if (location() instanceof Building)
					{
						safeNeighbours.add(lastPosition1);
						safeNeighbours.addAll(fireNeighbours);
					}
				}

			}

			Collections.shuffle(safeNeighbours, random);
			boolean found = false;
			for (StandardEntity next : safeNeighbours)
			{
				if (seen.contains(next))
				{
					continue;
				}
				current = next;
				found = true;
				break;
			}
			if (!found)
			{
				// We reached a dead-end.
				break;
			}
		}
		return result;
	}

	public final Condition REACHABLE_C = new Condition()
	{
		@Override
		public boolean eval(Object obj)
		{
			boolean reachable = false;
			if (obj instanceof Human)
			{
				Human mv = (Human) obj;

				if (mv.getPosition(model) == null)
					return false;
				reachable = model.getState().CG.isReachable(mv
						.getPosition(model));
			} else
				reachable = model.getState().CG.isReachable((Area) obj);

			if (!reachable && !useSpeak)
				return Math.random() < 0.3;

			return reachable;
		}
	};

	public Condition FULL_BUILDING_C = new Condition()
	{
		@Override
		public boolean eval(Object obj)
		{
			if (obj instanceof Building)
			{
				return fullBuildings.contains(obj);
			} else
				return false;
		}
	};

	Comparator allocationProbComparator = new Comparator()
	{
		public int compare(Object obj1, Object obj2)
		{
			Building b1 = (Building) obj1;
			Building b2 = (Building) obj2;

			double v1 = b1.getState().getCivPossibility(model) * 100
					+ getVisibleBldgProb(b1);
			double v2 = b2.getState().getCivPossibility(model) * 100
					+ getVisibleBldgProb(b2);

			if (v1 < v2)
				return 1;
			else
				return -1;
		}
	};

	protected double getVisibleBldgProb(Building bld)
	{
		if (bld == null)
			return 0.5;
		double hearable = 0.8 * bld.getState().getVisibleBuildings().size()
				/ model.getState().m_maxVisibleBuildingsPerBuilding
				+ bld.getState().civPossiblity;

		// hearable = Math.sqrt(hearable)/2;

		double heard = (double) bld.getState().getHearableBuildings().size()
				/ (double) model.getState().m_maxHearableBuildingsPerBuilding;

		// world.logFile.writeString(bld+"  vsb= "+hearable
		// +"  hed= "+heard, 63, "search");

		return Math.min(0.999, 0.5 + hearable / 2d + heard / 8d);
	}

	// ============================ decision methods
	// =======================================//

	/**
	 * 
	 * By JiangYh 2010-6-11
	 */
	private void checkAgentCurrentPosition()
	{
		// if (lastPosition2 != null){
		// lastPosition3 = lastPosition2;
		// }
		// else{
		// lastPosition3 = null;
		// }
		// if (lastPosition1 != null){
		// lastPosition2 = lastPosition1;
		// }
		// else{
		// lastPosition2 = null;
		// }
		// if (currentPosition != null){
		// lastPosition1 = currentPosition;
		// }
		// else{
		// lastPosition1 = null;
		// }
		if (location() instanceof Refuge)
		{
			Refuge re = (Refuge) location();
			if (!usefulRefuges.contains(re))
				usefulRefuges.add(re);
			if (unKnownRefuges.contains(re))
				unKnownRefuges.remove(re);
		}

		// for(Human plt : model.getState().platoonAgents){
		// if(plt.getPosition(model) instanceof Refuge){
		// modelState.usefulRefuges.add((Refuge)location());
		// }
		// }

		// lastPosition3 = lastPosition2;
		lastPosition2 = lastPosition1;
		lastPosition1 = currentPosition;
		if (location() instanceof Area)
		{
			currentPosition = (Area) location();
		} else
		{
			currentPosition = (Area) ((Human) location()).getPosition(model);
		}

		if (me() instanceof PoliceForce)
		{
			if (movePath.contains(currentPosition.getID()))
			{
				int end = movePath.indexOf(currentPosition.getID());
				for (int i = 0; i <= end; i++)
				{
					if (model.getEntity(movePath.get(i)) instanceof Road)
					{
						Road r = (Road) model.getEntity(movePath.get(i));
						if (!model.getState().checkedRoads.contains(r))
						{
							model.getState().checkedRoads.add(r);
						}
					}
				}
			}
		}

		// if(lastPosition == null){
		// lastPosition = currentPosition;
		// }

		if (currentPosition instanceof Area)
			if (currentPosition instanceof Building)
			{
				((Building) currentPosition).getState().setAreaChecked(true);
			} else if (currentPosition instanceof Road)
			{
				((Road) currentPosition).getState().setAreaChecked(true);
				((Road) currentPosition).getState().setPassTime(getTime());
			}
	}

	private void stepWorldCG()
	{
		model.getState().CG.step();

	}

	protected void checkFullBuildings()
	{
		Building bldg;
		int fullTime, time = getTime();
		for (Iterator it = fullBuildings.iterator(); it.hasNext();)
		{
			bldg = (Building) it.next();
			fullTime = fullBuildingsTimeMap.get(bldg);
			if (time - fullTime > MIN_TIME_TO_CLEAR_FULL_BUILDINGS)
				it.remove();
		}

		if (me() instanceof FireBrigade)
		{
			FireBrigade fb = (FireBrigade) me();
			if (!fb.getState().HasMovedInLastCycle(model) && submittedMoveAct)
			{
				if (pathEnd instanceof Building)
				{
					bldg = (Building) pathEnd;
					if (!(bldg instanceof Refuge))
					{
						setBuildingIsFull(bldg, time);
					}
				}
			}
			Human mo;
			for (Iterator it = model.getState().updatedHumans.iterator(); it
					.hasNext();)
			{
				mo = (Human) it.next();
				if (mo instanceof FireBrigade)
				{
					if (mo.getPosition(model) instanceof Building)
					{
						bldg = (Building) mo.getPosition(model);
						if (!(bldg instanceof Refuge))
						{
							setBuildingIsFull(bldg, time);
						}
					}
				}
			}
		}
	}

	protected void setBuildingIsFull(Building bldg, int time)
	{
		fullBuildingsTimeMap.remove(bldg);
		fullBuildingsTimeMap.put(bldg, time);
	}

	// TODO
	// protected void moveToAvailableBldgsIfRequire() {
	// Human requirer = giveWayRequirer;
	// giveWayRequirer = null;
	// StandardEntity pos = location();
	// if(pos instanceof Building) return;
	//		
	//		
	// List<Building> nearBuilding = new ArrayList<Building>();
	// if(me() instanceof FireBrigade)
	// nearBuilding =
	// REACHABLE_C.and(distancePrp.lte(EXTINGUISHABLE_DISTANCE*2))
	// .extract(world.buildings);
	// else
	// nearBuilding = BURNING_C.not().and(REACHABLE_C)
	// .and(distancePrp.lte(EXTINGUISHABLE_DISTANCE*2))
	// .extract(world.buildings);
	// if(!nearBuilding.isEmpty()) move(nearBuilding);
	//		
	// else if(pos instanceof Node){
	// Node meNode = (Node)pos;
	// List<Building> cBuilding =
	// BURNING_C.not().extract(meNode.getConnectedBuildings());
	// if(cBuilding.isEmpty()) moveBackToTheRequirer(requirer);
	// else move(cBuilding);
	// }
	// else if(pos instanceof Road){
	// moveBackToTheRequirer(requirer);
	//			
	// Road meRoad = (Road)pos;
	// int posExtra = me().getPositionExtra();
	// if(me() instanceof AmbulanceTeam){
	// if(NOT_PASSABLE_C.eval(meRoad)){
	// boolean toHead = posExtra < meRoad.getLength()/2;
	// Node nn;
	// if(toHead) nn = (Node)meRoad.getPointHead();
	// else nn = (Node)meRoad.getPointTail();
	// List<Building> cBuilding =
	// BURNING_C.not().extract(nn.getConnectedBuildings());
	// if(cBuilding.isEmpty()) moveBackToTheRequirer(requirer);
	// else move(cBuilding);
	// }
	// else{
	// Node nn = (Node)meRoad.getPointHead();
	// List<Building> cBuilding =
	// BURNING_C.not().extract(nn.getConnectedBuildings());
	// if(!cBuilding.isEmpty())
	// move(cBuilding);
	// nn = (Node)meRoad.getPointTail();
	// cBuilding = BURNING_C.not().extract(nn.getConnectedBuildings());
	// if(!cBuilding.isEmpty())
	// move(cBuilding);
	// moveBackToTheRequirer(requirer);
	// }
	// }
	// else if(me() instanceof FireBrigade) moveBackToTheRequirer(requirer);
	// }
	// }
	//	
	// protected void moveBackToTheRequirer(Human requirer){
	// //TODO
	// StandardEntity mpos = requirer.getPosition(model);
	// List<StandardEntity> neighbor = new
	// ArrayList<StandardEntity>(BURNING_C.not()
	// .and(mpos.distancePrp.lte(EXTINGUISHABLE_DISTANCE*2))
	// .extract(world.buildings));
	// Area farest = (Area)getFarestObject(neighbor);
	//		
	// // if(!SeuConstants.IS_COMPETETION_MODE)
	// // world.logFile.writeString(me()+"   move to=  "
	// // + farest, 77, "required");
	//		
	// if(farest != null) move(farest);
	// }

	public StandardEntity getFarestObject(List<StandardEntity> objects)
	{
		if (objects == null || objects.isEmpty())
		{
			System.err.println("getFarestObject : objects is empty");
			return null;
		}
		StandardEntity result = objects.get(0);
		int maxDist = model.getDistance(me(), result);
		for (StandardEntity ro : objects)
		{
			if (model.getDistance(me(), ro) > maxDist)
			{
				result = ro;
				maxDist = model.getDistance(me(), ro);
			}
		}
		return result;
	}

	protected boolean restIfIsDamaging()
	{
		if (me() instanceof FireBrigade)
			return false;
		if (me() instanceof Human)
		{
			Human tempMe = (Human) me();
			if (tempMe.getDamage() > 50 || tempMe.getHP() < 5000
					&& tempMe.getDamage() > 5)
			{
				restAtRefuge();
				return true;
			}
		}
		return false;
	}

	protected void restAtRefuge()
	{
		if (location() instanceof Refuge)
		{
			rest();
		} else
		{
			moveToRefuge();
		}
	}

	protected void extinguishNearFire()
	{
		if (me() instanceof FireBrigade)
		{
			FireBrigade self = (FireBrigade) me();
			List<Building> nearFires = new ArrayList<Building>();
			nearFires.clear();
			for (Building b : model.getState().burningBuildings)
			{
				if (model.getDistance(me(), b) < extinguishDistance)
					nearFires.add(b);
			}
			if (nearFires.isEmpty() || self.getWater() == 0)
				return;

			Building fire = null;
			int max = 4;
			for (Building b : nearFires)
			{
				if (b.getFieryness() < max)
				{
					max = b.getFieryness();
					fire = b;
				}
			}
			extinguish(fire.getID());
		}
	}

	// protected void tryToGetOutOfBlockades(){
	// // TODO
	// if(me() instanceof AmbulanceTeam){
	// AmbulanceTeam at = (AmbulanceTeam) me();
	// if (!at.getState().getWereLockedByBlockade())
	// return;
	// }
	// else if(me() instanceof FireBrigade){
	// FireBrigade fb = (FireBrigade) me();
	// if (!fb.getState().getWereLockedByBlockade())
	// return;
	// }
	//		
	//		
	//		
	// if(me().nearBlockNodes == null){
	// me().nearBlockNodes = new ArrayList<Node>();
	//			
	// ArrayList<Node> temp = new ArrayList<Node>();
	// if(world.CG.myGraph != null){
	// for(Integer id : world.CG.myGraph.vertices)
	// temp.add((Node)world.get(id));
	// } else {
	// temp.addAll(world.nodes);
	// }
	//			
	// temp = REACHABLE_C.extract(temp);
	// Collections.sort(temp, distanceToSelfComparator);
	// int maxCount = 5;
	// for(int i=0;i<maxCount && i<temp.size();i++){
	// me().nearBlockNodes.add(temp.get(i));
	// }
	// }
	// int size = me().nearBlockNodes.size();
	// if(size == 0) return;
	// int index = getTime() % size;
	//		
	// move(me().nearBlockNodes.get(index));
	//
	// }

	protected void searchNearBuilding()
	{
		List<Building> nearBldg = REACHABLE_C.and(BURNING_C.not()).and(
				BURNED_C.not()).extract(model.getState().notSeenBuildings);

		if (!nearBldg.isEmpty())
			if (nearBldg.size() > 2)
				move(nearBldg.get(getTime() % 4));
			else
			{
				move(nearBldg.get(0));
			}
	}

	protected void updateBuildingSet()
	{
		if (getTime() <= 4)
			return;
		visibleBuildingsIC.clear();
		for (Building b : model.getState().updatedBuildings)
		{
			if (model.getDistance(me(), b) <= seu.utilities.SimulatorsConstants.VISIBLE_DISTANCE)
			{
				visibleBuildingsIC.add(b);
			}
		}

		model.getState().notSeenBuildings.removeAll(visibleBuildingsIC);

		model.getState().visitedBuildings.addAll(visibleBuildingsIC);

		for (Building b : visibleBuildingsIC)
		{
			ArrayList<Human> temp = new ArrayList<Human>(
					model.getState().updatedCivilians);
			temp.addAll(model.getState().buriedAgents);
			boolean hasHum = false;
			for (Human h : temp)
			{
				if (h.getPosition(model) == b)
				{
					hasHum = true;
					continue;
				}
			}
			if (hasHum)
				continue;
			model.getState().emptyBuildings.add(b);
			b.getState().setEmptyTime(getTime());

			b.getState().civPossiblity = 0;
		}

		// update visible buildings
		for (Building bld : model.getState().notSeenBuildings)
		{
			bld.getState().removeVisibleBuildings(visibleBuildingsIC);
			bld.getState().removeHearableBuildings(visibleBuildingsIC);

		}

		ArrayList<Building> hearable = new ArrayList<Building>();
		hearable.clear();
		for (Building b : model.getState().buildings)
		{
			if (model.getDistance(me(), b) <= seu.utilities.SimulatorsConstants.HEARABLE_DISTANCE)
			{
				hearable.add(b);
			}
		}
		for (Building b : hearable)
		{
			if (model.getState().notSeenBuildings.contains(b))
				b.getState().hearableCount++;
			if (b.getState().isBurned())
				b.getState().hearableCount = 0;
		}

	}

	protected void updateForCivilianSearch()
	{

		double p0 = 80.0 / model.getState().buildings.size();

		for (Iterator<Building> it = model.getState().buildings.iterator(); it
				.hasNext();)
		{
			Building b = it.next();
			b.getState().civPossiblity = 0;

			if (b.getState().isBurned() || b.getState().isBurning())
				continue;

			if (!model.getState().notSeenBuildings.contains(b))
				continue;

			if (b.getState().hearableCount < 60)
				b.getState().civPossiblity = p0
						* Math.pow(0.8, b.getState().hearableCount);

			b.getState().civPossiblity += b.getState().getCivPossibility(model);

		}

		// for(Iterator<Node> it = model.getState().nodes.iterator();
		// it.hasNext(); ){
		// Node n = it.next();
		// n.civPossiblity = 0;
		//			
		// for(Building b : n.getVisibleBuildings()){
		// if(!world.notSeenBuildings.contains(b))
		// continue;
		//				
		// n.civPossiblity += b.civPossiblity;
		// }
		//			
		// }

	}

	// ============================= report for messages
	// ==========================================//
	private int reportedBuried = -1;

	private void report(AgentMessageManager agentMessageManager)
	{
		// TODO
		if (getTime() <= config
				.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY))
			return;
		if (agentMessageManager == null)
			return;
		Human me = null;
		if (me() instanceof Human)
		{
			me = (Human) me();
			if (me.isPositionDefined())
			{
				agentMessageManager.reportPosition(me);
				if (me.getBuriedness() > 0)
				{
					reportedBuried = 0;
					agentMessageManager.reportVictim(me);
				} else if (reportedBuried >= 0 && reportedBuried < 2)
				{
					reportedBuried++;
					agentMessageManager.reportVictim(me);
				}
			}
		}

		// =======report whether me is locked by blockade
		if (!(me instanceof PoliceForce) && getTime() > 5)
		{
			if (checkLockedInBlockades(me))
			{
				// System.out.println(me()+"  pos : "+me().motionlessPosition()+"  time : "+time());
				if (modelState.reportLocked.getCount() <= 1)
				{
					modelState.reportUnlocked.reset();
					modelState.setWereLockedByBlockade(true, getTime());
					modelState.reportLocked.update(getTime());
					modelState.report.update(getTime());

					if (!modelState.lockedByBlockadeAgentMap.values().contains(
							me()))
						modelState.lockedByBlockadeAgentMap.put(me.getID()
								.getValue(), me);

					agentMessageManager.reportLockedByBlockade();
				}

			} else
			{
				modelState.lockedByBlockadeAgentMap.remove(me());
				modelState.reportLocked.reset();

				if (modelState.reportUnlocked.getCount() <= 1)
				{
					modelState.reportUnlocked.update(getTime());
					modelState.report.update(getTime());
					agentMessageManager.reportUnlockedByBlockade();
				}
			}
		}
		// TODO
		// if(me instanceof FireBrigade){
		// agentMessageManager.reportRemainedWater();
		// Building bld = ((FireBrigade)me).extinguish ;
		// if(bld != null){
		// agentMessageManager.reportFireWater(bld, 1000);
		// }
		// }

		// === Reporting Fiery Buildings :
		for (Building b : modelState.updatedBuildings)
		{
			if (model.getDistance(b.getID(), me.getID()) > 30000)
				continue;
			if (b.isOnFire())
			{
				agentMessageManager.reportFire(b);
				// TODO
				/*
				 * if(b.tempChanged) agentMessageManager.reportTemperature(b);
				 */
				b.getState().report.update(getTime());
			}
		}

		// === Reporting Civilians
		for (Civilian civ : modelState.foundVictimCivilianIC)
		{
			agentMessageManager.reportFoundCivilian(civ);// must be victim
			civ.getState().report.update(getTime());
		}
		ArrayList<Civilian> victims = new ArrayList<Civilian>(
				modelState.updatedCivilians);
		victims.removeAll(modelState.foundVictimCivilianIC);
		for (Civilian civ : victims)
		{
			if (civ.getPosition(model) instanceof Refuge)
				continue;
			if (civ.isPositionDefined())
			{
				agentMessageManager.reportVictim(civ);
				civ.getState().report.update(getTime());
			}
		}

		// === Reporting Roads (Clears and UnClears and Blocks)
		for (Iterator i = modelState.updatedRoads.iterator(); i.hasNext();)
		{
			Road road = (Road) i.next();
			// TODO
			/*
			 * if(road.getCondition() == Road.CONDITION_PASSABLE){
			 * if(road.reportClear.getCount() > 1) continue;
			 * road.reportClear.update(getTime()); }
			 * 
			 * if(road.getCondition() == Road.CONDITION_BLOCKED){
			 * if(road.reportUnClear.getCount() > 1) continue;
			 * road.reportUnClear.update(getTime()); }
			 * 
			 * agentMessageManager.reportRoad(road, road.getCondition());
			 */
		}

		// TODO
		// report position History Roads
		/*
		 * for(Iterator it = positionHistoryRoadsForReport.iterator();
		 * it.hasNext();){ Road road = (Road)it.next();
		 * if(road.isReportedAsClear()) continue;
		 * road.setIsReportedAsClear(true); agentMessageManager.reportRoad(road,
		 * road.getCondition()); }
		 */

		// TODO
		// === reporting empty building
		/*
		 * for (Iterator iterator =
		 * visibleBuildingsIC.iterator();iterator.hasNext();) { Building
		 * building = (Building) iterator.next(); if
		 * (building.emptyBuildingReport.getCount() > 1 ||
		 * building.isReportedAsEmpty()) continue;
		 * if(!world.emptyBuildings.contains(building)) continue;
		 * 
		 * agentMessageManager.reportEmptyBuilding(building);
		 * building.emptyBuildingReport.update(getTime());
		 * building.report.update(getTime()); }
		 */

		/* ================ report when channel collapsed =================== */

		// TODO
		/*
		 * if(isRadioChannelCollapsed){
		 * 
		 * if (checkLockedInBlockades(me())) { me().locked = true;
		 * agentMessageManager.reportLockedByBlockade(); } else{
		 * if(me().locked){ agentMessageManager.reportUnlockedByBlockade();
		 * me().locked = false; } }
		 * 
		 * for(Civilian civ : world.foundVictimCivilianIC){
		 * if(world.getTime()<civ.getTime()+30){
		 * agentMessageManager.reportFoundCivilian(civ);//must be victim
		 * civ.report.update(getTime()); } } ArrayList<Civilian> victims2 = new
		 * ArrayList<Civilian>(world.updatedCivilians);
		 * victims2.removeAll(world.foundVictimCivilianIC); for(Civilian civ :
		 * victims2){ if(world.getTime()<civ.getTime()+30){
		 * if(civ.getRealPosition() instanceof Refuge) continue;
		 * agentMessageManager.reportVictim(civ); civ.report.update(getTime());
		 * } }
		 * 
		 * // === Reporting Roads (Clears and Blocks) for (Iterator i =
		 * world.roads.iterator(); i.hasNext();) { Road road = (Road) i.next();
		 * if(road.getCondition() == Road.CONDITION_PASSABLE ||
		 * road.getCondition() == Road.CONDITION_BLOCKED) {
		 * if(world.getTime()<road.getTime()+5){
		 * agentMessageManager.reportRoad(road, road.getCondition()); } }
		 * 
		 * }
		 * 
		 * // === reporting empty building for (Iterator iterator =
		 * world.emptyBuildings.iterator();iterator.hasNext();) { Building
		 * building = (Building) iterator.next();
		 * if(world.getTime()<building.getTime()+5){
		 * agentMessageManager.reportEmptyBuilding(building); } }
		 * 
		 * // === reporting fires for (Iterator iterator =
		 * world.burningBuildings.iterator();iterator.hasNext();) { Building
		 * building = (Building) iterator.next();
		 * if(world.getTime()<building.getTime()+10){
		 * agentMessageManager.reportFire(building); } }
		 * 
		 * }
		 */
	}

	/**
	 * @param me
	 * @return By JiangYh 2010-5-10
	 */
	private boolean checkLockedInBlockades(Human me)
	{
		StandardEntity mo = me.getPosition(model);
		if (mo == null || mo instanceof AmbulanceTeam)
		{
			return false;
		}
		if (mo instanceof PoliceForce)
		{
			return false;
		}

		if (!model.getEntitiesOfType(StandardEntityURN.REFUGE).isEmpty()
				&& search.breadthFirstSearch(mo.getID(), getRefugesID()) != null)
		{
			return false;
		}

		ArrayList testDestination = new ArrayList();
		ArrayList<PoliceForce> police = new ArrayList<PoliceForce>();
		for (Iterator it = modelState.policeForces.iterator(); it.hasNext();)
		{
			PoliceForce pf = (PoliceForce) it.next();
			if (pf.isBuriednessDefined() && pf.getBuriedness() == 0)
			{
				police.add(pf);
			}
		}
		for (PoliceForce pf : police)
		{
			if (pf.getPosition(model) == null)
				continue;

			if (search.breadthFirstSearch(me.getID(), Collections.singleton(pf
					.getPosition())) != null)
				return false;
		}

		if (getRefuges().isEmpty())
		{
			testDestination = new ArrayList<StandardEntity>();
			testDestination.addAll(model
					.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE));
			testDestination.addAll(model
					.getEntitiesOfType(StandardEntityURN.FIRE_STATION));
			testDestination.addAll(model
					.getEntitiesOfType(StandardEntityURN.POLICE_OFFICE));

			for (Object o : testDestination)
			{
				if (search.breadthFirstSearch(me.getPosition(), Collections
						.singleton(((StandardEntity) o).getID())) != null)
					return false;
			}

		}

		return true;
	}

	public void updateModel(Collection<Command> heard)
	{
		for (Command next : heard)
		{
			if (next instanceof AKSpeak)
			{
				AKSpeak message = (AKSpeak) next;
				// System.out.println(getTime()+" "+message.getContent().length);
				if (message.getAgentID() == null
						|| message.getContent().length == 0)
					return;
				StandardEntity sender = model.getEntity(message.getAgentID());
				// System.out.println(me()+" : Sender = "+sender+" Message = "+new
				// String(message.getContent()));
				if (sender != me())
				{
					// hasHeard = true;
					if (agentMessageManager != null)
						agentMessageManager.importMessage(message.getContent(),
								getTime(), sender);
				}
			} else if (next instanceof AKSay)
			{
				AKSay message = (AKSay) next;
				if (message.getAgentID() == null)
					return;
				StandardEntity sender = model.getEntity(message.getAgentID());
				if (sender == null)
				{
					sender = StandardEntityFactory.INSTANCE.makeEntity(
							StandardEntityURN.CIVILIAN, message.getAgentID());

					model.addEntity(sender);
					model.getState().add(message.getAgentID());
					sender.getState().setTime(0);

					if (!SPECIAL_BUILDING_C.eval(me()))
					{
						if (me() instanceof AmbulanceTeam)
						{
							AmbulanceTeam at = (AmbulanceTeam) me();
							((Civilian) sender).getState().addHearPosition(
									model, (Area) location(),
									at.getState().getPositionExtra(), true);

							((Civilian) sender).getState()
									.updatePossiblePositions(model);
						} else if (me() instanceof FireBrigade)
						{
							FireBrigade fb = (FireBrigade) me();
							((Civilian) sender).getState().addHearPosition(
									model, (Area) location(),
									fb.getState().getPositionExtra(), true);

							((Civilian) sender).getState()
									.updatePossiblePositions(model);
						} else if (me() instanceof PoliceForce)
						{
							PoliceForce pf = (PoliceForce) me();
							((Civilian) sender).getState().addHearPosition(
									model, (Area) location(),
									pf.getState().getPositionExtra(), true);

							((Civilian) sender).getState()
									.updatePossiblePositions(model);
						}

					}
				}

				else if (sender != me())
				{
					if (sender instanceof Civilian)
					{
						Civilian civ = (Civilian) sender;
						if (civ.getState().hasFound())
							return;

						if (location() instanceof Area)
						{
							Area mo = (Area) location();
							if (me() instanceof AmbulanceTeam)
							{
								AmbulanceTeam at = (AmbulanceTeam) me();
								if (civ.getPosition(model) == null)
								{
									if (!model.getState().heardCivilians
											.contains(civ))
									{
										model.getState().heardCivilians
												.add(civ);
										if (agentMessageManager != null)
										{
											agentMessageManager
													.reportHeardCivilian(
															civ,
															mo,
															at
																	.getState()
																	.getPositionExtra());
										}
									}
								}
								civ.getState().addHearPosition(model, mo,
										at.getState().getPositionExtra(), true);
								civ.getState().updatePossiblePositions(model);
							} else if (me() instanceof FireBrigade)
							{
								FireBrigade fb = (FireBrigade) me();
								if (civ.getPosition(model) == null)
								{
									if (!model.getState().heardCivilians
											.contains(civ))
									{
										model.getState().heardCivilians
												.add(civ);
										if (agentMessageManager != null)
										{
											agentMessageManager
													.reportHeardCivilian(
															civ,
															mo,
															fb
																	.getState()
																	.getPositionExtra());
										}
									}
								}
								civ.getState().addHearPosition(model, mo,
										fb.getState().getPositionExtra(), true);
								civ.getState().updatePossiblePositions(model);
							} else if (me() instanceof PoliceForce)
							{
								PoliceForce pf = (PoliceForce) me();
								if (civ.getPosition(model) == null)
								{
									if (!model.getState().heardCivilians
											.contains(civ))
									{
										model.getState().heardCivilians
												.add(civ);
										if (agentMessageManager != null)
										{
											agentMessageManager
													.reportHeardCivilian(
															civ,
															mo,
															pf
																	.getState()
																	.getPositionExtra());
										}
									}
								}
								civ.getState().addHearPosition(model, mo,
										pf.getState().getPositionExtra(), true);
								civ.getState().updatePossiblePositions(model);
							}
						}
						return;
					} else
					{
						if (sayMessageManager != null)
						{
							if (message.getContent().length == 0)
								return;
							sayMessageManager.importMessage(message
									.getContent(), getTime(), sender);
						}
					}
				}
			}
		}
	}

	// ======================== the action used by agents
	// ========================================//
	protected void clear(EntityID target)
	{
		sendClear(getTime(), target);
		action = true;
		isClear = true;
		// System.out.println(me()+"isClear");
	}

	protected void extinguish(EntityID target)
	{
		sendExtinguish(getTime(), target, extinguishPower);
		action = true;
		isExtinguish = true;
	}

	protected void load(Human target)
	{
		sendLoad(getTime(), target.getID());
		myLoadedVictim = target;
		action = true;
	}

	private StandardEntity pathEnd = null;

	protected void move(StandardEntity target)
	{
		List<EntityID> path = search.breadthFirstSearch(location().getID(),
				target.getID());
		if (path != null)
		{
			movePath.clear();
			movePath.addAll(path);
			pathEnd = model.getEntity(path.get(path.size() - 1));
			sendMove(getTime(), path);
			action = true;
		}
	}

	protected void move(List<EntityID> path)
	{
		if (path != null)
		{
			movePath.clear();
			movePath.addAll(path);
			pathEnd = model.getEntity(path.get(path.size() - 1));
			sendMove(getTime(), path);
			action = true;
		}
	}

	protected void move(List<EntityID> path, int destX, int destY)
	{
		movePath.clear();
		movePath.addAll(path);
		sendMove(getTime(), path, destX, destY);
		action = true;
	}

	protected void rescue(StandardEntity target)
	{
		sendRescue(getTime(), target.getID());
		action = true;
		isRescue = true;
	}

	protected void rest()
	{
		sendRest(getTime());
		action = true;
	}

	protected void unload()
	{
		sendUnload(getTime());
		action = true;
	}

	// =====
	protected void moveToRefuge()
	{
		List<EntityID> path = search.breadthFirstSearch(location().getID(),
				getRefugesID());
		if (path == null)
			return;
		move(path);
	}

	// =============================
	protected boolean commandIgnore(int time)
	{
		return time == config
				.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY);
	}

	// ======================================searchCivilian===============================/

	protected void setSearchPartition()
	{

		if (myPartition != null && !useSpeak)
		{
			return;
		}

		myPartition = null;

		// System.out.println(" search partitions : "+searchPartitions.size());

		if (searchPartitions == null || searchPartitions.isEmpty())
		{
			// System.out.println(" search finished  in : "+getTime());
			return;
		}

		for (Iterator it = searchPartitions.iterator(); it.hasNext();)
		{
			List<Building> partiBlds = ((RadialPartition) it.next())
					.getContainedBuildings();
			partiBlds.removeAll(model.getState().emptyBuildings);

			if (partiBlds == null || partiBlds.isEmpty())
			{
				it.remove();
			}
		}
		int searchSize = searchPartitions.size();

		for (int i = 0; i < searchSize; i++)
		{
			if (myExploreTeams.indexOf(me()) == i)
			{
				myPartition = searchPartitions.get(i);
				break;
			}
		}
		if (myPartition == null && searchSize > 0)
		{
			myPartition = searchPartitions.get(random.nextInt(searchSize));
		}
	}

	private Civilian m_lastHearedCivTarget = null;

	protected void gotoPossibleCivBuildings(boolean onlyInPartition)
	{
		List<Civilian> heared = new ArrayList<Civilian>();
		for (Civilian civ : model.getState().civilians)
		{
			if (civ.getPosition(model) == null)
				heared.add(civ);
		}
		if (heared.isEmpty())
			return;
		if (m_lastHearedCivTarget != null
				&& heared.contains(m_lastHearedCivTarget))
		{
			moveHearedCivilian(m_lastHearedCivTarget);

		}
		m_lastHearedCivTarget = null;

		Civilian civ, tmp = null;
		List res;
		do
		{
			tmp = null;
			for (Iterator it = heared.iterator(); it.hasNext();)
			{
				civ = (Civilian) it.next();
				if (civ.getState().getMostPossibleBuilding() == null)
				{
					it.remove();
					continue;
				}

				if (model.getState().getNearestObjectInRange(
						civ.getState().getMostPossibleBuilding(),
						myExploreTeams) == me())
				{
					tmp = civ;
					break;
				}
			}
			if (tmp == null)
				return;
			heared.remove(tmp);
			res = correctListToCheck(tmp.getState().getPossiblePositions());
		} while (res.isEmpty());
		m_lastHearedCivTarget = tmp;
		moveHearedCivilian(m_lastHearedCivTarget);

	}

	private Building m_CivBldg = null;

	private void moveHearedCivilian(Civilian civ)
	{
		if (civ.getPosition(model) != null)
		{
			throw new Error("WTF");
		}

		if (m_CivBldg != null)
			checkShouldContinueAndMove(m_CivBldg, true);

		List<Building> possible = m_lastHearedCivTarget.getState()
				.getPossiblePositions();
		Collections.sort(possible, allocationProbComparator);

		for (Building bldg : possible)
		{
			if (bldg == null)
				continue;
			if (bldg.getState().isBurning() || bldg.getState().isBurned()
					|| !model.getState().notSeenBuildings.contains(bldg)
					|| location() == bldg || model.getState().isInTraffic(bldg)
					|| !REACHABLE_C.eval(bldg))
				continue;

			m_CivBldg = bldg;
			break;
		}
		move(m_CivBldg);
	}

	protected List<Building> getSearchBuildings(RadialPartition partition)
	{
		if (partition == null)
			return null;

		if (!useSpeak)
		{
			ArrayList<Building> tmpBuildings = new ArrayList<Building>();
			tmpBuildings.addAll(partition.getContainedBuildings());
			tmpBuildings.removeAll(model.getState().emptyBuildings);

			return tmpBuildings;
		}

		List<Building> result = new ArrayList<Building>();
		List<Building> best = new ArrayList<Building>();
		if (me() instanceof PoliceForce)
		{
			ArrayList<Building> tmpBuildings = new ArrayList<Building>();
			tmpBuildings.addAll(partition.getContainedBuildings());
			tmpBuildings.removeAll(model.getState().emptyBuildings);

			return tmpBuildings;
		}

		for (Iterator<Building> it = model.getState().notSeenBuildings
				.iterator(); it.hasNext();)
		{
			Building b = it.next();
			if (!REACHABLE_C.eval(b))
				continue;

			if (partition.isInPartition(b))
				best.add(b);

			result.add(b);

		}

		if (!best.isEmpty())
			return best;
		return result;

	}

	protected void checkShouldContinueAndMove(Area mlo, boolean mustBeNew)
	{
		if (mlo == null)
		{
			return;
		}
		if (location() == mlo)
		{
			return;
		}
		if (mlo instanceof Building)
		{
			Building bldg = (Building) mlo;
			if (bldg.getState().isBurning() || bldg.getState().isBurned())
			{
				return;
			}

			if (mustBeNew && !model.getState().notSeenBuildings.contains(bldg))
			{
				return;
			}

			if (FULL_BUILDING_C.eval(bldg))
				return;

			try
			{
				List<Human> vAgs = new ArrayList<Human>();
				for (AmbulanceTeam at : model.getState().ambulanceTeams)
				{
					if (at.getPosition(model) == bldg)
					{
						vAgs.add(at);
					}
				}
				if (Util.getConditionCount(vAgs, AMBULANCE_TEAM_C) > 0
						&& useSpeak)
					return;

				for (Human hum : model.getState().platoonAgents)
				{
					if (hum.getPosition(model) == bldg)
					{
						vAgs.add(hum);
					}
				}
				if (vAgs.size() > 0 && useSpeak)
					return;

			} catch (Exception e)
			{
				System.out.println(me() + "  " + bldg + "");
				e.printStackTrace();
			}
		}

		if (!(me() instanceof PoliceForce)
				&& (mlo.getState().isVisible(model) || !REACHABLE_C.eval(mlo)))
			return;

		if (me() instanceof AmbulanceTeam)
		{
			AmbulanceTeam at = (AmbulanceTeam) me();
			if (at.getState().getStuckCycles() > 2)
				return;
		} else if (me() instanceof FireBrigade)
		{
			FireBrigade fb = (FireBrigade) me();
			if (fb.getState().getStuckCycles() > 2)
				return;
		} else if (me() instanceof PoliceForce)
		{
			PoliceForce pf = (PoliceForce) me();
			if (pf.getState().getStuckCycles() > 2)
				return;
		}

		// System.out.println(me()+"  check and move ... ");

		move(mlo);
	}

	private final int QUAD_PARTITION = 0;
	private final int AMORPH_PARTITION = 1;

	private Area m_searchTarget = null;
	private List<Building> m_partiBuildings = new ArrayList<Building>();

	protected void partitionSearch(int type)
	{

		if (m_searchTarget != null
				&& (model.getState().notSeenBuildings.contains(m_searchTarget)))
			checkShouldContinueAndMove(m_searchTarget, true);

		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("#######################      "+getTime(),
		// 13, "search");

		switch (type)
		{
		case QUAD_PARTITION:
			;
		case AMORPH_PARTITION:

			Area lastTarget = m_searchTarget;

			m_searchTarget = null;

			if (searchPartitions == null || searchPartitions.isEmpty()
					|| model.getState().notSeenBuildings.isEmpty())
			{
				// System.err.println(me()+"[ "+myExploreTeams.indexOf(me())+" ]    search Partitions finished    in : "+getTime());
				return;
			}
			if (myPartition == null)
			{
				// System.err.println(me()+"[ "+myExploreTeams.indexOf(me())+" ]  my partition is null, reselecting ...  in :  "+getTime());
				setSearchPartition();
			}

			if (myPartition == null)
			{
				m_searchTarget = null;
				// System.out.println(me()+"[ "+myExploreTeams.indexOf(me())+" ]   finished search  in :  "+getTime());
				m_partiBuildings = new ArrayList<Building>();
				return;
			}
			// else {
			// m_partiBuildings = getSearchBuildings(myPartition);
			//					
			// }

			List tmp = new ArrayList(myPartition.getContainedBuildings());
			tmp.removeAll(model.getState().emptyBuildings);

			if (tmp == null || tmp.isEmpty())
			{
				// if(!SeuConstants.IS_COMPETETION_MODE)
				// world.logFile.writeString(" remove partition ", 13,
				// "search");

				searchPartitions.remove(myPartition);// no need to search, so
				// remove
				myPartition = null;
				m_searchTarget = null;
				// System.err.println(me()+"[ "+myExploreTeams.indexOf(me())+" ] reselected in :  "+getTime());
				// return;
			}

			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString(getTime()+"  part= "+myPartition.toString(),
			// 13, "search");

			// long t = System.currentTimeMillis();

			// List partiBlds = correctListToCheck(m_partiBuildings);

			// if(partiBlds == null || partiBlds.isEmpty()){
			// //
			// System.err.println(me()+"[ "+myExploreTeams.indexOf(me())+" ]   partiBlds == null || partiBlds.isEmpty() in :  "+getTime());
			//					
			// // if(!SeuConstants.IS_COMPETETION_MODE)
			// //
			// world.logFile.writeString(" partiBlds == null || partiBlds.isEmpty() ",
			// 13, "search");
			//					
			// return;
			// }

			// System.err.println(me()+" get searching target ... ");

			// m_searchTarget = getSortedByProbability(partiBlds);

			m_searchTarget = getSearchTargets(myPartition, lastTarget);

			if (m_searchTarget != null)
			{
				move(m_searchTarget);
			}

			break;

		default:
			System.err.println("   ****  no such partition type!  ***");
			break;
		}
	}

	protected void searchCivilians(boolean onlyInPartition)
	{

		if (getTime() < agentIgnoreTime)
			return;

		gotoPossibleCivBuildings(onlyInPartition);
		if (action)
			return;

		partitionSearch(AMORPH_PARTITION);
		if (action)
			return;
		// simpleSearch(onlyInPartition);

		searchBurntCivilians();
		if (action)
			return;

		move(randomWalk());
		if (action)
			return;
	}

	protected Area getSearchTargets(RadialPartition partition, Area exception)
	{
		Area result = null;

		if (partition == null)
			return null;

		double totalValue = 0;

		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("##########################         "+getTime(),
		// 10, "searchTargets");

		ArrayList<Area> targets = new ArrayList<Area>();
		for (Iterator<Building> it = model.getState().buildings.iterator(); it
				.hasNext();)
		{
			Building b = it.next();

			b.getState().civTotalPossiblity = 0;

			if (partition != null && !partition.isInPartition(b))
				continue;

			if (b.getState().isBurned() || b.getState().isBurning())
				continue;

			if (!model.getState().notSeenBuildings.contains(b))
				continue;

			if (b.getState().civPossiblity > 0.000001)
			{
				// if(!SeuConstants.IS_COMPETETION_MODE)
				// world.logFile.writeString(b+"  value = "+b.civPossiblity, 10,
				// "searchTargets");

				for (Building bld : b.getState().getVisibleBuildings())
				{
					if (!model.getState().notSeenBuildings.contains(bld))
						continue;
					b.getState().civTotalPossiblity += bld.getState().civPossiblity;
				}

				if (me() instanceof PoliceForce || !isReachable(b.getID()))
				{
					totalValue += (b.getState().civPossiblity > 0.04 ? b
							.getState().civPossiblity : 0.04);
					targets.add(b);
				}
			}
		}

		for (Iterator<Road> it = model.getState().roads.iterator(); it
				.hasNext();)
		{
			Road n = it.next();

			n.getState().civTotalPossiblity = n.getState().civPossiblity;

			if (partition != null && !partition.isInPartition(n))
				continue;

			if (n.getState().civPossiblity > 0.000001)
			{
				// if(!SeuConstants.IS_COMPETETION_MODE)
				// world.logFile.writeString(n+"  value = "+n.civPossiblity, 10,
				// "searchTargets");

				if (me() instanceof PoliceForce || REACHABLE_C.eval(n))
				{
					totalValue += (n.getState().civPossiblity > 0.04 ? n
							.getState().civPossiblity : 0.04);
					targets.add(n);
				}

			}
		}

		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("total Value =  "+totalValue, 10,
		// "searchTargets");

		if (targets.isEmpty())
		{
			// System.out.println(" empty targets ... ");

			for (Building b : model.getState().notSeenBuildings)
			{

				if (b.getState().isBurning() || b.getState().isBurned())
					continue;

				if (exception == b)
					continue;

				if (me() instanceof PoliceForce || REACHABLE_C.eval(b))
					targets.add(b);

			}

			totalValue = 0.04 * targets.size();
		}

		double rnd = Math.random();
		double p = 0;
		for (int i = 0; i < targets.size(); i++)
		{
			Area tar = targets.get(i);
			if (tar == location())
				continue;

			if (exception == tar)
				continue;

			if (tar instanceof Building)
			{
				Building b = (Building) tar;
				p += (b.getState().civPossiblity > 0.04 ? b.getState().civPossiblity
						: 0.04)
						/ totalValue;
				if (p >= rnd)
				{
					result = tar;
					break;
				}
			} else if (tar instanceof Road)
			{
				Road r = (Road) tar;
				p += (r.getState().civPossiblity > 0.04 ? r.getState().civPossiblity
						: 0.04)
						/ totalValue;
				if (p >= rnd)
				{
					result = tar;
					break;
				}
			}
		}

		if ((result == null || result == location()) && useSpeak)
			result = model.getState().notSeenBuildings.get(random.nextInt(model
					.getState().notSeenBuildings.size()));

		if (result == null)
			return null;

		// System.out.println(" go to : "+result);

		List<EntityID> route = search.breadthFirstSearch(location().getID(),
				Collections.singleton(result.getID()));

		// if(route == null)
		// System.out.println(" NULL route  to : "+result);

		int i = 0;
		double maxValue = -1000;
		for (EntityID eId : route)
		{
			Area mo = model.getEntity(eId) instanceof Area ? (Area) model
					.getEntity(eId) : null;
			if (mo == null)
				continue;
			if (!(mo instanceof Road))
				continue;

			if (exception == mo)
				continue;

			if (mo instanceof Building)
			{
				Building b = (Building) mo;
				if (b.getState().civPossiblity < 0.00001)
					continue;
			} else if (mo instanceof Road)
			{
				Road r = (Road) mo;
				if (r.getState().civPossiblity < 0.00001)
					continue;
			}

			double v = ((Road) mo).getState().civPossiblity / (i + 1.0);

			if (v > maxValue)
			{
				maxValue = v;
				result = mo;
			}

			for (Building b : ((Road) mo).getState().getConnectedBuildings(
					model))
			{
				double tmp = 0;

				for (Building bld : b.getState().getHearableBuildings())
				{
					if (!model.getState().notSeenBuildings.contains(bld))
						continue;
					if (exception == bld)
						continue;

					tmp += bld.getState().civPossiblity;
				}

				v = tmp / (i + 1.5);

				if (v > maxValue)
				{
					maxValue = v;
					result = b;
				}
			}

			i += 2;
		}

		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("finally, go to:  "+result, 10,
		// "searchTargets");

		return result;

	}

	private Building m_burntCivilianBuilding = null;

	protected void searchBurntCivilians()
	{
		if (m_burntCivilianBuilding != null
				&& !m_burntCivilianBuilding.getState().isVisible(model))
			moveToBurntBuilding(m_burntCivilianBuilding);

		m_burntCivilianBuilding = null;

		List<Building> burntBuildings = new ArrayList<Building>(model
				.getState().buildings);
		burntBuildings.removeAll(model.getState().visitedBuildings);

		for (Building b : burntBuildings)
		{
			if (b.getState().isVisible(model) || b.getState().isBurning())
				continue;
			if (!isReachable(b.getID()))
				continue;

			if (b.getState().isBurned())
			{
				m_burntCivilianBuilding = b;
				break;
			}

		}

		if (m_burntCivilianBuilding != null)
			move(m_burntCivilianBuilding);
		else
		{
			if (burntBuildings.isEmpty())
				return;
			List<Building> walkBuildings = new ArrayList<Building>(
					burntBuildings);
			// System.out.println("b = "+burntBuildings.size());
			for (Building b : burntBuildings)
			{
				if (b.getState().isVisible(model) || b.getState().isBurning())
				{
					walkBuildings.remove(b);
					continue;
				}

				if (!isReachable(b.getID()))
				{
					walkBuildings.remove(b);
					continue;
				}

				if (model.getDistance(location(), b) < MAX_HEARING_DISTANCE
						&& Math.random() < 0.5)
				{
					move(b);
					return;
				}
			}

			Collections.sort(walkBuildings, distanceToMeComparator);
			if (burntBuildings.size() > 7)
			{
				move(burntBuildings.get(getTime() % 8));
			} else if (burntBuildings.size() > 3)
			{
				move(burntBuildings.get(getTime() % 4));
			} else if (!burntBuildings.isEmpty())
			{
				move(burntBuildings.get(0));
			}
		}
	}

	protected void moveToBurntBuilding(Building target)
	{

		if (FULL_BUILDING_C.eval(target) || target.getState().isBurning()
				|| target.getState().isVisible(model))
			return;

		try
		{
			List<Human> vAgs = new ArrayList<Human>();
			for (Human hum : model.getState().platoonAgents)
			{
				if (hum.getPosition(model) == target)
				{
					vAgs.add(hum);
				}
			}

			if (vAgs.size() > 0)
				return;

		} catch (Exception e)
		{
			e.printStackTrace();
		}

		if (!(me() instanceof PoliceForce)
				&& (target.getState().isVisible(model) || !REACHABLE_C
						.eval(target)))
			return;

		move(target);
	}

	protected List correctListToCheck(Collection bldgs)
	{
		Building tmp;
		if (bldgs == null || bldgs.isEmpty())
			return null;

		List res = new ArrayList();
		for (Iterator it = bldgs.iterator(); it.hasNext();)
		{
			Object obj = it.next();
			if (obj instanceof Building)
			{
				tmp = (Building) obj;
				if (tmp.getState().isBurning() || tmp.getState().isBurned()
						|| !model.getState().notSeenBuildings.contains(tmp)
						|| FULL_BUILDING_C.eval(tmp)
				// || (!iPolice && !REACHABLE_C.eval(tmp))
				)
				{
					continue;
				}
				res.add(tmp);
			}
		}
		return res;
	}

	// public final Condition IN_TRAFFIC_C = new Condition() {
	// final Set trafficSet = model.getState().inTraffic;
	// public boolean eval(Object obj) {
	// return trafficSet.contains(obj);
	// }
	// };
	//	
	// public final Condition NOT_IN_TRAFFIC_C = IN_TRAFFIC_C.not();

	protected void doSomethingIfWereStuckByHumanoids()
	{
		Area pos = (Area) location();
		if (pos instanceof Road && NOT_PASSABLE_C.eval(pos))
			return;
		int time = getTime();
		if (time <= agentIgnoreTime + 1)
			return;

		if (me() instanceof AmbulanceTeam)
		{
			AmbulanceTeam at = (AmbulanceTeam) me();
			if (at.getState().getStuckCycles() > 0)
			{
				if (pos instanceof Building)
					return;
				model.getState().computeTraffic();
			}
		} else if (me() instanceof FireBrigade)
		{
			FireBrigade fb = (FireBrigade) me();
			if (fb.getState().getStuckCycles() > 0)
			{
				if (pos instanceof Building)
					return;
				model.getState().computeTraffic();

				if (me() instanceof PoliceForce)
				{
					// System.err.println(me()+"   send give way!! ");
					// world.logFile.writeString(me()+"  say give way", 78,
					// "giveway");
					sayMessageManager.reportGiveWay();
				}
			}
		} else if (me() instanceof PoliceForce)
		{
			PoliceForce pf = (PoliceForce) me();
			if (pf.getState().getStuckCycles() > 0)
			{
				if (pos instanceof Building)
					return;
				model.getState().computeTraffic();

				sayMessageManager.reportGiveWay();
			}
		}

	}

	// ======================================temp use
	// config===============================/

	public final static int MAX_HEARING_DISTANCE = 30000;
	public static final int HUMANOID_SEARCH_RADIUS = 30000;
	public static final int MIN_TIME_TO_CLEAR_FULL_BUILDINGS = 3;
	public static final int MIN_TIME_INTEVAL_TO_CHECK_CIV = 10;

	// =======================================temp use for check
	// reachable=======================//
	private Map<EntityID, Set<EntityID>> graph;

	// public boolean isReachable(EntityID goal){
	// if(location().getID() == goal)return true;
	// List<EntityID> open = new LinkedList<EntityID>();
	// Map<EntityID, EntityID> ancestors = new HashMap<EntityID, EntityID>();
	// open.add(location().getID());
	// EntityID next = null;
	// boolean found = false;
	// ancestors.put(location().getID(), location().getID());
	// do {
	// next = open.remove(0);
	// if (next == goal) {
	// found = true;
	// break;
	// }
	// // System.out.println(location());
	// // System.out.println(next);
	// // if(ture){
	// // System.out.println(graph.get(next));
	// // return false;
	// // }
	// Collection<EntityID> neighbours = graph.get(next);
	// if (neighbours.isEmpty()) {
	// continue;
	// }
	// for (EntityID neighbour : neighbours) {
	// if (neighbour == goal) {
	// ancestors.put(neighbour, next);
	// next = neighbour;
	// found = true;
	// break;
	// }
	// else {
	// if (!ancestors.containsKey(neighbour)) {
	// open.add(neighbour);
	// ancestors.put(neighbour, next);
	// }
	// }
	// }
	// } while (!found && !open.isEmpty());
	// if (!found) {
	// // No path
	// return false;
	// }
	// return true;
	// // // Walk back from goal to start
	// // EntityID current = next;
	// // List<EntityID> path = new LinkedList<EntityID>();
	// // do {
	// // path.add(0, current);
	// // current = ancestors.get(current);
	// // if (current == null) {
	// // throw new
	// RuntimeException("Found a node with no ancestor! Something is broken.");
	// // }
	// // } while (current != start);
	// // if(path.)
	// // return path;
	// }

	// public void updateMyGraph(StandardWorldModel world){
	// Map<EntityID, Set<EntityID>> neighbours = new LazyMap<EntityID,
	// Set<EntityID>>() {
	// @Override
	// public Set<EntityID> createValue() {
	// return new HashSet<EntityID>();
	// }
	// };
	// List<Road> graphRoads = new
	// ArrayList<Road>(world.getState().passableRoads);
	// graphRoads.addAll(world.getState().unKnownRoads);
	// if(location() instanceof Road)
	// graphRoads.add((Road)location());
	// for(Road next : graphRoads){
	// Collection<EntityID> areaNeighbours = ((Area)next).getNeighbours();
	// neighbours.get(next.getID()).addAll(areaNeighbours);
	// }
	// graph = neighbours;
	// }

	Comparator distanceToMeComparator = new Comparator()
	{
		public int compare(Object obj1, Object obj2)
		{
			Building civ1 = (Building) obj1;
			Building civ2 = (Building) obj2;
			if (model.getDistance(me(), civ1) > model.getDistance(me(), civ2))
				return 1;
			else if (model.getDistance(me(), civ1) == model.getDistance(me(),
					civ2))
				return 0;
			else
				return -1;
		}
	};

	// ====

	public void checkAround(ChangeSet changed)
	{
		for (EntityID next : changed.getChangedEntities())
		{
			// update Roads
			if (model.getEntity(next) instanceof Road)
			{
				Road r = (Road) model.getEntity(next);

				if (!passedRoads.contains(r))
					passedRoads.add(r);

				if (unKnownRoads.contains(r))
					unKnownRoads.remove(r);

				if (r.isBlockadesDefined() && r.getBlockades().isEmpty())
				{
					if (!passableRoads.contains(r))
						passableRoads.add(r);
					if (unpassableRoads.contains(r))
						unpassableRoads.remove(r);
				} else
				{
					if (passableRoads.contains(r))
						passableRoads.remove(r);
					if (!unpassableRoads.contains(r))
						unpassableRoads.add(r);
				}
			}

			// update Buildings
			if (model.getEntity(next) instanceof Building)
			{
				Building b = (Building) model.getEntity(next);

				if (!visitedBuildings.contains(b))
					visitedBuildings.add(b);

				if (b.isOnFire())
				{
					if (safeBuildings.contains(b))
						safeBuildings.remove(b);
					if (!dangerBuildings.contains(b))
						dangerBuildings.add(b);
				} else
				{
					if (!safeBuildings.contains(b))
						safeBuildings.add(b);
					if (dangerBuildings.contains(b))
						dangerBuildings.remove(b);
				}
			}

		}
	}

	public boolean isReachable(EntityID goals)
	{
		List<EntityID> passPath = search.breadthFirstSearch(location().getID(),
				goals);
		if (passPath == null)
			return false;

		for (EntityID next : passPath)
		{
			StandardEntity en = model.getEntity(next);
			if (unpassableRoads.contains(en))
				return false;
		}
		return true;
	}

	List<Road> passedRoads = new ArrayList<Road>();
	List<Road> passableRoads = new ArrayList<Road>();
	List<Road> unpassableRoads = new ArrayList<Road>();
	List<Road> unKnownRoads = new ArrayList<Road>();
	List<Building> unVisitedBuildings = new ArrayList<Building>();
	List<Building> visitedBuildings = new ArrayList<Building>();
	List<Building> safeBuildings = new ArrayList<Building>();
	List<Building> dangerBuildings = new ArrayList<Building>();
	List<Refuge> unKnownRefuges = new ArrayList<Refuge>();
	List<Refuge> usefulRefuges = new ArrayList<Refuge>();
}
