//SEU2010

package seu.agent;

import static rescuecore2.misc.Handy.objectsToIDs;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import rescuecore2.log.Logger;
import rescuecore2.messages.Command;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import seu.entities.BlockedPath;
import seu.entities.Path;
import seu.messages.AgentMessageManager;
import seu.messages.SayMessageManager;
import seu.states.PoliceForceState;
import seu.states.RoadState;
import seu.utilities.TrafficEvaluation;
import seu.utilities.Partition.RadialPartition;

/**
 * A sample police force agent.
 */
public class PoliceForceAgent extends PlatoonAgent<PoliceForce> {

	private PoliceForceState myState;
	
	private TrafficEvaluation tfEvaluation = null;
	
	public List<Human> needHelpAgents = null;

	@Override
	public String toString() {
		return "Seu police force";
	}

	public PoliceForceState getState() {
		return myState;
	}

	
//	List<Refuge> usefulRefuges = new ArrayList<Refuge> ();
	@Override
	protected void postConnect() {
		super.postConnect();
		model.indexClass(StandardEntityURN.ROAD);
	}

	@Override
	protected void think(int time, ChangeSet changed, Collection<Command> heard) {
		
		prepareForAct(time, changed, heard);

		act(time);
	}
	
	public void prepareForAct(int time, ChangeSet changed,
			Collection<Command> heard) {
		if (time < agentIgnoreTime)
			return;

		if (time == agentIgnoreTime) {
			try{
				initialChannels(time);
			}catch (Exception e) {
				System.out.println("initialChannel failed");
				e.printStackTrace();
			}
		}
//		System.out.println(getTime()+" pf "+me()+" 1");
		// update world message
		
		try{
			updateModel(heard);
		}catch (Exception e) {
			System.out.println(me()+"updateModel failed");
			e.printStackTrace();
		}
		
//		System.out.println(getTime()+" pf "+me()+" 2");
		super.prepareForAct(time, changed);
//		System.out.println(getTime()+" pf "+me()+" 3");
		modelState = model.getState();
		myState = me().getState();
		
//		setClearPartition();
//		System.out.println(getTime()+" pf "+me()+" 4");
		
		try{
			if (agentMessageManager != null) {
				agentMessageManager.sendMessages();
				if (agentMessageManager.getMessageBytes() != null) {
//					System.out.println("at"+agentMessageManager.getMessageBytes().length);
					sendSpeak(time, speakChannel, agentMessageManager.getMessageBytes());
					// System.out.println("at send =  "+(agentMessageManager.getMessageBytes()));
				}
			}
		}catch (Exception e) {
			System.out.println(me()+" sendMessage failed");
			e.printStackTrace();
		}	
//		System.out.println(getTime()+" pf "+me()+" 5");
	}
		
	public void act(int time) {
		if (time < agentIgnoreTime)
			return;

		super.act();
//		System.out.println(getTime()+" pf "+me()+" 6");
//		System.out.println(getTime()+" "+modelState.needHelpAgents.size()+" "+modelState.platoonAgents );
//		System.out.println(me()+" "+getTime()+" 1");
//		System.out.println(me()+" "+getTime()+" 1");
//		System.out.println(getTime()+" pf "+me()+" 7");
		clearHere();
		if (action)return;
//		System.out.println(me()+" "+getTime()+" 2");
//		System.out.println(getTime()+" pf "+me()+" 8");
//		clearAssgined();
//		if (action)return;
//		System.out.println(me()+" "+getTime()+" 3");
		
		clearForAgent();
		if (action)return;
		
		clearToRefuge();
		if (action)return;
//		System.out.println(me()+" "+getTime()+" 4");
//		System.out.println(getTime()+" pf "+me()+" 9");
		clearForCivilian();
		if (action)return;
//		System.out.println(me()+" "+getTime()+" 5");
//		System.out.println(getTime()+" pf "+me()+" 10");
		
//		selectTarget();
//		moveForClearing();
//		if (action)return;
//		System.out.println(getTime()+" pf "+me()+" 11");
//		System.out.println(me()+" "+getTime()+" "+usefulRefuges);
//		System.out.println(me()+" "+getTime()+" "+location());
//		clearInRange();
//		if (action)return;
//		List<EntityID> path1 = search.breadthFirstSearch(me().getPosition(), getRefugesID());
//        if (path1 != null) {
////            Logger.info("Moving to target");
////            Road r = (Road)model.getEntity(path1.get(path1.size() - 1));
////            Blockade b = getTargetBlockade(r, -1);
//            move(path1);
////            Logger.debug("Path: " + path);
////            Logger.debug("Target coordinates: " + b.getX() + ", " + b.getY());
//            return;
//        }
//		
//		List<EntityID> path2 = search.breadthFirstSearch(me().getPosition(), getBlockedRoadsID());
//        if (path2 != null) {
////            Logger.info("Moving to target");
//            Road r = (Road)model.getEntity(path2.get(path2.size() - 1));
//            Blockade b = getTargetBlockade(r, -1);
//            sendMove(time, path2, b.getX(), b.getY());
////            Logger.debug("Path: " + path);
////            Logger.debug("Target coordinates: " + b.getX() + ", " + b.getY());
//            return;
//        }
//        Logger.debug("Couldn't plan a path to a blocked road");
//        Logger.info("Moving randomly");
//        sendMove(time, randomWalk());
   
		
//		searchCivilians(true);
//		if (action)return;
//		System.out.println(getTime()+" pf "+me()+" 12");
		
		pfSearch();
		if (action)return;
		
		System.out.println("random start");
		
		move(randomWalk());
//		System.out.println(me()+" walk+ "+location());
		if (action)return;
		
		restAtRefuge();
		if (action)return;
	}

	public void clearToRefuge(){
		
		if(unKnownRefuges.isEmpty())return;

		int index = modelState.policeForces.indexOf(me());
		if(usefulRefuges.size() < unKnownRefuges.size()){
			List<EntityID> path = search.breadthFirstSearch(location().getID(),
					objectsToIDs(unKnownRefuges));
			move(path);
			return;
		}
		
		for(Refuge rf : unKnownRefuges){
			if(model.getDistance(location(), rf) < Constants.SEE_DISTANCE * 10){
				move(rf);
				return;
			}
		}
		
	}
	
	public void clearForAgent(){
		if(getTime()%25 == 0){
			modelState.needHelpAgents.clear();
			modelState.needHelpAgents.addAll(modelState.ambulanceTeams);
			modelState.needHelpAgents.addAll(modelState.fireBrigades);
		}
		
//		if(modelState.needHelpAgents.contains(me())){
//			modelState.needHelpAgents.remove(me());
//		}
		
		for(Human agent : modelState.platoonAgents){
			if(location() == agent.getPosition(model)){
				if(location() instanceof Road){
					Road ro = (Road)location();
					if(ro.getBlockades().isEmpty()){
						modelState.needHelpAgents.remove(agent);
						break;
					}
				}
			}
			
			if(agent.isPositionDefined() && model.getDistance(location(), agent.getPosition(model)) < MAX_HEARING_DISTANCE 
					&& !isReachable(agent.getPosition())){
				if(agent.getPosition(model) != null){
//					if(location() == agent.getPosition(model)){
//						if(location() instanceof Road){
//							if(((Road)location()).getBlockades().isEmpty()){
//								agent.setPosition(null);
//								return;
//							}
//						}
//					}
					move(agent.getPosition(model));
//					System.out.println(me()+" "+getTime()+" for agent "+agent);
					break;
				}
			}
		}
	}
	
	public void clearForCivilian(){
		for(Civilian civ : modelState.civilians){
			if(model.getDistance(location(), civ.getPosition(model)) < seu.utilities.SimulatorsConstants.VISIBLE_DISTANCE 
					&& !isReachable(civ.getPosition()) && civ.getPosition(model) instanceof Building){
				move(civ.getPosition(model));
				System.out.println("5.1");
				break;
			}
		}
	}
	
	public void clearAssgined(){
		if(me().getState().getCenterAssignedTarget() != null ){
			System.out.println(me()+" : "+me().getState().getCenterAssignedTarget());
//			System.out.println("3.1");
			if(me().getState().getCenterAssignedTarget() instanceof Road){
				Road ro =(Road) me().getState().getCenterAssignedTarget();
				if(!ro.getState().getAreaChecked()){
					move(me().getState().getCenterAssignedTarget());
				}
			}
			else if(me().getState().getCenterAssignedTarget() instanceof Building){
				Building b =(Building) me().getState().getCenterAssignedTarget();
				if(!b.getState().getAreaChecked()){
					move(me().getState().getCenterAssignedTarget());
				}
			}
		}
	}
	
//		int lowLvAct = -1;
//		lowLvAct = this.reflect(time, heard);
//		this.reasoning(time, heard);
//		this.intend(lowLvAct, time);
		
//		if (time == config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY)) {
//            // Subscribe to channel 1
//            sendSubscribe(time, 1);
//        }
//        for (Command next : heard) {
//            Logger.debug("Heard " + next);
//        }
//		if(time < agentIgnoreTime) return;
//		
//		if (time == config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY)) {
//    		initialChannels(time);
//        }
//    	
//    	//update world message
//    	updateModel(heard);
//    	
//    	super.prepareForAct(time);
//    	
//    	modelState = model.getState();
//    	myState = me().getState();
//    	
//    	if (agentMessageManager != null){
//			agentMessageManager.sendMessages();
//			if(agentMessageManager.getMessageBytes() != null){
//				sendSpeak(time, 2 ,agentMessageManager.getMessageBytes());
////				System.out.println("pf send =  "+(agentMessageManager.getMessageBytes()));
//			}
//		}
//		
//        // Am I near a blockade?
//        Blockade target = getTargetBlockade();
//        if (target != null) {
//            Logger.info("Clearing blockade " + target);
////            sendSpeak(time, 1, ("Clearing " + target).getBytes());
//            sendClear(time, target.getID());
//            return;
//        }
//        // Plan a path to a blocked area
//        List<EntityID> path = search.breadthFirstSearch(me().getPosition(model), getBlockedRoads());
//        if (path != null) {
//            Logger.info("Moving to target");
//            Road r = (Road)model.getEntity(path.get(path.size() - 1));
//            Blockade b = getTargetBlockade(r, -1);
//            sendMove(time, path, b.getX(), b.getY());
//            Logger.debug("Path: " + path);
//            Logger.debug("Target coordinates: " + b.getX() + ", " + b.getY());
//            return;
//        }
//        Logger.debug("Couldn't plan a path to a blocked road");
//        Logger.info("Moving randomly");
//        sendMove(time, randomWalk());
//	}

//	/**
//	 * 
//	 * By JiangYh 2010-4-23
//	 */
//	private void intend(int low, int time) {
//		System.out.println(low);
//		switch (low) {
//		case COMMAND_IGNORE:
//			sendSubscribe(time, 1);
//			break;
//		case REFUGE_REST:
//			moveToRefuge();
//			break;
//		case BLOCKADE_CLEAR:
//			clearHere(time);
//			break;
//		case ESCAPE_BURNING_BUILDING:
//			exitIfIsBurning();
//			break;
//		case CENTRE_ASSIGNMENT:
//			//TODO
//			break;
//		default:
//			moveForClearing();
//			//searchCivilians(true);
//		}
//	}
	/**
	 * 
	 * By JiangYh 2010-6-14
	 */
	private void moveForClearing() {
		ArrayList<EntityID> targetLocations;
		
		targetLocations = myState.pathMove(myState.targetPath,model);
		if(targetLocations == null)return;
		List<EntityID> movePath = search.breadthFirstSearch(location().getID(), targetLocations);
		targetLocations.remove(0);
		movePath.addAll(targetLocations);
		this.move(movePath);
	}

//	/**
//	 * 
//	 * By JiangYh 2010-4-23
//	 * 
//	 * @param heard
//	 * @param time
//	 * @return If agent get a reflection which is based on its current
//	 *         situation, return true. Otherwise return false.
//	 */
//	private int reflect(int time, Collection<Command> heard) {
//
//		if (this.commandIgnore(time))
//			return 0;
//		// RestInRefuge
//		if (this.restIfIsDamaging())
//			return 1;
//		// ClearHere
//		if (this.clearHere(time))
//			return 2;
//
//		if (this.exitIfIsBurning())
//			return 3;
//
//		if (this.hasBeenAssigned())
//			return 4;
//		return -1;
//	}

//	/**
//	 * @return By JiangYh 2010-6-12
//	 */
//	private boolean hasBeenAssigned() {
//		boolean Areachecked = false;
//		StandardEntity target = me().getState().getCenterAssignedTarget();
//		if (target instanceof Road)
//			Areachecked = ((Road) target).getState().getAreaChecked();
//		if (target instanceof Building)
//			Areachecked = ((Building) target).getState().getAreaChecked();
//		return me().getState().getCenterAssignedTarget() != null
//				&& !Areachecked;
//	}

	/**
	 * @return By JiangYh 2010-6-12
	 */
	private boolean exitIfIsBurning() {
		Area pos = null;
		if (me().getPosition(model) instanceof Area) {
			pos = (Area) me().getPosition(model);
		} else
			return false;
		if (pos instanceof Building) {
			if (pos instanceof Refuge)
				return false;
			Building bldg = (Building) pos;
			if (bldg.isOnFire())
				return true;
		}
		return false;
	}

	/**
	 * If a police force agent is near a blockade, the agent will clean the
	 * blockade. By JiangYh 2010-4-27
	 * 
	 * @param time
	 */
	private void clearHere() {
		Blockade nearbyTarget = getTargetBlockadeInRange();
		if (nearbyTarget != null) {
			Logger.info("Clearing blockade " + nearbyTarget);
//			sendSpeak(time, 1, ("Clearing " + nearbyTarget).getBytes());
			clear(nearbyTarget.getID());
		}
	}
	
	private void clearInRange() {
		Blockade nearbyTarget = getTargetBlockadeInRange();
		if (nearbyTarget != null) {
			Logger.info("Clearing blockade " + nearbyTarget);
//			sendSpeak(time, 1, ("Clearing " + nearbyTarget).getBytes());
			clear(nearbyTarget.getID());
		}
	}

//	/**
//	 * 
//	 * By JiangYh 2010-4-23
//	 * 
//	 * @param heard
//	 * @param time
//	 */
//	private void reasoning(int time, Collection<Command> heard) {
//		
//		if (time == config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY)) {
//    		initialChannels(time);
//        }
//    	
//    	//update world message
//    	updateModel(heard);
//    	
//    	super.prepareForAct(time);
//    	
//    	modelState = model.getState();
//    	myState = me().getState();
//			
//    	if (agentMessageManager != null){
//			agentMessageManager.sendMessages();
//			if(agentMessageManager.getMessageBytes() != null){
//				sendSpeak(time, 2 ,agentMessageManager.getMessageBytes());
//				System.out.println("pf send =  "+(agentMessageManager.getMessageBytes()));
//			}
//		}
//
//		// TODO Auto-generated method stub
//		setClearPartition();
//
//		actAsPFLeader();
//
//		selectTarget();
//
//	}

	/**
	 * 
	 * By JiangYh 2010-6-12
	 */
	private void selectTarget() {
		{
			myState.initBlkPath();

			if (getTime() < config
					.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY))
				return;

			if (myState.myClearPartition == null) {
				myState.targetPath = null;
				return;
			}

			// in partitions
			for (Iterator<Path> it = myState.myClearPartition
					.getClearingPaths().iterator(); it.hasNext();) {
				Path p = it.next();

				int count[] = p.getNumOfBlockadesOnPath(getTime());
				if (count[0] > 0) {
					BlockedPath blockedPath = new BlockedPath(p);
					blockedPath.directDistance = myState.getDirectDistance(p,
							model);
					myState.mapBlockedPaths.put(p, blockedPath);
				}

				if (count[0] == 0 && count[1] == 0)
					it.remove();
			}

			for (Iterator<Human> it = model.getState().lockedByBlockadeAgentMap
					.values().iterator(); it.hasNext();) {
				Human hum = it.next();
				if (hum.getPosition(model) == null)
					continue;
				Path p = null;
				if (hum.getPosition(model) instanceof Area)
					if (hum.getPosition(model) instanceof Road)
						p = ((Road) hum.getPosition(model)).getState()
								.getContainedPaths(model).get(0);
					else
						p = ((Building) hum.getPosition(model)).getState()
								.getContainedPaths(model).get(0);
				else
					return;

				if (myState.mapBlockedPaths.containsKey(p)) {
					myState.mapBlockedPaths.get(p).numOfTrappedAgent++;
					myState.mapBlockedPaths.get(p).directDistance /= 3;
				}
			}

			for (Iterator<Human> it = model.getState().buriedAgents.iterator(); it
					.hasNext();) {
				Human hum = it.next();
				if (hum.getPosition(model) == null)
					continue;
				Path p = null;
				if (hum.getPosition(model) instanceof Building)
					p = ((Building) hum.getPosition(model)).getState()
							.getContainedPaths(model).get(0);
				else
					return;

				if (myState.mapBlockedPaths.containsKey(p)) {
					myState.mapBlockedPaths.get(p).numOfTrappedAgent++;
					myState.mapBlockedPaths.get(p).directDistance /= 3;
				}
			}

			if (myState.targetPath != null) {
				int[] count = myState.targetPath
						.getNumOfBlockadesOnPath(getTime());
				if (count[0] > 0 || count[1] > 0) {
					if (myState.mapBlockedPaths.containsKey(myState.targetPath)) {
						myState.mapBlockedPaths.get(myState.targetPath).isPrevTask = 1;
					}
				}
			}

			// reset targetPath
			myState.targetPath = null;
			//

			if (model.getState().CG.myGraph != null) {
				for (Iterator<Integer> it = model.getState().CG.connectedBlockedPaths
						.iterator(); it.hasNext();) {
					int id = it.next();
					Path p = model.getState().paths.get(id);
					if (myState.mapBlockedPaths.containsKey(p)) {
						BlockedPath info = myState.mapBlockedPaths.get(p);
						if (model.getState().CG.m_additionalPathValue.get(p) != null)
							info.additional = model.getState().CG.m_additionalPathValue
									.get(p);

						info.directDistance = myState.getDirectDistance(p,
								model);
					}
				}
			}

			// near fires
			List<Building> aroundFire = new ArrayList<Building>();

			for (Building fire : model.getState().burningBuildings) {
				aroundFire.add(fire);
				for (Building b : fire.getState().getExtinguishableNeighbours(
						model)) {
					if (!aroundFire.contains(b))
						aroundFire.add(b);
				}
			}

			//
			for (Building b : aroundFire) {
				for (Path path : b.getState().getContainedPaths(model)) {

					if (myState.mapBlockedPaths.containsKey(path)) {

						myState.mapBlockedPaths.get(path).civValue += b
								.getState().civPossiblity;
						myState.mapBlockedPaths.get(path).numOfFiredBldgs++;
					}
				}
			}
			//
			for (Civilian civ : model.getState().civilians) {
				if (civ.getPosition(model) == null
						|| civ.getPosition(model) instanceof Refuge)
					continue;
				if (isReachable(civ.getPosition()))
					continue;

				// if(!SeuConstants.IS_COMPETETION_MODE)
				// world.logFile.writeString(civ+"  not reachale !!!", 62,
				// "pfEva");

				List<Path> cuPaths = null;
				if (civ.getPosition(model) instanceof Road)
					cuPaths = ((Road) civ.getPosition(model)).getState()
							.getContainedPaths(model);
				if (civ.getPosition(model) instanceof Building)
					cuPaths = ((Building) civ.getPosition(model)).getState()
							.getContainedPaths(model);
				for (Path p : cuPaths) {
					if (myState.mapBlockedPaths.containsKey(p)) {
						myState.mapBlockedPaths.get(p).numOfTrappedCivilian += 3;
						myState.mapBlockedPaths.get(p).directDistance = myState.mapBlockedPaths
								.get(p).directDistance / 3;
					}
				}

			}

			// calculate values
			for (Iterator it = myState.mapBlockedPaths.values().iterator(); it
					.hasNext();) {
				BlockedPath blockedPath = (BlockedPath) it.next();
				blockedPath.calculateValue(true);
				myState.blockedPaths.add(blockedPath);
			}

			HashSet<Path> pathOmitted = new HashSet<Path>();
			if (!model.getState().CG.m_additionalPathValue.isEmpty())
				for (Iterator it = model.getState().CG.m_additionalPathValue
						.keySet().iterator(); it.hasNext();) {
					Path p = (Path) it.next();

					if (myState.mapBlockedPaths.containsKey(p)) {
						Road pHead = p.getHeadNode();
						Road pTail = p.getTailNode();
						if (pHead.getState().getConnectedRoads(model).size() < 2
								&& !model.getState().CG.containedMovingAgents
										.contains(pHead)) {
							myState.mapBlockedPaths.get(p).additional = 0;
							pathOmitted.add(p);
						} else if (pTail.getState().getConnectedRoads(model)
								.size() < 2
								&& !model.getState().CG.containedMovingAgents
										.contains(pTail)) {
							myState.mapBlockedPaths.get(p).additional = 0;
							pathOmitted.add(p);
						} else
							myState.mapBlockedPaths.get(p).additional = model
									.getState().CG.m_additionalPathValue.get(p);
					}
				}

			Collections.sort(myState.blockedPaths,
					new PoliceForceState.blockedPathComparator());

			ArrayList<BlockedPath> deleted = new ArrayList<BlockedPath>();
			for (int i = 0; i < myState.blockedPaths.size(); i++) {
				BlockedPath bp = myState.blockedPaths.get(i);
				if (bp.getValue() < 0 && pathOmitted.contains(bp.m_path)) {
					deleted.add(bp);
				}
			}

			myState.blockedPaths.removeAll(deleted);

			ArrayList<Double> roundTable = new ArrayList<Double>();
			double totalValue = 0;
			for (Iterator<BlockedPath> it = myState.blockedPaths.iterator(); it
					.hasNext();) {
				BlockedPath bp = it.next();
				roundTable
						.add(roundTable.size() == 0 ? bp.getValue()
								: roundTable.get(roundTable.size() - 1)
										+ bp.getValue());
				totalValue += bp.getValue();
			}

			double dice = Math.random() - myState.randomCoef;

			for (int i = 0; i < roundTable.size(); i++)
				if (dice <= (roundTable.get(i) / totalValue)) {
					myState.targetPath = myState.blockedPaths.get(i).m_path;
					myState.randomCoef += (1 - dice) / 4;
					myState.randomCoef = myState.randomCoef >= 1 ? 0
							: myState.randomCoef;
					break;
				}

			// if(!blockedPaths.isEmpty()){
			// targetPath = blockedPaths.get(0).m_path;
			//			
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString("finally: "+blockedPaths.get(0).m_path+"  value= "+blockedPaths.get(0).value,
			// 62, "pfEva");
			// }

		}

	}

	/**
	 * 
	 * By JiangYh 2010-6-11
	 */
	protected void actAsPFLeader() {
		if (!me().isVirtualCenter)
			return;

		if (getTime() == 2)
			tfEvaluation = new TrafficEvaluation(model);

		if (tfEvaluation != null) {
			tfEvaluation.evaluate();
			tfEvaluation.assignTargets();
			// sendPFAssigns();
		}
	}

	/**
	 * 
	 * By JiangYh 2010-6-11
	 */
	// TODO
	// private void sendPFAssigns() {
	// if (centerMessageManager != null) {
	//
	// for (Iterator it = tfEvaluation.m_assign.keySet().iterator(); it
	// .hasNext();) {
	// StandardEntity real = (StandardEntity) it.next();
	//					
	// Area mlo = null;
	//					
	// if(real instanceof Human)
	// if()
	// mlo = ((Human)real).getPosition(model);
	// else
	// mlo = (Area)real;
	//
	// PoliceForce pf = tfEvaluation.m_assign.get(real);
	//
	// if (mlo == null || pf == null) {
	// continue;
	// }
	//					
	// centerMessageManager.reportAssignedClearTargets(mlo, pf, true);
	//
	// }
	// }
	// }
	/**
	 * 
	 * By JiangYh 2010-6-11
	 */
	private void setClearPartition() {
		if (myState.myClearPartition != null) {
			boolean needDel = true;

			List freePFs = BURIED_C.not()
					.extract(model.getState().policeForces);

			if (freePFs.size() == model.getState().policeForces.size()
					|| !useSpeak)
				return;

			for (Road rd : myState.myClearPartition.getContainedRoads()) {
				if (rd.getState().getCondition() == RoadState.CONDITION_BLOCKED
						|| rd.getState().getCondition() == RoadState.CONDITION_UNKNOWN) {
					needDel = false;
					break;
				}
			}

			if (needDel) {
				myState.clearPartitions.remove(myState.myClearPartition);
				myState.myClearPartition = null;
			}

		}

		if (myState.myClearPartition == null)
			return;

		if (myState.clearPartitions.isEmpty())
			return;

		int searchSize = myState.clearPartitions.size();

		if (myState.myClearPartition == null && searchSize > 0) {
			myState.myClearPartition = (RadialPartition) myState.clearPartitions
					.get(random.nextInt(searchSize));
		}

	}

	@Override
	protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.POLICE_FORCE);
	}

	private List<Road> getBlockedRoads() {
		Collection<StandardEntity> e = model
				.getEntitiesOfType(StandardEntityURN.ROAD);
		List<Road> result = new ArrayList<Road>();
		for (StandardEntity next : e) {
			Road r = (Road) next;
			if (r.isBlockadesDefined() && !r.getBlockades().isEmpty()) {
				result.add(r);
			}
		}
		return result;
	}

	private Blockade getTargetBlockade() {
		//logger.debug("Looking for target blockade");
		Area location = (Area) location();
		//logger.debug("Looking in current location");
		Blockade result = getTargetBlockade(location, clearDistance);
		if (result != null) {
			return result;
		}
		//logger.debug("Looking in neighbouring locations");
		for (EntityID next : location.getNeighbours()) {
			location = (Area) model.getEntity(next);
			if(location instanceof Road){
				if(((Road)location).getState().m_roadnode.size() < 2){	
					continue;
				}
				System.out.println(location+" m_roadnode < 2 "+ ((Road)location).getState().m_roadnode.size());
			}
			result = getTargetBlockade(location, clearDistance);
			if (result != null) {
				return result;
			}
		}
		return null;
	}
	
	private Blockade getTargetBlockadeInRange() {
		//logger.debug("Looking for target blockade");
        Area location = (Area)location();
        //logger.debug("Looking in current location");
        Blockade result = getTargetBlockade(location, clearDistance);
        if (result != null) {
            return result;
        }
        //logger.debug("Looking in neighbouring locations");
        for (EntityID next : location.getNeighbours()) {
            location = (Area)model.getEntity(next);
            result = getTargetBlockade(location, clearDistance);
            if (result != null) {
                return result;
            }
        }
        return null;
	}

	private Blockade getTargetBlockade(Area area, int maxDistance) {
		//logger.debug("Looking for nearest blockade in " + area);
		if (!area.isBlockadesDefined()) {
			//logger.debug("Blockades undefined");
			return null;
		}
		List<EntityID> ids = area.getBlockades();
		// Find the first blockade that is in range.
		int x = me().getX();
		int y = me().getY();
		for (EntityID next : ids) {
			Blockade b = (Blockade) model.getEntity(next);
			double d = findDistanceTo(b, x, y);
			//logger.debug("Distance to " + b + " = " + d);
			if (maxDistance < 0 || d < maxDistance) {
				//logger.debug("In range");
				return b;
			}
		}
		//logger.debug("No blockades in range");
		return null;
	}

	private int findDistanceTo(Blockade b, int x, int y) {
		//logger.debug("Finding distance to " + b + " from " + x + ", " + y);
		List<Line2D> lines = GeometryTools2D.pointsToLines(GeometryTools2D
				.vertexArrayToPoints(b.getApexes()), true);
		double best = Double.MAX_VALUE;
		Point2D origin = new Point2D(x, y);
		for (Line2D next : lines) {
			Point2D closest = GeometryTools2D.getClosestPointOnSegment(next,
					origin);
			double d = GeometryTools2D.getDistance(origin, closest);
			//logger.debug("Next line: " + next + ", closest point: " + closest
//					+ ", distance: " + d);
			if (d < best) {
				best = d;
				//logger.debug("New best distance");
			}

		}
		return (int) best;
	}

	/*
	 * private void clearSelectedTarget() throws ActionCommandException {
	 * 
	 * if (myState.myPath != null) {
	 * 
	 * int[] count = myPath.getNumOfBlockadesOnPath(getTime()); if (count[0] ==
	 * 0 && count[1] == 0) { finishedPaths.add(myPath);
	 * partiPaths.remove(myPath); myPath = null; return; }
	 * 
	 * 
	 * List<PoliceForce> samePath = getSamePathPFs();
	 * 
	 * if (samePath.size() > 1) { Collections.sort(samePath, Util.idComparator);
	 * 
	 * myPath = null; return; } }
	 * 
	 * if (myPath.isOnPath(me().getMotionlessPosition())) { pathMove(myPath); }
	 * else { Route rt = getSimplePathMoveRoute(Collections.singleton(myPath
	 * .getCenter())); if (rt != null) move(myPath.getCenter()); } } }
	 */

	/**
	 * Get the blockade that is nearest this agent.
	 * 
	 * @return The EntityID of the nearest blockade, or null if there are no
	 *         blockades in the agents current location.
	 */
	/*
	 * public EntityID getNearestBlockade() { return
	 * getNearestBlockade((Area)location(), me().getX(), me().getY()); }
	 */

	/**
	 * Get the blockade that is nearest a point.
	 * 
	 * @param area
	 *            The area to check.
	 * @param x
	 *            The X coordinate to look up.
	 * @param y
	 *            The X coordinate to look up.
	 * @return The EntityID of the nearest blockade, or null if there are no
	 *         blockades in this area.
	 */
	/*
	 * public EntityID getNearestBlockade(Area area, int x, int y) { double
	 * bestDistance = 0; EntityID best = null;
	 * //logger.debug("Finding nearest blockade"); if (area.isBlockadesDefined())
	 * { for (EntityID blockadeID : area.getBlockades()) {
	 * //logger.debug("Checking " + blockadeID); StandardEntity entity =
	 * model.getEntity(blockadeID); //logger.debug("Found " + entity); if (entity
	 * == null) { continue; } Pair<Integer, Integer> location =
	 * entity.getLocation(model); //logger.debug("Location: " + location); if
	 * (location == null) { continue; } double dx = location.first() - x; double
	 * dy = location.second() - y; double distance = Math.hypot(dx, dy); if
	 * (best == null || distance < bestDistance) { bestDistance = distance; best
	 * = entity.getID(); } } } ////logger.debug("Nearest blockade: " + best); return
	 * best; }
	 */

//	/*
//	 * =========================================CONSTANT==========================
//	 * ==========
//	 */
//	private static final int COMMAND_IGNORE = 0;
//	private static final int REFUGE_REST = 1;
//	private static final int BLOCKADE_CLEAR = 2;
//	private static final int ESCAPE_BURNING_BUILDING = 3;
//	private static final int CENTRE_ASSIGNMENT = 4;
	
	//==================================================
	
	protected void initialChannels(int time){
		if(useSpeak){
    		if(model.getState().channelCount > 3){
        		sendSubscribe(time, 3);
        		speakChannel = 2;
        	}
        	else if(model.getState().channelCount > 2){
        		sendSubscribe(time, 1);
        		speakChannel = 2;
        	}
        	else if(model.getState().channelCount > 1){
        		sendSubscribe(time, 1);
        		speakChannel = -1;
        	}
    		agentMessageManager = new AgentMessageManager(model, me().getID());
    	}
		else {
			sayMessageManager = new SayMessageManager(model, me().getID());
		}
//    	System.out.println(me() +" areas = "+modelState.areas.size()+" "
//    			+" Buildings = "+modelState.buildings.size()+" "
//    			+" Roads = "+modelState.roads.size()+" "
//    			+model.getEntitiesOfType(StandardEntityURN.BUILDING).size()
//    			+model.getEntitiesOfType(StandardEntityURN.ROAD).size()
//    			+model.getEntitiesOfType(StandardEntityURN.BUILDING,StandardEntityURN.ROAD).size());
    }
	private List<EntityID> getBlockedRoadsID() {
        Collection<StandardEntity> e = model.getEntitiesOfType(StandardEntityURN.ROAD);
        List<EntityID> result = new ArrayList<EntityID>();
        for (StandardEntity next : e) {
            Road r = (Road)next;
            if (r.isBlockadesDefined() && !r.getBlockades().isEmpty()) {
                result.add(r.getID());
            }
        }
        return result;
    }
	
	protected void pfSearch(){
		if(unKnownRoads.isEmpty()) return;
		Collections.sort(unKnownRoads, distanceToMeComparator);
		for(Road next : unKnownRoads){
			if(model.getDistance(location(), next) < Constants.SEE_DISTANCE * 2) continue;
			move(next);
			return;
		}
	}
	
	Comparator distanceToMeComparator = new Comparator()
	{
		public int compare(Object obj1, Object obj2)
		{
			Road r1 = (Road) obj1;
			Road r2 = (Road) obj2;
			if (model.getDistance(location(),r1) > model.getDistance(location(),r2))
				return 1;
			else if (model.getDistance(location(),r1) == model.getDistance(location(),r2))
				return 0;
			else
				return -1;
		}
	};
	
}
