package seu.agent;

import java.util.Collection;
import java.util.EnumSet;
import java.util.Iterator;

import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.messages.Command;
import rescuecore2.log.Logger;

import rescuecore2.standard.entities.*;
import seu.messages.CenterMessageManager;
import seu.utilities.AmbulanceEvaluation;
import seu.utilities.TrafficEvaluation;

/**
   A PoliceOffice agent.
 */
public class PoliceOfficeAgent extends CentreAgent<PoliceOffice> {
	
	protected AmbulanceEvaluation ambulanceEvaluation = null;
	protected TrafficEvaluation policeEvaluation = null;

    protected void postConnect() {
        super.postConnect();
        
        policeEvaluation = new TrafficEvaluation(model);
		
		if(model.getEntitiesOfType(StandardEntityURN.FIRE_STATION).isEmpty()
				&& model.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE).isEmpty())
			ambulanceEvaluation = new AmbulanceEvaluation(model);
    }
    
    @Override
    protected void think(int time, ChangeSet changed, Collection<Command> heard) {
        
    	prepareForAct(time, changed, heard);
    	
    	act();
    	
    }

    public void prepareForAct(int time , ChangeSet changed, Collection<Command> heard) {
    	if(time < agentIgnoreTime)return;
    	
    	if (time == agentIgnoreTime) {
			try{
				initialChannels(time);
			}catch (Exception e) {
				System.out.println(me()+" initialChannel failed");
				e.printStackTrace();
			}
		}
    	
    	super.prepareForAct(heard);
    	
		if(ambulanceEvaluation != null){
			ambulanceEvaluation.evaluateForRescue();
			sendATAssigns();	
		}
		
		if(policeEvaluation != null){
			policeEvaluation.evaluate();
			policeEvaluation.assignTargets();
			sendPFAssigns();
		}
		
		try{
			if (centerMessageManager != null){
				centerMessageManager.sendMessages();
				sendSpeak(time, speakChannel ,centerMessageManager.getMessageBytes());
			}
		}catch (Exception e) {
			System.out.println(me()+" sendMessage failed");
			e.printStackTrace();
		}
    }
    
    public void act() {
		super.act();
	}
    
    protected void initialChannels(int time){
    	if(useSpeak){
    		if(model.getState().channelCount > 3){
        		sendSubscribe(time, 1, 2);
        		speakChannel = 3;
        		centerMessageManager = new CenterMessageManager(model, me().getID());
        	}
        	else if(model.getState().channelCount > 2){
        		sendSubscribe(time, 1, 2);
        		speakChannel = 1;
        		centerMessageManager = new CenterMessageManager(model, me().getID());
        	}
        	else if(model.getState().channelCount > 1){
        		if(model.getState().policeOffices.get(0) == me()){
        			sendSubscribe(time, 1);
            	speakChannel = 1;
            	centerMessageManager = new CenterMessageManager(model, me().getID());
        		}
        	}
    	}
    }
    
    private void sendATAssigns() {
		if (getTime() < config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY))
			return;
		if (centerMessageManager != null) {

			for (Iterator it = ambulanceEvaluation.m_assign.keySet().iterator(); it
					.hasNext();) {
				AmbulanceTeam at = (AmbulanceTeam) it.next();
				Human target = ambulanceEvaluation.m_assign.get(at);
				if (at == null || target == null)
					continue;

				centerMessageManager.reportAssignedRescueTargets(target, at, 1);
			}

			for (Iterator it = ambulanceEvaluation.m_secondAssign.keySet()
					.iterator(); it.hasNext();) {
				AmbulanceTeam at = (AmbulanceTeam) it.next();
				Human target = ambulanceEvaluation.m_secondAssign.get(at);
				if (at == null || target == null)
					continue;

				centerMessageManager.reportAssignedRescueTargets(target, at, 0);
			}
		}
	}
    
    private void sendPFAssigns() {
		if (centerMessageManager != null) {

			for (Iterator it = policeEvaluation.m_assign.keySet().iterator(); it
					.hasNext();) {
				StandardEntity real = (StandardEntity) it.next();
				StandardEntity pos = null;
				Area mlo = null;
				
				if(real instanceof Human){
					pos = ((Human)real).getPosition(model);
					if(pos instanceof Area){
						mlo = (Area)pos;
					}
					else if(pos instanceof Human){
						mlo = (Area)((Human)pos).getPosition(model);
					}
				}	
				else {
					mlo = (Area)real;
				}
				PoliceForce pf = policeEvaluation.m_assign.get(real);

				if (mlo == null || pf == null) {
					continue;
				}
				
				centerMessageManager.reportAssignedClearTargets(mlo, pf, true);

			}
		}
	}
    
    @Override
    public String toString() {
        return "PoliceOffice";
    }
    
    @Override
    protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
        return EnumSet.of(StandardEntityURN.POLICE_OFFICE);
    }
}
