/**************************************************************************
 * Copyright (C) 2010 Knight Guan <Knight.Guan@gmail.com>
 * All rights reserved.
 * Edited by Southeast University RoboCup.
 **************************************************************************
 * Last modify:
 * $Id: Util.java,v 2.0 2010 16:43 Knight Exp $
 */
package seu.agent;

import java.awt.Point;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;
import java.util.StringTokenizer;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.StandardEntity;
import seu.utilities.Condition;
import seu.utilities.Wall;
import seu.utilities.base.SingleData;

public class Util {
	public static final int distance(Area from, Area to) {
		float dx = to.getX() - from.getX();
		float dy = to.getY() - from.getY();
		return (int) Math.sqrt(dx * dx + dy * dy);
	}

	public static final int distance(Area from, int x, int y) {
		float dx = x - from.getX();
		float dy = y - from.getY();
		return (int) Math.sqrt(dx * dx + dy * dy);
	}

	public static final int distance(Area from, Point to) {
		float dx = to.x - from.getX();
		float dy = to.y - from.getY();
		return (int) Math.sqrt(dx * dx + dy * dy);
	}

	public static final int direction(Area from, Area to) {
		float dx = to.getX() - from.getX();
		float dy = to.getY() - from.getY();
		float theta = (float) Math.atan2(-dx, dy);
		if (theta < 0)
			theta += 2 * Math.PI;
		return (int) (theta * 360 * 60 * 60 / (2 * Math.PI));
	}

	/**
	 * For normalize directions to [0,360-1] in degree.<br/>
	 * 
	 * @param from
	 * @param to
	 * @return
	 */
	public static final int normalizedDirection(Area from, Area to) {
		float dx = to.getX() - from.getX();
		float dy = to.getY() - from.getY();
		if (dx == 0) {
			if (dy > 0)
				return 90;
			else
				return 270;
		}
		float theta = (float) Math.atan(dy / dx);
		if (theta < 0) {
			if (dx > 0)
				theta += 2 * Math.PI;
			else
				theta += Math.PI;
		} else {
			if (dx < 0)
				theta += Math.PI;
		}
		return (int) (theta * 360 / (2 * Math.PI));
	}

	/**
	 * For normalize directions to [0,360-1] in degree.<br/>
	 * 
	 * @param from
	 * @param to
	 * @return
	 */
	public static final int normalizedDirection(int startX, int startY,
			int endX, int endY) {
		float dx = endX - startX;
		float dy = endY - startY;
		if (dx == 0) {
			if (dy > 0)
				return 90;
			else
				return 270;
		}
		float theta = (float) Math.atan(dy / dx);
		if (theta < 0) {
			if (dx > 0)
				theta += 2 * Math.PI;
			else
				theta += Math.PI;
		} else {
			if (dx < 0)
				theta += Math.PI;
		}
		return (int) (theta * 360 / (2 * Math.PI));
	}

	// Guan Daqi Use
	public static final int normalizedDirection(double startX, double startY,
			double endX, double endY) {
		double dx = endX - startX;
		double dy = endY - startY;
		if (dx == 0) {
			if (dy > 0)
				return 90;
			else
				return 270;
		}
		double theta = Math.atan(dy / dx);
		if (theta < 0) {
			if (dx > 0)
				theta += 2 * Math.PI;
			else
				theta += Math.PI;
		} else {
			if (dx < 0)
				theta += Math.PI;
		}
		return (int) Math.round((theta * 360 / (2 * Math.PI)));
	}

	public static String classBaseName(Object obj) {
		StringTokenizer st = new StringTokenizer(obj.getClass().getName(), ".");
		String token = null;
		while (st.hasMoreTokens())
			token = st.nextToken();
		return token;
	}

	public static int max(int x, int y) {
		return x >= y ? x : y;
	}

	public static int min(int x, int y) {
		return x <= y ? x : y;
	}

	public static Object randomChoice(Collection col, Random random) {
		if (col.isEmpty())
			throw new Error("the col must not be empty.");
		Iterator it = col.iterator();
		for (int i = random.nextInt(col.size()); i > 0; i--)
			it.next();
		return it.next();
	}

	//
	public static final int distance(int x1, int y1, int x2, int y2) {
		float dx = x1 - x2;
		float dy = y1 - y2;
		return (int) Math.sqrt(dx * dx + dy * dy);
	}

	public static final int direction(int x1, int y1, int x2, int y2) {
		float dx = x2 - x1;
		float dy = y2 - y1;
		float theta = (float) Math.atan2(-dx, dy);
		if (theta < 0)
			theta += 2 * Math.PI;
		return (int) (theta * 360 * 60 * 60 / (2 * Math.PI));
	}

	public static int getConditionCount(Collection col, Condition cond) {
		int res = 0;
		for (Iterator it = col.iterator(); it.hasNext();) {
			if (cond.eval(it.next()))
				res++;
		}
		return res;
	}

	public static Comparator<StandardEntity> idComparator = new Comparator<StandardEntity>() {
		public int compare(StandardEntity o1, StandardEntity o2) {
			StandardEntity r1 = o1;
			StandardEntity r2 = o2;

			if (r1.getID().getValue() > r2.getID().getValue())
				return 1;
			if (r1.getID().getValue() == r2.getID().getValue())
				return 0;

			return -1;
		}
	};

	// public static Comparator idComparator2 = new Comparator() {
	// public int compare(Object o1, Object o2) {
	// if (o1 instanceof MapPartition && o2 instanceof MapPartition) {
	// MapPartition r1 = (MapPartition) o1;
	// MapPartition r2 = (MapPartition) o2;
	// if (r1.id > r2.id)
	// return 1;
	// if (r1.id == r2.id)
	// return 0;
	// return -1;
	// } else if (o1 instanceof Path && o2 instanceof Path) {
	// Path r1 = (Path) o1;
	// Path r2 = (Path) o2;
	// if (r1.id > r2.id)
	// return 1;
	// if (r1.id == r2.id)
	// return 0;
	// return -1;
	// } else if (o1 instanceof Entrance && o2 instanceof Entrance) {
	// Entrance r1 = (Entrance) o1;
	// Entrance r2 = (Entrance) o2;
	// if (r1.id > r2.id)
	// return 1;
	// if (r1.id == r2.id)
	// return 0;
	// return -1;
	// } else {
	// System.err.println("Invalid Compare : " + o1 + " With : " + o2);
	// return o1.hashCode() - o2.hashCode();
	// }
	// }
	// };

	public static boolean Position(double x0, double y0, double ang, double tg,
			Wall edg) {
		double a = edg.x1;
		double b = edg.y1;
		double c = edg.x2;
		double d = edg.y2;

		if (ang == 0f || ang == Math.PI) {
			if (isBetween(a, x0, c)) {
				if (a == c) {
					edg.distance = Util.distance((int) x0, (int) y0, (int) a,
							(int) b);
					if (a > x0) {
						edg.right = true;

					} else {
						edg.right = false;
					}
				} else {
					edg.distance = Util.distance((int) x0, (int) y0, (int) x0,
							(int) ((b - d) * (x0 - c) / (a - c) + d));
					edg.right = false;
				}
				return true;
			}
			edg.distance = 0;
			return false;
		} else {
			if (a == c) {
				double y = tg * (a - x0) + y0;
				if (isBetween(b, y, d)) {
					edg.distance = Util.distance((int) x0, (int) y0, (int) a,
							(int) y);
					if (a > x0) {
						edg.right = true;
					} else {
						edg.right = false;
					}
					return true;
				}
				edg.distance = 0;
				return false;
			}
			if (a * tg - c * tg + d - b == 0) {
				edg.distance = 0;
				return false;
			}
			double x = (-b * c + d * a + c * y0 + a * tg * x0 - a * y0 - c * tg
					* x0)
					/ (a * tg - c * tg + d - b);
			if (isBetween(a, x, c)) {
				if (a == c) {
					edg.distance = Util.distance((int) x0, (int) y0, (int) a,
							(int) b);
					if (a > x0) {
						edg.right = true;
					} else {
						edg.right = false;
					}
				} else {
					edg.distance = Util.distance((int) x0, (int) y0, (int) x,
							(int) ((b - d) * (x - c) / (a - c) + d));
					if (x > x0) {
						edg.right = true;
					} else {
						edg.right = false;
					}
				}
				return true;
			}
			edg.distance = 0;
			return false;
		}
	}

	// 
	// public static Comparator<MessageNode> capacityComparator = new
	// Comparator<MessageNode>() {
	// public int compare(MessageNode o1, MessageNode o2) {
	// MessageNode m1 = o1;
	// MessageNode m2 = o2;
	// if (m1.getCapacity() < m2.getCapacity())
	// return 1;
	// if (m1.getCapacity() == m2.getCapacity())
	// return 0;
	// return -1;
	// }
	// };

	// add for mutli-tree sorting 从大到小
	// public static Comparator<MultiTree> heightComparator = new
	// Comparator<MultiTree>() {
	// public int compare(MultiTree o1, MultiTree o2) {
	// MultiTree mt1 = o1;
	// MultiTree mt2 = o2;
	//
	// if (mt1.getHeight() < mt2.getHeight())
	// return 1;
	// if (mt1.getHeight() == mt2.getHeight())
	// return 0;
	// return -1;
	// }
	// };

	// public static Comparator<MessageNode> nodeIdComparator = new
	// Comparator<MessageNode>() {
	// public int compare(MessageNode o1, MessageNode o2) {
	// if (o1.getID() > o2.getID())
	// return 1;
	// if (o1.getID() == o2.getID())
	// return 0;
	// return -1;
	// }
	// };

	public static boolean isBetween(double a, double x, double b) {
		if (a > b) {
			double k = a;
			a = b;
			b = k;
		}
		if (x >= a && x <= b) {
			return true;
		}
		return false;
	}

	/**
	 * 
	 * 
	 * @param <T>
	 * @param list
	 * @param time
	 * @return
	 */
	public static <T> T searchSingleDataForTime(LinkedList<SingleData<T>> list,
			int time) {
		int left = 0, right = list.size() - 1;

		if (time <= list.getFirst().getTime())
			return list.getFirst().getObject();

		if (time >= list.getLast().getTime())
			return list.getLast().getObject();

		while (true) {
			int location = (left + right) / 2;
			int locationTime = list.get(location).getTime();

			if (location == list.size() - 1 && locationTime < time)
				return list.get(location).getObject();

			int locationTimeNext = list.get(location + 1).getTime();
			if (locationTime < time && locationTimeNext > time)
				return list.get(location).getObject();

			if (locationTime > time)
				right = location - 1;
			else if (locationTime < time)
				left = location + 1;
			else
				return list.get(location).getObject();
		}

	}

	public static <T> void updateSingleDataList(LinkedList<SingleData<T>> list,
			T data, int time, boolean isSee, int maxLength) {

		if (list.isEmpty()) {
			list.addLast(new SingleData<T>(data, time, isSee));
			return;
		}

		SingleData<T> lastData = list.getLast();
		if (lastData.getTime() < time) {
			if (list.size() >= maxLength) {
				list.removeFirst();
				list.addLast(new SingleData<T>(data, time, isSee));
			} else {
				list.addLast(new SingleData<T>(data, time, isSee));

			}
		} else if (lastData.getTime() == time) {
			if (!lastData.isSee() && isSee) {
				list.removeLast();
				list.addLast(new SingleData<T>(data, time, isSee));
			}

		}

	}

}
