//SEU2010

package seu.messages;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import rescuecore2.connection.Connection;
import rescuecore2.messages.components.IntComponent;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

/**
 * @author JiangYh 2010-05-01
 * 
 */
public class AgentMessageManager extends MessageManager {
	
	public AgentMessageManager(StandardWorldModel world, EntityID entityID) {
		super(world, entityID);
	}

	public void importMessage(byte[] message, int time, StandardEntity sender) {
		if(sender == null) return;
		if (sender.getID() == this.selfId) {
			return;
		}

		parseMessage(message, time, sender, false);
	}

	public void sendMessages() {
		if (noCenters == 0) {
			if (!messagesToCenters.isEmpty())
				sendMessages(0 , messagesToCenters);
			clearForNextCycle();
			messagesToAgents.clear();
			messagesToCenters.clear();
		} else {
			messagesToCenters.addAll(messagesToAgents);
			if (!messagesToCenters.isEmpty()) {
				sendMessages(0, messagesToCenters);
			}
			incrementDelayedMessageLife(messagesToCenters);
			clearForNextCycle();
			messagesToAgents.clear();
		}
		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("AgentMessageManager  ##############  above Cycle = "
		// + world.getTime(), 56, "sendMsg");
	}

	public void reportFoundCivilian(Human civ) {
		if (civ != null && civ.getPosition(world) != null) {
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString("FoundCiv = " + civ+"  pos= "
			// +civ.getMotionlessPosition()
			// +" damage = "+civ.getDamage()
			// +" bury = "+civ.getBuriedness(), 56, "sendMsg");
			SEUMessage message = new FoundCivilianMessage(civ);
			messagesToCenters.add(message);
		}
	}

	public void reportFoundCivilian(Collection civs) {
		for (Iterator it = civs.iterator(); it.hasNext();) {
			Civilian c = (Civilian) it.next();
			SEUMessage message = new FoundCivilianMessage(c);
			messagesToCenters.add(message);
		}
	}

	public void reportVictim(Human victim) {
		if (victim != null && victim.getPosition() != null) {
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString("Victim = " + victim+"  pos= "
			// +victim.getRealPosition()
			// +" dmg = "+victim.getDamage()
			// +" bury = "+victim.getBuriedness(), 56, "sendMsg");
			SEUMessage message = new VictimMessage(victim);
			messagesToCenters.add(message);
		}
	}

	public void reportVictim(Human victim, Area pos) {
		if (victim != null && pos != null) {
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString("Victim = " + victim+"  pos= "
			// +pos
			// +" bury = "+victim.getBuriedness(), 56, "sendMsg");
			SEUMessage message = new VictimMessage(victim, pos);
			messagesToCenters.add(message);
		}
	}

	// public void reportVictim(Collection victims)
	// {
	// for (Iterator i = victims.iterator(); i.hasNext();)
	// {
	// Human h = (Human) i.next();
	// Message message = new VictimMessage(h);
	// messagesToCenters.add(message);
	// }
	// }

	public void reportFire(Building building) {
		if (building != null) {
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString("Fire = " + building
			// + "  fiery= " + building.getFieryness()
			// + "  time = "+world.getTime()
			// // + "temperature = " + building.getTemperature()
			// , 56, "sendMsg");

			// TODO
			SEUMessage message = new FireMessage(world, building, worldState.agents
					.indexOf(world.getEntity(selfId)));

			messagesToCenters.add(message);
		}
	}

	public void reportFire(Collection<Building> building) {
		for (Iterator<Building> iterator = building.iterator(); iterator
				.hasNext();) {
			Building b = iterator.next();
			SEUMessage message = new FireMessage(world, b, worldState.agents
					.indexOf(world.getEntity(selfId)));
			messagesToCenters.add(message);
		}
	}

	public void reportRoad(Road road, int condition) {
		if (road != null) {
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString("Road = " + road + "  condition= "
			// + road.getCondition() + "  width= " + road.isVisible(), 56,
			// "sendMsg");

			SEUMessage message = new RoadMessage(road, condition);
			messagesToCenters.add(message);
		}
	}

	// TODO
	/*
	 * public void reportRoad(Collection roads) { for (Iterator iterator =
	 * roads.iterator(); iterator.hasNext();) { Area r = (Area) iterator.next();
	 * if (r instanceof Road) { Road road = (Road) r; //
	 * if(!SeuConstants.IS_COMPETETION_MODE) //
	 * world.logFile.writeString("Road = " + road + "  block= " // +
	 * road.getBlock(), 56, "sendMsg"); //TODO seuMessage message = new
	 * RoadMessage(road, road.getCondition()); messagesToCenters.add(message); }
	 * } }
	 */
	public void reportEmptyBuilding(Building building) {
		if (building != null) {
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString("EmptyBuilding = " + building, 56,
			// "sendMsg");
			SEUMessage message = new EmptyBuildingMessage(building);
			messagesToCenters.add(message);
		}
	}

	public void reportEmptyBuilding(Collection buildings) {
		for (Iterator iterator = buildings.iterator(); iterator.hasNext();) {
			Area b = (Area) iterator.next();
			if (b instanceof Building) {
				SEUMessage message = new EmptyBuildingMessage((Building) b);
				messagesToCenters.add(message);
			} else {
				;
			}
		}
	}

	public void reportPosition(Human hum) {

		if (hum == null || hum.getPosition(world) == null) {
			return;
		}
		StandardEntity pos = hum.getPosition(world);
		if (pos instanceof Human) pos = ((Human) pos).getPosition(world);
		// Collections.sort(world.agents, yab.agent.Util.idComparator);
		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("Position = " + hum.getRealPosition(), 56,
		// "sendMsg");
		boolean isOPHead = true;
		/*
		 * if (hum.getPosition(world) instanceof Road) { Road r = (Road)
		 * hum.getPosition(world); //TODO if
		 * (world.passedNodes.contains(r.getPointHead())) isOPHead = true; }
		 */
		SEUMessage message = new PositionMessage((Area) pos,
				worldState.agents.indexOf(world.getEntity(selfId)), isOPHead);
		messagesToCenters.add(message);
	}

	public void reportHeardCivilian(Civilian civilian, Area position,
			int positionExtra) {
		SEUMessage message = new HeardCivilianMessage(civilian, position,
				positionExtra);
		messagesToCenters.add(message);
	}

	public void reportLockedByBlockade() {
		if (!(world.getEntity(selfId) instanceof Human)) {
			return;
		}
		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("Locked = " + world.me, 56, "sendMsg");
		// Collections.sort(world.agents, seu.agent.Util.idComparator);
		SEUMessage message = new LockedByBlockadeMessage(worldState.agents
				.indexOf(world.getEntity(selfId)));
		messagesToCenters.add(message);
	}

	public void reportUnlockedByBlockade() {
		if (!(world.getEntity(selfId) instanceof Human)) return;
		// if(!SeuConstants.IS_COMPETETION_MODE)
		// world.logFile.writeString("UnLocked = " + world.me, 56, "sendMsg");
		// TODO
		Collections.sort(worldState.agents,
				seu.utilities.AgentUtil.idComparator);
		SEUMessage message = new UnLockedByBlockadeMessage(worldState.agents
				.indexOf(world.getEntity(selfId)));

		messagesToCenters.add(message);
	}

	// used while no centers
	public void reportAssignedRescueTargets(Human target, Human targetReceiver,
			int isFirstAssign) {
		if (targetReceiver == null || target == null) {
			return;
		}

		// System.out.println(" AC assign target  " + target
		// + "  to " + targetReceiver);

		SEUMessage message = new AssignRescueTargetMessage(target,
				worldState.ambulanceTeams.indexOf(targetReceiver),
				isFirstAssign);
		messagesToCenters.add(message);
	}

	public void reportAssignedClearTargets(Road target,
			Human targetReceiver, boolean isFirstAssign) {
		if (targetReceiver == null || target == null) {
			return;
		}

		// System.out.println(" Center assign target  " + target
		// + "  to " + targetReceiver);

		SEUMessage message = new AssignClearTargetMessage(target,
				worldState.policeForces.indexOf(targetReceiver),
				isFirstAssign);
		messagesToCenters.add(message);
	}

	public void reportAssignedExtinguishTargets(Building target,
			Human targetReceiver) {
		if (targetReceiver == null || target == null
				|| !(targetReceiver instanceof FireBrigade)) {
			return;
		}

		// System.out.println(" FB assign target  " + target
		// + "  to " + targetReceiver);

		SEUMessage message = new AssignExtinguishTargetMessage(target,
				worldState.fireBrigades.indexOf(targetReceiver));
		messagesToCenters.add(message);
	}

//	public void reportFireWater(Building bld, int waterQuantity) {
//		SEUMessage message = new FireWaterMessage(bld, waterQuantity);
//		messagesToCenters.add(message);
//	}
//
//	public void reportRemainedWater() {
//		if (time.getValue() < 3) return;
//		if (!(world.getEntity(selfId) instanceof FireBrigade)) return;
//
//		int waterQuantity = ((FireBrigade) (world.getEntity(selfId)))
//				.getWater();
//		SEUMessage message = new RemainWaterMessage(worldState.fireBrigades
//				.indexOf(world.getEntity(selfId)), waterQuantity);
//		messagesToCenters.add(message);
//	}
}
