package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import seu.SeuConstants;
import seu.messages.Type;
import seu.utilities.MessageUtil;


public class AssignClearTargetMessage extends SEUMessage
{
	protected int targetId;
	protected int receiverIndex;
	protected int isFirstAssign;
    
    static final int idSize = ID_SIZE;
    static final int indexSize = 5;
    static final int isFirstSize = 1;

    public AssignClearTargetMessage(Area target, int receiverIndex, boolean isFirstAssign)
    {
        super(Type.ASSIGN_CLEAR_TARGET_REPORT, null);
        this.targetId = target.getID().getValue();
        this.receiverIndex = receiverIndex;
        this.isFirstAssign = isFirstAssign ? 1:0;
    } 
    
    public AssignClearTargetMessage(char[] message, Map virtualIdToId)
    {
		super(Type.ASSIGN_CLEAR_TARGET_REPORT, null);
		this.receiverIndex = MessageUtil.charArrayToint(message, 0, indexSize);
		this.targetId = ((Integer) virtualIdToId.get(new Integer(MessageUtil.charArrayToint(message, indexSize, idSize)))).intValue();
		this.isFirstAssign = MessageUtil.charArrayToint(message, indexSize + idSize, isFirstSize);
		this.lifeCycle = MessageUtil.charArrayToint(message, indexSize + idSize + isFirstSize,
				lifeSize);
	}


	@Override
	public boolean equals(Object m)
    {
        if (!(m instanceof AssignClearTargetMessage)) return false;
        return ((AssignClearTargetMessage) m).targetId == targetId
        			&& ((AssignClearTargetMessage) m).receiverIndex == receiverIndex
        			&& ((AssignClearTargetMessage)m).isFirstAssign == isFirstAssign;
    }

	@Override
	public int hashCode()
    {
        return targetId;
    }

	@Override
	protected int length()
    {
        return staticLength();
    }

    protected static int staticLength()
    {
        return indexSize + idSize + isFirstSize + lifeSize;
    }

    char[] toCharArray(Map idToVirtualId)
    {
        char[] result = new char[length()];

        char[] temp = MessageUtil.intToCharArray(receiverIndex, indexSize);
        MessageUtil.copyArray(temp, result, 0, 0, indexSize);

        temp = MessageUtil.intToCharArray(((Integer) (idToVirtualId.get(new Integer(targetId)))).intValue(), idSize);
        MessageUtil.copyArray(temp, result, 0, indexSize, idSize);
        
        temp = MessageUtil.intToCharArray(isFirstSize, isFirstSize);
		MessageUtil.copyArray(temp, result, 0, indexSize + idSize, lifeSize);
		
        temp = MessageUtil.intToCharArray(lifeCycle, lifeSize);
		MessageUtil.copyArray(temp, result, 0, indexSize + idSize + isFirstSize, lifeSize);

        return result;
    }

	public boolean updateWorld(StandardWorldModel world, int time) {
		PoliceForce pf = (PoliceForce) world.getState().policeForces.get(receiverIndex);

		Area target = (Area) world.getEntity(new EntityID(targetId));

		pf.getState().setCenterAssignedTarget(target, time - lifeCycle);

		return false;
	}

}

