package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.WorldModel;
import seu.messages.Type;
import seu.utilities.MessageUtil;



public class AssignExtinguishTargetMessage extends SEUMessage {

	protected int targetId;
	protected int receiverIndex;
	
	static final int indexSize = 5;
	 
	public AssignExtinguishTargetMessage(Building target, int receiverIndex) {
		super(Type.ASSIGN_EXTINGUISH_TARGET_REPORT, target);
		targetId = target.getID().getValue();
		this.receiverIndex = receiverIndex;
	}

	public AssignExtinguishTargetMessage(char[] message, Map virtualIdToId){
	        super(Type.ASSIGN_EXTINGUISH_TARGET_REPORT, null);
	        this.receiverIndex = MessageUtil.charArrayToint(message, 0, indexSize);
			this.targetId = ((Integer) virtualIdToId.get(new Integer(MessageUtil
					.charArrayToint(message, indexSize, idSize)))).intValue();
			this.lifeCycle = MessageUtil.charArrayToint(message, indexSize + idSize,
					lifeSize);
	}
	
	@Override
	public boolean equals(Object m) {
		if (!(m instanceof AssignExtinguishTargetMessage)) return false;
        return ((AssignExtinguishTargetMessage) m).targetId == targetId
        			&& ((AssignExtinguishTargetMessage) m).receiverIndex == receiverIndex
        			&& ((AssignExtinguishTargetMessage) m).lifeCycle == lifeCycle;
	}

	@Override
	public int hashCode() {
		return targetId;
	}

	@Override
	protected int length() {
		return staticLength();
	}
	
	protected static int staticLength()
	{
	    return indexSize + idSize + lifeSize;
	}

	char[] toCharArray(Map idToVirtualIds) {
		 char[] result = new char[length()];

	        char[] temp = MessageUtil.intToCharArray(receiverIndex, indexSize);
	        MessageUtil.copyArray(temp, result, 0, 0, indexSize);

	        temp = MessageUtil.intToCharArray(((Integer) idToVirtualIds.get(new Integer(targetId))).intValue(), idSize);
	        MessageUtil.copyArray(temp, result, 0, indexSize, idSize);
	        
	        temp = MessageUtil.intToCharArray(lifeCycle, lifeSize);
			MessageUtil.copyArray(temp, result, 0, indexSize + idSize, lifeSize);

	        return result;
	}

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		if (!(world.getState().me instanceof FireBrigade)) {
			return true;
		}

		 FireBrigade me = (FireBrigade) world.getState().me;

		FireBrigade assigned = world.getState().fireBrigades.get(receiverIndex);

		Building target = (Building) world.getEntity(new EntityID(targetId));

		assigned.getState().setCenterAssignedTarget(target);

		return false;
	}
}
