package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import seu.messages.Type;
import seu.utilities.MessageUtil;



public class AssignRescueTargetMessage extends SEUMessage {

	protected int targetId;
	protected int receiverIndex;
	protected int isFirstAssign;	
	
	 static final int indexSize = 5;
	 static final int realIdSize = 32;
	 static final int isFirstSize = 1;

	public AssignRescueTargetMessage(Human target, int receiverIndex, int isFirstAssign)
    {
        super(Type.ASSIGN_RESCUE_TARGET_REPORT, target);
        this.targetId = target.getID().getValue();
        this.receiverIndex = receiverIndex;
        this.isFirstAssign = isFirstAssign;
    }
	
	
	public AssignRescueTargetMessage(char[] message, Map virtualIdToId){
	    super(Type.ASSIGN_RESCUE_TARGET_REPORT, null);
	    this.receiverIndex = MessageUtil.charArrayToint(message, 0, indexSize);
		this.targetId = MessageUtil.charArrayToint(message, indexSize, realIdSize);
		this.isFirstAssign = MessageUtil.charArrayToint(message, indexSize + realIdSize, isFirstSize);
		this.lifeCycle = MessageUtil.charArrayToint(message, indexSize + realIdSize + isFirstSize,
					lifeSize);
	}
	
	@Override
	public boolean equals(Object m) {
		 if (!(m instanceof AssignRescueTargetMessage)) return false;
	        return ((AssignRescueTargetMessage) m).targetId == targetId
	        			&& ((AssignRescueTargetMessage) m).receiverIndex == receiverIndex
	        			&& ((AssignRescueTargetMessage) m).isFirstAssign == isFirstAssign;
	}

	@Override
	public int hashCode() {
		return receiverIndex;
	}
	@Override
	protected int length()
    {
        return staticLength();
    }

    protected static int staticLength()
    {
        return indexSize + realIdSize + lifeSize + isFirstSize;
    }
    
	@Override
	char[] toCharArray(Map ids) {
		 char[] result = new char[length()];

	        char[] temp = MessageUtil.intToCharArray(receiverIndex, indexSize);
	        MessageUtil.copyArray(temp, result, 0, 0, indexSize);

	        temp = MessageUtil.intToCharArray(targetId, realIdSize);
	        MessageUtil.copyArray(temp, result, 0, indexSize, realIdSize);
	        
	        temp = MessageUtil.intToCharArray(isFirstAssign, isFirstSize);
	        MessageUtil.copyArray(temp, result, 0, indexSize + realIdSize, isFirstSize);
	        
	        temp = MessageUtil.intToCharArray(lifeCycle, lifeSize);
			MessageUtil.copyArray(temp, result, 0, indexSize + realIdSize + isFirstSize, lifeSize);

	        return result;
	}
	
	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		
		//TODO
		//if list the ambulanceTeams or not
		AmbulanceTeam assigned = world.getState().ambulanceTeams.get(receiverIndex);
        
        Civilian target = (Civilian) world.getState().get(new EntityID(targetId));
        
        assigned.getState().setAssignedCycle(time);
        
        if (isFirstAssign == 1) {
			assigned.getState().setCenterAssignedTarget(target, time);
		} else {
			assigned.getState().setCenterSecondAssignedTarget(target, time);
		}

        return false;
	}
}
