//SEU2010

package seu.messages;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import rescuecore2.connection.Connection;
import rescuecore2.messages.components.IntComponent;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;


public class CenterMessageManager extends MessageManager {
	
	public static final int C_CHANNEL_CENTER = 0;

	public static final int C_CHANNEL_AGENT = 1;
	
	
	public CenterMessageManager(StandardWorldModel world, EntityID entityID) {
		super(world, entityID);
	}

	public void importMessage(byte[] message, int time, StandardEntity sender) {
		if (sender.getID() == this.selfId) {
			return;
		}

		parseMessage(message, time, sender, false);
	}

	public void sendMessages() {
		sendMessages(STATION_MSG_COUNT - 1, messagesToCenters);
		
		sendMessages(1, messagesToAgents);
		clearForNextCycle();
		
		incrementDelayedMessageLife(messagesToAgents);
		incrementDelayedMessageLife(messagesToCenters);
	}
	
	public void reportAssignedClearTargets(Area target,
			Human targetReceiver, boolean isFirstAssign) {
		if (targetReceiver == null || target == null) {
			return;
		}

//		System.out.println(" Center assign target  " + target
//				+ "  to " + targetReceiver);

		SEUMessage message = new AssignClearTargetMessage(target, worldState.policeForces
				.indexOf(targetReceiver),isFirstAssign);
		messagesToAgents.add(message);
	}
	
	public void reportAssignedRescueTargets(Human target,
			Human targetReceiver, int isFirstAssign) {
		if (targetReceiver == null || target == null) {
			return;
		}

//		System.out.println(" AC assign target  " + target
//				+ "  to " + targetReceiver);

		SEUMessage message = new AssignRescueTargetMessage(target, worldState.ambulanceTeams
				.indexOf(targetReceiver), isFirstAssign);
		messagesToAgents.add(message);
	}
	
	public void reportAssignedExtinguishTargets(Building target,
			Human targetReceiver) {
		if (targetReceiver == null || target == null
				|| !(targetReceiver instanceof FireBrigade)) {
			return;
		}

//		System.out.println(" FB assign target  " + target
//				+ "  to " + targetReceiver);

		SEUMessage message = new AssignExtinguishTargetMessage(target, worldState.fireBrigades
				.indexOf(targetReceiver));
		messagesToAgents.add(message);
	}
	

	public void reportFire(Building building)
	{
		if (building != null)
		{
//			if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString("Fire = " + building 
//					+ "  fiery= " + building.getFieryness()
//					+ "temperature = " + building.getTemperature()
//					, 56, "sendMsg");

			SEUMessage message = new FireMessage(world, building, worldState.agents
					.indexOf(world.getEntity(selfId)));

			messagesToCenters.add(message);
			messagesToAgents.add(message);
		}
	}
	
//	public void reportTemperature(Building building){
//		if(building != null){
////			if(!SeuConstants.IS_COMPETETION_MODE)
////			world.logFile.writeString(building + "temperature = "
////					+ building.getTemperature() + " time = " 
////					+ building.getTemperatureList().getLast().getTime()
////					, 56, "sendMsg");
//			
//			Message message = new TemperatureMessage(building);
//			
//			messagesToCenters.add(message);
//			messagesToAgents.add(message);
//		}
//	}
	

	public void reportRoad(Road road, int condition)
	{
		if (road != null)
		{
//			if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString("Road = " + road + "  condition= "
//					+ road.getCondition() + "  width= " + road.isVisible(), 56,
//					"sendMsg");

			SEUMessage message = new RoadMessage(road, condition);
			messagesToCenters.add(message);
			messagesToAgents.add(message);
		}
	}
	

	public void reportEmptyBuilding(Building building)
	{
		if (building != null)
		{
//			if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString("EmptyBuilding = " + building, 56, "sendMsg");
			SEUMessage message = new EmptyBuildingMessage(building);
			messagesToCenters.add(message);
			messagesToAgents.add(message);
		}
	}

	

	public void reportVictim(Human victim)
	{
		if (victim != null && victim.getPosition(world) != null)
		{
//			if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString("Victim = " + victim+"  pos= "
//					+victim.getRealPosition()
//					+" bury = "+victim.getBuriedness(), 56, "sendMsg");
			SEUMessage message = new VictimMessage(victim);
			messagesToCenters.add(message);
			messagesToAgents.add(message);
		}
	}
	
	

	public void reportFoundCivilian(Human civ)
	{
		if (civ != null && civ.getPosition(world) != null)
		{
//			if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString("FoundCiv = " + civ+"  pos= "
//					+civ.getMotionlessPosition()
//					+" bury = "+civ.getBuriedness(), 56, "sendMsg");
			SEUMessage message = new FoundCivilianMessage(civ);
			messagesToCenters.add(message);
			messagesToAgents.add(message);
		}
	}
	
	
}
