package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.WorldModel;
import seu.messages.Type;
import seu.states.BuildingState;
import seu.utilities.MessageUtil;


public class EmptyBuildingMessage extends SEUMessage
{
    private int buildingId;

    public EmptyBuildingMessage(Building building)
    {
        super(Type.EMPTYBUILDING_REPORT, building);
        buildingId = building.getID().getValue();
    }
    public EmptyBuildingMessage(char[] message, Map virtualIdToId)
    {
        super(Type.EMPTYBUILDING_REPORT, null);
        this.buildingId = ((Integer) virtualIdToId.get(new Integer(MessageUtil.charArrayToint(message, 0, idSize)))).intValue();
        this.lifeCycle = MessageUtil.charArrayToint(message, idSize, lifeSize);
    }
    public EmptyBuildingMessage(int buildingId, Map virtualToIds)
    {
        super(Type.EMPTYBUILDING_REPORT, null);
        this.buildingId = ((Integer) virtualToIds.get(new Integer(buildingId))).intValue();
    }
    
	@Override
	public boolean equals(Object m)
    {
        if (!(m instanceof EmptyBuildingMessage)) return false;
        return buildingId == ((EmptyBuildingMessage) m).buildingId;
    }

	@Override
	public int hashCode() {
		return buildingId;
	}

	@Override
	protected int length() {
		return staticLength();
	}
	
	protected static int staticLength()
    {
        return idSize + lifeSize;
    }

	@Override
    char[] toCharArray(Map idToVirtualIds)  // id
    {
        char[] result = new char[length()];

        char[] temp = MessageUtil.intToCharArray(((Integer) idToVirtualIds.get(new Integer(buildingId))).intValue(), idSize);
        MessageUtil.copyArray(temp, result, 0, 0, idSize);

        temp = MessageUtil.intToCharArray(lifeCycle, lifeSize);
        MessageUtil.copyArray(temp, result, 0, idSize, lifeSize);
        
        return result;
    }

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		
		obj = world.getEntity(new EntityID(buildingId));
		
		int messageTime = time - lifeCycle;
		
        Building bldg = (Building) obj;
        BuildingState bst = bldg.getState();

        if(bst.emptyBuildingReport.getCount() > 1){
        	return false;
        }
       	 
        bst.emptyBuildingReport.update(time);
        
        bst.civPossiblity = 0;
        
        world.getState().notSeenBuildings.remove(bldg);
        
        if(!world.getState().emptyBuildings.contains(bldg)){
          	 world.getState().emptyBuildings.add(bldg);
        }

        bst.setEmptyTime(messageTime);
        bst.setSeenTime(messageTime);
        bst.setReportedAsEmpty();
        return true;
	}
}
