package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import seu.messages.FireMessage;
import seu.messages.Util;
import seu.messages.Type;
import seu.states.BuildingState;

public class FireMessage extends SEUMessage{
	
	static final int fierynessSize = 4;
	static final int indexSize = 7;
	static final int burnTimeSize = 8;
	static final int temperatureSize = 8;

	private final int buildingId;
	private final int senderIndex;
	public int fieryness;
	public int burnTime;
	public int temperature;
	
	public FireMessage(StandardWorldModel world, Building building, int senderIndex)
	{
		super(Type.FIRE_REPORT, building);
		buildingId = building.getID().getValue();
		fieryness = building.getFieryness();
		burnTime = building.getState().getBurningTimeToNow(world);
		this.senderIndex = senderIndex;
		//TODO
		//can we use building.getTemperature() ?
		this.temperature = building.getTemperature()/50;
	}

	public FireMessage(char[] message, Map virtualIdToId) // buildingId -
	// firestart -
	// threshold -
	// fieryness //-
	// time
	{
		super(Type.FIRE_REPORT, null);

		this.buildingId = ((Integer) virtualIdToId.get(new Integer(Util
				.charArrayToint(message, 0, idSize)))).intValue();
		this.senderIndex = Util.charArrayToint(message, idSize, indexSize);
		this.fieryness = Util.charArrayToint(message, idSize + indexSize,
				fierynessSize);
		
		this.burnTime = Util.charArrayToint(message, idSize + indexSize
				+ fierynessSize , burnTimeSize);

		this.lifeCycle = Util.charArrayToint(message, idSize + indexSize
				+ fierynessSize + burnTimeSize, lifeSize);
		
		this.temperature = Util.charArrayToint(message, indexSize + indexSize
				+ fierynessSize + burnTimeSize + lifeSize, temperatureSize); 
	}
	
	public int getBuildingId()
	{
		return buildingId;
	}
	
	@Override
	public boolean equals(Object m) {
		if (!(m instanceof FireMessage))
			return false;
		return (buildingId == ((FireMessage) m).buildingId)
				&& (fieryness == ((FireMessage)m).fieryness
				&& temperature == ((FireMessage) m).temperature);
	}

	@Override
	public int hashCode() {
		return buildingId;
	}

	@Override
	protected int length() {
		return staticLength();
	}
	
	static protected int staticLength()
	{
		return idSize + indexSize + fierynessSize +  burnTimeSize + lifeSize + temperatureSize;
	}

	@Override
	char[] toCharArray(Map idToVirtualIds)
	{
		char[] result = new char[length()];

		char[] temp;

		temp = Util.intToCharArray(((Integer) idToVirtualIds.get(new Integer(
				buildingId))).intValue(), idSize);
		Util.copyArray(temp, result, 0, 0, idSize);

		temp = Util.intToCharArray(senderIndex, indexSize);
		Util.copyArray(temp, result, 0, idSize, indexSize);

		temp = Util.intToCharArray(fieryness, fierynessSize);
		Util.copyArray(temp, result, 0, idSize + indexSize, fierynessSize);
		
		temp = Util.intToCharArray(burnTime, burnTimeSize);
		Util.copyArray(temp, result, 0, idSize + indexSize + fierynessSize , burnTimeSize);

		temp = Util.intToCharArray(lifeCycle, lifeSize);
		Util.copyArray(temp, result, 0, idSize + indexSize + fierynessSize + burnTimeSize,
				lifeCycle);
		
		temp = Util.intToCharArray(temperature, temperatureSize);
        Util.copyArray(temp, result, 0, idSize + indexSize + fierynessSize + burnTimeSize + 
        		lifeSize, temperatureSize);

		return result;
	}

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		
		obj = world.getEntity(new EntityID(buildingId));
		
		int messageTime = time - lifeCycle;
		
		Building bldg = (Building) obj;
		BuildingState bst = bldg.getState();
		if (!bst.shouldUpdate(messageTime))
			return false;

		//TODO
		//setIgno
		bst.setSeenTime(messageTime);

		bst.setProperty(world, messageTime, bldg.getFierynessProperty(),
				new int[] { fieryness });
		
		int old_time = bst.getTime();
        if(old_time < messageTime){
        	bst.setProperty(world, messageTime, bldg.getTemperatureProperty(), 
    				new int[] {temperature*50}); 
        }
		
		if (bst.m_justIgnited)
		{
			bst.setIgniteTime(Math.max(0, messageTime - burnTime));
		}
		if(bst.isBurned()){
			world.getState().emptyBuildings.add(bldg);
		}

		return true;
	}
}