package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import seu.SeuConstants;
import seu.states.*;


public class FoundCivilianMessage extends SEUMessage
{
    private int realId;
    protected int posId;
    protected int hp;
    protected int damage;
    protected int buriedness;

    static final int realIdSize = 32;
    static final int hpSize = 4;
    static final int damageSize = 8;
    static final int buriednessSize = 6;
    static final int HP_BASE = 1000;
    static final int DAMAGE_BASE = 10;

    public FoundCivilianMessage(Human victim)
    {
        super(Type.FOUND_CIVILIAN_REPORT, victim);
        this.realId = victim.getID().getValue();
        this.posId = victim.getPosition().getValue();
        this.hp = victim.getHP();
        this.damage = victim.getDamage();
        this.buriedness = victim.getBuriedness();
    }

    public FoundCivilianMessage(char[] message, Map virtualIdToId)
    {
        super(Type.FOUND_CIVILIAN_REPORT, null);
        this.realId = Util.charArrayToint(message, 0, realIdSize);
        this.posId = ((Integer) virtualIdToId.get(new Integer(Util.charArrayToint(message, realIdSize, idSize)))).intValue();
        this.hp = Util.charArrayToint(message, realIdSize + idSize, hpSize)*HP_BASE;
        this.damage = Util.charArrayToint(message, realIdSize + idSize + hpSize, damageSize)*DAMAGE_BASE;
        this.buriedness = (Util.charArrayToint(message, realIdSize + idSize + hpSize + damageSize, buriednessSize));
        this.lifeCycle = Util.charArrayToint(message, realIdSize + idSize + hpSize + damageSize + buriednessSize, lifeSize);
    }

	@Override
	public boolean equals(Object m) {
		if (!(m instanceof FoundCivilianMessage)) return false;
        return (realId == ((FoundCivilianMessage) m).realId)
        		&& posId == ((FoundCivilianMessage)m).posId
        		&& damage == ((FoundCivilianMessage)m).damage
        		&& hp == ((FoundCivilianMessage)m).hp
        		&& buriedness == ((FoundCivilianMessage)m).buriedness;
	}

	@Override
	public int hashCode() {
		return realId;
	}

	@Override
	protected int length() {
		return staticLength();
	}

	protected static int staticLength()
    {
        return realIdSize + idSize + hpSize + damageSize + buriednessSize + lifeSize;
    }
	
	@Override
	char[] toCharArray(Map idToVirtualIds) {
		char[] result = new char[length()];
        char[] temp;

        temp = Util.intToCharArray(realId, realIdSize);
        Util.copyArray(temp, result, 0, 0, realIdSize);

        temp = Util.intToCharArray(((Integer) idToVirtualIds.get(new Integer(posId))).intValue(), idSize);
        Util.copyArray(temp, result, 0, realIdSize, idSize);

        temp = Util.intToCharArray(hp/HP_BASE, hpSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize, hpSize);

        temp = Util.intToCharArray(damage/DAMAGE_BASE, damageSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize + hpSize, damageSize);

        temp = Util.intToCharArray(buriedness, buriednessSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize + hpSize + damageSize, buriednessSize);

        temp = Util.intToCharArray(lifeCycle, lifeSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize + hpSize + damageSize + buriednessSize, lifeSize);

        return result;
	}

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		obj = world.getEntity(new EntityID(realId));
        
        int messageTime = time - lifeCycle;
           
        if (obj == null){
            obj = StandardEntityFactory.INSTANCE.makeEntity(StandardEntityURN.CIVILIAN, new EntityID(realId));
            world.addEntity(obj);
            world.getState().add(new EntityID(realId));
            obj.getState().setTime(0);
        }
        
        if(!(obj instanceof Civilian)) {	
        	return false;
        }
        
        if(obj.getState().isVisible(world)) {
        	return false;
        }
        
        Civilian civ = (Civilian) obj;
        
        civ.getState().setParticle(world, time, hp, damage, buriedness, 1);
        
        world.getState().heardCivilians.remove(civ);
            
        if(!civ.getState().shouldUpdate(time)){
        	return false;
        }
        
        if(world.getState().savedHumanoids.contains(civ)){
        	return false;
        }
        
        civ.getState().setFound();
        civ.getState().setTime(time);
        
        civ.getState().setProperty(world, time, civ.getPositionProperty(), new int[]{posId});
        civ.getState().setProperty(world, time, civ.getHPProperty(), new int[]{hp});
        civ.getState().setProperty(world, time, civ.getDamageProperty(), new int[]{damage});
        civ.getState().setProperty(world, time, civ.getBuriednessProperty(), new int[]{buriedness});
        
        if(civ.getPosition(world) instanceof Refuge)
        	world.getState().savedHumanoids.add(civ);
        
        if(civ.getPosition(world) instanceof Building){
        	((Building)civ.getPosition(world)).getState().setSeenTime(time);
        }
        
        return true;
	}
}
