package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

public class HeardCivilianMessage extends SEUMessage
{
    static final int realIdSize = 32;
    static final int posExtraSize = 6;

    int realId;
    int posId;
    int posExtra;

    public HeardCivilianMessage(Human civilian, Area position , int posEx)
    {
        super(Type.HEARD_CIVILIAN_REPORT, civilian);
        realId = civilian.getID().getValue();
        posId = position.getID().getValue();
        posExtra = posEx / 1000;
    }

    public HeardCivilianMessage(char[] message, Map virtualIdToId)
    {
        super(Type.HEARD_CIVILIAN_REPORT, null);

        this.realId = Util.charArrayToint(message, 0, realIdSize);
        this.posId = ((Integer) virtualIdToId.get(new Integer(Util.charArrayToint(message, realIdSize, idSize)))).intValue();
        this.posExtra = Util.charArrayToint(message, realIdSize + idSize, posExtraSize);
        this.lifeCycle = Util.charArrayToint(message, realIdSize + idSize + posExtraSize, lifeSize);
    }

	@Override
	public boolean equals(Object m) {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public int hashCode() {
		return realId;
	}

	@Override
	protected int length() {
		return staticLength();
	}

	static protected int staticLength()
    {
        return realIdSize + idSize + posExtraSize + lifeSize;
    }
	
	@Override
	char[] toCharArray(Map idToVirtualIds) {
		char[] result = new char[length()];
        char[] temp;

        temp = Util.intToCharArray(realId, realIdSize);
        Util.copyArray(temp, result, 0, 0, realIdSize);

        temp = Util.intToCharArray(((Integer) idToVirtualIds.get(new Integer(posId))).intValue(), idSize);
        Util.copyArray(temp, result, 0, realIdSize, idSize);

        temp = Util.intToCharArray(posExtra,posExtraSize);
        Util.copyArray(temp, result, 0, realIdSize+ idSize,posExtraSize);
        
        temp = Util.intToCharArray(lifeCycle,lifeSize);
        Util.copyArray(temp, result, 0, realIdSize+ idSize + posExtraSize, lifeSize);

        return result;
	}

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		obj = world.getEntity(new EntityID(realId));
		
        if (obj == null){
        	obj = StandardEntityFactory.INSTANCE.makeEntity(StandardEntityURN.CIVILIAN, new EntityID(realId));
            world.addEntity(obj);
            obj.getState().setTime(0);
        }
        
        if(!(obj instanceof Civilian)) return false;
        
        Civilian civ = (Civilian)obj;
        
        if(civ.getState().hasFound()) return false;
        
        int sendTime = time - lifeCycle;
        if(!civ.getState().shouldUpdate(sendTime)){
        	return false;
        }
                
        Area heardPos = (Area) world.getEntity(new EntityID(posId));
        if (civ.getPosition(world) == null && world.getState().me instanceof Human)
        {
            world.getState().heardCivilians.add(civ);
            civ.getState().addHearPosition(world, heardPos, posExtra*1000, false);
            civ.getState().updatePossiblePositions(world);
        }
        return true;
	}
}
