package seu.messages;

import java.util.Collections;
import java.util.Map;

import rescuecore2.standard.entities.*;
import seu.utilities.MessageUtil;
import seu.messages.Type;

public class LockedByBlockadeMessage extends SEUMessage{
	
	private final int senderIndex;
    static final int indexSize = 7;

	public LockedByBlockadeMessage(int senderIndex)
	{
		super(Type.LOCKEDBYBLOCKADE_REPORT, null);
        this.senderIndex = senderIndex;
	}
	
	public LockedByBlockadeMessage(char[] message, Map virtualIdToId)
    {
        super(Type.LOCKEDBYBLOCKADE_REPORT, null);
        this.senderIndex = MessageUtil.charArrayToint(message, 0, indexSize);
        this.lifeCycle = MessageUtil.charArrayToint(message, indexSize, lifeSize);
    }

	@Override
	public boolean equals(Object m) {
		if (!(m instanceof LockedByBlockadeMessage)) return false;
        return ((LockedByBlockadeMessage) m).senderIndex == senderIndex;
	}

	@Override
	public int hashCode() {
		return senderIndex;
	}

	@Override
	protected int length() {
		return staticLength();
	}

	protected static int staticLength(){
        return indexSize + lifeSize;
    }

	@Override
	   char[] toCharArray(Map idToVirtualId)
    {
        char[] result = new char[length()];

        char[] temp = MessageUtil.intToCharArray(senderIndex, indexSize);
        MessageUtil.copyArray(temp, result, 0, 0, indexSize);

        temp = MessageUtil.intToCharArray(lifeCycle, lifeSize);
        MessageUtil.copyArray(temp, result, 0, indexSize, lifeSize);

        return result;
    }

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		Collections.sort(world.getState().agents, seu.agent.Util.idComparator);
		Human hum = (Human) world.getState().agents.get(senderIndex);
		if(!world.getState().lockedByBlockadeAgentMap.containsKey(hum.getID().getValue())){
    		world.getState().lockedByBlockadeAgentMap.put(hum.getID().getValue(), hum);
    		if(hum instanceof AmbulanceTeam){
    			AmbulanceTeam at = (AmbulanceTeam) hum;
    			at.getState().setWereLockedByBlockade(world, true);
    			return true;
    		}
    		else if(hum instanceof FireBrigade){
    			FireBrigade fb = (FireBrigade) hum;
    			fb.getState().setWereLockedByBlockade(world, true);
    			return true;
    		}		
    	}
		return false;
	}

}