//SEU2010

package seu.messages;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.standard.entities.Human;

import seu.messages.Type;
import seu.messages.MessageConstants;
import seu.states.WorldState;
import seu.utilities.Condition;
import seu.utilities.MessageUtil;

/**
 * @author Administrator
 *
 */
public class MessageManager implements MessageConstants {

	protected StandardWorldModel world;
	public WorldState worldState;
	
	byte[] messageBytes = null;
	byte[] sendBytes = null;
	
	private int pointer;

	protected EntityID selfId;
	
	protected int centreLength = 256;
	protected int ambulanceTeamLength;
	protected int fireBrigadeLength;
	protected int policeForceLength;
	protected int sendLength;
	protected int parseLength;

	protected int messageCount;

	public ArrayList reportedVictims;
	public ArrayList reportedFoundCivilian;
	public ArrayList reportedFires;
	public ArrayList reportedPositions;
	public ArrayList reportedRoads;
	public ArrayList reportedEmptyBuildings;
	public ArrayList reportedHeardCivilians;
	public ArrayList reportedLockedByBlockades;
	public ArrayList reportedUnLockedByBlockades;
//	public ArrayList reportedTargets;
//	public ArrayList reportedTargetAssigned;
//	public ArrayList reportedHelpTargets;
//	public ArrayList reportedRequestTargets;
//	public ArrayList reportedTasks;
//	public ArrayList reportedFinishedTasks;
//	public ArrayList reportAISCoefficient;
	public ArrayList reportedATAssign;
	public ArrayList reportedFBAssign;
	public ArrayList reportedPFAssign;
//	public ArrayList reportedTemperature;
//	public ArrayList reportedFireWater;
//	public ArrayList reportedRemaindWater;

	// ========================================

	public ArrayList centerReportedVictims;
	public ArrayList centerReportedFoundCivilian;
	public ArrayList centerReportedFires;
	public ArrayList centerReportedRoads;
	public ArrayList centerReportedPositions;
	public ArrayList centerReportedEmptyBuidings;
	public ArrayList centerReportedHeardCivilians;
	public ArrayList centerReportedLockedByBlockades;
	public ArrayList centerReportedUnLockedByBlockades;
//	public ArrayList centerReportedTargets;
//	public ArrayList centerReportedTargetAssigned;
//	public ArrayList centerReportedRequestTargets;
//	public ArrayList centerReportedTasks;
//	public ArrayList centerReportedHelpTargets;
//	public ArrayList centerReportedFinishedTasks;
//	public ArrayList centerReportAISCoefficient;
	public ArrayList centerReportedATAssign;
	public ArrayList centerReportedFBAssign;
	public ArrayList centerReportedPFAssign;
//	public ArrayList centerReportedTemperature;
//	public ArrayList centerReportedFireWater;
//	public ArrayList centerReportedRemaindWater;

	// ==============================================

	protected Map<Integer, Integer> idToVirtualId;

	protected Map<Integer, Integer> virtualIdToId;

	protected HashSet<SEUMessage> messagesToCenters;

	protected HashSet<SEUMessage> messagesToAgents;

	protected int noCenters = 0;

	protected static int MAX_LIFE_CYCLE = 8;

	public HashSet<SEUMessage> getMessagesToCenters() {
		return messagesToCenters;
	}

	public HashSet<SEUMessage> getMessagesToAgents() {
		return messagesToAgents;
	}

	public void clearMessagesToCenters() {
		messagesToCenters.clear();
	}

	public void clearMessagesToAgents() {
		messagesToAgents.clear();
	}
	/**
     * @param connection
     * @param world
     * @param tellChannels
     * @param entityID
     */
	public MessageManager(StandardWorldModel world, EntityID entityID) {

		this.selfId = entityID;
		if(world.getState().channelCount > 4){
			ambulanceTeamLength = world.getState().channel1/world.getState().ambulanceTeams.size();
			fireBrigadeLength = world.getState().channel2/(world.getState().fireBrigades.size()+world.getState().policeForces.size());
			policeForceLength = world.getState().channel2/(world.getState().fireBrigades.size()+world.getState().policeForces.size());
			if(world.getEntity(entityID) instanceof AmbulanceTeam){
				sendLength = Math.min(ambulanceTeamLength,centreLength);
				parseLength = centreLength;
			}
			else if(world.getEntity(entityID) instanceof FireBrigade){
				sendLength = Math.min(fireBrigadeLength, centreLength);
				parseLength = centreLength;
			}
			else if(world.getEntity(entityID) instanceof PoliceForce){
				sendLength = Math.min(policeForceLength, centreLength);
				parseLength = centreLength;
			}
			else{
				sendLength = centreLength;
				parseLength = centreLength;
			}
		}
		else if(world.getState().channelCount == 3){
			ambulanceTeamLength = (world.getState().channel1 - (world.getState().centerAgents.size()*centreLength))/world.getState().ambulanceTeams.size();
			fireBrigadeLength = world.getState().channel2/(world.getState().fireBrigades.size()+world.getState().policeForces.size());
			policeForceLength = world.getState().channel2/(world.getState().fireBrigades.size()+world.getState().policeForces.size());
			if(world.getEntity(entityID) instanceof AmbulanceTeam){
				sendLength = Math.min(ambulanceTeamLength,centreLength);
				parseLength = Math.max(ambulanceTeamLength, centreLength);
			}
			else if(world.getEntity(entityID) instanceof FireBrigade){
				sendLength = Math.min(fireBrigadeLength, centreLength);
				parseLength = centreLength;
			}
			else if(world.getEntity(entityID) instanceof PoliceForce){
				sendLength = Math.min(policeForceLength, centreLength);
				parseLength = centreLength;
			}
			else{
				sendLength = centreLength;
				parseLength = centreLength;
			}
		}
		else if(world.getState().channelCount == 2){
//			if(world.getState().channel1 > world.getState().centerAgents.size()*centreLength*2){
			ambulanceTeamLength = (world.getState().channel1 - (world.getState().centerAgents.size()*centreLength))/world.getState().ambulanceTeams.size();
//			}
//			else{
//				
//			}
			fireBrigadeLength = 0;
			policeForceLength = 0;
			if(world.getEntity(entityID) instanceof AmbulanceTeam){
				sendLength = Math.min(ambulanceTeamLength,centreLength);
				parseLength = Math.max(ambulanceTeamLength, centreLength);
			}
			else if(world.getEntity(entityID) instanceof FireBrigade){
				sendLength = 0;
				parseLength = Math.max(ambulanceTeamLength, centreLength);
			}
			else if(world.getEntity(entityID) instanceof PoliceForce){
				sendLength = 0;
				parseLength = Math.max(ambulanceTeamLength, centreLength);
			}
			else{
				sendLength = centreLength;
				parseLength = centreLength;
			}
		}
		
		
//		this.messageCount =
//		        world.getEntity(selfId) instanceof Human ? MessageConstants.AGENT_MSG_COUNT
//		                : MessageConstants.STATION_MSG_COUNT;
		this.world = world;
		this.worldState = world.getState();

		if(world.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE).isEmpty())
			noCenters++;

		if(world.getEntitiesOfType(StandardEntityURN.POLICE_OFFICE).isEmpty())
			noCenters++;

		if(world.getEntitiesOfType(StandardEntityURN.FIRE_STATION).isEmpty())
			noCenters++;
		
		idToVirtualId = new HashMap<Integer, Integer>();
		virtualIdToId = new HashMap<Integer, Integer>();

		int virtualId = 0;
		Collections.sort(worldState.areas, seu.agent.Util.idComparator);
		for (Iterator i = worldState.areas.iterator(); i.hasNext(); virtualId++) {
			Area ar = (Area) i.next();
			idToVirtualId.put(new Integer(ar.getID().getValue()),
			        new Integer(virtualId));
			virtualIdToId.put(new Integer(virtualId),
			        new Integer(ar.getID().getValue()));
		}

		reportedFoundCivilian = new ArrayList();
		reportedVictims = new ArrayList();
		reportedRoads = new ArrayList();
		reportedFires = new ArrayList();
		reportedPositions = new ArrayList();
		reportedEmptyBuildings = new ArrayList();
		reportedHeardCivilians = new ArrayList();
		reportedLockedByBlockades = new ArrayList();
		reportedUnLockedByBlockades = new ArrayList();
//		reportedTargets = new ArrayList();
//		reportedRequestTargets = new ArrayList();
//		reportedTasks = new ArrayList();
//		reportedFinishedTasks = new ArrayList();
//		reportedTargetAssigned = new ArrayList();
//		reportedHelpTargets = new ArrayList();
//		reportAISCoefficient = new ArrayList();
		reportedATAssign = new ArrayList();
		reportedFBAssign = new ArrayList();
		reportedPFAssign = new ArrayList();
//		reportedTemperature = new ArrayList();
//		reportedFireWater = new ArrayList();
//		reportedRemaindWater = new ArrayList();
		// ====================================

		centerReportedFoundCivilian = new ArrayList();
		centerReportedVictims = new ArrayList();
		centerReportedRoads = new ArrayList();
		centerReportedFires = new ArrayList();
		centerReportedPositions = new ArrayList();
		centerReportedEmptyBuidings = new ArrayList();
		centerReportedHeardCivilians = new ArrayList();
		centerReportedLockedByBlockades = new ArrayList();
		centerReportedUnLockedByBlockades = new ArrayList();
//		centerReportedTargets = new ArrayList();
//		centerReportedRequestTargets = new ArrayList();
//		centerReportedTasks = new ArrayList();
//		centerReportedFinishedTasks = new ArrayList();
//		centerReportedTargetAssigned = new ArrayList();
//		centerReportedHelpTargets = new ArrayList();
//		centerReportAISCoefficient = new ArrayList();
		centerReportedATAssign = new ArrayList();
		centerReportedFBAssign = new ArrayList();
		centerReportedPFAssign = new ArrayList();
//		centerReportedTemperature = new ArrayList();
//		centerReportedFireWater = new ArrayList();
//		centerReportedRemaindWater = new ArrayList();
		// ====================================

		messagesToCenters = new HashSet<SEUMessage>();
		messagesToAgents = new HashSet<SEUMessage>();

	}

	/**
	 * This method can be called only once per cycle! otherwise consult your
	 * psychiretrist
	 */
	protected void clearForNextCycle() {
		reportedFoundCivilian.clear();
		reportedVictims.clear();
		reportedRoads.clear();
		reportedFires.clear();
		reportedPositions.clear();
		reportedEmptyBuildings.clear();
		reportedHeardCivilians.clear();
		reportedLockedByBlockades.clear();
		reportedUnLockedByBlockades.clear();
//		reportedTargets.clear();
//		reportedRequestTargets.clear();
//		reportedTasks.clear();
//		reportedFinishedTasks.clear();
//		reportedTargetAssigned.clear();
//		reportedHelpTargets.clear();
//		reportAISCoefficient.clear();
		reportedATAssign.clear();
		reportedFBAssign.clear();
		reportedPFAssign.clear();
//		reportedTemperature.clear();
//		reportedFireWater.clear();
//		reportedRemaindWater.clear();
		// ====================================

		centerReportedFoundCivilian.clear();
		centerReportedVictims.clear();
		centerReportedRoads.clear();
		centerReportedFires.clear();
		centerReportedPositions.clear();
		centerReportedEmptyBuidings.clear();
		centerReportedHeardCivilians.clear();
		centerReportedLockedByBlockades.clear();
		centerReportedUnLockedByBlockades.clear();
//		centerReportedTargets.clear();
//		centerReportedRequestTargets.clear();
//		centerReportedTasks.clear();
//		centerReportedFinishedTasks.clear();
//		centerReportedTargetAssigned.clear();
//		centerReportedHelpTargets.clear();
//		centerReportAISCoefficient.clear();
		centerReportedATAssign.clear();
		centerReportedFBAssign.clear();
		centerReportedPFAssign.clear();
//		centerReportedTemperature.clear();
//		centerReportedFireWater.clear();
//		centerReportedRemaindWater.clear();

	}

	protected void sendMessages(int tempCount, Collection messages) {

//		if (tempCount == MessageConstants.MAX_MESSAGE_COUNT) {
//			System.out.println("EXCEEDED MAX_MESSAGE_COUNT in send messages!");
//			// if(!SeuConstants.IS_COMPETETION_MODE)
//			// world.logFile.writeString("EXCEEDED MAX_MESSAGE_COUNT in send messages!",
//			// 56, "sendMsg");
//			return;
//		}

//		boolean isChannelReged = false;
//		for (int i = 0; i < tellChannels.length; i++) {
//			if (tellChannels[i] == channel) {
//				isChannelReged = true;
//				break;
//			}
//		}
//		if (!isChannelReged) {
//			System.out
//			        .println("the channel " + channel + " is not registered!");
//			// if(!SeuConstants.IS_COMPETETION_MODE)
//			// world.logFile.writeString("the channel "+channel+" is not registered!",
//			// 56, "sendMsg");
//			return;
//		}

		// world.logFile.writeString(" ----- TIME = " + world.me.getTime() +
		// " send message "
		// + tempCount + " messages.size = " + messages.size() + " id = "
		// + selfId, 99, "recievedMessage");

		char[] message = new char[sendLength * 8];
		pointer = 0; // where to write in char [] message

		// remove some unnecessary message type
		addMessage(Type.ASSIGN_RESCUE_TARGET_REPORT, message, messages);
		addMessage(Type.POSITION_REPORT, message, messages);
		addMessage(Type.LOCKEDBYBLOCKADE_REPORT, message, messages);
		addMessage(Type.UNLOCKEDBYBLOCKADE_REPORT, message, messages);
		addMessage(Type.ASSIGN_CLEAR_TARGET_REPORT, message, messages);
		addMessage(Type.ASSIGN_EXTINGUISH_TARGET_REPORT, message, messages);

//		addMessage(Type.AIS_COEFFICIENT_REPORT, message, messages);
		// ============================================
		// addMessage(Type.TARGET_ASSIGNED_REPORT, message, messages);
		// addMessage(Type.HELP_TARGET_REPORT, message, messages);
		addMessage(Type.FIRE_REPORT, message, messages);
		addMessage(Type.FOUND_CIVILIAN_REPORT, message, messages);
		addMessage(Type.ROAD_REPORT, message, messages);
		addMessage(Type.EMPTYBUILDING_REPORT, message, messages);
		addMessage(Type.VICTIM_REPORT, message, messages);
		addMessage(Type.HEARD_CIVILIAN_REPORT, message, messages);

//		addMessage(Type.FIRE_WATER_REPORT, message, messages);
//		addMessage(Type.REMAINED_WATER_REPORT, message, messages);

		send(message);
		// log("	mmmmmmmmmmmm message length = "+pointer);
//		if (messages.size() > 0 && ++tempCount < messageCount) {
//			sendMessages(tempCount, messages);
//		}
		// log (" end " + tempCount + " messages.size = " + messages.size ());
	}

	private void addMessage(Type type, char[] message, Collection messages) {
		final Type t = type;
		Condition c = new Condition() {
			public boolean eval(Object obj) {
				if (((SEUMessage) obj).type == t)
					return true;
				return false;
			}
		};
		ArrayList temp = c.extractMAR(messages);
		int count = 0; // tedade message hayi ke khahad ferestad.

		if (!temp.isEmpty()) {
			int headerPos = pointer; // jayi ke bayad header va length ezafe
			// beshan ro nigah midare.
			pointer += MessageConstants.HEADER_SIZE + MessageConstants.LENGTH_SIZE; // baraye inke badan biad
			// header va length ro
			// benevise ja mizare.

			for (Iterator i = temp.iterator(); i.hasNext();) {
				SEUMessage m = (SEUMessage) i.next();

				if (sendLength * 8 - pointer - MessageConstants.HEADER_SIZE >= m.length()) // ja
				// baraye
				// m
				// hast
				{
					count++;
					i.remove();// msg e ferestade shode ro hazf mikone
					MessageUtil.copyArray(m.toCharArray(idToVirtualId), message, 0,
					        pointer, m.length());
					pointer += m.length();
				} else {
					// msg haye ferestade nashode hanooz too list hastan.
					messages.addAll(temp);
					// if(!SeuConstants.IS_COMPETETION_MODE)
					// world.logFile.writeString(" ########  MM- Message Full Messages, length = "+m.length(),
					// 56, "sendMsg");
					break;
				}
			}

			if (count > 0)// hade aghal ye msg zade
			{
				MessageUtil.copyArray(MessageUtil
				        .intToCharArray(type.intValue(), MessageConstants.HEADER_SIZE), message,
				        0, headerPos, MessageConstants.HEADER_SIZE);
				MessageUtil.copyArray(MessageUtil.intToCharArray(count, MessageConstants.LENGTH_SIZE),
				        message, 0, headerPos + MessageConstants.HEADER_SIZE, MessageConstants.LENGTH_SIZE);
			} else {
				pointer = headerPos;
			}
		}
		// world.logFile.writeString("(" + count + " " + t.toString() + ")", 99,
		// "recievedMessage");
	}

	private void send(char[] messageArray) {
		
		MessageUtil.copyArray(MessageUtil.intToCharArray(MessageConstants.HEADER_NULL, MessageConstants.HEADER_SIZE),
		        messageArray, 0, pointer, MessageConstants.HEADER_SIZE);
		messageBytes = new byte[sendLength];
		for (int i = 0; i < sendLength; i++) {
			byte b = 0;
			for (int j = 0; j < 8; j++) {
				b = (byte) (b << 1);
				if (messageArray[i * 8 + j] == '1') {
					b++;
				}
			}
			messageBytes[i] = b;
		}

	}
	
	public byte[] getMessageBytes(){
		return messageBytes;
	}
	
	
	
//TODO 
	private char[] byteArrayToCharArray(byte[] messageBytes) {
		char[] messageArray = new char[parseLength * 8];

		for (int i = 0; i < 8 * parseLength; i++) {
			messageArray[i] = '0';
		}
		for (int i = 0; i < Math.min(messageBytes.length, parseLength); i++) {
			byte b = messageBytes[i];
			for (int j = 7; j >= 0; j--) {
				messageArray[i * 8 + j] = (b & 1) == 1 ? '1' : '0';
				b = (byte) (b >> 1);
			}
		}

		// log("\n" + new String(messageArray) + "\n");
		return messageArray;
	}

	protected void parseMessage(byte[] message, int time, StandardEntity sender,
	        boolean isCenter) {
		/*
		if (!SeuConstants.IS_COMPETETION_MODE)
			world.logFile.writeString(" ============= TIME = " + time
			        + " ,=========== importMessage from = " + sender.id, 99,
			        "recievedMessage");
		 */
		char[] messageArray = byteArrayToCharArray(message);

		boolean fromCenter =
		        (sender instanceof Building || (sender instanceof Human && ((Human) sender).isVirtualCenter));

		pointer = 0;

		while (true) {
			int header =
			        MessageUtil.charArrayToint(readMessage(messageArray, MessageConstants.HEADER_SIZE),
			                0, MessageConstants.HEADER_SIZE);
			int length = 0;
			if (header != MessageConstants.HEADER_NULL) {
				length =
				        MessageUtil.charArrayToint(readMessage(messageArray,
				                MessageConstants.LENGTH_SIZE), 0, MessageConstants.LENGTH_SIZE);
			}
			SEUMessage m = null;

			switch (header) {
				case MessageConstants.HEADER_NULL: {
					return;
				}
				case MessageConstants.HEADER_ROAD: {
					for (int i = 0; i < length; i++) {
						m =
						        new RoadMessage(readMessage(messageArray,
						                RoadMessage.staticLength()),
						                virtualIdToId);
						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedRoads, centerReportedRoads,
						                true);

						/*
						if (!SeuConstants.IS_COMPETETION_MODE)
							world.logFile.writeString("Road = " + m.obj
							        + " condition= "
							        + ((RoadMessage) m).condition
							        + " delayed = " + m.lifeCycle
							        + "  addType = " + add, 99,
							        "recievedMessage");
							        */

					}
					break;
				}
				case MessageConstants.HEADER_VICTIM: {
					for (int i = 0; i < length; i++) {
						m =
						        new VictimMessage(readMessage(messageArray,
						                VictimMessage.staticLength()),
						                virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedVictims, centerReportedVictims,
						                true);

//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile.writeString("Victim "
//							        + m.obj
//							        + "  pos= "
//							        + ((Humanoid) m.obj)
//							                .getMotionlessPosition()
//							        + "  dmg = " + ((VictimMessage) m).damage
//							        + "  bury = "
//							        + ((VictimMessage) m).buriedness
//							        + "  lifeCycle = " + m.lifeCycle
//							        + "  addType = " + add, 99,
//							        "recievedMessage");

					}
					break;
				}
				case MessageConstants.HEADER_FOUND_CIVILIAN: {
					for (int i = 0; i < length; i++) {
						m =
						        new FoundCivilianMessage(readMessage(messageArray,
						                FoundCivilianMessage.staticLength()),
						                virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedFoundCivilian,
						                centerReportedFoundCivilian, true);

//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile.writeString("FoundCivilian (" + m.obj
//							        + ") " + "  pos = "
//							        + world.get(((FoundCivMessage) m).posId)
//							        + "  dmg = " + ((FoundCivMessage) m).damage
//							        + "  bury = "
//							        + ((FoundCivMessage) m).buriedness
//							        + "  life = " + m.lifeCycle
//							        + "  addType = " + add, 99,
//							        "recievedMessage");

					}
					break;
				}
				case MessageConstants.HEADER_FIRE: {
					for (int i = 0; i < length; i++) {
						m =
						        new FireMessage(readMessage(messageArray,
						                FireMessage.staticLength()),
						                virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedFires, centerReportedFires,
						                true);

//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile.writeString("Fire " + m.obj
//							        + "  fiery = "
//							        + ((FireMessage) m).fieryness + " life = "
//							        + m.lifeCycle + "  addType = " + add, 99,
//							        "recievedMessage");

					}
					break;
				}

//				case MessageConstants.HEADER_TEMPERATURE: {
//					for (int i = 0; i < length; i++) {
//						m =
//						        new TemperatureMessage(readMessage(
//						                messageArray, TemperatureMessage
//						                        .staticLength()), virtualIdToId);
//
//						if (!m.updateWorld(world, time))
//							continue;
//						// if(!SeuConstants.IS_COMPETETION_MODE)
//						// world.logFile.writeString("Temperature " + m.obj
//						// +"  temperature = "+((TemperatureMessage)m).temperature*50
//						// +" life = "
//						// + m.lifeCycle, 99, "recievedMessage");
//
//						addToLists(fromCenter, m, isCenter,
//						        reportedTemperature, centerReportedTemperature,
//						        true);
//					}
//					break;
//				}

				case MessageConstants.HEADER_POSITION: {
					for (int i = 0; i < length; i++) {
						m =
						        new PositionMessage(readMessage(messageArray,
						                PositionMessage.staticLength()),
						                virtualIdToId);
						// if(!SeuConstants.IS_COMPETETION_MODE)
						// world.logFile.writeString("before update  Position "
						// + ((PositionMessage) m).positionId
						// +" isCenter= "+isCenter
						// +" life = " + m.lifeCycle, 99, "recievedMessage");

						if (!m.updateWorld(world, time))
							continue;

						int tmp =
						        addToLists(fromCenter, m, isCenter,
						                reportedPositions,
						                centerReportedPositions, true);

//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile.writeString("Position "
//							        + ((PositionMessage) m).positionId
//							        // +" addToList= "+tmp
//							        + " life = " + m.lifeCycle + "  addType = "
//							        + tmp, 99, "recievedMessage");

					}
					break;
				}
				case MessageConstants.HEADER_EMPTYBUILDING: {
					for (int i = 0; i < length; i++) {
						m =
						        new EmptyBuildingMessage(readMessage(
						                messageArray, EmptyBuildingMessage
						                        .staticLength()), virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedEmptyBuildings,
						                centerReportedEmptyBuidings, true);
//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile.writeString("EmptyBuilding "
//							        + m.obj.id + " life = " + m.lifeCycle
//							        + " addType = " + add, 99,
//							        "recievedMessage");

					}
					break;
				}
				case MessageConstants.HEADER_HEARD_CIVILIAN: {
					for (int i = 0; i < length; i++) {
						m =
						        new HeardCivilianMessage(readMessage(
						                messageArray, HeardCivilianMessage
						                        .staticLength()), virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedHeardCivilians,
						                centerReportedHeardCivilians, true);

//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile.writeString("HeardCivilian "
//							        + m.obj.id + " life = " + m.lifeCycle
//							        + " addType = " + add, 99,
//							        "recievedMessage");

					}
					break;
				}
				case MessageConstants.HEADER_LOCKEDBYBLOCKADE: {
					for (int i = 0; i < length; i++) {
						m =
						        new LockedByBlockadeMessage(readMessage(
						                messageArray, LockedByBlockadeMessage
						                        .staticLength()), virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedLockedByBlockades,
						                centerReportedLockedByBlockades, true);

//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile
//							        .writeString(
//							                "Locked = "
//							                        + (world.agents
//							                                .get(((LockedByBlockadeMessage) m).index))
//							                        + " delay = " + m.lifeCycle
//							                        + "  addType = " + add, 99,
//							                "recievedMessage");

					}
					break;
				}
				case MessageConstants.HEADER_UNLOCKEDBYBLOCKADE: {
					for (int i = 0; i < length; i++) {
						m =
						        new UnLockedByBlockadeMessage(readMessage(
						                messageArray, UnLockedByBlockadeMessage
						                        .staticLength()));

						if (!m.updateWorld(world, time))
							continue;

						int add =
						        addToLists(fromCenter, m, isCenter,
						                reportedUnLockedByBlockades,
						                centerReportedUnLockedByBlockades, true);

//						if (!SeuConstants.IS_COMPETETION_MODE)
//							world.logFile
//							        .writeString(
//							                "Unlocked = "
//							                        + (world.agents
//							                                .get(((UnlockedByBlockadeMessage) m).index))
//							                        + " delay = " + m.lifeCycle
//							                        + "  addType = " + add, 99,
//							                "recievedMessage");

					}
					break;
				}
				case MessageConstants.HEADER_ASSIGN_CLEAR_TARGET: {
					for (int i = 0; i < length; i++) {
						m =
						        new AssignClearTargetMessage(readMessage(
						                messageArray, AssignClearTargetMessage
						                        .staticLength()), virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;
						// if(!SeuConstants.IS_COMPETETION_MODE)
						// world.logFile.writeString("AssignedTarget "
						// +
						// world.agents.get(((AssignClearTargetMessage)m).receiverIndex)
						// + "  to = " +
						// world.get(((AssignClearTargetMessage)m).targetId)
						// +" life = " + m.lifeCycle, 99, "recievedMessage");

						addToLists(fromCenter, m, isCenter,
						        reportedPFAssign,
						        centerReportedPFAssign, false);
					}
					break;
				}
				case MessageConstants.HEADER_ASSIGN_RESCUE_TARGET: {
					for (int i = 0; i < length; i++) {
						m =
						        new AssignRescueTargetMessage(readMessage(
						                messageArray, AssignRescueTargetMessage
						                        .staticLength()), virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						addToLists(fromCenter, m, isCenter, reportedATAssign,
						        centerReportedATAssign, false);
					}
					break;
				}
				case MessageConstants.HEADER_ASSIGN_EXTINGUISH_TARGET: {
					for (int i = 0; i < length; i++) {
						m =
						        new AssignExtinguishTargetMessage(readMessage(
						                messageArray,
						                AssignExtinguishTargetMessage
						                        .staticLength()), virtualIdToId);

						if (!m.updateWorld(world, time))
							continue;

						addToLists(fromCenter, m, isCenter, reportedFBAssign,
						        centerReportedFBAssign, false);
					}
					break;
				}
				//TODO
				/*
				case MessageConstants.HEADER_HELP_TARGET: {
					for (int i = 0; i < length; i++) {
						m =
						        new HelpTargetMessage(readMessage(messageArray,
						                HelpTargetMessage.staticLength()),
						                virtualIdToId);
						if (!m.updateWorld(world, time))
							continue;
						// if(!SeuConstants.IS_COMPETETION_MODE)
						// world.logFile.writeString("Help target "
						// + world.get(((AssignClearTargetMessage)m).targetId)
						// +" life = " + m.lifeCycle, 99, "recievedMessage");

						addToLists(fromCenter, m, isCenter,
						        reportedHelpTargets, centerReportedHelpTargets,
						        true);
					}
					break;
				}
					*/
//				case MessageConstants.HEADER_FIRE_WATER: {
//					for (int i = 0; i < length; i++) {
//						m =
//						        new FireWaterMessage(readMessage(messageArray,
//						                FireWaterMessage.staticLength()),
//						                virtualIdToId);
//
////						if (seu.agent.Constants.WATER_DEBUG) {
////							// if (world.logFile != null)
////							// world.logFile.writeString("recived fire water",
////							// 1,
////							// "watermessage");
////
////							// System.out.println("recived fire water");
////						}
//
//						if (!m.updateWorld(world, time))
//							continue;
//
//						addToLists(fromCenter, m, isCenter, reportedFireWater,
//						        centerReportedFireWater, true);
//					}
//					break;
//				}
//
//				case MessageConstants.HEADER_REMAINED_WATER: {
//					for (int i = 0; i < length; i++) {
//						m =
//						        new RemainWaterMessage(readMessage(
//						                messageArray, RemainWaterMessage
//						                        .staticLength()));
//
////						if (seu.agent.Constants.WATER_DEBUG) {
////							// if(world.logFile!=null)
////							// world.logFile.writeString("recived remained water",
////							// 1,
////							// "watermessage");
////							// System.out.println("recived remained water");
////						}
//
//						if (!m.updateWorld(world, time))
//							continue;
//
//						addToLists(fromCenter, m, isCenter,
//						        reportedRemaindWater,
//						        centerReportedRemaindWater, true);
//
//					}
//					break;
//				}

			}
		}
	}

	protected int addToLists(boolean fromCenter, SEUMessage m, boolean isCenter,
	        Collection agentList, Collection centerList, boolean sendToCenters) {
		if (isCenter && discardMessages(m, agentList))
			return 0;

		m.incrementLife();// and one life to this message

		if (fromCenter) {
			centerList.add(m);
		} else {
			agentList.add(m);
		}

		if (isCenter || noCenters > 0) {
			messagesToAgents.add(m);
			if (!fromCenter && sendToCenters) {
				messagesToCenters.add(m);
			}
			return 1;
		}

		return -1;
	}

	private boolean discardMessages(SEUMessage m, Collection agentList) {
		boolean discard = false;
		for (Iterator it = agentList.iterator(); it.hasNext();) {
			SEUMessage mg = (SEUMessage) it.next();
			if (m.equals(mg)) {
				return true;
			}
		}
		if (m.getLife() > MAX_LIFE_CYCLE)
			discard = true;
		return discard;
	}

	protected void incrementDelayedMessageLife(HashSet<SEUMessage> delayedMessages) {
		for (Iterator<SEUMessage> it = delayedMessages.iterator(); it.hasNext();) {
			SEUMessage m = it.next();
			if (m.lifeCycle > MAX_LIFE_CYCLE) {
				it.remove();
				continue;
			}
			m.incrementLife();
		}
	}

	private char[] readMessage(char[] message, int length) {
		char[] result = new char[length];
		if (!MessageUtil.copyArray(message, result, pointer, 0, length))
			;

		pointer += length;
		return result;
	}



}
