package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import seu.messages.Type;
import seu.utilities.MessageUtil;


public class PositionMessage extends SEUMessage {
	protected int positionId;

	boolean isOPHead = false;

	protected int senderIndex;

	static final int indexSize = 7;

	static final int flagSize = 1;

	public PositionMessage(Area position, int senderIndex, boolean isOPHead) {
		super(Type.POSITION_REPORT, position);
		this.positionId = position.getID().getValue();
		this.senderIndex = senderIndex;
		this.isOPHead = isOPHead;
	}
	
	public PositionMessage(char[] message, Map virtualIdToId) {
		super(Type.POSITION_REPORT, null);
		this.positionId = ((Integer) virtualIdToId.get(new Integer(MessageUtil
				.charArrayToint(message, 0, idSize)))).intValue();
		this.senderIndex = MessageUtil.charArrayToint(message, idSize, indexSize);
		this.isOPHead = message[idSize + indexSize] == '1';
//		TODO Francis may be some error!
		this.lifeCycle = MessageUtil.charArrayToint(message, idSize + indexSize + flagSize, lifeSize);
	}

	@Override
	public boolean equals(Object m) {
		if (!(m instanceof PositionMessage))
			return false;
		return (((PositionMessage) m).senderIndex == senderIndex)
				&& ((PositionMessage) m).positionId == positionId;
	}

	@Override
	public int hashCode() {
		return senderIndex;
	}

	@Override
	protected int length() {
		return staticLength();
	}

	protected static int staticLength() {
		return idSize + indexSize + flagSize + lifeSize;
	}
	
	@Override
	char[] toCharArray(Map idToVirtualId) {
		char[] result = new char[length()];

		char[] temp = MessageUtil.intToCharArray(((Integer) (idToVirtualId
				.get(new Integer(positionId)))).intValue(), idSize);
		MessageUtil.copyArray(temp, result, 0, 0, idSize);

		temp = MessageUtil.intToCharArray(senderIndex, indexSize);
		MessageUtil.copyArray(temp, result, 0, idSize, indexSize);

		temp = new char[1];

		if (isOPHead)
			temp[0] = '1';
		else
			temp[0] = '0';

		MessageUtil.copyArray(temp, result, 0, idSize + indexSize, flagSize);
		
		temp = MessageUtil.intToCharArray(lifeCycle, lifeSize);
		MessageUtil.copyArray(temp, result, 0, idSize + indexSize + flagSize , lifeSize);
		
		return result;
	}
	
	@Override
	//TODO
	public boolean updateWorld(StandardWorldModel world, int time) {
		Human hum = (Human) world.getState().agents.get(senderIndex);
		
		int messageTime = time - lifeCycle;

		if (hum.getState().isVisible(world)) {
			return false;
		}
		if(!hum.getState().shouldUpdate(messageTime)){
			return false;
		} else {
			hum.getState().setProperty(world, messageTime, hum.getPositionProperty(),
					new int[] { positionId });
			StandardEntity en = (StandardEntity) world.getEntity(new EntityID(positionId));
			if (en instanceof Road) {
				Road rd = (Road) en;
				
				//TODO
//				if (isOPHead) {
//					hum.getState().setProperty(world, messageTime, PROPERTY_POSITION_EXTRA,
//							new int[] { 0 });
//				} else {
//					hum.getState().setProperty(world, messageTime, PROPERTY_POSITION_EXTRA,
//							new int[] { rd.getState().getLength(world) - 1 });
//				}
			}
			
			hum.getState().setTime(messageTime);
			
			return true;
		}
	}
}
