package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import seu.messages.Type;
import seu.states.RoadState;
import seu.utilities.MessageUtil;

public class RoadMessage extends SEUMessage {

	private int roadId;
    /**	 0 is cleared; 1 is passable; 2 is blocked*/
	public int condition; 
	
	static final int conditionSize = 2;
	
	public RoadMessage(Road road, int condition) {
		super(Type.ROAD_REPORT, road);
		roadId = obj.getID().getValue();
		this.condition = condition;
	}

	public RoadMessage(char[] message, Map virtualIdToId)
    {
        super(Type.ROAD_REPORT, null);
        this.roadId = ((Integer) virtualIdToId.get(new Integer(MessageUtil.charArrayToint(message, 0, idSize)))).intValue();
        this.condition = MessageUtil.charArrayToint(message, idSize, conditionSize);
        this.lifeCycle = MessageUtil.charArrayToint(message, idSize + conditionSize, lifeSize);
    }
	
	@Override
	public boolean equals(Object m) {
		if (!(m instanceof RoadMessage)) return false;
//	     not needed to decide lifecyle.
        return (((RoadMessage) m).roadId == roadId)
        		&& ((RoadMessage) m).condition == condition;
	}

	@Override
	public int hashCode() {
		return roadId;
	}

	@Override
	protected int length() {
		return staticLength();
	}
	
	 protected static int staticLength()
	 {
	    return idSize + conditionSize + lifeSize;
	 }

	@Override
	 char[] toCharArray(Map idToVirtualId)
	  {
	        char[] result = new char[length()];

	        char[] temp = MessageUtil.intToCharArray(((Integer) (idToVirtualId.get(new Integer(roadId)))).intValue(), idSize);
	        MessageUtil.copyArray(temp, result, 0, 0, idSize);
	        
	        temp = MessageUtil.intToCharArray(condition, conditionSize);
	        MessageUtil.copyArray(temp, result, 0, idSize, conditionSize);
	        
	        temp = MessageUtil.intToCharArray(lifeCycle, lifeSize);
	        MessageUtil.copyArray(temp, result, 0, idSize + conditionSize, lifeSize);

	        return result;
	    }

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		obj = world.getState().get(new EntityID(roadId));
        Road road = (Road) obj;
        
        int messageTime = time - lifeCycle;
        
        if(!road.getState().shouldUpdate(messageTime)){
        	return false;
        }
        
        if (condition == RoadState.CONDITION_BLOCKED
				&& road.getState().getCondition() == RoadState.CONDITION_PASSABLE) {
			return false;
		}
        
        if (condition == RoadState.CONDITION_PASSABLE
				|| condition == RoadState.CONDITION_MAYBE_PASSABLE) {
			if (road.getState().reportClear.getCount() > 1)
				return false;
			road.getState().setBlock(0, messageTime,world);
			world.getState().blockRoads.remove(road);
			road.getState().setCondition(RoadState.CONDITION_PASSABLE);

			road.getState().reportClear.update(messageTime);
		}
		
		if(condition == RoadState.CONDITION_BLOCKED){
			if(road.getState().reportUnClear.getCount() > 1)
				return false;
			road.getState().setBlock(500, messageTime,world);
        	world.getState().blockRoads.add(road);
        	road.getState().setCondition(RoadState.CONDITION_BLOCKED);
        	
			road.getState().reportUnClear.update(messageTime);
		}
        if(!road.getState().getContainedPaths(world).isEmpty()){
        	road.getState().getContainedPaths(world).get(0).calculateLength(world);
        }
        
        
		return true;
	}

}
