//SEU2010

package seu.messages;


import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

import rescuecore2.connection.Connection;
import rescuecore2.connection.ConnectionException;
import rescuecore2.messages.components.IntComponent;
import rescuecore2.standard.entities.*;
import rescuecore2.standard.messages.AKSay;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.EntityID;
import seu.SeuConstants;
import seu.states.WorldState;
import seu.utilities.Condition;

public class SayMessageManager implements MessageConstants {

	protected StandardWorldModel world;
	public WorldState worldState;
	static final int MAX_LENGTH = 256;
	byte[] messageBytes = null;
	
	private int pointer;
	protected EntityID selfId;
	protected int[] tellChannels;
	protected int messageCount;
	
	public ArrayList reportedGiveWay;
	public ArrayList reportedVictims;
	public ArrayList reportedFoundCivilian;
	public ArrayList reportedFires;
	public ArrayList reportedClears;
	public ArrayList reportedPositions;
	public ArrayList reportedUnClears;
	public ArrayList reportedBlocks;
	public ArrayList reportedEmptyBuildings;
	public ArrayList reportedHeardCivilians;
	public ArrayList reportedLockedByBlockades;
	public ArrayList reportedRequestPartitionChangings;
//	public ArrayList reportedPartitionFinisheds;
//	public ArrayList reportedSetPartitions;
	public ArrayList reportedClearedPaths;
	public ArrayList reportedTargets;
	
	protected Map<Integer, Integer> idToVirtualId;

	protected Map<Integer, Integer> virtualIdToId;

	protected HashSet<SEUMessage> sayMessages;
	
	protected static int MAX_LIFE_CYCLE = 8;

	private int test_received_msg_when_only_say = 0;
	
	public HashSet<SEUMessage> getSayMessages(){return sayMessages;}

	private final boolean CLOSELOG = true;//for competition
	protected FileWriter fw;
	public static final String baseLogFileName = "c:/";
	

	public SayMessageManager(StandardWorldModel world, EntityID entityID) {
		
		this.selfId = entityID;
		this.messageCount = SAY_MSG_COUNT;
		this.world = world;
		this.worldState = world.getState();
		
		idToVirtualId = new HashMap<Integer, Integer>();
		virtualIdToId = new HashMap<Integer, Integer>();

		int virtualId = 0;
		for (Iterator i = world.getEntitiesOfType(StandardEntityURN.BUILDING, StandardEntityURN.ROAD).iterator(); i.hasNext(); virtualId++) {
			Area ar = (Area) i.next();
			idToVirtualId.put(new Integer(ar.getID().getValue()),
					new Integer(virtualId));
			virtualIdToId.put(new Integer(virtualId),
					new Integer(ar.getID().getValue()));
		}

		reportedGiveWay = new ArrayList();
		reportedFoundCivilian = new ArrayList();
		reportedVictims = new ArrayList();
		reportedClears = new ArrayList();
		reportedFires = new ArrayList();
		reportedPositions = new ArrayList();
		reportedUnClears = new ArrayList();
		reportedBlocks = new ArrayList();
		reportedEmptyBuildings = new ArrayList();
		reportedHeardCivilians = new ArrayList();
		reportedLockedByBlockades = new ArrayList();
		reportedClearedPaths = new ArrayList();
		reportedTargets = new ArrayList();
		
		sayMessages = new HashSet<SEUMessage>();
		
		int extra_max_cycles = 0;
		if (world.getEntitiesOfType(StandardEntityURN.FIRE_STATION).isEmpty()) {
			extra_max_cycles = Math.max(extra_max_cycles, 
					world.getEntitiesOfType(StandardEntityURN.FIRE_STATION).size()/4 
					+ (world.getEntitiesOfType(StandardEntityURN.FIRE_STATION).size()%4 > 0 ? 1 : 0));
		} 
		if (world.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE).isEmpty()) {
			extra_max_cycles = Math.max(extra_max_cycles,
					world.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE).size()/4 
					+ (world.getEntitiesOfType(StandardEntityURN.AMBULANCE_CENTRE).size()%4 > 0 ? 1 : 0));
		} 
		if (world.getEntitiesOfType(StandardEntityURN.POLICE_OFFICE).isEmpty()) {
			extra_max_cycles = Math.max(extra_max_cycles, 
					world.getEntitiesOfType(StandardEntityURN.POLICE_OFFICE).size()/4
					+ (world.getEntitiesOfType(StandardEntityURN.POLICE_OFFICE).size()%4 > 0 ? 1 : 0));
		}
		
		MAX_LIFE_CYCLE += extra_max_cycles;
		
		if(!CLOSELOG){
			
			try {
				
				String myType = "";
				if(worldState.me instanceof AmbulanceTeam)
					myType += "at-" + worldState.me.getID().getValue();
				else if(worldState.me instanceof PoliceForce)
					myType += "pf-" + worldState.me.getID().getValue();
				else if(worldState.me instanceof FireBrigade)
					myType += "fb-" + worldState.me.getID().getValue();
				
				String filePath = baseLogFileName+myType;
				
				fw = new FileWriter(new File(filePath+"-say.log"));
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
	}

	//
	//param path 
	public static boolean delAllFile() {
		boolean flag = false;
		File file = new File(baseLogFileName);
		if (!file.exists()) {
			return flag;
		}
		if (!file.isDirectory()) {
			return flag;
		}
		String[] tempList = file.list();
		File temp = null;
		for (int i = 0; i < tempList.length; i++) {
			if (baseLogFileName.endsWith(File.separator)) {
				temp = new File(baseLogFileName + tempList[i]);
			} else {
				temp = new File(baseLogFileName + File.separator + tempList[i]);
			}
			if (temp.isFile()) {
				temp.delete();
			}
		}
		return flag;
	}

	
	/**
	 * This method can be called only once per cycle! otherwise consult your
	 * psychiretrist
	 */
	protected void clearForNextCycle() {
		reportedGiveWay.clear();
		reportedFoundCivilian.clear();
		reportedVictims.clear();
		reportedClears.clear();
		reportedFires.clear();
		reportedPositions.clear();
		reportedUnClears.clear();
		reportedBlocks.clear();
		reportedEmptyBuildings.clear();
		reportedHeardCivilians.clear();
		reportedLockedByBlockades.clear();
		reportedTargets.clear();
	}

	public void sendMessages(int tempCount,  Collection messages) {
		//System.out.println(selfId+" "+tempCount+"at"+world.getTime());
		if (tempCount == SAY_MSG_COUNT) {
			return;
		}
		
		char[] message = new char[MAX_LENGTH * 8];
		pointer = 0; // where to write in char [] message
		
		addMessage(Type.GIVE_WAY_REPORT, message, messages);
		addMessage(Type.LOCKEDBYBLOCKADE_REPORT, message, messages);
		addMessage(Type.UNLOCKEDBYBLOCKADE_REPORT, message, messages);
		addMessage(Type.POSITION_REPORT, message, messages);
		addMessage(Type.VICTIM_REPORT, message, messages);
		addMessage(Type.FOUND_CIVILIAN_REPORT, message, messages);
		addMessage(Type.HEARD_CIVILIAN_REPORT, message, messages);
		addMessage(Type.FIRE_REPORT, message, messages);
		addMessage(Type.ROAD_REPORT, message, messages);
		addMessage(Type.EMPTYBUILDING_REPORT, message, messages);
		
		send(message);
		if (messages.size() > 0 && ++tempCount < messageCount) {
			sendMessages(tempCount, messages);
		}
	}

	protected void log(String message) {
		if(CLOSELOG) return;
		try {
			fw.write(message);
			fw.flush();
		} catch (IOException e) {
			e.printStackTrace(); 
		}
	}

	
	private void addMessage(Type type, char[] message, Collection messages) {
		final Type t = type;
		Condition c = new Condition() {
			public boolean eval(Object obj) {
				if (((SEUMessage) obj).type == t)
					return true;
				return false;
			}
		};
		ArrayList temp = c.extractMAR(messages);
		int count = 0; // tedade message hayi ke khahad ferestad.
		
		if (!temp.isEmpty()) {
			int headerPos = pointer; // jayi ke bayad header va length ezafe
										// beshan ro nigah midare.
			pointer += HEADER_SIZE + LENGTH_SIZE; // baraye inke badan biad
													// header va length ro
													// benevise ja mizare.

			for (Iterator i = temp.iterator(); i.hasNext();) {
				SEUMessage m = (SEUMessage) i.next();
				if (MAX_LENGTH * 8 - pointer - HEADER_SIZE >= m.length()) // ja
																			// baraye
																			// m
																			// hast
				{
					count++;
					i.remove();// msg e ferestade shode ro hazf mikone
					Util.copyArray(m.toCharArray(idToVirtualId), message, 0,
							pointer, m.length());
					pointer += m.length();
				} else {
					// msg haye ferestade nashode hanooz too list hastan.
					messages.addAll(temp);
					break;
				}
			}

			if (count > 0)// hade aghal ye msg zade
			{
				Util.copyArray(Util
						.intToCharArray(type.intValue(), HEADER_SIZE), message,
						0, headerPos, HEADER_SIZE);
				Util.copyArray(Util.intToCharArray(count, LENGTH_SIZE),
						message, 0, headerPos + HEADER_SIZE, LENGTH_SIZE);
			} else {
				pointer = headerPos;
			}
		}
		
	}

	private void send(char[] messageArray) {
		Util.copyArray(Util.intToCharArray(HEADER_NULL, HEADER_SIZE),
				messageArray, 0, pointer, HEADER_SIZE);
		messageBytes = new byte[MAX_LENGTH];
		for (int i = 0; i < MAX_LENGTH; i++) {
			byte b = 0;
			for (int j = 0; j < 8; j++) {
				b = (byte) (b << 1);
				if (messageArray[i * 8 + j] == '1') {
					b++;
				}
			}
			messageBytes[i] = b;
		}

	}

	public byte[] getMessageBytes(){
		return messageBytes;
	}
	
	private char[] byteArrayToCharArray(byte[] messageBytes) {
		char[] messageArray = new char[MAX_LENGTH * 8];

		for (int i = 0; i < 8 * MAX_LENGTH; i++) {
			messageArray[i] = '0';
		}

		for (int i = 0; i < messageBytes.length; i++) {
			byte b = messageBytes[i];
			for (int j = 7; j >= 0; j--) {
				messageArray[i * 8 + j] = (b & 1) == 1 ? '1' : '0';
				b = (byte) (b >> 1);
			}
		}

		return messageArray;
	}

	public void parseMessage(byte[] message, int time, StandardEntity sender) {
//		if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString(" ============= TIME = " + time + " ,=========== importMessage from = "
//				+ sender.id, 12, "sayMessage");

		char[] messageArray = byteArrayToCharArray(message);

		
		
		pointer = 0;

		while (true) {
			int header = Util.charArrayToint(readMessage(messageArray,
					HEADER_SIZE), 0, HEADER_SIZE);
			int length = 0;
			if (header != HEADER_NULL) {
				length = Util.charArrayToint(readMessage(messageArray,
						LENGTH_SIZE), 0, LENGTH_SIZE);
				
//				if(time>270)
//				log(time + "  "	+ world.me+"  "+test_received_msg_when_only_say+"\n");
			}
			SEUMessage m = null;
			switch (header) {
			case HEADER_NULL: {
				return;
			}
			case HEADER_ROAD: {
				for (int i = 0; i < length; i++) {
					m = new RoadMessage(readMessage(messageArray, RoadMessage
							.staticLength()), virtualIdToId);
					if(!m.updateWorld(world, time))
						return;
//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("Road = "+((RoadMessage)m).obj, 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedClears);
				}
				break;
			}
			case HEADER_VICTIM: {
				for (int i = 0; i < length; i++) {
					m = new VictimMessage(readMessage(messageArray,
							VictimMessage.staticLength()), virtualIdToId);
					if(!m.updateWorld(world, time))
						return;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("VictimMessage =  " + m.obj + " bury "
//							+ ((VictimMessage) m).buriedness, 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedVictims);
				}
				break;
			}
			case HEADER_FOUND_CIVILIAN: {
				for (int i = 0; i < length; i++) {
					m = new FoundCivilianMessage(readMessage(messageArray,
							FoundCivilianMessage.staticLength()), virtualIdToId);
					if(!m.updateWorld(world, time))
						return;

//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("FoundCivilianMessage " + m.obj.id + " position "
//							+ ((FoundCivMessage) m).posId + " buriedness "
//							+ ((FoundCivMessage) m).buriedness, 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedFoundCivilian);
				}
				break;
			}
			case HEADER_FIRE: {
				for (int i = 0; i < length; i++) {
					m = new FireMessage(readMessage(messageArray, FireMessage
							.staticLength()), virtualIdToId);
					if(!m.updateWorld(world, time))
						return;

//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("FireMessage " + m.obj, 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedFires);
					Building b = (Building) world.getEntity(
							new EntityID(((FireMessage) m).getBuildingId()));
				}
				break;
			}
			case HEADER_POSITION: {
				for (int i = 0; i < length; i++) {
					m = new PositionMessage(readMessage(messageArray,
							PositionMessage.staticLength()), virtualIdToId);
					if(!m.updateWorld(world, time))
						return;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("Position("
//							+ ((PositionMessage) m).index
//							+ "="
//							+ ((Humanoid) (world.agents
//									.get(((PositionMessage) m).index))).id
//							+ ") " + ((PositionMessage) m).positionId, 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedPositions);
				}
				break;
			}
			case HEADER_EMPTYBUILDING: {
				for (int i = 0; i < length; i++) {
					m = new EmptyBuildingMessage(readMessage(messageArray,
							EmptyBuildingMessage.staticLength()), virtualIdToId);
					if(!m.updateWorld(world, time))
						return;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("EmptyBuilding " + m.obj.id, 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedEmptyBuildings);
				}
				break;
			}
			case HEADER_HEARD_CIVILIAN: {
				for (int i = 0; i < length; i++) {
					m = new HeardCivilianMessage(readMessage(messageArray,
							HeardCivilianMessage.staticLength()), virtualIdToId);
					if(!m.updateWorld(world, time))
						return;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("HeardCivilian: " + m.obj.id + " in Position: "
//							+ ((HeardCivilianMessage) m).posId, 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedHeardCivilians);
				}
				break;
			}
			case HEADER_LOCKEDBYBLOCKADE: {
				for (int i = 0; i < length; i++) {
					m = new LockedByBlockadeMessage(readMessage(messageArray,
							LockedByBlockadeMessage.staticLength()),
							virtualIdToId);
					if(!m.updateWorld(world, time))
						return;
					
					test_received_msg_when_only_say ++;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("LockedByBlockade("
//							+ ((LockedByBlockadeMessage) m).index
//							+ "="
//							+ ((Humanoid) (world.agents
//									.get(((LockedByBlockadeMessage) m).index)))
//							+ ") ", 12, "sayMessage");
					

					addToLists(m, reportedLockedByBlockades);
				}
				break;
			}
			case HEADER_UNLOCKEDBYBLOCKADE: {
				for (int i = 0; i < length; i++) {
					m = new UnLockedByBlockadeMessage(readMessage(messageArray,
							UnLockedByBlockadeMessage.staticLength()));
					if(!m.updateWorld(world, time))
						return;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//						world.logFile.writeString("UnlockedByBlockade: "
//							+ ((Humanoid) (world.agents
//									.get(((UnlockedByBlockadeMessage) m).index))), 12, "sayMessage");
					
					test_received_msg_when_only_say ++;
					
					addToLists(m, reportedLockedByBlockades);
				}
				break;
			}
			case HEADER_GIVE_WAY: {
//				 log(" " + Type.CLEAR_REPORT.toString() + length);
				for (int i = 0; i < length; i++) {
					m = new GiveWayMessage(readMessage(messageArray, GiveWayMessage
							.staticLength()));
					if(!m.updateWorld(world, time))
						return;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//					world.logFile.writeString(world.me+"  received give way", 12, "sayMessage");
					
				}
				break;
			}
			}
		}
	}
	
	/** 1: give way, 0 : not give way. MUST be 1 when to report.
	 * @param giveWay
	 */
	public void reportGiveWay() {
			SEUMessage message = new GiveWayMessage();
//			if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString(world.me+" said give way", 56, "sendMsg");
			sayMessages.add(message);
	}
	

	public void reportFoundCivilian(Civilian civ) {
		if (civ != null) {
			SEUMessage message = new FoundCivilianMessage(civ);
			sayMessages.add(message);
		}
	}

	public void reportFoundCivilian(Collection civs) {
		for (Iterator it = civs.iterator(); it.hasNext();) {
			Civilian c = (Civilian) it.next();
			SEUMessage message = new FoundCivilianMessage(c);
			sayMessages.add(message);
		}
	}

	//   

	public void reportVictim(Human victim) {
		if (victim != null) {
			SEUMessage message = new VictimMessage(victim);
			sayMessages.add(message);
		}
	}
	
	public void reportFire(Building building) {
		if (building != null) {
			SEUMessage message = new FireMessage(world, building, worldState.agents
					.indexOf(worldState.me));
			sayMessages.add(message);
		}
	}
	
//	public void reportTemperature(Building building){
//		if(building != null){
//			if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString(building + "temperature = "
//					+ building.getTemperature() + " time = " 
//					+ building.getTemperatureList().getLast().getTime()
//					, 56, "sendMsg");
//			
//			SEUMessage message = new TemperatureMessage(building);
//			
//			sayMessages.add(message);
//		}
//	}

	public void reportFire(Collection building) {
		for (Iterator iterator = building.iterator(); iterator.hasNext();) {
			Building b = (Building) iterator.next();
			SEUMessage message = new FireMessage(world, b, worldState.agents
					.indexOf(worldState.me));
			sayMessages.add(message);
		}
	}
	
//	public void reportTemperature(Collection building) {
//		for (Iterator iterator = building.iterator(); iterator.hasNext();) {
//			Building b = (Building) iterator.next();
//			SEUMessage message = new TemperatureMessage(b);
//			sayMessages.add(message);
//		}
//	}

	public void reportRoad(Road road, int condition) {
		if (road != null) {
			SEUMessage message = new RoadMessage(road,condition);
			sayMessages.add(message);
		}
	}

	public void reportEmptyBuilding(Building building) {
		if (building != null) {
			SEUMessage message = new EmptyBuildingMessage(building);
			sayMessages.add(message);
		}
	}

	public void reportEmptyBuilding(Collection buildings) {
		for (Iterator iterator = buildings.iterator(); iterator.hasNext();) {
			Building b = (Building) iterator.next();
			if (b instanceof Building) {
				SEUMessage message = new EmptyBuildingMessage((Building) b);
				sayMessages.add(message);
			}
		}
	}

	public void reportPosition(Human hum) {
		if (hum == null || hum.getPosition(world) == null) {
			return;
		}

		// Collections.sort(world.agents, yab.agent.Util.idComparator);
		boolean isOPHead = false;
		if(((Human)worldState.me).getPosition(world) instanceof Road)
		{
			Road r = (Road) ((Human)worldState.me).getPosition(world);
			if(worldState.passedRoads.contains(r.getState().getPointHead(world)))
				isOPHead = true;
		}
		SEUMessage message = new PositionMessage((Area)hum.getPosition(world), worldState.agents
				.indexOf(worldState.me), isOPHead);
		sayMessages.add(message);
	}

	public void reportHeardCivilian(Civilian civilian,
			Area position, int positionExtra) {
		SEUMessage message = new HeardCivilianMessage(civilian, position,
				positionExtra);
		sayMessages.add(message);
	}

	public void reportLockedByBlockade() {
		if (!(worldState.me instanceof Human)) {
			return;
		}
		Collections.sort(worldState.agents,seu.agent.Util.idComparator);
		SEUMessage message = new LockedByBlockadeMessage(worldState.agents
				.indexOf(worldState.me));
		sayMessages.add(message);
	}

	public void reportUnlockedByBlockade() {
		if (!(worldState.me instanceof Human))
			return;
		Collections.sort(worldState.agents, seu.agent.Util.idComparator);
		SEUMessage message = new UnLockedByBlockadeMessage(worldState.agents
				.indexOf(worldState.me));
		sayMessages.add(message);
	}
	
	protected int addToLists(SEUMessage m, Collection agentList) {
		if(discardMessages(m,agentList)) return 0;
		
		if(m.type == Type.GIVE_WAY_REPORT) // giveway message donot report to others.
			return 0;
		
		m.incrementLife();// and one life to this message

		sayMessages.add(m);
		
		return 1;
	}
	
	private boolean discardMessages(SEUMessage m, Collection agentList){
		boolean discard = false;
		for(Iterator it=agentList.iterator();it.hasNext();){
			SEUMessage mg = (SEUMessage)it.next();
			if(m.equals(mg)){
				discard = true;
				break;
			}
		}
		if(m.getLife()>MAX_LIFE_CYCLE) discard=true;
		return discard;
	}

	private char[] readMessage(char[] message, int length) {
		char[] result = new char[length];
		if (!Util.copyArray(message, result, pointer, 0, length))
			;

		pointer += length;
		return result;
	}
	
	public void importMessage(byte[] message, int time, StandardEntity sender) {
		if (sender == worldState.me)
			return;
		parseMessage(message, time, sender);
	}

	public void sendMessages() {
		if(!sayMessages.isEmpty())
			sendMessages(0, sayMessages);
		clearForNextCycle();
		sayMessages.clear();
	}
	
	public void sendAdditionalMessages(HashSet<SEUMessage> addtionMsgs) {
		sayMessages.addAll(addtionMsgs);
		sendMessages(0, sayMessages);
		clearForNextCycle();
		sayMessages.clear();
	}
}
