package seu.messages;

import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import seu.messages.Type;
import seu.messages.Util;

public class VictimMessage extends SEUMessage
{
	protected int realId;
    protected int posId;
    protected int hp;
    protected int damage;
    protected int buriedness;
//    private int time;

    static final int realIdSize = 32;
    static final int hpSize = 4;
    static final int damageSize = 8;
    static final int buriednessSize = 6;
    static final int timeSize = 9;
    static final int HP_BASE = 1000;

    public VictimMessage(Human victim)
    {
        super(Type.VICTIM_REPORT, victim);
        this.realId = victim.getID().getValue();
        this.posId = victim.getPosition().getValue();
        this.hp = victim.getHP();
        this.damage = victim.getDamage();
        this.buriedness = victim.getBuriedness();
//        this.time = time;
    }
    
    public VictimMessage(Human victim, Area pos)
    {
        super(Type.VICTIM_REPORT, victim);
        this.realId = victim.getID().getValue();
        this.posId = pos.getID().getValue();
        this.hp = victim.getHP();
        this.damage = victim.getDamage();
        this.buriedness = victim.getBuriedness();
//        this.time = time;
    }

    public VictimMessage(char[] message, Map virtualIdToId)
    {
        super(Type.VICTIM_REPORT, null);
        this.realId = Util.charArrayToint(message, 0, realIdSize);
        this.posId = ((Integer) virtualIdToId.get(new Integer(Util.charArrayToint(message, realIdSize, idSize)))).intValue();
        this.hp = Util.charArrayToint(message, realIdSize + idSize, hpSize)*HP_BASE;
        this.damage = Util.charArrayToint(message, realIdSize + idSize + hpSize, damageSize);
        this.buriedness = (Util.charArrayToint(message, realIdSize + idSize + hpSize + damageSize, buriednessSize));
//        this.time = Util.charArrayToint(message, realIdSize + idSize + hpSize + damageSize + buriednessSize, timeSize);
        this.lifeCycle = Util.charArrayToint(message, realIdSize + idSize + hpSize + damageSize + buriednessSize, lifeSize);
    }
    
	@Override
	public boolean equals(Object m) {
		if (!(m instanceof VictimMessage)) return false;
        return realId == ((VictimMessage) m).realId
        		&& posId == ((VictimMessage) m).posId
        		&& damage == ((VictimMessage) m).damage
        		&& hp == ((VictimMessage)m).hp
        		&& buriedness == ((VictimMessage)m).buriedness;
	}

	@Override
	public int hashCode() {
		return realId;
	}

	@Override
	protected int length() {
		return staticLength();
	}

	protected static int staticLength()
    {
        return realIdSize + idSize + hpSize + damageSize + buriednessSize + lifeSize;
    }
	
	@Override
	char[] toCharArray(Map idToVirtualIds) {
		char[] result = new char[length()];
        char[] temp;

        temp = Util.intToCharArray(realId, realIdSize);
        Util.copyArray(temp, result, 0, 0, realIdSize);
        
        temp = Util.intToCharArray(((Integer) idToVirtualIds.get(new Integer(posId))).intValue(), idSize);
        Util.copyArray(temp, result, 0, realIdSize, idSize);

        temp = Util.intToCharArray(hp/HP_BASE, hpSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize, hpSize);

        temp = Util.intToCharArray(damage, damageSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize + hpSize, damageSize);

        temp = Util.intToCharArray(buriedness, buriednessSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize + hpSize + damageSize, buriednessSize);

//        temp = Util.intToCharArray(time, timeSize);
//        Util.copyArray(temp, result, 0, realIdSize + idSize + hpSize + damageSize + buriednessSize, timeSize);
        
        temp = Util.intToCharArray(lifeCycle, lifeSize);
        Util.copyArray(temp, result, 0, realIdSize + idSize + hpSize + damageSize + buriednessSize, lifeSize);

        return result;
	}

	@Override
	public boolean updateWorld(StandardWorldModel world, int time) {
		obj = world.getEntity(new EntityID(realId));

		int messageTime = time - lifeCycle;
		
		if (obj == null){
            obj = StandardEntityFactory.INSTANCE.makeEntity(StandardEntityURN.CIVILIAN, new EntityID(realId));
            world.addEntity(obj);
            world.getState().add(new EntityID(realId));
            obj.getState().setTime(0);
        }
		
		if(world.getState().savedHumanoids.contains(obj))
			return false;
		
		if (!((Human) obj).getState().shouldUpdate(messageTime))
			return false;
		
		if(obj instanceof Human){
			Human hum = (Human) obj;
			hum.getState().setTime(messageTime);
	      hum.getState().setProperty(world, messageTime, hum.getPositionProperty(), posId);
			hum.getState().setProperty(world, messageTime, hum.getHPProperty(), hp);
			hum.getState().setProperty(world, messageTime, hum.getDamageProperty(), damage);
			hum.getState().setProperty(world, messageTime, hum.getBuriednessProperty(), buriedness);
			
			if (obj instanceof Civilian) {
				Civilian civ = (Civilian) obj;
				civ.getState().setFound();// important
				
				if(civ.getPosition(world) instanceof Refuge)
					world.getState().savedHumanoids.add(civ);
				
				civ.getState().setParticle(world, messageTime, hp, damage, buriedness, 1);
			}
		}
		
		

		return true;
	}

}
