//SEU2010

package seu.states;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.Property;
import seu.SeuConstants;
import seu.SeuParameter;
import seu.agent.Constants;
import seu.agent.Util;
import seu.entities.Path;
import seu.entities.Report;
import seu.utilities.ConstantConditions;
import seu.utilities.Wall;
import seu.utilities.base.SingleData;

public class BuildingState extends AreaState {
	private Building me;

	public int moreExtinguishCount = 0;
	public int extraExtinguishTime = 0;
	int m_totalDistanceToVictims;
	private SingleData<Integer> waterQuantity;
	
	public BuildingState(Building me) {
		super(me);
		this.me = me;
		estimatedEnergy = new SingleData<Float>(0f, -1000);
		estimatedFuel = new SingleData<Float>(0f, -1000);
		estimatedTemperature = new SingleData<Float>(0f, -1000);
		waterQuantity = new SingleData<Integer>(0, -1000);
		m_fieryness = new LinkedList<SingleData<Integer>>();
		m_temperature = new LinkedList<SingleData<Integer>>();
	}

//	public int getType() {
//		return SeuConstants.TYPE_BUILDING;
//	}

	// from BaseBuilding
	// ================================ basic Variables of building
	// ==================================//
	protected LinkedList<SingleData<Integer>> m_fieryness;
	protected LinkedList<SingleData<Integer>> m_temperature;
	private int[] m_buildingApexes;
	
	// public void setIgnition(int value) { m_ignition = (value != 0); }
	public void setFieryness(int value, int time){
		Util.updateSingleDataList(m_fieryness, value, time, true, 
				SeuParameter.LIST_LENGTH_BUILDING_FIERYNESS);
	}

	public void setTemperature(int value, int time){
		if (value <20) value = 20;
		Util.updateSingleDataList(m_temperature, value, time, true,
				SeuParameter.LIST_LENGTH_BUILDING_TEMPERATURE);
	}
	
	//apexes
	public void setBuildingApexes(int[] value){
		m_buildingApexes = value;
	}
	
	public int[] getBuildingApexes(){
		return m_buildingApexes;
	}
	
	public LinkedList<SingleData<Integer>> getTemperatureList (){
		return m_temperature;
	}

	/*
	 * curCyc for temperature
	 */
	public int getTemperature(){
		return m_temperature.getLast().getObject();
	}

	public int getTemperature(int time){
		return Util.searchSingleDataForTime(m_temperature, time);
	}
	
	public List<SingleData<Integer>> getAllTemperature(){
		return m_temperature;
	}

	/*
	 * curCyc for fieryness
	 */
	public int getFieryness(){
		return m_fieryness.getLast().getObject();
	}

	public int getFieryness(int time){
		return Util.searchSingleDataForTime(m_fieryness, time);
	}

	public List<SingleData<Integer>> getAllFieryness(){
		return m_fieryness;
	}
		
	public boolean isNoBurned(){
		if(!me.isFierynessDefined()) return true;
		return (me.getFieryness() == 0);
	}

//	public boolean isNoBurned(int time){
//		return (getFieryness(time) == 0);
//	}

	public boolean isBurning(){
		if(!me.isFierynessDefined()) return false;
		return (1 <= me.getFieryness() && me.getFieryness() <= 3);
	}

//	public boolean isBurning(int time){
//		if(!me.isFierynessDefined()) return false;
//		return (1 <= getFieryness(time) && getFieryness(time) <= 3);
//	}

	public boolean isUnburned(){
		if(!me.isFierynessDefined()) return true;
		return me.getFieryness() == 0 || me.getFieryness() == 4;
	}

	public boolean isPutOut(){
		if(!me.isFierynessDefined()) return false;
		return 5 <= me.getFieryness() && me.getFieryness() <= 7;
	}

	public boolean isBurned(){
		if(!me.isFierynessDefined()) return false;
		return me.getFieryness() > 7;
	}

	@Override
	public String toString(){
		return super.toString();
	}
	
	// ==============================================================================================//

	@Override
	public List<Path> getContainedPaths(StandardWorldModel world) {
		if (m_containedPaths == null) {
			m_containedPaths = new ArrayList<Path>();
			for (Path path : world.getState().paths) {
				if (path.getContainedRoads().contains(this)) {
					m_containedPaths.add(path);
				}
			}
		}
		return m_containedPaths;
	}
	public List<Path> getContainedPaths(List<Path> pathss) {
		if (m_containedPaths == null) {
			m_containedPaths = new ArrayList<Path>();
			for (Path path : pathss) {
				if (path.getContainedRoads().contains(this)) {
					m_containedPaths.add(path);
				}
			}
		}
		return m_containedPaths;
	}
	@Override
	public Collection getNeighborhood(StandardWorldModel world) {
		return null;
	}

	//===================================================================================================//

	//
	public void setTotalDistanceToVictims(int distance){
		m_totalDistanceToVictims = distance;
	}

	private List<Path> m_containedPaths = null;

	private List<Building> m_ExtinguishableNeighbours = null;

	public List<Building> getExtinguishableNeighbours(StandardWorldModel world)
	{
		if (m_ExtinguishableNeighbours == null)
		{
			m_ExtinguishableNeighbours = new ArrayList<Building>();
			for (Building b : world.getState().buildings)
			{
				if (world.getDistance(me,b) <= Constants.EXTINGUISHABLE_DISTANCE)
					m_ExtinguishableNeighbours.add(b);
			}
		}
		return m_ExtinguishableNeighbours;
	}

	public Report emptyBuildingReport = new Report();

	private final HashSet<Civilian> possibleCivilians = new HashSet<Civilian>();

	public void addPossibleCivilians(StandardWorldModel world, Civilian civ)
	{
		if (civ.getPosition(world) != null)
		{
			if (possibleCivilians.contains(civ))
				possibleCivilians.remove(civ);
			return;
		}
		possibleCivilians.add(civ);
	}

	public double getCivPossibility(StandardWorldModel world)
	{
		if (!world.getState().notSeenBuildings.contains(this) || this.isBurned())
			return 0;

		double base_value = 80.0 / world.getState().buildings.size();

		if (hearableCount < 50)
			return m_civPossiblity + base_value * Math.pow(0.8, hearableCount);

		return m_civPossiblity;
	}

	private double m_civPossiblity = 0;

	public void setCivPossiblity(double value)
	{
		m_civPossiblity = value;
	}

	public List<Building> getVisibleBuildings()
	{
		return m_visibleBuildings;
	}

	private final List<Building> m_visibleBuildings = new ArrayList<Building>();

	public void addVisibleBuildings(List<Building> blds)
	{
		m_visibleBuildings.addAll(blds);
	}

	public void addVisibleBuildings(Building blds)
	{
		m_visibleBuildings.add(blds);
	}

	private final List<Building> m_hearableBuildings = new ArrayList<Building>();

	public void addHearableBuildings(List<Building> hearable)
	{
		m_hearableBuildings.addAll(hearable);
	}

	public void addHearableBuildings(Building hearable)
	{
		m_hearableBuildings.add(hearable);
	}

	public List<Building> getHearableBuildings()
	{
		return m_hearableBuildings;
	}

	public void removeVisibleBuildings(Collection<Building> blds)
	{
		m_visibleBuildings.removeAll(blds);
	}

	public void removeHearableBuildings(Collection<Building> blds)
	{
		m_hearableBuildings.removeAll(blds);
	}

	public int hearableCount = 0;
	
	public int heardCount = 0;

	private int m_ignitedTime = Integer.MAX_VALUE;
	private int m_reignitedTime = Integer.MAX_VALUE;
	public int putoutTime = 300;

	public boolean m_justIgnited = false;

	private int totalBurningTime = 0;
	//
	public int reIgnitionCount = 0;

	public void setIgniteTime(int value)
	{
		m_ignitedTime = value;
	}

	public int getIgniteTime()
	{
		return m_ignitedTime;
	}

	public int getReigniteTime()
	{
		return m_reignitedTime;
	}

	public int getBurningTime(StandardWorldModel world)
	{
		return Math.max(0, world.getState().getTime() - m_ignitedTime);
	}

	public int getReburningTime(StandardWorldModel world)
	{
		return Math.max(0, world.getState().getTime() - m_reignitedTime);
	}

	public int getBurningTimeToNow(StandardWorldModel world)
	{
		if (getReburningTime(world) > 0)
		{
			return getReburningTime(world);
		}
		return getBurningTime(world);
	}

	public int getLastBurningTime(StandardWorldModel world){
		if (m_reignitedTime > 0){
			return getReburningTime(world);
		} 
		else{
			return getBurningTime(world);
		}
	}

	public boolean tempChanged = false;

	public void setProperty(StandardWorldModel world, int time, Property type, int[] value)
	{
		WorldState wst = world.getState();
		boolean wasBurned = false;
		boolean wasUnburned = true;
		boolean isExtinguished = false;
		if(me.isFierynessDefined()){
			wasBurned = isBurning();
			wasUnburned = (me.getFieryness() == 0);
			isExtinguished = isPutOut();
		}
		
		int tempLastChangeTime = getTime();
		int fireyLastChangeTime = getTime();
		tempChanged = false;

		if(type == me.getBrokennessProperty()){
			me.setBrokenness(value[0]);
		}
		else if(type == me.getBuildingAttributesProperty()){
			me.setBuildingAttributes(value[0]);
		}
		else if(type == me.getBuildingCodeProperty()){
			me.setBuildingCode(value[0]);
		}
		else if(type == me.getFierynessProperty()){
			me.setFieryness(value[0]);
			setFieryness(value[0], time);
		}
		else if(type == me.getFloorsProperty()){
			me.setFloors(value[0]);
		}	
		else if(type == me.getGroundAreaProperty()){
			me.setGroundArea(value[0]);
		}
		else if(type == me.getIgnitionProperty()){
			//TODO
//			me.setIgnition(value[0]);
		}
		else if(type == me.getImportanceProperty()){
			me.setImportance(value[0]);
		}
		else if(type == me.getTemperatureProperty()){
			me.setTemperature(value[0]);
			setTemperature(value[0], time);
		}
		else if(type == me.getTotalAreaProperty()){
			me.setTotalArea(value[0]);
		}
		else{
			super.setProperty(world, time, type, value);
		}
		
//		switch (type){
//
//		case ProtocolConstants.PROPERTY_FLOORS:
//			me.setFloors(value[0]);
//			break;
//		case ProtocolConstants.PROPERTY_BUILDING_ATTRIBUTES:
//			me.setBuildingAttributes(value[0]);
//			break;
//		case ProtocolConstants.PROPERTY_FIERYNESS:
//			me.setFieryness(value[0]);
//			setFieryness(value[0], time);
//			break;
//		case ProtocolConstants.PROPERTY_BROKENNESS:
//			me.setBrokenness(value[0]);
//			break;
//		case ProtocolConstants.PROPERTY_BUILDING_CODE:
//			me.setBuildingCode(value[0]);
//			break;
//		case ProtocolConstants.PROPERTY_BUILDING_AREA_GROUND:
//			me.setGroundArea(value[0]);
//			break;
//		case ProtocolConstants.PROPERTY_BUILDING_AREA_TOTAL:
//			me.setTotalArea(value[0]);
//			break;
//		case ProtocolConstants.PROPERTY_BUILDING_APEXES:
//			setBuildingApexes(value);
//			break;
//		case ProtocolConstants.PROPERTY_BUILDING_TEMPERATURE:
//			me.setTemperature(value[0]);
//			setTemperature(value[0], time);
//			break;
//		default:
//			super.setProperty(world, time, type, value);
//			break;
//		}
		
		
		if (tempLastChangeTime < getTime())
		{
			tempChanged = true;

			if (wst.getTime() > 3){
				newUpdatedTemperature();
			}
		}
		if (fireyLastChangeTime < getTime())
		{
			if (wst.getTime() > 3){
				newUpdatedFieryness();
			}
		}

		m_justIgnited = false;

		if (wasUnburned && isBurning())
		{
			m_justIgnited = true;
			wst.burningBuildings.add(me);
		}

		if (isExtinguished && isBurning())
		{
			m_justIgnited = true;
			m_reignitedTime = world.getState().getTime();
			reIgnitionCount++;

			wst.burningBuildings.add(me);
			boolean reBurnCause = true;
			for (Building b : connectedBuildings)
			{
				if (b.getState().isBurning() && b.getState().getBurningTimeToNow(world) > 2)
				{
					reBurnCause = false;
					break;
				}
			}
			if (reBurnCause)
			{
				wst.reBurningBuildings.add(me);
			}
		}

		if (wasBurned && isPutOut())
		{
			m_ignitedTime = Integer.MAX_VALUE;
			m_reignitedTime = Integer.MAX_VALUE;
			putoutTime = wst.getTime();
			wst.burningBuildings.remove(me);
			totalBurningTime += getBurningTimeToNow(world);
		}

		if (getFieryness() == 8)
		{
			wst.reBurningBuildings.remove(me);
			wst.burningBuildings.remove(me);
		}
	}

	private int m_emptyTime = 0;

	public void setEmptyTime(int time){
		m_emptyTime = time;
	}

	public boolean isEmpty(){
		return m_emptyTime > Constants.TIME_STARTING_ACTION;
	}

	private boolean isReportedAsEmpty = false;

	public boolean isReportedAsEmpty(){
		return isReportedAsEmpty;
	}

	public void setReportedAsEmpty(){
		isReportedAsEmpty = true;
	}

	private int m_seenTime = -1;

	public void setSeenTime(int time){
		m_seenTime = time;
	}

	public int getSeenTime(){
		return m_seenTime;
	}

	public float initialCivPosibility(StandardWorldModel world){
		return 0.1f / (1 + me.getTotalArea()
				/ world.getState().mapBuildingAreaAverage);
	}

	public Map<Building, Float> distanceValue = new HashMap<Building, Float>();
	public ArrayList<Building> neighbourBuildings = new ArrayList<Building>();
	public ArrayList<Building> nearBuildings = new ArrayList<Building>();
	public ArrayList<Building> connectedBuildings = new ArrayList<Building>();
	
	public int counter = 0;
	public float[] connectedValues;
	public HashMap<Building, Float> connectedValueMap = new HashMap<Building, Float>();

	public ArrayList<Building> inDangerNeighbors()
	{
		if (me.getFieryness() == 4 || me.getFieryness() == 0 || !me.isFierynessDefined())
		{
			return null;
		}
		ArrayList<Building> ret = new ArrayList<Building>();
		if(connectedBuildings==null)System.out.println("empty");
		for (Building b : connectedBuildings)
		{
			if (b.getFieryness() == 0)
			{
				if (!(b instanceof Refuge) && !(b instanceof AmbulanceCentre)
						&& !(b instanceof PoliceOffice)
						&& !(b instanceof FireStation))
				{
					ret.add(b);
				}
			} else if (b.getFieryness() >= 4 && b.getFieryness() <= 7)
			{
				ret.add(b);
			}
		}
		return ret;
	}

	static final int FIRE_PASSING_DISTANCE = 40000;

	public float simpleCalculateConnectValue()
	{
		float sum = 0;
		for (int i = 0; i < connectedBuildings.size(); i++)
		{
			Building b = connectedBuildings.get(i);
            BuildingState bst = b.getState();
			if (bst.connectedBuildings.contains(this)){
				if (ConstantConditions.SPECIAL_BUILDING_C.eval(b)){
					continue;
				}

				float connectValue = connectedValues[i]
						* ((FIRE_PASSING_DISTANCE - bst.distanceValue.get(me)) / FIRE_PASSING_DISTANCE)
						* ((float) me.getTotalArea() / (float) me.getTotalArea());

				double burningValue = b.getTotalArea() / 100d;

				double value = burningValue * connectValue;
				if (bst.isUnburned() || bst.isPutOut()){
					sum += value;
				} 
				else if (bst.isBurning()){
					sum -= value;
				} 
			}
		}
		return sum;
	}

	public int connectBurnableCount()
	{
		int sum = 0;
		for (Building b : connectedBuildings)
		{
			BuildingState bst = b.getState();
			if (bst.isUnburned() || bst.isPutOut()){
				if (ConstantConditions.SPECIAL_BUILDING_C.eval(b)){
					continue;
				}
				sum++;
			}
		}
		return sum;
	}

	private float m_totalWallArea;
	private final LinkedList<Wall> m_walls = new LinkedList<Wall>();

	public float getTotalWallArea(){
		return m_totalWallArea;
	}

	public void initWalls()
	{
		m_totalWallArea = 0;
		int[] apexes;
		if ((apexes = me.getApexList()) == null)
			return;
		int fx = apexes[0];
		int fy = apexes[1];
		int lx = fx;
		int ly = fy;
		for (int n = 2; n < apexes.length - 1; n++)
		{
			int tx = apexes[n];
			int ty = apexes[++n];
			Wall w = new Wall(lx, ly, tx, ty, me);
			if (w.validate()){
				m_walls.add(w);
				m_totalWallArea += FLOOR_HEIGHT * 1000 * w.length;
			}
			lx = tx;
			ly = ty;
		}
		Wall w = new Wall(lx, ly, fx, fy, me);
		m_walls.add(w);
		m_totalWallArea = m_totalWallArea / 1000000f;
	}

	// added for fire prediction by francis
	private final int FLOOR_HEIGHT = 3;
	private float m_volume;
	private float m_initFuel;

	private float m_capacity;

	private int needWaterQuantity;
	private int fireCycle;
	public float deltaEnergy = 0;

	// For each cycle's prediction
	public float energy;
	public float fule;
	public float temperature;

	// Estimated result for this cycle
	private SingleData<Float> estimatedEnergy;
	private SingleData<Float> estimatedTemperature;
	private SingleData<Float> estimatedFuel;

	public void initializeFireFigures()
	{

		initWalls();
		
		this.m_volume = (me.getGroundArea() * me.getFloors() * FLOOR_HEIGHT) / 10;
		this.m_initFuel = ((getFuelDensity() * me.getGroundArea() 
				* me.getFloors() * FLOOR_HEIGHT) / 10);
		this.m_capacity = this.m_volume * getThermoCapacity();
		setWaterQuantity(0, 0);

		this.needWaterQuantity = 0;
		this.fireCycle = seu.agent.Constants.INITIAL_FIRE_CYCLES;

		setEstimatedEnergy(seu.agent.Constants.INITIAL_TEMPERATURE
				* getCapacity(), 0);
		setEstimatedFuel(m_initFuel, 0);
		setEstimatedTemperature(seu.agent.Constants.INITIAL_TEMPERATURE, 0);
	}

	public float getFuelDensity(){
		switch (me.getBuildingCode()){
		case 0:   return seu.agent.Constants.WOOD_ENERGIE;
		case 1:   return seu.agent.Constants.STEEL_ENERGIE;
		default:  return seu.agent.Constants.CONCRETE_ENERGIE;
		}
	}

	private float getThermoCapacity(){
		switch (me.getBuildingCode()){
		case 0:   return seu.agent.Constants.WOOD_CAPACITY;
		case 1:   return seu.agent.Constants.STEEL_CAPACITY;
		default:  return seu.agent.Constants.CONCRETE_CAPACITY;
		}
	}

	public float getIgnitionPoint(){
		switch (me.getBuildingCode()){
		case 0:   return seu.agent.Constants.WOOD_IGNITION;
		case 1:   return seu.agent.Constants.STEEL_IGNITION;
		default:  return seu.agent.Constants.CONCRETE_IGNITION;
		}
	}

	public SingleData<Float> getEstimatedEnergy(){
		return estimatedEnergy;
	}

	public void setEstimatedEnergy(float value, int time){
		estimatedEnergy = new SingleData<Float>(value, time);
	}

	public SingleData<Float> getEstimatedFuel(){
		return estimatedFuel;
	}

	public void setEstimatedFuel(float value, int time){
		estimatedFuel = new SingleData<Float>(value, time);
	}

	public SingleData<Float> getEstimatedTemperature(){
		return estimatedTemperature;
	}

	public void setEstimatedTemperature(float value, int time){
		// if(time>= this.estimatedTemperature.getTime())
		estimatedTemperature = new SingleData<Float>(value, time);

	}

	public int getWaterQuantity(){
		return this.waterQuantity.getObject().intValue();
	}

	public float getCapacity(){
		return this.m_capacity;
	}

	public float getInitFuel(){
		return this.m_initFuel;
	}

	public int getNeedWaterQuantity(){
		return this.needWaterQuantity;
	}

	public void setNeedWaterQuantity(int value){
		this.needWaterQuantity = value;
	}

	public int getFireCycle(){
		return fireCycle;
	}

	public void setFireCycle(int fireCycle){
		this.fireCycle = fireCycle;
	}

	public void newUpdatedTemperature(){
		float temp = estimatedTemperature.getObject().floatValue();
		if (temp < getTemperature() - 25f || temp > getTemperature() + 25f)
		{
			temp = getTemperature();

			if (isBurning() && temp < getIgnitionPoint())
				temp = getIgnitionPoint() * 1.1f;

			int up_time = m_temperature.getLast().getTime();
			setEstimatedTemperature(temp, up_time);
			setEstimatedEnergy(temp * getCapacity(), up_time);
		}
	}

	private final float DELTA = 0.08f;

	public void newUpdatedFieryness()
	{

		int fiery = getFieryness();
		int es_fiery = calculateFieryness(estimatedFuel.getObject()
				.floatValue());
		int es_time = estimatedFuel.getTime();

		if (es_fiery == fiery)
			return;

		int up_time = m_fieryness.getLast().getTime();
		float delta = es_time > up_time ? es_time - up_time : up_time - es_time;
		delta /= 10;

		float fuel = -1;

		if (fiery == 0 || fiery == 4){
			setEstimatedFuel(m_initFuel, up_time);
			fuel = 1;
		} 
		else if (fiery > 7){
			setEstimatedFuel(0, up_time);
			fuel = 0;
		} 
		else{
			switch (fiery){
			case 1:
			case 5:
			{
				if (es_fiery == 2)
				{
					setEstimatedEnergy(getEstimatedEnergy().getObject()
							.floatValue()
							* (1 - delta - DELTA), up_time);
					setEstimatedTemperature(getEstimatedTemperature()
							.getObject().floatValue()
							* (1 - delta - DELTA), up_time);

					float temp = getEstimatedFuel().getObject().floatValue()
							* (1 + delta);

					if (temp > m_initFuel)
						temp = m_initFuel * 0.9f;
					setEstimatedFuel(temp, up_time);
					fuel = temp / m_initFuel;
				} 
				else
				{
					setEstimatedFuel((m_initFuel * 0.8f), up_time);
					fuel = 0.8f;
				}
				break;
			}
			case 2:
			case 6:
			{
				if (es_fiery == 1)
				{
					setEstimatedEnergy(getEstimatedEnergy().getObject()
							.floatValue()
							* (1 + delta + DELTA), up_time);
					setEstimatedTemperature(getEstimatedTemperature()
							.getObject().floatValue()
							* (1 + delta + DELTA), up_time);

					float temp = getEstimatedFuel().getObject().floatValue()
							* (1 - delta);

					if (temp < m_initFuel * 0.33f)
						temp = m_initFuel * 0.33f;

					setEstimatedFuel(temp, up_time);
					fuel = temp / m_initFuel;
				} 
				else if (es_fiery == 3)
				{
					setEstimatedEnergy(getEstimatedEnergy().getObject()
							.floatValue()
							* (1 - delta - DELTA), up_time);
					setEstimatedTemperature(getEstimatedTemperature()
							.getObject().floatValue()
							* (1 - delta - DELTA), up_time);

					float temp = getEstimatedFuel().getObject().floatValue()
							* (1 + delta);

					if (temp > 0.66f * m_initFuel)
						temp = m_initFuel * 0.66f;

					setEstimatedFuel(temp, up_time);
					fuel = temp / m_initFuel;
				} 
				else
				{
					setEstimatedFuel((m_initFuel * 0.5f), up_time);
					fuel = 0.5f;
				}
				break;
			}
			case 3:
			case 7:
			{
				if (es_fiery == 2)
				{
					setEstimatedEnergy(getEstimatedEnergy().getObject()
							.floatValue()
							* (1 + delta + DELTA), up_time);
					setEstimatedTemperature(getEstimatedTemperature()
							.getObject().floatValue()
							* (1 + delta + DELTA), up_time);

					float temp = getEstimatedFuel().getObject().floatValue()
							* (1 - delta);

					if (temp < 0)
						temp = 0.05f * m_initFuel;

					setEstimatedFuel(temp, up_time);
					fuel = temp / m_initFuel;
				} 
				else
				{
					setEstimatedFuel((m_initFuel * 0.15f), up_time);
					fuel = 0.15f;
				}
				break;
			}
			default:
				break;
			}
		}
	}

	public int calculateFieryness(float fuel)
	{
		if (isBurning())
		{
			if (fuel >= m_initFuel * 0.66f)
				return 1; // burning, slightly damaged
			if (fuel >= m_initFuel * 0.33f)
				return 2; // burning, more damaged
			if (fuel > 0)
				return 3; // burning, severly damaged
		}
		if (fuel == m_initFuel)
			if (getWaterQuantity() > 0)
				return 4; // not burnt, but watered-damaged
			else
				return 0; // not burnt, no water damage
		if (fuel >= m_initFuel * 0.66)
			return 5; // extinguished, slightly damaged
		if (fuel >= m_initFuel * 0.33)
			return 6; // extinguished, more damaged
		if (fuel > 0)
			return 7; // extinguished, severely damaged
		return 8; // completely burnt down
	}

	public int getWaterNeededToExtinguishFire()
	{
		return getNeedWaterQuantity();
	}

	public void setWaterQuantity(int waterQuantity, int time)
	{
		this.waterQuantity = new SingleData<Integer>(waterQuantity, time);
	}

	/**
	 * @param ox
	 * @param oy
	 * @return
	 */
	public boolean isBuilding(int x, int y)
	{
		return (me.getX() == x && me.getY() == y);
	}
	//===================================================================================================//
//	private boolean m_isTargetFinished = false;
//	
//	public void setTargetFinished(){
//		m_isTargetFinished = true;
//	}
//	
//	public boolean isTargetFinished(){
//		return m_isTargetFinished;
//	}
	
	//========================================================================================
	
}