/**************************************************************************
 * Copyright (C) 2008 All rights reserved.
 * Created by Southeast University(SEU) RoboCup Simulation League.
 **************************************************************************
 * Last modify:
 * $Id: AmbulanceCenterAgent.java,v 1.3 2007/06/28 09:15:34 nobel Exp $
 */

package seu.states;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import rescuecore2.standard.entities.*;
import seu.SeuConstants;
import seu.SeuParameter;
import seu.agent.Constants;
import seu.agent.Util;
import seu.utilities.ConstantConditions;
import seu.utilities.base.SingleData;

public class CivilianState extends HumanState {

	private Civilian me;
	
	private LinkedList<SingleData<Integer>> m_estimatedHp = new LinkedList<SingleData<Integer>>(); 
   private LinkedList<SingleData<Integer>> m_estimatedDamage = new LinkedList<SingleData<Integer>>();
   private LinkedList<SingleData<Integer>> m_estimatedLife = new LinkedList<SingleData<Integer>>();
   public int rescueValue = -1; 
   
	public CivilianState(Civilian me) {
		super(me);
		this.me = me;
		m_estimatedHp.addLast(new SingleData<Integer>(10000,-1));
		m_estimatedDamage.addLast(new SingleData<Integer>(0,-1));
		m_estimatedLife.addLast(new SingleData<Integer>(300, -1));
	}
	
//	public int getType(){
//		return SeuConstants.TYPE_CIVILIAN;
//	}
	
	public HashSet<Point> hearPositions = new HashSet<Point>();
	private Point m_lastHeardPoint;
	
	//TODO
	//Wait for JYH
	public void addHearPosition(StandardWorldModel world, Area heardPos, int posExtra, boolean selfHeard){
		if (selfHeard) {
			if(world.getState().me instanceof Human){
				Human self = (Human) world.getState().me;
				int x = self.getX();
				int y = self.getY();
				hearPositions.add(new Point(x, y));
				m_lastHeardPoint = new Point(x, y);
			}
		} 
		else {
			int x, y;
			if (heardPos instanceof Road) {
				Road rd = (Road) heardPos;
				x = rd.getState().getPointHead(world).getX() + (posExtra / rd.getState().getLength(world))
						* (rd.getState().getPointTail(world).getX() - rd.getState().getPointHead(world).getX());
				y = rd.getState().getPointHead(world).getY() + (posExtra / rd.getState().getLength(world))
						* (rd.getState().getPointTail(world).getY() - rd.getState().getPointHead(world).getY());
			} else {
				x = heardPos.getX();
				y = heardPos.getY();
			}
			hearPositions.add(new Point(x, y));
			m_lastHeardPoint = new Point(x, y);
		}
	}

	private List<Building> m_possiblePositions = new ArrayList<Building>();	
	public List<Building> getPossiblePositions(){
		return m_possiblePositions;
	}
	
	private Building m_mostPossibleBldg;
	public Building getMostPossibleBuilding(){
		return m_mostPossibleBldg;
	}
	
	public void updatePossiblePositions(StandardWorldModel world){
		if(me.getPosition(world) != null){
			return;
		}
		
		List<Building> possibles = new ArrayList<Building>();
		for(Building b : world.getState().notSeenBuildings){
			if(ConstantConditions.SPECIAL_BUILDING_C.eval(b))
				continue;
			if(b.getState().isBurned())
				continue;
			if(Util.distance(b, m_lastHeardPoint) < Constants.HEAR_DISTANCE){
				b.getState().heardCount ++;
				possibles.add(b);
			}
		}
		
		m_possiblePositions.retainAll(possibles);
		
		if(m_possiblePositions.isEmpty()
			|| m_possiblePositions.size() == 0){
			return;
		}
		
		double sizeValue = 1.0/possibles.size()*80.0/world.getState().buildings.size();
		
		double maxValue = 0;
		
		for(Building bld : m_possiblePositions){
			if(ConstantConditions.SPECIAL_BUILDING_C.eval(bld)){
				continue;
			}
			BuildingState bst = bld.getState();
			bst.addPossibleCivilians(world,me);
			double value = sizeValue;// + bld.getCivPossibility();//must add last possibility
			value *= Math.pow(0.8, Math.max(0,(bst.hearableCount - bst.heardCount)));
			if(bst.isBurning()) 
				bst.setCivPossiblity(value/2f/bld.getFieryness());
			else if(bst.isBurned()) 
				bst.setCivPossiblity(0);
			else bst.setCivPossiblity(value);
			
			// used for most possible building 
			if(bst.getCivPossibility(world) > maxValue){
				maxValue = bst.getCivPossibility(world);
				m_mostPossibleBldg = bld;
			}
		}
	}
	
	// estimate life 
	private double[][] m_particles = null;
	
	private int[] m_deadTime = new int[60];
	
	private int m_averageDeadTime = 300;
	
	public boolean m_propertyChanged = false;
	public boolean m_particleNeedResample = false;
	public boolean m_needKeepLifeTime = false;
	
	public int m_fireDeadTime = 300;
	public Building m_nearestBurningBuilding = null;
	
	private static final int HP_4K = 4500;
	
	private static final int HP_1K = 1500;
	
	private static final int MAX_PARTICLE_NUM = 600;
	private int m_updatedTime = -1;
	private int m_time_to_refresh = -1;
	private int m_particle_time = -1;
	
	public int m_1stDeadTime = 300;
	
	public int m_2ndDeadTime = 299;
	
	private int[] m_4kDeadTime = new int[60];
	
	private int[] m_1kDeadTime = new int[60];
	
	public int m_4kHpDeadTime = 299;
	
	public int m_4kHpDeadTime_backup = new Integer(299);
	
	public int m_1kHpDeadTime = 299;
	
	public int getDeadTime(){
		return m_1stDeadTime;
	}
	
	public void setDeadTime(int t){
		m_1stDeadTime = t;
	}
	
	public int m_currentHp = 10000;
	public double m_currnetHpCoef = 0.0;
	
	private int m_hp = 0;
	private int m_damage = 0;
	private int m_buriedness = 0;
	
	public void setParticle(StandardWorldModel world, int time, int hp, int dmg, int bury, int isSee){
		
		if(me.getPosition(world) instanceof Refuge)
			return;
		
		m_hp = hp;
		m_damage = dmg;
		m_buriedness = bury;
		
		m_updatedTime = time;
		
		if(dmg != getDamage(time-1))
			m_propertyChanged = true;
		
		if(hp != getHp(time-1))
			m_propertyChanged = true;	
	}
	
	public void calculateDeadTime(StandardWorldModel world, int time){
		if(m_particles == null && m_updatedTime > 0)
		{
			particleInit();
		}
		
		if(m_propertyChanged || time > m_time_to_refresh)
		{
			updateParticles(time);
		}
		
		calculateHp(world);
	}
	
	private void updateParticles(int currentTime){
		int time = m_updatedTime - m_particle_time;
		if(time <= 0)
			return;
		
		boolean propertyChanged = m_propertyChanged;
		m_propertyChanged = false;
		
		ArrayList<double[]> newParticle = new ArrayList<double[]>();
		for(int i=0;i<600;i++)
		{
			double[] newTmp = calculateLifeSpan(m_particles[i],time);
			if(checkParticle(newTmp))
			{
				newParticle.add(newTmp);
			}
		}
		//
		if(newParticle.size()==0)
		{
			if(propertyChanged && m_particleNeedResample)
			{
				m_particleNeedResample = false;
				particleInit();
				return;
			}

			double[][] tmpParticle = initTempParticles();
			ArrayList<Integer> deadTime = new ArrayList<Integer>();
			int total = 0;
			for(int i=0;i<60;i++)
			{
				int index=0;
				index = (int)(Math.random()*600);
				
				int t = calculateDeadTime(tmpParticle[index],m_updatedTime);
				deadTime.add(t);
				total += t;
			}
			int avg = total/60;
			total = 0;
			for(int i=0;i<60;i++)
				total += Math.abs(deadTime.get(i)-avg);
			int scatterNew = total;
			//
			deadTime = new ArrayList<Integer>();
			total = 0;
			for(int i=0;i<60;i++)
			{
				int index=0;
				index = (int)(Math.random()*600);
				
				int t = calculateDeadTime(m_particles[index]);
				deadTime.add(t);
				total += t;
			}
			avg = total/60;
			total = 0;
			for(int i=0;i<60;i++)
				total += Math.abs(deadTime.get(i)-avg);
			int scatterOld = total;
			//
			if(scatterNew < scatterOld)
			{
				particleInit();
				return;
			}
			//
			m_time_to_refresh = currentTime + 15;
			return;
		}

		m_time_to_refresh = currentTime + 40;
		m_particles = new double[600][3];
		for(int i=0;i<newParticle.size();i++)
			m_particles[i] = newParticle.get(i);
		for(int i = newParticle.size();i<600;i++)
		{
			int index=0;
			index = (int)(Math.random()* ((double)newParticle.size())  );
			
			m_particles[i] = newParticle.get(index);
		}
		m_particle_time = m_updatedTime;
		
		m_deadTime = calculateDeadTime();
		
		m_4kDeadTime = calculatePredicted4kDeadTime();
		
		
	}
	
	
	private boolean checkParticle(double[] status)
	{
		return (status[0] > m_hp-500 && status[0] < m_hp+500
				&& status[1] > m_damage-5 && status[1] < m_damage+5);
	}
	
	private double[] calculateLifeSpan(double[] status , int time)
	{
		double[] result = new double[3];
		double hp = status[0];
		double dmg = status[1];
		double brokenDmg = dmg * status[2]/100.0;
		double buryDmg = dmg - brokenDmg;
		for(int i=0;i<time;i++)
		{
			buryDmg += buryDmg*buryDmg*0.000035;
			buryDmg += 0.01;
			brokenDmg += brokenDmg*brokenDmg*0.00025;
			brokenDmg += 0.01;
			hp -= buryDmg;
			hp -= brokenDmg;
		}
		result[0] = hp;
		result[1] = buryDmg+brokenDmg;
		result[2] = brokenDmg/result[1]*100;
		return result;
	}
	
	

	private double[][] initTempParticles()
	{
		double[][] result = new double[600][3];
		double[] hpTable = new double[6];
		hpTable[0] = m_hp - 499;
		hpTable[1] = m_hp - 299;
		hpTable[2] = m_hp - 99;
		hpTable[3] = m_hp + 101;
		hpTable[4] = m_hp + 301;
		hpTable[5] = m_hp + 499;
		//
		double[] dmgTable = new double[10];
		double step = 9.98/9;
		for(int i=0;i<10;i++)
			dmgTable[i] = ((double)m_damage)-4.99+step*(double)i;
		//
		double[] brokenRateTable = new double[10];
		for(int i=0;i<10;i++)
			brokenRateTable[i] = 5+10*i;
		//
		for(int i=0;i<600;i++)
		{
			int a = i/100;
			int b = (i%100)/10;
			int c = i%10;
			result[i][0] = hpTable[a];
			if(m_damage == 0)
				result[i][1] = 3.0;
			else
				result[i][1] = dmgTable[b];
			if(m_buriedness == 0)
				result[i][2] = 100.0;
			else
				result[i][2] = brokenRateTable[c];
		}
		//
		return result;
	}
	//
	
	private int[] calculateDeadTime()
	{
		ArrayList<Integer> result = new ArrayList<Integer>();
		for(int i=0;i<60;i++)
		{
			int index=0;
			index = (int)(Math.random()*600);
		
			result.add(calculateDeadTime(m_particles[index]));
		}
		
		//
		double dead = 0;
		for(int i = 0; i < m_particles.length; i ++){
			dead += calculateDeadTime(m_particles[i]);
		}
		
		m_averageDeadTime = (int)dead / m_particles.length;
		//
		
		Collections.sort(result);
		int[] finalResult = new int[result.size()];
		for(int i=0;i<result.size();i++)
			finalResult[i]=result.get(i);
		
		return finalResult;
		
	}
	
	
	private int calculateDeadTime(double[] status)
	{
		double hp = status[0];
		double dmg = status[1];
		
		double brokenDmg = dmg * status[2]/100.0;
		double buryDmg = dmg - brokenDmg;
		int time = 0;
		while(hp > 0)
		{
			buryDmg += buryDmg*buryDmg*0.000035;
			buryDmg += 0.01;
			brokenDmg += brokenDmg*brokenDmg*0.00025;
			brokenDmg += 0.01;
			hp -= buryDmg;
			hp -= brokenDmg;
			
			time++;
			if(time > 400)
				return m_particle_time+400;
		}

		return m_particle_time + time;
	}
	
	private int calculateDeadTime(double[] status,int baseTime)
	{
		double hp = status[0];
		double dmg = status[1];
		double brokenDmg = dmg * status[2]/100.0;
		double buryDmg = dmg - brokenDmg;
		int time = 0;
		while(hp > 0)
		{
			buryDmg += buryDmg*buryDmg*0.000035;
			buryDmg += 0.01;
			brokenDmg += brokenDmg*brokenDmg*0.00025;
			brokenDmg += 0.01;
			hp -= buryDmg;
			hp -= brokenDmg;
			time++;
			if(time > 400)
				return baseTime+400;
		}
		
		return baseTime + time;
	}
	
	private int[] calculatePredicted4kDeadTime()
	{
		ArrayList<Integer> result = new ArrayList<Integer>();
		
		ArrayList<Integer> oneKHp = new ArrayList<Integer>();
		
		for(int i=0;i<60;i++)
		{
			int index=0;
			index = (int)(Math.random()*600);
			
			double[] predHp = calculatePredictedHp(m_particles[index]);
			
			for(int t = m_updatedTime; t < predHp.length-1; t ++){
				if(predHp[t+1] < HP_4K){
					result.add((int)t);
//					break;
				}
				
				if(predHp[t+1] < HP_1K){
					oneKHp.add((int)t);
					break;
				}
			}
		}
		
		//
		double dead = 0;
		for(int i = 0; i < m_particles.length; i ++){
			dead += calculateDeadTime(m_particles[i]);
		}
		
		m_averageDeadTime = (int)dead / m_particles.length;
		// 1k
		Collections.sort(oneKHp);
		
		for(int i = 0; i < oneKHp.size(); i ++)
		{
			m_1kDeadTime[i] = oneKHp.get(i);
		}
		
		//
		Collections.sort(result);
		int[] finalResult = new int[result.size()];
		for(int i=0;i<result.size();i++)
			finalResult[i]=result.get(i);
		
		return finalResult;
	}
	
	
	private double[] calculatePredictedHp(double[] status)
	{
		double hp = status[0];
		double dmg = status[1];
		
		double brokenDmg = dmg * status[2]/100.0;
		double buryDmg = dmg - brokenDmg;
		int time = 0;
		
		double[] result = new double[1000];
		
		while(hp > 0)
		{
			buryDmg += buryDmg*buryDmg*0.000035;
			buryDmg += 0.01;
			brokenDmg += brokenDmg*brokenDmg*0.00025;
			brokenDmg += 0.01;
			hp -= buryDmg;
			hp -= brokenDmg;
			
			result[m_particle_time + time] = hp;
			
			time++;
			if(time > 400)
				return result;
		}

		return result;
	}
	
	private void calculateHp(StandardWorldModel world){
		double estimatedHp = 0;
		
		if(m_particles == null){
			m_currentHp = m_hp - (world.getState().getTime()-m_updatedTime)*500;
			return;
		}
		
		for(int i = 0; i < m_particles.length; i ++){
			
			// 
			double hp = m_particles[i][0];
			double dmg = m_particles[i][1];
			
			double brokenDmg = dmg * m_particles[i][2]/100.0;
			double buryDmg = dmg - brokenDmg;
			for(int time = m_particle_time; time <= world.getState().getTime() && hp > 0; time ++){
				buryDmg += buryDmg*buryDmg*0.000035;
				buryDmg += 0.01;
				brokenDmg += brokenDmg*brokenDmg*0.00025;
				brokenDmg += 0.01;
				hp -= buryDmg;
				hp -= brokenDmg;
			}
			//
			estimatedHp += hp;
		}
		
		m_currentHp = (int)(estimatedHp/m_particles.length);
		
		if(m_currentHp > 7000)
			m_currnetHpCoef = 0.5;
		else if(m_currentHp > 4000)
			m_currnetHpCoef = 2.0;
		else if(m_currentHp > 1000)
			m_currnetHpCoef = 2.5;
		else 
			m_currnetHpCoef = 0.00001;
		
		
	}
	
	public int[] getPredictedDeadTime()
	{
		return m_deadTime;
	}
	
	public void particleInit(){
		
		m_propertyChanged = false;
		
		m_time_to_refresh = 25 + m_updatedTime + (int)(Math.random()*20);
		
		m_particle_time = m_updatedTime;
		
		if(m_hp == 0){
			m_particles = null;
			m_deadTime = new int[60];
			m_4kDeadTime = new int[60];
			m_1kDeadTime = new int[60];
			return;
		}
		
		m_particles = new double[MAX_PARTICLE_NUM][3];
		
		
		double[] hpTable = new double[6];
		{
			hpTable[0] = m_hp - 499;
			hpTable[1] = m_hp - 299;
			hpTable[2] = m_hp - 99;
			hpTable[3] = m_hp + 101;
			hpTable[4] = m_hp + 301;
			hpTable[5] = m_hp + 499;
		}
		
		double[] dmgTable = new double[10];
		double step = 9.98/9;
		for(int i=0;i<10;i++)
			dmgTable[i] = m_damage-4.99+step*(double)i;
		
		double[] brokenRateTable = new double[10];
		for(int i=0;i<10;i++)
			brokenRateTable[i] = 5+10*i;
		
		for(int i=0;i<MAX_PARTICLE_NUM;i++)
		{
			int a = i/100;
			int b = (i%100)/10;
			int c = i%10;
			
			m_particles[i][0] = hpTable[a];
			if(m_damage == 0)
				m_particles[i][1] = 3.0;
			else 
				m_particles[i][1] = dmgTable[b];
			
			if(m_buriedness == 0)
				m_particles[i][2] = 100.0;
			else
				m_particles[i][2] = brokenRateTable[c];

		}
		
		m_deadTime = calculateDeadTime();
		
		m_4kDeadTime = calculatePredicted4kDeadTime();
		
	}
	
	public double[][] tmpParticleInit(){
		double[][] result = new double[600][3];
		
		double[] hpTable = new double[6];
		{
			hpTable[0] = m_hp - 499;
			hpTable[1] = m_hp - 299;
			hpTable[2] = m_hp - 99;
			hpTable[3] = m_hp + 101;
			hpTable[4] = m_hp + 301;
			hpTable[5] = m_hp + 499;
		}
		
		double[] dmgTable = new double[10];
		double step = 9.98/9;
		for(int i=0;i<10;i++)
			dmgTable[i] = m_damage-4.99+step*(double)i;
		
		double[] brokenRateTable = new double[10];
		for(int i=0;i<10;i++)
			brokenRateTable[i] = 5+10*i;
		
		for(int i=0;i<MAX_PARTICLE_NUM;i++)
		{
			int a = i/100;
			int b = (i%100)/10;
			int c = i%10;
			result[i][0] = hpTable[a];
			if(m_damage == 0)
				result[i][1] = 3.0;
			else
				result[i][1] = dmgTable[b];
			if(m_buriedness == 0)
				result[i][2] = 100.0;
			else
				result[i][2] = brokenRateTable[c];
		}
		
		return result;
		
	}
	
	

	public void updateCivilianDeadTime(StandardWorldModel world)
	{
		if(m_needKeepLifeTime == true)
		{
			if( !(me.getPosition(world) instanceof Building))
				return;
			Building b = (Building)me.getPosition(world);
			if(b.getState().isBurning())
				return;
			m_needKeepLifeTime = false;
			m_particleNeedResample = true;
		}
		
		calculateDeadTime(world,world.getState().getTime());
		
		
		int injuryDeadTime, fireDeadTime;
		int offset = 60 - getBuriedness();
		if(offset < 5)
			offset = 5;
		if(offset > 55)
			offset = 55;
		
		injuryDeadTime = getPredictedDeadTime()[offset];
		
//		if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString(world.getTime()+" "+injuryDeadTime+" "+m_averageDeadTime+" "+getBuriedness(), 10, ""+this.id);
		
		
		Building nearestBurningBuilding = getNearestBurningBuilding(world);
		if(nearestBurningBuilding == null)
		{
			fireDeadTime = 500;
			m_fireDeadTime = 500;
			m_nearestBurningBuilding = null;
		}
		else if(nearestBurningBuilding == m_nearestBurningBuilding)
			fireDeadTime = m_fireDeadTime;
		else
		{
			m_nearestBurningBuilding = nearestBurningBuilding;
			
			int dist = world.getDistance(me, nearestBurningBuilding);
			int additionalTime = 10 - (world.getState().getTime() 
					- nearestBurningBuilding.getState().getIgniteTime());
			if(additionalTime < 0)
				additionalTime = 0;
			m_fireDeadTime = world.getState().getTime() + dist/1000 + additionalTime;

			m_fireDeadTime += 20;
			fireDeadTime = m_fireDeadTime;
		}
		//
		if(civilianBurning(world,me))
		{
			m_needKeepLifeTime = true;
			if(getBuriedness() > 20)
			{
				setDeadTime(world.getState().getTime());
				m_2ndDeadTime = world.getState().getTime();
			}
			else
			{
				setDeadTime(world.getState().getTime()+20);
				m_2ndDeadTime = world.getState().getTime()+20;
			}
		}
		else
		{
			setDeadTime(Math.min(injuryDeadTime, fireDeadTime));
			m_2ndDeadTime = Math.min(getPredictedDeadTime()[30], fireDeadTime);
		}
		
		m_4kHpDeadTime = Math.min(m_4kDeadTime[offset], m_1stDeadTime);
		
		m_4kHpDeadTime_backup = m_4kHpDeadTime;
		
		m_1kHpDeadTime = Math.min(m_1kDeadTime[offset], m_1stDeadTime);
		
	}
	

	private Building getNearestBurningBuilding(StandardWorldModel world){
		int minDist = Integer.MAX_VALUE;
		Building result = null;
		for(Iterator it = world.getState().burningBuildings.iterator();it.hasNext();)
		{
			Building b = (Building)it.next();
			int dist = world.getDistance(me, b);
			if(dist < minDist)
			{
				minDist = dist;
				result = b;
			}
		}
		return result;
	}
	
	
	private boolean civilianBurning(StandardWorldModel world, Civilian c)
	{
		if( !(c.getPosition(world) instanceof Building) ) 
			return false;
		Building b = (Building)c.getPosition(world);
		if(b.getState().isBurning())
			return true;
		else
			return false;
	}
	
	private int m_rescuePriority = -1;
	
	public int getRescuePriority(){
		return m_rescuePriority;
	}
	
	public void setRescuePriority(int p){
		m_rescuePriority = p;
	}
	
	public void setEstimatedDamage(int value,int time){
    	Util.updateSingleDataList(m_estimatedDamage, value, time, true,
    			SeuParameter.LIST_LENGTH_CIVILIAN_DAMAGE);
    }
	
	public int getEstimatedDamage(int time){
    	return Util.searchSingleDataForTime(m_estimatedDamage,time);
    }
	
	public int getEstimatedDamage(){
		return m_estimatedDamage.getLast().getObject().intValue();
	}
	
	public void setEstimatedHp(int value,int time){
    	Util.updateSingleDataList(m_estimatedHp, value, time, true,
    			SeuParameter.LIST_LENGTH_CIVILIAN_HP);
    }
	
	public int getEstimatedHp(int time){
    	return Util.searchSingleDataForTime(m_estimatedHp,time);
    }
	
	public int getEstimatedHp(){
		return m_estimatedHp.getLast().getObject().intValue();
	}
	
	public void setEstimatedLife(int value,int time){
    	Util.updateSingleDataList(m_estimatedLife, value, time, true,
    			SeuParameter.LIST_LENGTH_CIVILIAN_LIFE);
    }
	
	public int getEstimatedLife(int time){
    	return Util.searchSingleDataForTime(m_estimatedLife,time);
    }
	
	public double random(double start,double end)
	{
		return Math.random()*(end - start) +start;
	}
	
	private boolean hasFound = false;
	
	public boolean hasFound(){
		return hasFound;
	}
	
	public void setFound(){
		hasFound = true;
	}
		
	public int m_rescueATs = 0;
	
	
	public int value = 0;
	
	
	@Override
	public boolean needLoad(StandardWorldModel world) {
		return hasFound() && super.needLoad(world);
	}
	
	@Override
	public boolean needRescue(StandardWorldModel world) {
		return hasFound() && super.needRescue(world);
	}
	
	public int getRescueValue(){
		return rescueValue;
	}
	
}
