//SEU2010

package seu.states;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.Property;
import seu.SeuConstants;
import seu.SeuParameter;
import seu.agent.Util;
import seu.entities.Node;
import seu.entities.Path;
import seu.entities.Report;
import seu.utilities.base.SingleData;

public abstract class HumanState extends EntityState {

	private Human me;

	public static final int REST_ACT = 0x80;
	public static final int MOVE_ACT = 0x81;
	public static final int LOAD_ACT = 0x82;
	public static final int UNLOAD_ACT = 0x83;
	public static final int EXTINGUISH_ACT = 0x86;
	public static final int RESCUE_ACT = 0x88;
	public static final int CLEAR_ACT = 0x89;

	private LinkedList<SingleData<Integer>> m_hp = new LinkedList<SingleData<Integer>>();
	private LinkedList<SingleData<Integer>> m_damage = new LinkedList<SingleData<Integer>>();
	private LinkedList<SingleData<Integer>> m_buriedness = new LinkedList<SingleData<Integer>>();

	public HumanState(Human me) {
		super(me);
		this.me = me;
		m_hp.addLast(new SingleData<Integer>(10000, -1));
		m_damage.addLast(new SingleData<Integer>(0, -1));
		m_buriedness.addLast(new SingleData<Integer>(0, -1));
	}

	// from MovingObject
	
	
	// getTime()

	// =======================================================================================//

	// from BaseHumanoid

	// public int getStamina() { return m_stamina; }
	public int getHp() {
		return m_hp.getLast().getObject();
	}

	public int getDamage() {
		return m_damage.getLast().getObject();
	}

	public int getBuriedness() {
		return m_buriedness.getLast().getObject();
	}

	public int getHp(int time) {
		return Util.searchSingleDataForTime(m_hp, time);
	}

	public List<SingleData<Integer>> getAllHp() {
		return m_hp;
	}

	public int getDamage(int time) {
		return Util.searchSingleDataForTime(m_damage, time);
	}

	public List<SingleData<Integer>> getAllDamage() {
		return m_damage;
	}

	public int getBuriedness(int time) {
		return Util.searchSingleDataForTime(m_buriedness, time);
	}

	public List<SingleData<Integer>> getAllBuriedness() {
		return m_buriedness;
	}

	public void setHp(StandardWorldModel world, int value, int time) {
		if (me instanceof Civilian) {
			if (value != getHp()) {
				((Civilian) me).getState().m_propertyChanged = true;
			}
		}
		Util.updateSingleDataList(m_hp, value, time, true,
				SeuParameter.LIST_LENGTH_HUMANOID_HP);
	}

	public void setDamage(StandardWorldModel world, int value, int time) {
		if (me instanceof Civilian) {
			if (value != getDamage()) {
				((Civilian) me).getState().m_propertyChanged = true;
			}
		}
		Util.updateSingleDataList(m_damage, value, time, true,
				SeuParameter.LIST_LENGTH_HUMANOID_DAMAGE);
	}

	public void setBuriedness(StandardWorldModel world, int value, int time) {
		Util.updateSingleDataList(m_buriedness, value, time, true,
				SeuParameter.LIST_LENGTH_HUMANOID_BURIEDNESS);
	}

	public void setProperty(StandardWorldModel world, int time, Property type , int[] value) {
		if(type == me.getPositionProperty()){
			me.setPosition(new EntityID(value[0]));
			setPosition(world, value[0], time);
		}
		else if(type == me.getHPProperty()){
			me.setHP(value[0]);
			setHp(world, value[0], time);
		}
		else if(type == me.getDamageProperty()){
			me.setDamage(value[0]);
			setDamage(world, value[0], time);
		}
		else if(type == me.getBuriednessProperty()){
			me.setBuriedness(value[0]);
			setBuriedness(world, value[0], time);
		}
		else if(type == me.getPositionHistoryProperty()){
			me.setPositionHistory(value);
			setPositionHistory(value);
		}
		else{
			super.setProperty(world, time, type, value);
		}
//		switch (type) {
//		default:
//			super.setProperty(world, time, type, value);
//			break;
//		// case PROPERTY_STAMINA: setStamina(value[0]); break;
//		case ProtocolConstants.PROPERTY_POSITION:
//			me.setPosition(new EntityID(value[0]));
//			setPosition(world, value[0], time);
//			break;
//		case ProtocolConstants.PROPERTY_HP:
//			me.setHP(value[0]);
//			setHp(world, value[0], time);
//			break;
//		case ProtocolConstants.PROPERTY_DAMAGE:
//			me.setDamage(value[0]);
//			setDamage(world, value[0], time);
//			break;
//		case ProtocolConstants.PROPERTY_BURIEDNESS:
//			me.setBuriedness(value[0]);
//			setBuriedness(world, value[0], time);
//			break;
//		case ProtocolConstants.PROPERTY_POSITION_EXTRA:
//			setPositionExtra(world, value[0], time);
//			break;
//		case ProtocolConstants.PROPERTY_POSITION_HISTORY:
//			setPositionHistory(value);
//		}
	}

	public String toString() {
		return super.toString();
	}

	// end

	// from Humanoid

	public boolean needLoad(StandardWorldModel world) {
		StandardEntity pos = me.getPosition(world);
		return (me.getBuriedness() == 0 && me.getDamage() >= 0
				&& me.getDamage() < 999 && pos != null
				&& !(me.getPosition(world) instanceof Refuge)
				&& me instanceof Civilian && pos instanceof Building);
	}

	public boolean needLoadDead(StandardWorldModel world) {
		StandardEntity pos = me.getPosition(world);
		return (me.getBuriedness() == 0 && me.getDamage() >= 0 && pos != null
				&& !(me.getPosition(world) instanceof Refuge)
				&& me instanceof Civilian && pos instanceof Building);
	}

	public boolean needRescue(StandardWorldModel world) {
		StandardEntity pos = me.getPosition(world);
		return (me.getHP() > 0 && me.getBuriedness() > 0 && pos != null && !(me
				.getPosition(world) instanceof Refuge));
	}

	public boolean needRescueDead(StandardWorldModel world) {
		StandardEntity pos = me.getPosition(world);
		return (me.getBuriedness() > 0 && pos != null && !(me
				.getPosition(world) instanceof Refuge));
	}

	public boolean isBuried() {
		return (me.getBuriedness() > 0);
	}

	public Report reportLocked = new Report();
	public Report reportUnlocked = new Report();
	//
	public boolean isAtLeader = false;
	public boolean isPfLeader = false;
	public boolean isFbLeader = false;

	public boolean isVirtualCenter = false;

	private boolean needGiveWay = false;

	public boolean locked = false;

	public void updateGiveWay(boolean giveWay) {
		needGiveWay = giveWay;
	}

	public boolean giveWay() {
		if (me instanceof PoliceForce)
			return false;
		return needGiveWay;
	}

	// TODO
	// public void cloneTo(HumanState o){
	// super.cloneTo(o);
	// }

	// ================estimate used============
	public boolean isRescueAble(StandardWorldModel world) {
		if (!(me instanceof Civilian)) {
			return true;
		}

		int atNum = world.getState().ambulanceTeams.size() - 2;
		if (atNum <= 1)
			return true;
		int allowanceCycle = 4;
		int eLifespan = ((Civilian) me).getState().getDeadTime()
				- world.getState().getTime();
		int elapseTime = 5;
		int buriedness = getBuriedness();
		int diff = (eLifespan - elapseTime - allowanceCycle)
				- (int) Math.ceil((double) buriedness / (double) atNum);

		return diff > 0;
	}

	public int m_sayToTime = -1;

	// ===================AIS used==================
	private int m_lastUpdateTime = 2;

	public int lastUpdateTime() {
		return m_lastUpdateTime;
	}

	public void setLastUpdateTime(int value) {
		m_lastUpdateTime = value;
	}

	private float m_lastStimValue = 0;

	public float lastStimValue() {
		return m_lastStimValue;
	}

	public void setLastStimValue(float value) {
		m_lastStimValue = value;
	}

	private float m_lastConcentration = 0.5f;

	public float lastConcentration() {
		return m_lastConcentration;
	}

	public void setLastConcentration(float value) {
		m_lastConcentration = value;
	}

	private float m_concentration = 0.5f;

	public float concentration() {
		return m_concentration;
	}

	public void setConcentration(float value) {
		m_concentration = value;
	}

	// ====================================================================================
	public Area getAreaPosition(StandardWorldModel world) {
		Area mlo = null;
		StandardEntity pos = ((Human) me).getPosition(world);
		if (pos instanceof Area) {
			mlo = (Area) pos;
		} else if (pos instanceof Human) {
			mlo = (Area) ((Human) pos).getPosition(world);
		}
		return mlo;
	}

	// public Area getAreaPosition(StandardWorldModel world , int time) {
	// if(getRealPosition(world , time) == null)
	// return null;
	// return getRealPosition(world, time).getState().getAreaPosition(world);
	// }
	//
	// public StandardEntity getRealPosition(StandardWorldModel world) {
	// return (StandardEntity) world.getState().get(getIdPosition());
	// }
	//	
	// public StandardEntity getRealPosition(StandardWorldModel world , int
	// time) {
	// return (StandardEntity) world.getState().get(getIdPosition(time));
	// }

	protected LinkedList<SingleData<Integer>> m_position = new LinkedList<SingleData<Integer>>();

	// private LinkedList<SingleData<Integer>> m_positionHistory = new
	// LinkedList<SingleData<Integer>>();

	private LinkedList<SingleData<Integer>> m_positionExtra = new LinkedList<SingleData<Integer>>();

	private LinkedList<SingleData<Integer>> m_direction = new LinkedList<SingleData<Integer>>();

	public int getIdPosition() {
		return m_position.getLast().getObject();
	}

	// public int getIdPositionHistory() {
	// return m_positionHistory.getLast().getObject();
	// }

	public int getPositionExtra() {
		return m_positionExtra.getLast().getObject();
	}

	public List<SingleData<Integer>> getAllPosExtra() {
		return m_positionExtra;
	}

	public int getDirection() {
		return m_direction.getLast().getObject();
	}

	public int getIdPosition(int time) {
		return Util.searchSingleDataForTime(m_position, time);
	}

	public int getPositionExtra(int time) {
		return Util.searchSingleDataForTime(m_positionExtra, time);
	}

	public int getDirection(int time) {
		return Util.searchSingleDataForTime(m_direction, time);
	}

	public void setPosition(StandardWorldModel world, int value, int time) {
		Util.updateSingleDataList(m_position, value, time, true, world
				.getState().getParameter().LIST_LENGTH_MOVINGOBJECT_POSITION);
	}

	public void setPositionExtra(StandardWorldModel world, int value, int time) {
		Util
				.updateSingleDataList(
						m_positionExtra,
						value,
						time,
						true,
						world.getState().getParameter().LIST_LENGTH_MOVINGOBJECT_POSITIONEXTRA);
	}

	public void setDirection(StandardWorldModel world, int value, int time) {
		Util.updateSingleDataList(m_direction, value, time, true, world
				.getState().getParameter().LIST_LENGTH_MOVINGOBJECT_DIRECTION);

	}

	private int[] m_positionHistory;

	public void setPositionHistory(int[] value) {
		m_positionHistory = value;
	}

	public int[] getPositionHistory() {
		return m_positionHistory;
	}

	public LinkedList<List> getPositionHistoryNodes() {
		return positionHistoryMap;
	}

	private final int SIM_TIME = seu.agent.Constants.SIMULATING_TIME;

	private int[][] positionHistory = new int[SIM_TIME + 2][];

	public void updatePositionHistory(StandardWorldModel world) { // Should Be
		// Run Each
		// Cycle
		// ONCE
		if (world.getState().getTime() > 300 || me != world.getState().me)
			return;
		int[] ph = getPositionHistory();
		int time = world.getState().getTime();
		positionHistory[time] = ph;
		moHistory[time] = getAreaPosition(world);

		hasMovedInPrevCycle[time] = hasMovedInLastCycle(world);
		stuckCycles[time] = calStuckCycles(world);
		m_poshistObjs.clear();
		m_poshistRoads.clear();
		m_poshistNodes.clear();
		if (ph[0] == 0) {
			System.out.println("aa");
			return;
		}
		Road node;
		for (int i = 0; i < ph.length; i++) {
			node = (Road) world.getEntity(new EntityID(ph[i]));
			m_poshistNodes.add(node);
		}

		if (!m_poshistNodes.isEmpty())
			positionHistoryMap.addLast(m_poshistNodes);

		m_poshistObjs.addAll(m_poshistNodes);
		if (ph.length < 2)
			return;

	}

	public int[] getPositionHistory(int time) {
		return positionHistory[time];
	}

	private final List m_poshistObjs = new ArrayList();

	public List getPosHistObjs() {
		return m_poshistObjs;
	}

	private final List m_poshistRoads = new ArrayList();

	public List getPosHistRoads() { // Doesnt Include Roads That Aren't
		// Completely Travelled
		return m_poshistRoads;
	}

	private final List m_poshistNodes = new ArrayList();

	protected LinkedList<List> positionHistoryMap = new LinkedList<List>();

	public List getPosHistNodes() {
		return m_poshistNodes;
	}

	public Node getLatestPositionHistoryNodes(Road rd, StandardWorldModel world) {

		if (positionHistoryMap.isEmpty())
			return null;

		List<Node> historyNodes = positionHistoryMap.getLast();
		if (historyNodes.contains(rd.getState().getPointHead(world)))
			return (Node) rd.getState().getPointHead(world);
		if (historyNodes.contains(rd.getState().getPointTail(world)))
			return (Node) rd.getState().getPointTail(world);

		return null;
	}

	private boolean hasMovedInLastCycle(StandardWorldModel world) {
		// a new method to determine whether move in last cycle
		Area posNow = getAreaPosition(world);
		Area posPre = moHistory[world.getState().getTime() - 1];
		if (posNow != posPre)
			return true;
		else {
			if (posNow instanceof Road || posNow instanceof Building)
				return false;
			else if (posNow instanceof Road) {
				int posExtraNow = getPositionExtra();
				int posExtraPre = getPositionExtra(world.getState().getTime() - 1);
				if (Math.abs(posExtraNow - posExtraPre) > 20)
					return true;
				else
					return false;
			} else
				return false;
		}

	}

	private boolean[] hasMovedInPrevCycle = new boolean[SIM_TIME + 2];

	public boolean HasMovedInPrevCycle(int time) {
		return hasMovedInPrevCycle[time];
	}

	public boolean HasMovedInLastCycle(StandardWorldModel world) {
		return hasMovedInPrevCycle[world.getState().getTime()];
	}

	private Area moHistory[] = new Area[SIM_TIME + 2];

	public boolean getSubmittedMoveAct(int time) {
		return subMittedAct[time] == MOVE_ACT;
	}

	public boolean submittedMoveAct() {
		return subMittedAct[getTime() - 1] == MOVE_ACT;
	}

	public boolean rested() {
		return subMittedAct[getTime() - 1] == REST_ACT;
	}

	public boolean getRested(int time) {
		return subMittedAct[time] == REST_ACT;
	}

	private int[] subMittedAct = new int[SIM_TIME + 2];

	public void setAct(int act, int targetId) {
		if (!isValidAct(act))
			throw new Error("WHA");
		subMittedAct[getTime()] = act;
	}

	public boolean submittedAction() {
		return subMittedAct[getTime() - 1] > MOVE_ACT;
	}

	private boolean isValidAct(int act) {
		if (act < REST_ACT || act > CLEAR_ACT)
			return false;
		return true;
	}

	public int getSubmittedAction(int time) {
		return subMittedAct[time];
	}

	private int[] stuckCycles = new int[SIM_TIME + 2];

	private int calStuckCycles(StandardWorldModel world) {
		int prevVal = stuckCycles[world.getState().getTime() - 1];
		if (!submittedMoveAct()) {
			// if(!SeuConstants.IS_COMPETETION_MODE)
			// world.logFile.writeString(" not move:   "+getTime(), 10,
			// "stuck");
			return 0;
		} else {
			if (hasMovedInPrevCycle[world.getState().getTime()]) {
				// if(!SeuConstants.IS_COMPETETION_MODE)
				// world.logFile.writeString("has moved in:  "+getTime(), 10,
				// "stuck");
				return 0;
			} else {
				// if(!SeuConstants.IS_COMPETETION_MODE)
				// world.logFile.writeString("stuck in:  "+getTime(), 10,
				// "stuck");
				return prevVal + 1;
			}
		}
	}

	public int getStuckCycles(int time) {
		return stuckCycles[time];
	}

	public int getStuckCycles() {
		return stuckCycles[getTime()];
	}

	private boolean[] wereLockedByBlockade = new boolean[SIM_TIME + 2];
	public ArrayList<Node> nearBlockNodes = null;

	public void setWereLockedByBlockade(StandardWorldModel world,
			boolean isLocked) {
		wereLockedByBlockade[world.getState().getTime()] = isLocked;
	}

	public boolean getWereLockedByBlockade(int time) {
		return wereLockedByBlockade[time];
	}

	public boolean getWereLockedByBlockade() {
		return wereLockedByBlockade[getTime()];
	}

	public ArrayList<EntityID> pathMove(Path path,StandardWorldModel model){
		if(path == null)
			return null;
		
		if(!path.isOnPath(me.getPosition(model)))
			for(Road rd : path.getContainedRoads()){
				if(rd.getState().getCondition() == RoadState.CONDITION_PASSABLE)
					continue;
				return new ArrayList<EntityID>(Collections.singleton(rd.getID()));
			}

		
		StandardEntity pos = outsidePosition(me,model);
		
		ArrayList<EntityID> route = path.toRoute(true);
		
		int headValue = 0;
		int tailValue = 0;
		boolean toTail = false;
		
		for(EntityID eId:route){
			
			if(eId == pos.getID())
				break;
			
			if(model.getEntity(eId) instanceof Road && ((Road)model.getEntity(eId)).getState().getCondition() != RoadState.CONDITION_PASSABLE){
				headValue ++;
			}
			
		}
		
		
		route = path.toRoute(false);
		for(EntityID eId:route){
			
			if(eId == pos.getID())
				break;
			
			if(model.getEntity(eId) instanceof Road && ((Road)model.getEntity(eId)).getState().getCondition() != RoadState.CONDITION_PASSABLE){
				tailValue ++;
			}
			
		}
		
		if(headValue > tailValue){
			toTail = true;
		}
		
		route = path.toRoute(toTail);
		
		if(route!=null)
			return route;
		
		return null;
		
	}
	public static Road outsidePosition(Human mo,StandardWorldModel world) {
		StandardEntity mop = mo.getPosition(world);
		if (mop == null)
			return null;
		if (mop instanceof Road) {
			Road rd = (Road) mop;
			return rd;
		}
		if (mop instanceof Building){
			for(Edge eg:((Building)mop).getEdges())
				if(eg.isPassable() && world.getEntity(eg.getNeighbour()) instanceof Road)
					return (Road)world.getEntity(eg.getNeighbour());
		}
		throw new Error("Should Be Returned Already");
	}
}
