//SEU2010

package seu.states;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import seu.SeuConstants;
import seu.entities.Node;
import seu.entities.Path;
import seu.entities.Report;
import seu.utilities.AgentUtil;
import seu.utilities.base.SingleData;

public class RoadState extends AreaState {

	private Road me;
	private boolean isConnector;
	private boolean roadChecked=false;

	public RoadState(Road me) {
		super(me);
		this.me = me;
	}

	/**
	 * 
	 * By JiangYh 2010-6-12
	 */
	public void initNode(StandardWorldModel world) {
		if (m_node.size() > 0)
			return;
		for (Edge eg : me.getEdges())
			if (eg.isPassable())
				m_node.add(new Node(me, eg));

		for (Node nd : m_node)
			if (nd.getNeighborhood(world) instanceof Road)
				m_roadnode.add(nd);
	}
	public void initRoadNode(StandardWorldModel world){
		if(roadChecked) return;
		HashSet<Node> needless=new HashSet<Node>();
		for(Node nd:m_roadnode)
			if(((Road)nd.getNeighborhood(world)).getState().m_roadnode.size()==1)
				needless.add(nd);
		m_roadnode.removeAll(needless);
		roadChecked=true;
	}

//	public int getType() {
//		return SeuConstants.TYPE_ROAD;
//	}

	public ArrayList<Node> m_roadnode = new ArrayList<Node>();
	public Report reportClear = new Report();

	public Report reportUnClear = new Report();

	private boolean m_isReportedAsClear = false;

	public boolean isReportedAsClear() {
		return m_isReportedAsClear;
	}

	public void setIsReportedAsClear(boolean isReportedAsClear) {
		this.m_isReportedAsClear = isReportedAsClear;
	}

	private boolean m_isReportedAsBlock = false;

	public boolean isReportedAsBlock() {
		return m_isReportedAsBlock;
	}

	public void setIsReportedAsBlock(boolean isReportAsBlock) {
		m_isReportedAsBlock = isReportAsBlock;
	}

	private ArrayList<Road> m_ConnectedRoads;

	public ArrayList getConnectedRoads(StandardWorldModel world) { // not include
		// self
		if (m_ConnectedRoads == null) {
			m_ConnectedRoads=new ArrayList<Road>();
			if(m_roadnode==null) {initNode(world);initRoadNode(world);}
			for(Node nd:m_roadnode)
				m_ConnectedRoads.add((Road)nd.getNeighborhood(world));
		}
		return m_ConnectedRoads;
	}

	public List<Road> getConnectedRoads(int HEAD_TAIL, StandardWorldModel world) {
		// 0:getPointHead
		// 1:getPointTail
		List<Road> hCol = ((Node) me.getState().getPointHead(world))
				.getConnectedRoads(world);
		List<Road> tCol = null;
		if (me.getState().getPointTail(world) != null)
			tCol = ((Node) me.getState().getPointTail(world))
					.getConnectedRoads(world);

		if (HEAD_TAIL == 0)
			return hCol;
		return tCol;
	}

	private List<Building> m_ConnectedBuildings;

	public List<Building> getConnectedBuildings(StandardWorldModel world) {
		if (m_ConnectedBuildings == null) {
			m_ConnectedBuildings = new ArrayList<Building>();
			for (Node nd : m_node) {
				if (nd.getNeighborhood(world) instanceof Building)
					m_ConnectedBuildings.add((Building) nd
							.getNeighborhood(world));
				if (nd.getNeighborhood(world) instanceof Road
						&& ((Road) nd.getNeighborhood(world)).getState().m_roadnode
								.size() == 1)
					m_ConnectedBuildings.addAll(((Road) nd
							.getNeighborhood(world)).getState()
							.getConnectedBuildings(world));

			}
		}
		return m_ConnectedBuildings;
	}

	private List<Path> m_containedPaths = null;

	Node headpoint = null, tailpoint = null;

	/**
	 * @return By JiangYh 2010-5-18
	 */
	public Node getPointTail(StandardWorldModel world) {
		if (tailpoint == null) {
			initNode(world);
			initRoadNode(world);
		}
		if (m_roadnode.size() == 0)
			return null;
		for (Node nd : m_roadnode){
			Road rd = (Road) nd.getNeighborhood(world);
			rd.getState().initNode(world);
			rd.getState().initRoadNode(world);
			if (rd.getState().m_roadnode.size() > 1
					&& (headpoint == null || headpoint != nd)) {
				tailpoint = nd;
				break;
			}
		}
		return tailpoint;
	}

	/**
	 * @return Head node of this Road By JiangYh 2010-5-18
	 */
	public Node getPointHead(StandardWorldModel world) {
		if (headpoint == null) {
			initNode(world);
			initRoadNode(world);
		}
		// System.out.println("m_roadnode.size() = "+m_roadnode.size());
		if (m_roadnode.size() == 0)
			return null;
		for (Node nd : m_roadnode) {
			Road rd = (Road) nd.getNeighborhood(world);
			rd.getState().initNode(world);
			rd.getState().initRoadNode(world);
			if (rd.getState().m_roadnode.size() > 1
					&& (tailpoint == null || tailpoint != nd)) {
				headpoint = nd;
				break;
			}
		}
		// if (headpoint == null) System.out.println(me.toString() +
		// "head null");
		return headpoint;
	}

	private int m_passTime = 0;

	public void setPassTime(int time) {
		m_passTime = time;
	}

	public int getPassTime() {
		return m_passTime;
	}

	public boolean hasBeenPassed() {
		return m_passTime >= seu.agent.Constants.TIME_STARTING_ACTION;
	}

	// public boolean isClearable() {
	// return (getBlockedLines() >= 1);
	// }

	public static final int CONDITION_UNKNOWN = 0;
	public static final int CONDITION_PASSABLE = 1;
	public static final int CONDITION_BLOCKED = 2;
	public static final int CONDITION_MAYBE_PASSABLE = 3;

	private int m_condition = CONDITION_UNKNOWN;

	public ArrayList<Node> m_node = new ArrayList<Node>();

	public void setCondition(int condition) {
		m_condition = condition;
	}

	public int getCondition() {
		return m_condition;
	}

	// ============================================================================================//

	public List<Path> getContainedPaths(StandardWorldModel world) {
		if (m_containedPaths == null) {
			m_containedPaths = new ArrayList<Path>();
			for (Path path : world.getState().paths) {
				if (path.getContainedRoads().contains(me)) {
					m_containedPaths.add(path);
				}
			}
		}
		return m_containedPaths;
	}

	public List<Path> getContainedPaths(List<Path> pathss) {
		if (m_containedPaths == null) {
			m_containedPaths = new ArrayList<Path>();
			for (Path path : pathss) {
				if (path.getContainedRoads().contains(me)) {
					m_containedPaths.add(path);
				}
			}
		}
		return m_containedPaths;
	}

	public Road getOtherEndNode(Road one, StandardWorldModel world) {
//		System.out.println("+++++++++++++++++++++++++++++++");
		for (Node nd : me.getState().m_roadnode){
			if (nd.me().getNeighbour().getValue() != one.getID().getValue()){

//				System.out.println(nd.me().getNeighbour());
//				System.out.println(one.getID());
				return (Road) nd.getNeighborhood(world);
			}
		}
		return null;
	}

	private ArrayList<Road> m_neighb = null;

	public Collection getNeighborhood(StandardWorldModel world) {
		if (m_neighb == null) {
			m_neighb = new ArrayList<Road>();
			for (Node nd : m_roadnode)
				m_neighb.add((Road) nd.getNeighborhood(world));
		}
		return m_neighb;
	}

	private int length;

	/**
	 * @return By JiangYh 2010-5-18
	 */
	public int getLength(StandardWorldModel world) {
		if (length != 0)
			return length;
		if(this.getPointHead(world) == null){
//			System.out.println("no head");
			return 0;
		}
		if(this.getPointTail(world) == null){
//			System.out.println("no head");
			return 0;
		}
		return AgentUtil.distance(this.getPointHead(world).me(), this
				.getPointTail(world).me());
	}

	protected LinkedList<SingleData<Integer>> m_block = new LinkedList<SingleData<Integer>>();

	protected LinkedList<SingleData<Integer>> m_repairCost = new LinkedList<SingleData<Integer>>();

	public void setBlock(int value, int time, StandardWorldModel world) {

		AgentUtil.updateSingleDataList(m_block, value, time, true, world
				.getState().getParameter().LIST_LENGTH_ROAD_BLOCK);

	}
}
