//SEU2010

package seu.states;

//import static rescuecore2.standard.Constants.ENTITY_URN_PREFIX;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rescuecore2.registry.Registry;
import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.Property;
import seu.SeuParameter;
import seu.agent.Constants;
import seu.agent.Util;
import seu.entities.Node;
import seu.entities.Path;
import seu.entities.Report;
import seu.utilities.AgentUtil;
import seu.utilities.ConnectedComponents;
import seu.utilities.ConstantConditions;
import seu.utilities.SimulatorsConstants;
import seu.utilities.Wall;
import seu.utilities.Partition.RadialPartition;

public class WorldState implements ConstantConditions{

	private StandardWorldModel world;

	public WorldState(StandardWorldModel world, EntityID selfId) {
		this.world = world;
		mparameter = new SeuParameter();
		if (selfId.getValue() > 0) {
			update(world.getAllEntities(), 0);
			
			// System.out.println(" : worldStateData : agents="+agents.size()
			// +" centers="+centerAgents.size()
			// +" platoons="+platoonAgents.size()
			// +" specialBuildings="+specialBuildings.size()
			// +" emptyBuildings="+emptyBuildings.size()
			// +" burningBuildings="+burningBuildings.size()
			// +" paths="+paths.size()
			// +" nodes="+nodes.size()
			// +" passedNodes="+passedNodes.size()
			// +" savedHumanoids="+savedHumanoids.size());
			me = world.getEntity(selfId);
			calculateAverageAreas();
			
			mapWidth = getBoundary(3) - getBoundary(2);
			mapHeight = getBoundary(0) - getBoundary(1);

			notSeenBuildings.addAll(buildings);
			notSeenBuildings.removeAll(specialBuildings);

			// if (!SeuConstants.IS_COMPETETION_MODE)
			// createLogFile();
			// long t = System.currentTimeMillis();
			initialAllPaths();
			// long t1 = System.currentTimeMillis();
			// if((t1-t)>50)
			// System.out.println("create path used  "+(t1-t));

//			initializeVisibleBuildingsPerBuilding();
//			System.out.println("3");

			// createPartitions();
			// long t2 = System.currentTimeMillis();
			// if ((t2 - t1) > 50)
			// System.out.println("visible used== " + (t2 - t1));

//			generatePartitions();

			// long t3=System.currentTimeMillis();
			// if((t3-t2)>50)
			// System.out.println("partition used === "+(t3-t2));

			computePrimitiveCalcs();

			// long t4=System.currentTimeMillis();
			// if((t4-t3)>50)
			// System.out.println("compute used== "+(t4-t3));

			CG = new ConnectedComponents(world, paths);

			// pathPlanner = new ShortestPathPlanner(this);
		} else {
			me = null;
			// pathPlanner = null;
			CG = null;
		}
	}

	// ============================= basic world variables
	// ========================================//
	private final HashMap<EntityID, StandardEntity> idEntityMap = new HashMap<EntityID, StandardEntity>();
	public EntityID selfId;

	private SeuParameter mparameter;

	public int channelCount;
	public int channel1;
	public int channel2;
	public int channel3;
	
	public final List<Area> areas = new ArrayList<Area>();
	public final List<Road> roads = new ArrayList<Road>();
	public final List<Building> buildings = new ArrayList<Building>();
	public final List<Building> specialBuildings = new ArrayList<Building>();
	public final List<Refuge> refuges = new ArrayList<Refuge>();
	public List<Refuge> usefulRefuges = new ArrayList<Refuge>();
	public List<Road> checkedRoads = new ArrayList<Road>();
	public final List<Blockade> blockades = new ArrayList<Blockade>();
	
	public final List<Road> passableRoads = new ArrayList<Road>();
	public final List<Road> unPassableRoads = new ArrayList<Road>();
	public final List<Road> unKnownRoads = new ArrayList<Road>();

	public final List<Human> humans = new ArrayList<Human>();
	public final List<Civilian> civilians = new ArrayList<Civilian>();
	public final List<AmbulanceTeam> ambulanceTeams = new ArrayList<AmbulanceTeam>();
	public final List<FireBrigade> fireBrigades = new ArrayList<FireBrigade>();
	public final List<PoliceForce> policeForces = new ArrayList<PoliceForce>();
	public final ArrayList<AmbulanceCentre> ambulanceCenters = new ArrayList<AmbulanceCentre>();
	public final ArrayList<FireStation> fireStations = new ArrayList<FireStation>();
	public final ArrayList<PoliceOffice> policeOffices = new ArrayList<PoliceOffice>();
	public final List<StandardEntity> agents = new ArrayList<StandardEntity>();
	public final List<Building> centerAgents = new ArrayList<Building>();
	public final List<Human> platoonAgents = new ArrayList<Human>();
	public final List<Human> needHelpAgents = new ArrayList<Human>();
	

	public final ConnectedComponents CG;

	protected int m_time = 0;

	private int m_MaxX = 0;

	private int m_MaxY = 0;

	private int m_MinX = Integer.MAX_VALUE;

	private int m_MinY = Integer.MAX_VALUE;

	// TODO
	// public LogFileWriter logFile;

	// ============================= basic world methods
	// ========================================//
	public void setTime(int time) {
		this.m_time = time;
	}

	public int getTime() {
		return m_time;
	}

	public StandardEntity get(EntityID id) {
		return idEntityMap.get(id);
	}

//	public int getSelfType() {
//		return world.getEntity(selfId).getState().getType();
//	}

	public Collection<StandardEntity> getEntities() {
		return idEntityMap.values();
	}

	// TODO
	// public void updateProperty(int time, StandardEntity obj, int type, int[]
	// val)
	// {
	// obj.setProperty(time, type, val);
	// }

	public SeuParameter getParameter() {
		return mparameter;
	}

	public int getMaxX() {
		return m_MaxX;
	}

	public int getMaxY() {
		return m_MaxY;
	}

	public int getMinX() {
		return m_MinX;
	}

	public int getMinY() {
		return m_MinY;
	}

	public void add(EntityID id) {
		idEntityMap.put(id, world.getEntity(id));
		StandardEntity entity = world.getEntity(id);
		if (getTime() == 0) {
			if (entity instanceof Area) {
				areas.add((Area) entity);

				if (entity instanceof Road) {
					// System.out.println(entity.toString());
					roads.add((Road) entity);
					Node nd = ((Road) entity).getState().getPointHead(world);
					if (nodes.contains(nd))
						nodes.add(nd);
					nd = ((Road) entity).getState().getPointTail(world);
					if (nodes.contains(nd))
						nodes.add(nd);
				} else if (entity instanceof Building) {
					buildings.add((Building) entity);
					if (entity instanceof Refuge) {
						refuges.add((Refuge) entity);
						specialBuildings.add((Refuge) entity);
					} else if (entity instanceof AmbulanceCentre) {
						agents.add((AmbulanceCentre) entity);
						ambulanceCenters.add((AmbulanceCentre) entity);
						specialBuildings.add((AmbulanceCentre) entity);
						centerAgents.add((AmbulanceCentre) entity);
					} else if (entity instanceof FireStation) {
						agents.add((FireStation) entity);
						fireStations.add((FireStation) entity);
						specialBuildings.add((FireStation) entity);
						centerAgents.add((FireStation) entity);
					} else if (entity instanceof PoliceOffice) {
						agents.add((PoliceOffice) entity);
						policeOffices.add((PoliceOffice) entity);
						specialBuildings.add((PoliceOffice) entity);
						centerAgents.add((PoliceOffice) entity);
					}
				}
			} else if (entity instanceof Human) {
				humans.add((Human) entity);
				if (entity instanceof Civilian)
					civilians.add((Civilian) entity);

				else if (entity instanceof AmbulanceTeam) {
					agents.add((AmbulanceTeam) entity);
					ambulanceTeams.add((AmbulanceTeam) entity);
					platoonAgents.add((AmbulanceTeam) entity);
				} else if (entity instanceof FireBrigade) {
					agents.add((FireBrigade) entity);
					fireBrigades.add((FireBrigade) entity);
					platoonAgents.add((FireBrigade) entity);
				} else if (entity instanceof PoliceForce) {
					agents.add((PoliceForce) entity);
					policeForces.add((PoliceForce) entity);
					platoonAgents.add((PoliceForce) entity);
				}

			}
		} else {
			if (entity instanceof Human) {
				humans.add((Human) entity);
				if (entity instanceof Civilian)
					civilians.add((Civilian) entity);
			}
		}

	}

	public void setWorldRange() {
		Iterator<StandardEntity> it = idEntityMap.values().iterator();
		while (it.hasNext()) {
			Object obj = it.next();
			if (!(obj instanceof Area))
				continue;
			Area po = (Area) obj;
			if (m_MinX > po.getX())
				m_MinX = po.getX();
			if (m_MaxX < po.getX())
				m_MaxX = po.getX();
			if (m_MinY > po.getY())
				m_MinY = po.getY();
			if (m_MaxY < po.getY())
				m_MaxY = po.getY();
		}
	}

	// ============================== Self-defining Variables
	// ====================================//

	public StandardEntity me;

	private static final int SIMULATING_TIME = seu.agent.Constants.SIMULATING_TIME;

	public final List<Building> emptyBuildings = new ArrayList<Building>();
	public final List<Building> notSeenBuildings = new ArrayList<Building>();
	public final HashSet<Building> visitedBuildings = new HashSet<Building>();
	public final ArrayList<Human> buriedAgents = new ArrayList<Human>();
	public final HashMap<Integer, Human> lockedByBlockadeAgentMap = new HashMap<Integer, Human>();
	public final HashSet<Building> burningBuildings = new HashSet<Building>();
	public final HashSet<Building> reBurningBuildings = new HashSet<Building>();

	public final List<Human> updatedHumans = new ArrayList<Human>();
	public List<Building> updatedBuildings = new ArrayList<Building>();
	public final List<Civilian> updatedCivilians = new ArrayList<Civilian>();
	public List<Road> updatedRoads = new ArrayList<Road>();
	public List<Blockade> updatedBlockades = new ArrayList<Blockade>();

	public final List<Path> paths = new ArrayList<Path>();

	public final List<Node> nodes = new ArrayList<Node>();

	public ArrayList<Civilian> foundVictimCivilianIC = new ArrayList<Civilian>();
	public final HashSet<Civilian> heardCivilians = new HashSet<Civilian>();
	public final HashSet<Human> savedHumanoids = new HashSet<Human>();

	public final HashSet<Node> passedNodes = new HashSet<Node>(); // used for
	// dijkstra
	public final Set<Area> inTraffic = new HashSet<Area>();// TODO
	// update
	// in
	// method
	// of
	// computeTraffic

	public static Node Center = null;
	public static int mapWidth = 0;
	public static int mapHeight = 0;

	public final HashSet<Road> blockRoads = new HashSet<Road>();

	// private MapPartitioner partitioner;

	// public final ShortestPathPlanner pathPlanner;

	// public final List<MapPartition> partitions = new
	// ArrayList<MapPartition>();

	public final List<RadialPartition> radialParitions = new ArrayList<RadialPartition>();

	// public final List<FireZone> fireZones = new ArrayList<FireZone>();
	//
	// public final HashMap<Building, FireZone> zoneMap = new HashMap<Building,
	// FireZone>();

	public int m_maxVisibleBuildingsPerNode = 0;
	public int m_maxHearableBuildingsPerNode = 0;
	public int m_maxVisibleBuildingsPerBuilding = 0;
	public int m_maxHearableBuildingsPerBuilding = 0;

	public static float mapBuildingAreaTotal = 0;
	public static float mapBuildingAreaAverage = 0;

	//
	private Long hashValue = null;
	int RAY_DOMAIN = 3;
	int FIRE_PASSING_DISTANCE = 50000;

	public final HashSet<Road> passedRoads = new HashSet<Road>();
	// ===============================================================================================//

	public Report reportLocked = new Report();
	public Report reportUnlocked = new Report();
	public Report report = new Report();

	private boolean[] wereLockedByBlockade = new boolean[SIMULATING_TIME + 2];

	private boolean areaCheck = false;

	public void setWereLockedByBlockade(boolean locked, int time) {
		wereLockedByBlockade[time] = locked;
	}

	public boolean getWereLockedByBlockade(int time) {
		return wereLockedByBlockade[time];
	}

	public void setAreaChecked(boolean check) {
		areaCheck = check;
	}

	public boolean getAreaChecked() {
		return areaCheck;
	}

	// ======================================================================================

	/**
	 * @return the hashValue
	 */
	public Long getHashValue() {
		if (null == hashValue)
			return hash();
		return hashValue;
	}

	public long hash() {
		if (hashValue == null) {
			long sum = 0;
			for (Iterator<Node> i = nodes.iterator(); i.hasNext();) {
				Node node = i.next();
				if (Long.MAX_VALUE - sum <= node.getX()) {
					sum = 0;
				}
				sum += node.getX();

				if (Long.MAX_VALUE - sum <= node.getY()) {
					sum = 0;
				}
				sum += node.getY();

				// int[] ap = b.getBuildingApexes();
				// for (int c = 0; c < ap.length; c++)
				// {
				// if (Long.MAX_VALUE - sum <= ap[c])
				// {
				// sum = 0;
				// }
				// sum += ap[c];
				// }
			}
			hashValue = new Long(sum);
		}
		// System.err.println("Hash Vlaue of Map is = " +
		// hashValue.longValue());
		return hashValue.longValue();
	}

	/**
	 * @param ox
	 * @param oy
	 * @return (x,y)Building
	 */
	public Building getBuilding(int ox, int oy) {
		// TODO Knight Guan
		for (Iterator<Building> i = buildings.iterator(); i.hasNext();) {
			Building b = i.next();
			if (b.getState().isBuilding(ox, oy)) {
				return b;
			}
		}
		System.out.println("parser error");
		throw new NullPointerException();
	}

	// ===========================================================================
	public static final Comparator ID_COMP = new Comparator() {
		final Comparator<StandardEntity> idComparator = Util.idComparator;

		public int compare(Object obj1, Object obj2) {
			if (obj1 instanceof StandardEntity
					&& obj2 instanceof StandardEntity) {
				return idComparator.compare((StandardEntity) obj1,
						(StandardEntity) obj2);
			}
			System.err.println("Invalid Compare : " + obj1 + " With : " + obj2);
			return (obj1.hashCode() - obj2.hashCode());
		}
	};

	public void calculateAverageAreas() {
		if (Center == null)
			Center = getCenter();
		for (Building b : buildings) {
			mapBuildingAreaTotal += b.getTotalArea();
		}
		mapBuildingAreaAverage = mapBuildingAreaTotal / buildings.size();
	}

	public Node getCenter() {
		if (Center != null)
			return Center;
		if (nodes.size() == 0)
			return null;
		long x = 0, y = 0;
		Node result;
		for (int i = nodes.size() - 1; i >= 0; --i) {
			Node node = nodes.get(i);
			x += (node.me().getEndX() + node.me().getStartX()) / 2;
			y += (node.me().getEndY() + node.me().getStartY()) / 2;
		}
		x /= nodes.size();
		y /= nodes.size();
		result = nodes.get(0);
		for (int i = 1; i < nodes.size(); ++i) {
			Node temp = nodes.get(i);
			if (AgentUtil.distance((int) x, (int) y,
					(result.me().getEndX() + result.me().getStartX()) / 2,
					(result.me().getEndY() + result.me().getStartY()) / 2) > AgentUtil
					.distance((int) x, (int) y, (temp.me().getEndX() + temp
							.me().getStartX()) / 2, (temp.me().getEndY() + temp
							.me().getStartY()) / 2)) {
				result = temp;
			}
		}
		Center = result;
		return result;
	}

	private int getBoundary(int dir) { // 0 : Up , 1 : Down , 2 : Left , 3 :
		// Right
		int result = -1;
		int overlap = 100;
		int max;

		if (this == null || this.areas.isEmpty())
			return -1;

		if (dir == 0 || dir == 1)
			max = this.areas.get(0).getY();
		else if (dir == 2 || dir == 3)
			max = this.areas.get(0).getX();
		else
			max = -1;
		switch (dir) {
		case 0:
			for (Iterator it = this.areas.iterator(); it.hasNext();) {
				Area res = (Area) it.next();
				if (res.getY() > max)
					max = res.getY();
			}
			result = max + overlap;
			break;
		case 1:
			for (Iterator it = this.areas.iterator(); it.hasNext();) {
				Area res = (Area) it.next();
				if (res.getY() < max)
					max = res.getY();
			}
			result = max - overlap;
			break;
		case 2:
			for (Iterator it = this.areas.iterator(); it.hasNext();) {
				Area res = (Area) it.next();
				if (res.getX() < max)
					max = res.getX();
			}
			result = max - overlap;
			break;
		case 3:
			for (Iterator it = this.areas.iterator(); it.hasNext();) {
				Area res = (Area) it.next();
				if (res.getX() > max)
					max = res.getX();
			}
			result = max + overlap;
			break;
		default:
			break;
		}
		return result;
	}

	/**
	 * initialize visible buildings of every building
	 * 
	 */
	@SuppressWarnings("unchecked")
	protected void initializeVisibleBuildingsPerBuilding() {
		long hash = hash();
		boolean loaded = false;
		String str = "visble/";
		String fname = str + hash + ".vb";

		try {
			File f = new File(fname);
			BufferedReader br = new BufferedReader(new FileReader(f));

			String nl;

			// long t1 = System.currentTimeMillis();
			while (null != (nl = br.readLine())) {
				// 
				int quantity = Integer.parseInt(nl);
				int x = Integer.parseInt(br.readLine());
				int y = Integer.parseInt(br.readLine());

				// 
				Building bl = getBuilding(x, y);

				for (int c = 0; c < quantity; c++) {
					int ox = Integer.parseInt(br.readLine());
					int oy = Integer.parseInt(br.readLine());
					bl.getState().addVisibleBuildings(getBuilding(ox, oy));
				}

				m_maxVisibleBuildingsPerBuilding = Math.max(
						m_maxVisibleBuildingsPerBuilding, bl.getState()
								.getVisibleBuildings().size());

				quantity = Integer.parseInt(br.readLine());

				for (int c = 0; c < quantity; c++) {
					int ox = Integer.parseInt(br.readLine());
					int oy = Integer.parseInt(br.readLine());
					bl.getState().addHearableBuildings(getBuilding(ox, oy));
				}

				m_maxHearableBuildingsPerBuilding = Math.max(
						m_maxHearableBuildingsPerBuilding, bl.getState()
								.getHearableBuildings().size());

			}
			long t2 = System.currentTimeMillis();
			// System.out
			// .println("Load visible building file used : " + (t2 - t1));
			loaded = true;
			// System.out
			// .println("loaded visible building file \"" + fname + "\"");
		} catch (Exception e) {
			// System.err.println("unable to load visible building file \" "
			// + fname);

			Iterator it = buildings.iterator();
			while (it.hasNext()) {
				Building b = (Building) it.next();

				// visible
				ArrayList<Building> see = new ArrayList<Building>();
				ArrayList<Building> hearable = new ArrayList<Building>();
				for(Building next : buildings){
					if(world.getDistance(me, next) < Constants.SEE_DISTANCE && !specialBuildings.contains(next)){
						see.add(next);
						b.getState().addVisibleBuildings(see);
						m_maxVisibleBuildingsPerBuilding = Math.max(
								m_maxVisibleBuildingsPerBuilding, see.size());
					}
					if(world.getDistance(me, next) < Constants.HEAR_DISTANCE && !specialBuildings.contains(next)){
						hearable.add(next);
						b.getState().addHearableBuildings(next);
						m_maxHearableBuildingsPerBuilding = Math.max(
								m_maxHearableBuildingsPerBuilding, hearable.size());
					}
				}				

			}

		}
		try {
			long t1 = System.currentTimeMillis();
			if (!loaded) {
				File f = new File(fname);
				f.createNewFile();
				BufferedWriter bw = new BufferedWriter(new FileWriter(f));

				for (Iterator<Building> i = buildings.iterator(); i.hasNext();) {
					Building b = i.next();

					bw.write(b.getState().getVisibleBuildings().size() + "\n");

					bw.write(b.getX() + "\n");
					bw.write(b.getY() + "\n");

					for (int c = 0; c < b.getState().getVisibleBuildings()
							.size(); c++) {
						bw.write(b.getState().getVisibleBuildings().get(c)
								.getX()
								+ "\n");
						bw.write(b.getState().getVisibleBuildings().get(c)
								.getY()
								+ "\n");
					}

					bw.write(b.getState().getHearableBuildings().size() + "\n");

					for (int c = 0; c < b.getState().getHearableBuildings()
							.size(); c++) {
						bw.write(b.getState().getHearableBuildings().get(c)
								.getX()
								+ "\n");
						bw.write(b.getState().getHearableBuildings().get(c)
								.getY()
								+ "\n");
					}
				}
				bw.close();
				// System.out.println("wrote vb file \"" + fname + "\"");
			}
			// long t2 = System.currentTimeMillis();
			// System.out.println("Write vb file used : " + (t2 - t1));
		} catch (Exception e) {
			// System.out
			// .println("error while writting vb file \"" + fname + "\"");
		}

	}

	/**
	 * @param x
	 * @param y
	 * @return Node
	 */
	private Node getNode(int x, int y) {
		// TODO Knight Guan
		for (Iterator<Node> i = nodes.iterator(); i.hasNext();) {
			Node n = i.next();
			if (n.isNode(x, y)) {
				return n;
			}
		}
		// System.out.println("parser error");
		throw new NullPointerException();
	}

	private boolean computePrimitiveCalcs() {
		Road road;
		Building bldg;
		for (Iterator it = roads.iterator(); it.hasNext();) {
			road = (Road) it.next();
			road.getState().getNeighborhood(world);
			road.getState().getConnectedBuildings(world);
			road.getState().getConnectedRoads(world);
			road.getState().getContainedPaths(this.paths);
		}
//		Node node;
//		for (Iterator it = nodes.iterator(); it.hasNext();) {
//			node = (Node) it.next();
//			node.getNeighborhood(world);
//			node.getConnectedBuildings(world);
//			node.getConnectedRoads(world);
//			node.getContainedPaths(world);
//		}

		for (Iterator it = buildings.iterator(); it.hasNext();) {
			bldg = (Building) it.next();
			setNeighbourBuildings(bldg);
			bldg.getState().getNeighborhood(world);
			bldg.getState().getContainedPaths(this.paths);
		}
//		for (Path path : paths) {
//			path.getCenter();
//			path.toRoute(true);
//			path.toRoute(false);
//		}

		return true;
	}

	private void setNeighbourBuildings(Building build) {
		for (Iterator it = buildings.iterator(); it.hasNext();) {
			Building b = (Building) it.next();
			if (world.getDistance(b, build) < SimulatorsConstants.EXTINGUISHABLE_DISTANCE)
				build.getState().neighbourBuildings.add(b);
		}
	}

	private void generatePartitions() {
		Node center = getCenter();
		List<Node> tmpNodes = new ArrayList<Node>(nodes);

		// calculate max partition nums.
		int maxPartNum = 1;

		if (me instanceof PoliceForce || me instanceof PoliceOffice) {
			maxPartNum = policeForces.size();
		} else if (me instanceof FireBrigade || me instanceof FireStation) {
			maxPartNum = fireBrigades.size();
		} else if (me instanceof AmbulanceCentre) {
			float bili = (float) (world.getBounds().getWidth() / world
					.getBounds().getHeight());
			if (bili < 1) {
				bili = 1 / bili;
			}
			if (bili > 2) {
				maxPartNum = (int) (bili + 1);
			} else {
				maxPartNum = (int) (bili * 3) * 3;
			}
		} else {
			maxPartNum = ambulanceTeams.size();
		}

		// create radial paritions
		double dir_base = 360.0 / maxPartNum;

		for (int i = 0; i < maxPartNum; i++) {
			RadialPartition rp = new RadialPartition(i, i * dir_base, (i + 1)
					* dir_base, world);
			radialParitions.add(rp);
		}

		// fill neighbour partitions
		for (int i = 0; i < radialParitions.size(); i++) {
			RadialPartition rp = radialParitions.get(i);
			rp.setNeighbourPartition(radialParitions
					.get((i + 1) >= radialParitions.size() ? 0 : (i + 1)));
			rp.setNeighbourPartition(radialParitions.get((i - 1) < 0 ? 0
					: (i - 1)));
		}

		// add nodes to radialParitions
		for (Iterator<Node> it = tmpNodes.iterator(); it.hasNext();) {
			Node node = it.next();
			int dir = AgentUtil.normalizedDirection(center, node);
			addToPartition(dir, node);
		}

		for (RadialPartition rp : radialParitions) {
			rp.fillPartitions();
		}

	}

	private void addToPartition(int dir, Node node) {
		for (RadialPartition rp : radialParitions) {
			if (dir >= rp.startDir && dir < rp.endDir) {
				rp.nodes.add(node);
			}
		}
	}

	// private void createPartitions() {
	// PartitionerInit();
	// partitions.addAll(partitioner.getMapPartitions());
	// Collections.sort(partitions, Util.idComparator2);
	//		
	//		
	// Color[] rgb = new Color[10];
	// rgb[0] = Color.BLUE;
	// rgb[1] = Color.CYAN;
	// rgb[2] = Color.GREEN;
	// rgb[3] = Color.MAGENTA;
	// rgb[4] = Color.ORANGE;
	// rgb[5] = Color.PINK;
	// rgb[6] = Color.RED;
	// rgb[7] = Color.WHITE;
	// rgb[8] = Color.YELLOW;
	// rgb[9] = new Color(224, 159, 120);
	//
	// int i = 0;
	// Color color = rgb[i];
	//
	// if(!SeuConstants.IS_COMPETETION_MODE)
	// for (Iterator it = partitions.iterator(); it.hasNext();)
	// {
	// AmorphPartition pt = (AmorphPartition)it.next();
	//
	// for(Building b : pt.getContainedBuildings())
	// logFile.logIDObject(b.id, color, "partition");
	//			
	// i++;
	//
	// color = rgb[i % 10];
	//
	// }
	//		
	// }

	// private void PartitionerInit()
	// {
	// if (me instanceof AmbulanceCentre)
	// {
	// float bili = mapWidth / (float) mapHeight;
	// if (bili < 1)
	// {
	// bili = 1 / bili;
	// }
	// if (bili > 2)
	// {
	// partitioner = new MapPartitioner(
	// SeuConstants.partitionType.AMORPH, this);
	// partitioner.partition(new int[]
	// { (int) (bili + 1), 2 });
	// } else
	// {
	// dynamicPartitionWNAISP(false, (int) (bili * 3) * 3);
	// }
	// } else
	// dynamicPartition();
	// }

	// private void dynamicPartition()
	// {
	// if (me instanceof PoliceForce || me instanceof PoliceOffice)
	// {
	// int maxPartNum = policeForces.size() / 2;
	// int pfCount = policeForces.size();
	//
	// maxPartNum = (pfCount);
	// if (maxPartNum == 0)
	// maxPartNum = 1;
	// dynamicPartitionWNAISP(true, maxPartNum);
	// } else if (me instanceof FireBrigade || me instanceof FireStation)
	// {
	// int maxPartNum = fireBrigades.size() / 2;
	// int fbCount = fireBrigades.size();
	//
	// maxPartNum = fbCount;
	// if (maxPartNum == 0)
	// maxPartNum = 1;
	//
	// dynamicPartitionWNAISP(false, maxPartNum);
	// } else
	// {
	// int maxPartNum = ambulanceTeams.size() / 2;
	// int atCount = ambulanceTeams.size();
	//
	// maxPartNum = atCount;
	// if (maxPartNum == 0)
	// maxPartNum = 1;
	// dynamicPartitionWNAISP(false, maxPartNum);
	// }
	// }

	// private void dynamicPartitionWNAISP(boolean isPolice, int count)
	// {
	// if (count == 0)
	// count = 1;
	// int i = count;
	// double sqrt = Math.sqrt(i);
	// int floor = (int) sqrt, ceiling = (count / floor);
	// if (mapHeight > mapWidth)
	// {
	// int temp = floor;
	// floor = ceiling;
	// ceiling = temp;
	// }
	// if ((floor) == sqrt)
	// {
	// System.out.println(" ****** ColNum is : " + floor
	// + " , RowNum is : " + floor);
	// partitioner = new MapPartitioner(SeuConstants.partitionType.AMORPH,
	// this);
	// partitioner.partition(new int[]
	// { floor, floor });
	//
	// } else if (floor * ceiling <= count)
	// {
	// System.out.println("****** ColNum is : " + ceiling
	// + " , RowNum is : " + floor);
	//
	// partitioner = new MapPartitioner(SeuConstants.partitionType.AMORPH,
	// this);
	// partitioner.partition(new int[]
	// { ceiling, floor });
	//
	// } else
	// {
	// System.err.println("partition error!");
	// System.out.println("floor : " + floor);
	// partitioner = new MapPartitioner(SeuConstants.partitionType.QUAD,
	// this);
	// partitioner.partition(new int[]
	// { floor, floor });
	// }
	// }

	public void update(Collection<StandardEntity> objs, int time) {
		m_time = time;
//		System.out.println(objs.size());
		for (Iterator it1 = objs.iterator(); it1.hasNext();) {
			StandardEntity obj = (StandardEntity) it1.next();
			// if (obj == null) {
			// // Construct a new entity
			// obj = (StandardEntity)
			// Registry.getCurrentRegistry().createEntity(obj.getURN(),
			// obj.getID());
			// world.addEntity(obj);
			// add((StandardEntity) obj);
			// }
			add(obj.getID());
			obj.getState().setTime(time);

			for (Iterator it2 = obj.getProperties().iterator(); it2.hasNext();) {
				Property pro = (Property) it2.next();
				obj.getState().setProperty(world, time, pro, pro.getValue());
			}

			if(obj instanceof Road){
				unKnownRoads.add((Road)obj);
			}
			
			if (obj instanceof Human) {
				updatedHumans.add((Human) obj);
				if(obj instanceof AmbulanceTeam){
					AmbulanceTeam at =(AmbulanceTeam) obj;
					at.getState().m_sayToTime = time;
				}
				else if(obj instanceof FireBrigade){
					FireBrigade fb =(FireBrigade) obj;
					fb.getState().m_sayToTime = time;
				}
				else if(obj instanceof PoliceForce){
					PoliceForce pf =(PoliceForce) obj;
					pf.getState().m_sayToTime = time;
				}

				else if (obj instanceof Civilian) {

					Civilian civ = (Civilian) obj;
					civ.getState().m_sayToTime = time;
					civ.getState().setFound();

					civ.getState().setParticle(world, time,
							civ.getState().getHp(), civ.getState().getDamage(),
							civ.getState().getBuriedness(), 1);

					if (civ.getPosition(world) instanceof Building
							&& !(civ.getPosition(world) instanceof Refuge)) {
						foundVictimCivilianIC.add(civ);
						heardCivilians.remove(civ);
					}
				}
			}

			if (obj instanceof Building) {
				Building bldg = (Building) obj;
				updatedBuildings.add((Building) (obj));

				if (bldg.getState().m_justIgnited) {
					int dist = world.getDistance(me, bldg);
					int value = time
							- dist
							/ SimulatorsConstants.FIRE_COGNITION_SPREADING_SPEED
							- 1;

					// if(time > 2)
					// logFile.writeString("##############       "+getTime()+"\n"
					// +bldg+ "  ignite: "+value
					// +"  dist= "+dist, 78, "fireTime");
					bldg.getState().setIgniteTime(Math.max(0, value));
				}

				bldg.getState().setSeenTime(time);

				if (bldg.isFierynessDefined()) {
					if (bldg.getState().isBurned())
						notSeenBuildings.remove(bldg);
				}

			} else if (obj instanceof Road) {
				Road rd = (Road) obj;

				updatedRoads.add(rd);

				// used for Dijkstra
				if (time > 1) {
					if (PASSABLE_C.eval(rd)) {
						rd.getState()
								.setCondition(RoadState.CONDITION_PASSABLE);
						blockRoads.remove(rd);
					} else {
						rd.getState().setCondition(RoadState.CONDITION_BLOCKED);
						blockRoads.add(rd);
					}

					rd.getState().getContainedPaths(world).get(0)
							.calculateLength(world);
				}
			}

//			if (obj instanceof Human) {
//				 if (obj == me) {
//					 if (time > 2 && ((Human)obj).isPositionHistoryDefined()) {
//						 if(obj instanceof AmbulanceTeam){
//								AmbulanceTeam at =(AmbulanceTeam) obj;
//								at.getState().updatePositionHistory(world);
//						    }
//							else if(obj instanceof FireBrigade){
//								FireBrigade fb =(FireBrigade) obj;
//								fb.getState().updatePositionHistory(world);
//							}
//							else if(obj instanceof PoliceForce){
//								PoliceForce pf =(PoliceForce) obj;
//								pf.getState().updatePositionHistory(world);
//							}
//							else if (obj instanceof Civilian) {
//								Civilian civ = (Civilian) obj;
//								civ.getState().updatePositionHistory(world);
//							}
//					 }
//						  
//				 }
//			 }
		}

		updateValsAndCols();
		// System.out.println("........... areas = "+areas.size()+" "
		// +" Buildings = "+buildings.size()+" "
		// +" Roads = "+roads.size()+" ");
	}

	public void updateChangeSet(ChangeSet changeSet, int time) {
		if (time < 2)
			return;
		m_time = time;

		foundVictimCivilianIC.clear();
		updatedHumans.clear();
		updatedBuildings.clear();
		updatedCivilians.clear();
		updatedRoads.clear();

		for (EntityID e : changeSet.getChangedEntities()) {

			StandardEntity obj = world.getEntity(e);
			if (obj == null) {
				// Construct a new entity
				obj = (StandardEntity) Registry.getCurrentRegistry()
						.createEntity(changeSet.getEntityURN(e), e);
				world.addEntity(obj);
				add(e);
			}
			
			obj.getState().setTime(time);

			for (Property p : changeSet.getChangedProperties(e)) {
				Property existingProperty = obj.getProperty(p.getURN());
				existingProperty.takeValue(p);
				obj.getState().setProperty(world, time, p, p.getValue());
			}
			
			if(obj instanceof Road){
				Road ro = (Road) obj;
				unKnownRoads.remove(ro);
				if(ro.getBlockades().isEmpty() & !passableRoads.contains(ro)){
					passableRoads.add(ro);
					if(unPassableRoads.contains(ro))unPassableRoads.remove(ro);
//					System.out.println("Time = "+getTime()+"A "+obj+" "+ro.isBlockadesDefined()+" : "+ro.getBlockades()+" ");
				}
				else if(!unPassableRoads.contains(ro)){
					unPassableRoads.add(ro);
					if(passableRoads.contains(ro))passableRoads.remove(ro);
//					System.out.println("Time = "+getTime()+" B "+obj+" "+ro.isBlockadesDefined()+" : "+ro.getBlockades()+" ");
				}
				
			}

			// civilians.clear();
			// humans.clear();
			// for (Iterator it =
			// world.getEntitiesOfType(StandardEntityURN.CIVILIAN).iterator();
			// it.hasNext();) {
			// Civilian civ = (Civilian) it.next();
			// civilians.add(civ);
			// }
			// humans.addAll(civilians);
			// humans.addAll(platoonAgents);
			// Collections.sort(civilians, seu.agent.Util.idComparator);
			// Collections.sort(humans, seu.agent.Util.idComparator);

			if (obj instanceof Human) {
				updatedHumans.add((Human) obj);

				if(obj instanceof AmbulanceTeam){
					AmbulanceTeam at =(AmbulanceTeam) obj;
					at.getState().m_sayToTime = time;
				}
				else if(obj instanceof FireBrigade){
					FireBrigade fb =(FireBrigade) obj;
					fb.getState().m_sayToTime = time;
				}
				else if(obj instanceof PoliceForce){
					PoliceForce pf =(PoliceForce) obj;
					pf.getState().m_sayToTime = time;
				}

				else if (obj instanceof Civilian) {

					Civilian civ = (Civilian) obj;
					civ.getState().m_sayToTime = time;
					civ.getState().setFound();

					civ.getState().setParticle(world, time,
							civ.getState().getHp(), civ.getState().getDamage(),
							civ.getState().getBuriedness(), 1);

					if (civ.getPosition(world) instanceof Building
							&& !(civ.getPosition(world) instanceof Refuge)) {
						foundVictimCivilianIC.add(civ);
						heardCivilians.remove(civ);
					}
				}
			}

			if (obj instanceof Building) {
				Building bldg = (Building) obj;
				updatedBuildings.add((Building) (obj));

				if (bldg.getState().m_justIgnited) {
					int dist = world.getDistance(me, bldg);
					int value = time
							- dist
							/ SimulatorsConstants.FIRE_COGNITION_SPREADING_SPEED
							- 1;

					bldg.getState().setIgniteTime(Math.max(0, value));
				}

				bldg.getState().setSeenTime(time);

				if (bldg.isFierynessDefined()) {
					if (bldg.getState().isBurned())
						notSeenBuildings.remove(bldg);
				}

			} else if (obj instanceof Road) {
				Road rd = (Road) obj;

				updatedRoads.add(rd);

				// used for Dijkstra
				if (time > 1) {
					if (PASSABLE_C.eval(rd)) {
						rd.getState()
								.setCondition(RoadState.CONDITION_PASSABLE);
						blockRoads.remove(rd);
					} else {
						rd.getState().setCondition(RoadState.CONDITION_BLOCKED);
						blockRoads.add(rd);
					}
					if (rd.getState().getContainedPaths(paths).size() > 0)
						rd.getState().getContainedPaths(paths).get(0)
								.calculateLength(world);
				}
			}
			// TODO
//			 if (obj instanceof Human) {
//				 if (obj == me) {
//					 if (time > 2 && ((Human)obj).isPositionHistoryDefined()) {
//						 if(obj instanceof AmbulanceTeam){
//								AmbulanceTeam at =(AmbulanceTeam) obj;
//								at.getState().updatePositionHistory(world);
//						    }
//							else if(obj instanceof FireBrigade){
//								FireBrigade fb =(FireBrigade) obj;
//								fb.getState().updatePositionHistory(world);
//							}
//							else if(obj instanceof PoliceForce){
//								PoliceForce pf =(PoliceForce) obj;
//								pf.getState().updatePositionHistory(world);
//							}
//							else if (obj instanceof Civilian) {
//								Civilian civ = (Civilian) obj;
//								civ.getState().updatePositionHistory(world);
//							}
//					 }
//						  
//				 }
//			 }
		}

		updateValsAndCols();

	}

	private void updateValsAndCols() {
		switch (m_time) {
		case 2:
			sortByID();
			needHelpAgents.addAll(platoonAgents);
			for (Building b : specialBuildings) {
				if (b.isFierynessDefined()) {
					if (b.getFieryness() != 0)
						b.getState().setProperty(world, b.getState().getTime(),
								b.getFierynessProperty(), new int[] { 0 });
				}

			}

			// 
			if (ambulanceCenters.isEmpty() && policeOffices.isEmpty()
					&& fireStations.isEmpty()) {
				if(me instanceof AmbulanceTeam){
					AmbulanceTeam self = (AmbulanceTeam) me;
					if (self == ambulanceTeams.get(0))
						self.getState().isAtLeader = true;
				}
				else if(me instanceof FireBrigade){
					FireBrigade self = (FireBrigade) me;
					if (self ==  fireBrigades.get(0))
						self.getState().isFbLeader = true;
				}
				else if(me instanceof PoliceForce){
					PoliceForce self = (PoliceForce) me;
					if (self == policeForces.get(0))
						self.getState().isPfLeader = true;
				}
			}

			break;
		}

		buriedAgents.clear();
		for (Iterator it = platoonAgents.iterator(); it.hasNext();) {
			Human plt = (Human) it.next();
			if (plt.isBuriednessDefined()) {
				if (plt.getBuriedness() > 0) {
					buriedAgents.add(plt);
				}
			}
		}
		
		
		// buriedAgents.addAll(BURIED_C.extract(platoonAgents));

		burningBuildings.clear();
		for (Iterator it = buildings.iterator(); it.hasNext();) {
			Building b = (Building) it.next();
			if (b.isFierynessDefined()) {
				if (b.isOnFire()) {
					burningBuildings.add(b);
				}
			}

		}
		// burningBuildings.addAll(BURNING_C.extract(buildings));

	}

	// @Override
	// public void cloneTo(DisasterSpace o)
	// {
	// super.cloneTo(o);
	// if (selfId > 0)
	// o.me = o.get(selfId);
	//
	// o.sortByID();
	// }

	public void sortByID() {
		Collections.sort(platoonAgents, seu.agent.Util.idComparator);
		Collections.sort(centerAgents, seu.agent.Util.idComparator);
		Collections.sort(agents, seu.agent.Util.idComparator);
		Collections.sort(nodes, AgentUtil.idComparator3);
		Collections.sort(roads, seu.agent.Util.idComparator);
		Collections.sort(buildings, seu.agent.Util.idComparator);
		Collections.sort(areas, seu.agent.Util.idComparator);
		Collections.sort(ambulanceTeams, seu.agent.Util.idComparator);
		Collections.sort(fireBrigades, seu.agent.Util.idComparator);
		Collections.sort(policeForces, seu.agent.Util.idComparator);
		Collections.sort(civilians, seu.agent.Util.idComparator);
	}

	private void initialAllPaths() {
		HashSet<Road> roadSet = new HashSet<Road>(roads);
		for(Road rd:roads)
			rd.getState().initNode(world);
		
		for(Road rd:roads)
			rd.getState().initRoadNode(world);
		int pathIdCount = 0;
		while (!roadSet.isEmpty()) {
			Path path = Path.cartographerPath(pathIdCount, roadSet, world);
			if (path != null) {

				if (path.getContainedNodes().size() < 3) {
					path.setOneEndDead();
				} else if (path.getTailNode().getState().getConnectedRoads(
						world).size() < 2
						|| path.getHeadNode().getState().getConnectedRoads(
								world).size() < 2) {
					path.setOneEndDead();
				}
				paths.add(path);
//				System.out.println("path "+pathIdCount);
				pathIdCount++;
			}
		}

		// sort all path to a sequence
		for (Path p : paths) {
			if (p.getHeadNode().getID().getValue() > p.getTailNode().getID()
					.getValue()) {
				p.reverse();
			}
		}

		final Comparator pathHeadComparator = new Comparator() {
			public int compare(Object o1, Object o2) {
				Path p1 = (Path) o1;
				Path p2 = (Path) o2;

				int p1HeadId = p1.getHeadNode().getID().getValue();
				int p2HeadId = p2.getHeadNode().getID().getValue();
				if (p1HeadId > p2HeadId)
					return 1;
				else if (p1HeadId == p2HeadId)
					return 0;
				else
					return -1;
			}
		};

		Collections.sort(paths, pathHeadComparator);

		for (int i = 0; i < paths.size(); ++i) {
			Path p = paths.get(i);
			p.setId(i);
		}

	}

	public static void sortByID(List inp) {
		Collections.sort(inp, ID_COMP);
	}

	public void computeTraffic() {
		// TODO
		int range = 30 * 1000;
		inTraffic.clear();
		inTraffic.addAll(me.getState().distancePrp.lte(range).extract(
				this.roads));
		for (Iterator it = inTraffic.iterator(); it.hasNext();) {
			Road r = (Road) it.next();
			if (NOT_PASSABLE_C.eval(r)) {
				it.remove();
				continue;
			}
			int agentCount = 0;
			for (Human at : platoonAgents)
				if (at.getPosition(world) instanceof Road)
					if ((Road) at.getPosition(world) == r)
						agentCount++;
			if (agentCount <= 5)
				it.remove();
		}
	}

	public boolean isInTraffic(Area mo) {
		return inTraffic.contains(mo);
	}

	public StandardEntity getNearestObjectInRange(StandardEntity entity,
			Collection range) {
		StandardEntity result = null;
		int distance = Integer.MAX_VALUE;
		for (Iterator it = range.iterator(); it.hasNext();) {
			StandardEntity ro = (StandardEntity) it.next();
			try {
				if (distance > world.getDistance(entity, ro)) {
					distance = world.getDistance(entity, ro);
					result = ro;
				}
			} catch (Exception e) {

			}
		}
		return result;
	}

	public int getDistToNearestObjectInRange(StandardEntity entity,
			Collection range) {
		int distance = Integer.MAX_VALUE;
		if (range.isEmpty() || range.size() == 0)
			return distance;
		for (Iterator it = range.iterator(); it.hasNext();) {
			StandardEntity ro = (StandardEntity) it.next();
			if (ro == me)
				continue;

			int tmpDist = Integer.MAX_VALUE;
			try {
				tmpDist = world.getDistance(entity, ro);
			} catch (Exception e) {
				// TODO: handle exception
			}
			if (distance > tmpDist) {
				distance = tmpDist;
			}
		}
		return distance;
	}

	public void setBuildingConnections() {
		for (Iterator it = buildings.iterator(); it.hasNext();) {
			setConnectedNeighbours((Building) it.next());
		}
	}

	public final Map m_objSetKdTreeMap = new HashMap();

	public final Map m_objSetObjNeighboursMapMap = new HashMap();

	public void initializeNearBuildings() {
		for (Iterator<Building> it = buildings.iterator(); it.hasNext();) {
			Building bld1 = it.next();
			for (Iterator<Building> it2 = buildings.iterator(); it2.hasNext();) {
				Building tmp = it2.next();
				if (world.getDistance(tmp, bld1) < FIRE_PASSING_DISTANCE)
					bld1.getState().nearBuildings.add(tmp);
			}
		}
	}

	// public void initializeExtinguishNeighbours()
	// {
	// final Set buildingSet = new HashSet<Building>(buildings);
	//
	// for (Iterator<Building> it = buildings.iterator(); it.hasNext();)
	// {
	// Building bld1 = it.next();
	// bld1.getState().nearBuildings = getAround(bld1, buildingSet, 30000);
	// }
	// }
	//
	// public ArrayList getAround(StandardEntity obj, Set motionlessObjectSet,
	// int radius)
	// {
	// KdTree kdTree = (KdTree) m_objSetKdTreeMap.get(motionlessObjectSet);
	// Map objNeighboursMap = (Map) m_objSetObjNeighboursMapMap
	// .get(motionlessObjectSet);
	// if (kdTree == null)
	// {
	// System.out.println("" + motionlessObjectSet.size());
	// kdTree = new KdTree(motionlessObjectSet);
	// m_objSetKdTreeMap.put(motionlessObjectSet, kdTree);
	// objNeighboursMap = new HashMap();
	// m_objSetObjNeighboursMapMap.put(motionlessObjectSet,
	// objNeighboursMap);
	// }
	// ArrayList result = (ArrayList) objNeighboursMap.get(obj);
	// if (result == null)
	// {
	// result = kdTree.search(obj, radius);
	// objNeighboursMap.put(obj, result);
	// }
	// return result;
	// }
	//
	// boolean needWrite = true;

	private void setConnectedNeighbours(Building building) {
		ArrayList<Wall> edges = new ArrayList<Wall>();
		for (Iterator<Building> it = building.getState().neighbourBuildings
				.iterator(); it.hasNext();) {
			Building b = it.next();
			if (building != b) {
				b.getState().counter = 0;
				int[] apexes = b.getApexList();
				for (int j = 0; j < apexes.length - 2; j += 2) {
					edges.add(new Wall(apexes[j], apexes[j + 1], apexes[j + 2],
							apexes[j + 3], b));
				}
			}
		}

		ArrayList<Wall> myEdge = new ArrayList<Wall>();
		int apexes[] = building.getApexList();
		for (int i = 0; i < apexes.length - 2; i += 2) {
			myEdge.add(new Wall(apexes[i], apexes[i + 1], apexes[i + 2],
					apexes[i + 3], building));
		}

		HashSet connected = new HashSet(64);

		for (int i = 0; i < 180; i += RAY_DOMAIN) {
			double ang = i * Math.PI / 180;
			double tg = Math.tan(ang + Math.PI / 2);

			ArrayList<Wall> rightN = new ArrayList<Wall>();
			ArrayList<Wall> leftN = new ArrayList<Wall>();
			for (int j = 0; j < edges.size(); ++j) {
				Wall edg = edges.get(j);
				if (AgentUtil.Position(building.getX(), building.getY(), ang,
						tg, edg)) {
					if (edg.right) {
						rightN.add(edg);
					} else {
						leftN.add(edg);
					}
				}
			}

			if (rightN.size() > 0) {
				Wall minEdg = rightN.get(0);
				for (int k = 1; k < rightN.size(); ++k) {
					Wall edg = rightN.get(k);
					if (minEdg.distance > edg.distance) {
						minEdg = edg;
					}
				}
				if (minEdg.distance > 0) {
					connected.add(minEdg.b);
					minEdg.b.getState().counter++;

					Float min = building.getState().distanceValue.get(minEdg.b);
					if (min == null || min > minEdg.distance) {
						building.getState().distanceValue.put(minEdg.b,
								new Float(minEdg.distance));
					}

				}
			}
			if (leftN.size() > 0) {
				Wall minEdg = leftN.get(0);
				for (int k = 1; k < leftN.size(); ++k) {
					Wall edg = leftN.get(k);
					if (minEdg.distance > edg.distance) {
						minEdg = edg;
					}
				}
				if (minEdg.distance > 0) {
					connected.add(minEdg.b);
					minEdg.b.getState().counter++;

					Float min = building.getState().distanceValue.get(minEdg.b);
					if (min == null || min > minEdg.distance) {
						building.getState().distanceValue.put(minEdg.b,
								new Float(minEdg.distance));
					}

				}
			}
		}
		building.getState().connectedBuildings.clear();
		building.getState().connectedBuildings.addAll(connected);
		building.getState().connectedValues = new float[building.getState().connectedBuildings
				.size()];
		for (int i = building.getState().connectedBuildings.size() - 1; i >= 0; --i) {
			building.getState().connectedValues[i] = (building.getState().connectedBuildings
					.get(i)).getState().counter
					* RAY_DOMAIN / 450f;
			building.getState().connectedValueMap.put(
					building.getState().connectedBuildings.get(i), building
							.getState().connectedValues[i]);
		}
	}

	/*
	 * private void setConnectedNeighbours(Building building) {
	 * ArrayList<Circle> edges = new ArrayList<Circle>(); for
	 * (Iterator<Building> it = building.nearBuildings.iterator(); it
	 * .hasNext();) { Building b = it.next(); if (building != b) { b.counter =
	 * 0;
	 * 
	 * int radius = (int)(30*Math.sqrt(b.getBuildingAreaGround()));
	 * 
	 * edges.add(new Circle(b.getX(), b.getY(), radius, b)); } }
	 * 
	 * ArrayList<Circle> myEdge = new ArrayList<Circle>();
	 * 
	 * int radius = (int)(30*Math.sqrt(building.getBuildingAreaGround()));
	 * myEdge.add(new Circle(building.getX(), building.getY(), radius,
	 * building));
	 * 
	 * if(!SeuConstants.IS_COMPETETION_MODE && needWrite){
	 * 
	 * //
	 * System.out.println(building+"  area =  "+building.getBuildingAreaGround
	 * ()); //
	 * logFile.writeString(building.getX()+" "+building.getY()+" "+radius, 10,
	 * "radius");
	 * 
	 * // logFile.writeString(building.toString(), 10, "testBuilding");
	 * 
	 * 
	 * for (int j = 0; j < edges.size(); ++j) { Circle edg = edges.get(j);
	 * 
	 * if(!SeuConstants.IS_COMPETETION_MODE && needWrite){
	 * logFile.writeString(edg.x1+" "+edg.y1+" "+edg.radius, 10, "radius");
	 * logFile.writeString(edg.b.toString(), 10, "testBuilding"); } } needWrite
	 * = false; }
	 * 
	 * HashSet connected = new HashSet(64);
	 * 
	 * for (int i = 0; i < 360; i += RAY_DOMAIN) { ArrayList<Circle> jointed =
	 * new ArrayList<Circle>(); for (int j = 0; j < edges.size(); ++j) { Circle
	 * edg = edges.get(j); if(edg.radius+radius >= edg.b.getDistance(building))
	 * { jointed.add(edg); } else if (Util.IntectedWithCircle(building.getX(),
	 * building.getY(), i, edg)) { jointed.add(edg); } }
	 * 
	 * if (jointed.size() > 0) { Circle minEdg = jointed.get(0); for (int k = 1;
	 * k < jointed.size(); ++k) { Circle edg = jointed.get(k); if
	 * (minEdg.distance > edg.distance) { minEdg = edg; } } if (minEdg.distance
	 * > 0) { connected.add(minEdg.b); minEdg.b.counter++;
	 * 
	 * Float min = building.distanceValue.get(minEdg.b); if (min == null || min
	 * > minEdg.distance) { building.distanceValue.put(minEdg.b, new Float(
	 * minEdg.distance)); }
	 * 
	 * } } } building.connectedBuildings.clear();
	 * building.connectedBuildings.addAll(connected); building.connectedValues =
	 * new float[building.connectedBuildings.size()]; for (int i =
	 * building.connectedBuildings.size() - 1; i >= 0; --i) {
	 * building.connectedValues[i] =
	 * (building.connectedBuildings.get(i)).counter RAY_DOMAIN / 450f;
	 * 
	 * if(!SeuConstants.IS_COMPETETION_MODE)
	 * logFile.writeString(building.connectedValues
	 * [i]+"  counter = "+building.connectedBuildings.get(i).counter, 10,
	 * "building");
	 * building.connectedValueMap.put(building.connectedBuildings.get(i),
	 * building.connectedValues[i]); } }
	 */
	public void initialBuildingsForPrediction() {
		for (Iterator<Building> it = buildings.iterator(); it.hasNext();) {
			Building bld = it.next();
			bld.getState().initializeFireFigures();
		}
	}

	public Collection<Building> getDangerousBuildings() {
		HashSet<Building> bldList = new HashSet<Building>();

		bldList.addAll(burningBuildings);

		// if (seu.agent.Constants.BUILDING_PREDICTION_DEBUG)
		// {
		// if (logFile != null)
		// {
		// logFile.writeString("Time " + getTime() + " : "
		// + burningBuildings.size() + " burning buildings", 1,
		// seu.agent.Constants.PREDICTION_LOG);
		// }
		// }

		for (Building bld : burningBuildings) {
			bldList.addAll(bld.getState().neighbourBuildings);
		}

		// if (seu.agent.Constants.BUILDING_PREDICTION_DEBUG)
		// {
		// String str = "Time " + getTime() + " : " + bldList.size()
		// + " buildings are in danger!";
		// logFile.writeString(str, 1, seu.agent.Constants.PREDICTION_LOG);
		// }

		return bldList;
	}

}
