package seu.utilities;

import java.awt.Color;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import seu.SeuConstants;
import seu.agent.Constants;
import seu.states.CivilianState;
import seu.states.WorldState;

public class AmbulanceEvaluation {

	protected final StandardWorldModel world;
	protected final WorldState wst;
	protected PathMaker search;
	
	private static final int TIME_COST_ON_ROADS = 4;
	private static final int TIME_COST_TO_REFUGE = 5;
	
	public List<AmbulanceTeam> m_freeATs;
	
	public HashMap<AmbulanceTeam,Human> m_assign;
	public HashMap<AmbulanceTeam, Human> m_secondAssign;
	public HashMap<AmbulanceTeam,Human> m_greedyAssign;
	
	private HashMap<Human, Integer> m_assignBoundary;
	private HashMap<Human,Integer> m_rescueBoundary;
	
	public HashMap<AmbulanceTeam,Human> m_lastAssigned;

	public List<Civilian> m_victims;
	private List<Civilian> m_victims_backup;
	
	private HashMap<Integer, Integer> m_underRescued;
	
	public AmbulanceEvaluation(StandardWorldModel world){
		this.world = world;
		this.wst = world.getState();
		m_freeATs = new ArrayList<AmbulanceTeam>();
		m_assign = new HashMap<AmbulanceTeam, Human>();
		m_secondAssign = new HashMap<AmbulanceTeam, Human>();
		m_greedyAssign = new HashMap<AmbulanceTeam, Human>();
		m_lastAssigned = new HashMap<AmbulanceTeam, Human>();
		m_victims = new ArrayList<Civilian>();
		m_victims_backup = new ArrayList<Civilian>();
		m_underRescued = new HashMap<Integer, Integer>();
		m_rescueBoundary = new HashMap<Human, Integer>();
		m_assignBoundary = new HashMap<Human, Integer>();
	}
	
	public void evaluateForRescue(){
		
		getRescueVictims();
		
		updateVictimDeadTime();
		
		assignTargets();
		
		assignSecondTarget();
	}
	
	public void getRescueVictims(){
		
		m_freeATs.clear();
		for (Iterator it = wst.ambulanceTeams.iterator(); it.hasNext();) {
			AmbulanceTeam at = (AmbulanceTeam) it.next();
			if(at.isBuriednessDefined()){
				if(at.getBuriedness() > 0)continue;
				else m_freeATs.add(at);
			}
		}
		
		if (wst.getTime() < Constants.TIME_STARTING_ACTION)
			return;
		
		m_victims = new ArrayList<Civilian>();
		
		m_underRescued = new HashMap<Integer, Integer>();
		
		ArrayList<Civilian> longLifeTimeCiv = new ArrayList<Civilian>();
		
		for (Iterator it = wst.civilians.iterator(); it.hasNext();) {
			Civilian target = (Civilian) it.next();
			CivilianState tst = (CivilianState) target.getState();

			if (!(tst.needLoad(world) || tst.needRescue(world))) {
				continue;
			}
			
			if (wst.savedHumanoids.contains(target))
				continue;

			if(target.getPosition(world) instanceof Building){
				Building b = (Building) target.getPosition(world);
				if((b.getState().isBurning() && target.getBuriedness() > 30) 
						|| b.getState().isBurned())
					continue;
			}
			
			updateRescueATs(target);

			tst.updateCivilianDeadTime(world);

			int life = tst.getDeadTime();
			
			if (life > wst.getTime() + 2 && life < 400) {
				m_victims.add(target);
				if (tst.m_rescueATs > 0)
					m_underRescued.put(target.getID().getValue(), tst.m_rescueATs);
			} else if (life >= 400) {
				longLifeTimeCiv.add(target);
			}

		}


		if(wst.refuges.size()==0)
		{
			longLifeTimeCiv = new ArrayList<Civilian>();
			m_victims = new ArrayList<Civilian>();
			for(Iterator it = wst.civilians.iterator();it.hasNext();)
			{
				Civilian civ = (Civilian)it.next();
				if(!(civ.getPosition(world) instanceof Building))
					continue;
				Building b=(Building) civ.getPosition(world);
				if(!wst.savedHumanoids.contains(civ)
						&& civ.getHP()>0 &&!b.getState().isBurning())
				{
					if(civ.getState().getPredictedDeadTime()[55] < 300)
						continue;
					m_victims.add(civ);
				}
			}
			
		}
		//
		
		
		if(m_victims.size() < 3 )
		{
			m_victims.addAll(longLifeTimeCiv);
		}
		
		
//		System.out.println(" ****************************** ");
//		System.out.print(" totalCiv = "+m_victims.size());
		
		
//		System.out.println("before sim :  "+m_victims.size());
		
//		long t = System.currentTimeMillis();
		
		Collections.sort(m_victims, new civilianLifeComparator());
		
		while(!m_victims.isEmpty()){
			
			HashSet<Integer> saved = new HashSet<Integer>();
			int ATs = wst.ambulanceTeams.size();
			
			m_rescueBoundary = new HashMap<Human, Integer>();
			
			BinaryMinHeap simTasks = new BinaryMinHeap(40);
			
			int index = 0;
			int time = wst.getTime();
			
			while(ATs > 0 && index < m_victims.size()){
				while(saved.contains(m_victims.get(index).getID().getValue()))
					index ++;
				CivilianSimulator cs;
				Civilian civ = m_victims.get(index);
				if(simTasks.containsKey(civ.getID().getValue())){
					cs = (CivilianSimulator)simTasks.get(civ.getID().getValue());
				} else {
					cs = new CivilianSimulator(civ, civ.getState().getDeadTime());
				}
				
				int numATNeeded = cs.getNumATNeeded(time+TIME_COST_ON_ROADS);
				if(numATNeeded <= ATs)
				{
					ATs -= numATNeeded;
					cs.addATSkillful(numATNeeded, time+TIME_COST_ON_ROADS);
					if(simTasks.containsKey(cs.m_civ.getID().getValue()))
					{
						simTasks.decreaseKey(cs, cs.m_savedTime);
					}
					else
					{
						simTasks.insert(cs);
					}
					index++;
					saved.add(cs.m_civ.getID().getValue());
				}
				else
				{
					cs.addATSkillful(ATs, time+TIME_COST_ON_ROADS);
					ATs = 0;
					if(simTasks.containsKey(cs.m_civ.getID().getValue()))
					{
						simTasks.decreaseKey(cs, cs.m_savedTime);
					}
					else
					{
						simTasks.insert(cs);
					}
				}
				
			}
			
			boolean needReturn = true;
			
			while(!simTasks.isEmpty()){

				CivilianSimulator csSaved = (CivilianSimulator)simTasks.deleteMin();
				
				if(csSaved.m_savedTime + TIME_COST_TO_REFUGE > csSaved.m_deadTime)
				{
//					if(!SeuConstants.IS_COMPETETION_MODE)
//					world.logFile.writeString("DIE!!  = "+csSaved.m_civ
//							+" saveTime= "+csSaved.m_savedTime
//							+" AT: "+csSaved.m_numAT
//							+" time: "+csSaved.m_time
//							+" buried: "+csSaved.m_workLeft, 53, "Life");
					
//					System.err.println(" DIE!! "+" = "+csSaved.m_civ+" saveTime= "+csSaved.m_savedTime
//							+" AT: "+csSaved.m_numAT+"  "+csSaved.m_time+"  buried: "+csSaved.m_workLeft);
					
					Civilian savedCiv = csSaved.m_civ;
					Civilian died = null;
					int maxBury = -100;
					
					
					for(int i=0;i<m_victims.size();i++)
					{
						Civilian c = m_victims.get(i);
						
						if(c.getBuriedness()-4*c.getState().m_rescueATs > maxBury)
						{
							maxBury = c.getBuriedness()-4*c.getState().m_rescueATs;
							died = c;
						}
						if(c == savedCiv)
							break;
					}
//					if(!SeuConstants.IS_COMPETETION_MODE && died != null)
//					world.logFile.writeString("Deleted  = "+died+"  power= "+died.m_rescueATs+"  bury = "+died.getBuriedness(), 53, "Life");
//					
//					if(!SeuConstants.IS_COMPETETION_MODE && died != null)
//					world.logFile.logOval(new Point(died.getX(),died.getY()), 10000, Color.RED, false, "diedCivilian");
//					
					m_victims.remove(died);
					needReturn = false;
					break;
				}
				else
				{
					ATs = csSaved.m_numAT;
					saved.add(csSaved.m_civ.getID().getValue());
					
					m_rescueBoundary.put(csSaved.m_civ, csSaved.m_deadTime - csSaved.m_savedTime - TIME_COST_TO_REFUGE);
					
					while(ATs > 0 && index < m_victims.size())
					{
						while(index < m_victims.size() && saved.contains(m_victims.get(index).getID().getValue()))
							index++;
						
						if(index == m_victims.size())
							break;
						
						CivilianSimulator cs;
						Civilian civ = m_victims.get(index);
						if(simTasks.containsKey(civ.getID().getValue()))
							cs = (CivilianSimulator)simTasks.get(civ.getID().getValue());
						else
						{
							cs = new CivilianSimulator(civ, civ.getState().getDeadTime());
						}
						int numATNeeded = cs.getNumATNeeded(csSaved.m_savedTime+TIME_COST_ON_ROADS);
						if(numATNeeded <= ATs)
						{
							ATs -= numATNeeded;
							cs.addAT(numATNeeded, csSaved.m_savedTime+TIME_COST_ON_ROADS);
							if(simTasks.containsKey(cs.m_civ.getID().getValue()))
							{
								simTasks.decreaseKey(cs, cs.m_savedTime);
							}
							else
							{
								simTasks.insert(cs);
							}
							index++;
							saved.add(cs.m_civ.getID().getValue());
						}
						else
						{
							cs.addAT(ATs, csSaved.m_savedTime+TIME_COST_ON_ROADS);
							ATs = 0;
							if(simTasks.containsKey(cs.m_civ.getID().getValue()))
							{
								simTasks.decreaseKey(cs, cs.m_savedTime);
							}
							else
							{
								simTasks.insert(cs);
							}
						}
					}
				}
			
			}
			
			if(needReturn)
			{
//				t = System.currentTimeMillis() - t;
//				if(t > 100)
//					System.err.println(" sim civ use: "+t);
				
//				System.out.println("   After = "+m_victims.size()+"    time = "+world.getTime());
				
				m_victims_backup = new ArrayList<Civilian>();
				m_victims_backup.addAll(m_victims);
				return;
			}
			
		}
				
	}
	
	
	
	private void updateVictimDeadTime(){
		
		while(!m_victims.isEmpty()){
			
			Collections.sort(m_victims, new civilianLifeComparator());
			
			HashSet<Integer> saved = new HashSet<Integer>();
			int ATs = wst.ambulanceTeams.size();
			
			m_rescueBoundary = new HashMap<Human, Integer>();
			
			BinaryMinHeap simTasks = new BinaryMinHeap(40);
			
			int index = 0;
			int time = wst.getTime();
			
			while(ATs > 0 && index < m_victims.size()){
				while(saved.contains(m_victims.get(index).getID().getValue()))
					index ++;
				CivilianSimulator cs;
				Civilian civ = m_victims.get(index);
				if(simTasks.containsKey(civ.getID().getValue())){
					cs = (CivilianSimulator)simTasks.get(civ.getID().getValue());
				} else {
					cs = new CivilianSimulator(civ, civ.getState().getDeadTime());
				}
				
				int numATNeeded = cs.getNumATNeeded(time+TIME_COST_ON_ROADS);
				if(numATNeeded <= ATs)
				{
					ATs -= numATNeeded;
					cs.addATSkillful(numATNeeded, time+TIME_COST_ON_ROADS);
					if(simTasks.containsKey(cs.m_civ.getID().getValue()))
					{
						simTasks.decreaseKey(cs, cs.m_savedTime);
					}
					else
					{
						simTasks.insert(cs);
					}
					index++;
					saved.add(cs.m_civ.getID().getValue());
				}
				else
				{
					cs.addATSkillful(ATs, time+TIME_COST_ON_ROADS);
					ATs = 0;
					if(simTasks.containsKey(cs.m_civ.getID().getValue()))
					{
						simTasks.decreaseKey(cs, cs.m_savedTime);
					}
					else
					{
						simTasks.insert(cs);
					}
				}
				
			}
			
			boolean needReturn = true;
			
			while(!simTasks.isEmpty()){

				CivilianSimulator csSaved = (CivilianSimulator)simTasks.deleteMin();
				
				if(csSaved.m_savedTime + TIME_COST_TO_REFUGE > csSaved.m_deadTime)
				{
//					if(!SeuConstants.IS_COMPETETION_MODE)
//					world.logFile.writeString("DIE!!  = "+csSaved.m_civ
//							+" saveTime= "+csSaved.m_savedTime
//							+" AT: "+csSaved.m_numAT
//							+" time: "+csSaved.m_time
//							+" buried: "+csSaved.m_workLeft, 53, "Life");
					
//					System.err.println(" DIE!! "+" = "+csSaved.m_civ+" saveTime= "+csSaved.m_savedTime
//							+" AT: "+csSaved.m_numAT+"  "+csSaved.m_time+"  buried: "+csSaved.m_workLeft);
					
					Civilian savedCiv = csSaved.m_civ;
					Civilian died = null;
					int minBury = 1000;
					
					
					for(int i=m_victims.size()-1;i>=0;i--)
					{
						Civilian c = m_victims.get(i);
						
						if(c.getState().m_4kHpDeadTime == c.getState().getDeadTime())
							continue;
						
//						if(c.getBuriedness()-4*c.m_rescueATs < minBury)
//						{
//							minBury = c.getBuriedness()-4*c.m_rescueATs;
//							died = c;
//						}
						
						if(c == savedCiv)
							break;
					}
//					if(!SeuConstants.IS_COMPETETION_MODE && died != null)
//					world.logFile.writeString("Deleted  = "+died+"  power= "+died.m_rescueATs+"  bury = "+died.getBuriedness(), 53, "Life");
//					
//					if(!SeuConstants.IS_COMPETETION_MODE && died != null)
//					world.logFile.logOval(new Point(died.getX(),died.getY()), 10000, Color.RED, false, "diedCivilian");
//					
//					
					if(died == null){
						break;
					}
					else {
						died.getState().m_4kHpDeadTime = died.getState().getDeadTime();
					}
					needReturn = false;
					break;
				}
				else
				{
					ATs = csSaved.m_numAT;
					saved.add(csSaved.m_civ.getID().getValue());
					
					
					m_rescueBoundary.put(csSaved.m_civ, csSaved.m_deadTime - csSaved.m_savedTime - TIME_COST_TO_REFUGE);
					
					while(ATs > 0 && index < m_victims.size())
					{
						while(index < m_victims.size() && saved.contains(m_victims.get(index).getID().getValue()))
							index++;
						
						if(index == m_victims.size())
							break;
						
						CivilianSimulator cs;
						Civilian civ = m_victims.get(index);
						if(simTasks.containsKey(civ.getID().getValue()))
							cs = (CivilianSimulator)simTasks.get(civ.getID().getValue());
						else
						{
							cs = new CivilianSimulator(civ, civ.getState().m_4kHpDeadTime);
						}
						int numATNeeded = cs.getNumATNeeded(csSaved.m_savedTime+TIME_COST_ON_ROADS);
						if(numATNeeded <= ATs)
						{
							ATs -= numATNeeded;
							cs.addAT(numATNeeded, csSaved.m_savedTime+TIME_COST_ON_ROADS);
							if(simTasks.containsKey(cs.m_civ.getID().getValue()))
							{
								simTasks.decreaseKey(cs, cs.m_savedTime);
							}
							else
							{
								simTasks.insert(cs);
							}
							index++;
							saved.add(cs.m_civ.getID().getValue());
						}
						else
						{
							cs.addAT(ATs, csSaved.m_savedTime+TIME_COST_ON_ROADS);
							ATs = 0;
							if(simTasks.containsKey(cs.m_civ.getID().getValue()))
							{
								simTasks.decreaseKey(cs, cs.m_savedTime);
							}
							else
							{
								simTasks.insert(cs);
							}
						}
					}
				}
			
			}
			
			if(needReturn)
			{
//				t = System.currentTimeMillis() - t;
//				if(t > 100)
//					System.err.println(" sim civ use: "+t);
				
				Collections.sort(m_victims, new civilianLifeComparator());
				
				for(int i=0;i<m_victims.size();i++)
					m_victims.get(i).getState().setRescuePriority(i);
				
//				System.out.println("   After = "+m_victims.size()+"    time = "+world.getTime());
				
				return;
			}
			
		}
	}
	
	private void updateRescueATs(Civilian c){
		c.getState().m_rescueATs = 0;
		for(AmbulanceTeam at : wst.ambulanceTeams){
			if(at.getPosition(world) == null)
				continue;
			if(at.getPosition(world) == c.getPosition(world))
				c.getState().m_rescueATs ++;
		}
	}
	
	private boolean checkLastAssign(AmbulanceTeam at){
		Human civ = m_lastAssigned.get(at);
		if(civ == null)
			return false;
		
		int distance = 0;
		
		try {
			List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
					civ.getPosition());
			for(EntityID id : path){
				StandardEntity next = world.getEntity(id);
				if(next instanceof Road){
					distance += ((Road) next).getState().getLength(world);
				}
			}
			
//			distance = (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(civ.getRealPosition()), 
//					HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/2000;
		} catch (Exception e) {
			distance = (int)(Math.random()*500);
			return false;
		}
	
		if(distance >= Integer.MAX_VALUE/1000){
			return false;
		}
		
		if(!m_victims.contains(civ))
			return false;
		
		if(wst.savedHumanoids.contains(civ))
			return false;
		
		return true;
	}
	
	private boolean checkSecondAssign(AmbulanceTeam at){
		if(m_secondAssign.containsKey(at))
			return false;
		Human civ = m_secondAssign.get(at);
		if(!m_victims.contains(civ))
			return false;
		
		return true;
	}
	
	private Civilian getNearestTarget(AmbulanceTeam at, AssignSimulator assignInfo){
		Civilian result = null;
		
		HashSet<Human> saved = assignInfo.getSavedCiv();
		
		int minDist = Integer.MAX_VALUE;
		
		int minBest = Integer.MAX_VALUE;
		
		int time_simulation = 305;
		
		int time_urgent = 140;
		int dist_half_map = Math.max(world.getState().mapWidth/2000, world.getState().mapHeight/2000);
		dist_half_map *= 1.5;
		double coef = ((double)dist_half_map)/((double)time_urgent);
		
		Civilian best = null;
		
		for(Civilian civ : m_victims){
			if(civ.getPosition(world) == null)
				continue;
			
			if(at.getPosition(world) == null)
				return civ;
			
			int distance = 0;
			
			try {
				List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
						civ.getPosition());
				for(EntityID id : path){
					StandardEntity next = world.getEntity(id);
					if(next instanceof Road){
						distance += ((Road) next).getState().getLength(world);
					}
				}
				
//				distance = (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(civ.getRealPosition()), 
//						HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/1200;
			} catch (Exception e) {
				distance = (int)(Math.random()*500);
				continue;
			}
			
//			if(!SeuConstants.IS_COMPETETION_MODE && world.ambulanceTeams.get(0) == at)
//				System.out.println(world.getTime()+"   "+at+"   --->  "+civ+"     dist = "+distance);
			
//			String str = world.getTime()+"   "+at+"   --->  "+civ+"  deadT = "+civ.getDeadTime()
//						+"     1st dist = "+distance;
			
			if(civ.getState().getDeadTime() > time_simulation)
			{
				distance += dist_half_map + civ.getBuriedness()*8;
			}
			else if(civ.getState().getDeadTime() > time_urgent + wst.getTime())
			{
				distance += dist_half_map;
			}
			else
			{
				double a = coef*(civ.getState().getDeadTime() - wst.getTime());
				distance = distance + (int)a;
			}
			
//			str += "  final dist = "+distance;
			
//			if(!SeuConstants.IS_COMPETETION_MODE)
//				world.logFile.writeString(str, 10, "tar_dist");
			
			if(distance < minDist){
				minDist = distance;
				result = civ;
			}
			
			if(!saved.contains(civ)){
				if(distance < minBest){
					minBest = distance;
					best = civ;
				}
			}
		}
		if(minBest < Integer.MAX_VALUE/1200){
//			if(!SeuConstants.IS_COMPETETION_MODE)
//				world.logFile.writeString(" best tar = "+best, 10, "tar_dist");
			
			return best;
		}
		
//		if(!SeuConstants.IS_COMPETETION_MODE)
//			world.logFile.writeString(" not best tar = "+result, 10, "tar_dist");
		
		return result;
	}
	
	public void assignTargets() {
		
//		System.out.println(" victims = "+m_victims.size());
		
		for(Civilian c : m_victims){
			c.getState().setDeadTime(c.getState().m_2ndDeadTime);
		}
		
		if (m_victims.isEmpty()){
			m_lastAssigned = new HashMap<AmbulanceTeam, Human>();
			m_assign = new HashMap<AmbulanceTeam, Human>();
			return;
		}
		
//		System.out.println("assign ... ");
		
		AssignSimulator assignInfo = new AssignSimulator(m_victims);
		
		for (Iterator it = wst.ambulanceTeams.iterator(); it.hasNext();) {
			AmbulanceTeam at = (AmbulanceTeam) it.next();
			
			if(checkLastAssign(at)){
				assignInfo.addAssign(at, m_lastAssigned.get(at));
			} else if(checkSecondAssign(at)){
				assignInfo.addAssign(at, m_secondAssign.get(at));
			} else {
				Civilian target = getNearestTarget(at, assignInfo);
				assignInfo.addAssign(at, target);
			}
		}
		
		m_greedyAssign.putAll(assignInfo.getAssignation());
		
		
		
		
//		if(!SeuConstants.IS_COMPETETION_MODE)
//		world.logFile.writeString("###########   "+world.getTime(), 69, "assignSim");
		
		
		HashSet saved = new HashSet();
		
		BinaryMinHeap simTasks = new BinaryMinHeap(40);
		
		int deadIndex = -1;
		
		while (true){
			
			m_assignBoundary = new HashMap<Human, Integer>();
			
			saved = new HashSet();
			
			simTasks = new BinaryMinHeap(40);
			
			int index = 0;
			int time = wst.getTime()+5;
			
			for(Human assigned : assignInfo.getCivBeAssigned().keySet()){
				Civilian civ = (Civilian) assigned;
//				CivilianSimulator cs = new CivilianSimulator(civ, civ.getDeadTime());
				
				CivilianSimulator cs = new CivilianSimulator(civ, civ.getState().getDeadTime());
				
				cs.addATSkillful(assignInfo.getCivBeAssigned().get(civ).value, time+1);
				
				if(cs.m_savedTime + TIME_COST_TO_REFUGE < cs.m_deadTime){
					saved.add(cs.m_civ);
					simTasks.insert(cs);
				} else {
					simTasks.insert(cs);
				}
			}
			
			boolean needReturn = true;
			
			
			
			while(!simTasks.isEmpty()){

				CivilianSimulator csSaved = (CivilianSimulator)simTasks.deleteMin();
				
//				if(!SeuConstants.IS_COMPETETION_MODE)
//				world.logFile.writeString("==  "+csSaved.m_civ.id
//						+" saveTime= "+csSaved.m_savedTime
//						+" deadTime= "+csSaved.m_deadTime
//						+" AT: "+csSaved.m_numAT
//						+" life: "+csSaved.m_civ.getDeadTime()
//						+" 4kLife: "+csSaved.m_civ.m_4kHpDeadTime
//						+" buried: "+csSaved.m_civ.getBuriedness(), 53, "Life");
//				
//				
				
				if(csSaved.m_savedTime + TIME_COST_TO_REFUGE > csSaved.m_deadTime)
				{
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//					world.logFile.writeString("Not Saved= "+csSaved.m_civ.id
//							+" saveTime= "+csSaved.m_savedTime
//							+" deadTime= "+csSaved.m_deadTime
//							+" AT: "+csSaved.m_numAT
//							+" life: "+csSaved.m_civ.getDeadTime()
//							+" 4kLife: "+csSaved.m_civ.m_4kHpDeadTime
//							+" buried: "+csSaved.m_workLeft, 53, "Life");
//					
//					System.err.println(" DIE after assign !! "+" = "+csSaved.m_civ+" saveTime= "+csSaved.m_savedTime
//							+" "+csSaved.m_numAT+" "+csSaved.m_time+" "+csSaved.m_workLeft);
					
					deadIndex = csSaved.m_civ.getState().getRescuePriority();
					
					needReturn = false;
					break;
				}
				else
				{
					int ATs = csSaved.m_numAT;
					saved.add(csSaved.m_civ);
					
					m_assignBoundary.put(csSaved.m_civ, csSaved.m_deadTime - csSaved.m_savedTime - TIME_COST_TO_REFUGE);
					
					while(ATs > 0 && index < m_victims.size())
					{
						if(saved.contains(m_victims.get(index)))
						{
							index++;
							continue;
						}
						
						CivilianSimulator cs;
						Civilian civ = m_victims.get(index);
						if(simTasks.containsKey(civ.getID().getValue()))
							cs = (CivilianSimulator)simTasks.get(civ.getID().getValue());
						else
						{
//							cs = new CivilianSimulator(civ, civ.m_4kHpDeadTime);
							cs = new CivilianSimulator(civ, civ.getState().getDeadTime());
						}
						int numATNeeded = cs.getNumATNeeded(csSaved.m_savedTime+TIME_COST_ON_ROADS);
						if(numATNeeded <= ATs)
						{
							ATs -= numATNeeded;
							cs.addAT(numATNeeded, csSaved.m_savedTime+TIME_COST_ON_ROADS);
							if(simTasks.containsKey(cs.m_civ.getID().getValue()))
							{
								simTasks.decreaseKey(cs, cs.m_savedTime);
							}
							else
							{
								simTasks.insert(cs);
							}
							index++;
							saved.add(cs.m_civ);
						} else
						{
							cs.addAT(ATs, csSaved.m_savedTime+TIME_COST_ON_ROADS);
							ATs = 0;
							if(simTasks.containsKey(cs.m_civ.getID().getValue()))
							{
								simTasks.decreaseKey(cs, cs.m_savedTime);
							}
							else
							{
								simTasks.insert(cs);
							}
						}
					}
				}
			
			}
			
			if(needReturn)
			{
				assignInfo.dealBoundary();
				assignInfo.adjust();
				assignInfo.dealWithSpecial();
				
				m_assign = assignInfo.getAssignation();
				
				m_lastAssigned = new HashMap<AmbulanceTeam, Human>();
				for(Iterator it = m_assign.keySet().iterator();it.hasNext();)
				{
					AmbulanceTeam at = (AmbulanceTeam)it.next();
					m_lastAssigned.put(at, m_assign.get(at));
				}
				
				return;
			} else {
				
				// adjust 
				
				ArrayList<AmbulanceTeam> candiATs = new ArrayList<AmbulanceTeam>();
				ArrayList<Civilian> target = new ArrayList<Civilian>();
				
				for(AmbulanceTeam at : wst.ambulanceTeams){
					Civilian civ = (Civilian) assignInfo.getAssignation().get(at);
					if(civ == null)
						continue;
					
					if(civ.getState().getRescuePriority() > deadIndex){
						candiATs.add(at);
					}
				}
				
				if(candiATs.size()==0)
				{
//					System.err.println("ASSIGN FAILED!! = " + deadIndex);
					
					assignInfo.dealBoundary();
					assignInfo.adjust();
					assignInfo.dealWithSpecial();
					
					m_assign = assignInfo.getAssignation();
					m_lastAssigned = new HashMap<AmbulanceTeam, Human>();
					
					for(Iterator it = m_assign.keySet().iterator();it.hasNext();)
					{
						AmbulanceTeam at = (AmbulanceTeam)it.next();
						m_lastAssigned.put(at, m_assign.get(at));
					}
					
					return;
				}
				
				for(int i = 0; i <= deadIndex; i ++){
					Civilian civ = m_victims.get(i);
					target.add(civ);
				}
				
				AmbulanceTeam bestAT = null;
				Civilian bestTarget = null;
				double bestScore = -10000;
				
				
				for(Iterator it = candiATs.iterator();it.hasNext();)
				{
					AmbulanceTeam at = (AmbulanceTeam)it.next();
					Civilian civ = (Civilian)assignInfo.getAssignation().get(at);
					
					int[] id_dist = getNearestCivilianInRange(at, target, assignInfo.getSavedCiv());
					
					int dist2 = 0;
					try {
						List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
								civ.getPosition());
						for(EntityID id : path){
							StandardEntity next = world.getEntity(id);
							if(next instanceof Road){
								dist2 += ((Road) next).getState().getLength(world);
							}
						}
						
//						dist2 = (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(civ.getRealPosition()), 
//								HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/2000;
					} catch (Exception e) {
						dist2 = (int)(Math.random()*500);
					}
				
					int distCost = id_dist[1] - dist2;


					double score = ((double)civ.getBuriedness())/assignInfo.getCivBeAssigned().get(civ).value;

					score = score/distCost;
					
					if(score < 0)
						score=1000;
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//					world.logFile.writeString(at+" -->  newly= "+id_dist[0]
//							+"  oldDist = "+dist2
//							+"  newDist = "+id_dist[1]
//							+"  old= "+civ
//							+"   bury "+civ.getBuriedness()
//							+"   life = "+civ.getDeadTime()
//							+"   4kLife: "+civ.m_4kHpDeadTime
//							+"   assigned= "+assignInfo.getCivBeAssigned().get(civ).value
//							+"   score= "+score, 69, "assignSim");
					
//					if(!SeuConstants.IS_COMPETETION_MODE)
//					world.logFile.writeString(at+" -->  newly= "+id_dist[0]
//							+"  oldDist = "+dist2
//							+"  newDist = "+id_dist[1]
//							+"  old= "+civ
//							+"   bury "+civ.getBuriedness()
//							+"   life = "+civ.getDeadTime()
//							+"   4kLife: "+civ.m_4kHpDeadTime
//							+"   assigned= "+assignInfo.getCivBeAssigned().get(civ).value
//							+"   score= "+score, 53, "Life");
					
					if(score > bestScore)
					{
						bestScore = score;
						bestAT = at;
						bestTarget = (Civilian)world.getEntity(new EntityID(id_dist[0]));
					}
				}
				
//				if(bestTarget != null && !SeuConstants.IS_COMPETETION_MODE){
//				world.logFile.writeString("  ---> finally   "+bestAT+"  --->  "+bestTarget
//						+"  life = "+bestTarget.getDeadTime()
//						+" 4kLife: "+bestTarget.m_4kHpDeadTime, 53, "Life");
////				
//				world.logFile.writeString("  ---> finally   "+bestAT+"  --->  "
//						+bestTarget+"  life = "+bestTarget.getDeadTime(), 69, "assignSim");
//				}
				
				// remove the old one and put the adjusted one 
				assignInfo.remove(bestAT);
				assignInfo.addAssign(bestAT,bestTarget);
			}
			
		}
		
	}
	
	
	private HashMap<Civilian, ArrayList> civilianNeedAt; 
	public ArrayList atsNeedSecondAssign;
	
	private void findATsNeedSecondAssign()
	{
		civilianNeedAt = new HashMap<Civilian, ArrayList>();
		
		atsNeedSecondAssign = new ArrayList();

		HashMap temp=new HashMap();
		temp.putAll(m_assign);
		
		for(Iterator it=m_assign.values().iterator();it.hasNext();)
		{
			Civilian c=(Civilian)it.next();
			if(civilianNeedAt.containsKey(c))
				continue;
			ArrayList atList=new ArrayList();
			for(Iterator it2=temp.keySet().iterator();it2.hasNext();)
			{
				AmbulanceTeam at =(AmbulanceTeam)it2.next();
				if((Civilian)temp.get(at) == c)
				{
					atList.add(at);
				}
			}
			civilianNeedAt.put(c, atList);
			
			if(c.getState().m_rescueATs*2 + 3 >= c.getBuriedness())
			{
				for(Iterator it3=atList.iterator();it3.hasNext();)
				{
					AmbulanceTeam at =( AmbulanceTeam)it3.next();
					atsNeedSecondAssign.add(at);
				}
			}
		}
		//
		for(Iterator it4=civilianNeedAt.keySet().iterator();it4.hasNext();)
		{
			Civilian c=(Civilian)it4.next();
			if(c.getState().m_rescueATs*2 + 4 >= c.getBuriedness())
				m_victims.remove(c);
		}
	
	}
	
	public void assignSecondTarget()
	{
		m_secondAssign = new HashMap<AmbulanceTeam, Human>();
		
		findATsNeedSecondAssign();
				
		if(m_victims.isEmpty())
		{
			return;
		}
		
//		System.out.println("2nd assign ...");
		
		AssignSimulator assignInfo = new AssignSimulator(m_victims);
		
		
		for(Iterator it= m_assign.keySet().iterator();it.hasNext();)
		{
			AmbulanceTeam at = (AmbulanceTeam)it.next();
			Human value = m_assign.get(at);
			if(!atsNeedSecondAssign.contains(at))
				assignInfo.addAssign(at, value);
		}
		
		for(Iterator it = atsNeedSecondAssign.iterator();it.hasNext();)
		{
			AmbulanceTeam at = (AmbulanceTeam)it.next();
			Civilian civ = getNearestTarget(at, assignInfo);
			assignInfo.addAssign(at, civ);
		}
		
		assignInfo.dealBoundary();
		
		assignInfo.adjust();
		
		assignInfo.dealWithSpecial();
		
		HashSet<Human> excepts = new HashSet<Human>();
		for(AmbulanceTeam at : assignInfo.m_assign.keySet()){
			if(atsNeedSecondAssign.contains(at)){
				m_secondAssign.put(at, assignInfo.m_assign.get(at));
			}
				
		}
				
	}
	
	
	/** int[0] is civilian ID, and int[1] is distance. 
	 * @param hum
	 * @param targets
	 * @param excepts
	 * @param dist
	 * @return
	 */
	private int[] getNearestCivilianInRange(Human hum, List<Civilian> targets, HashSet<Human> excepts){
		Civilian result = null;
		int minDist = Integer.MAX_VALUE;
		
		int minBest = Integer.MAX_VALUE;
		Civilian best = null;
		
		for(Civilian civ : targets){
			
			int distance = 0;
			try {
				List<EntityID> path = search.breadthFirstSearch(hum.getPosition(),
						civ.getPosition());
				for(EntityID id : path){
					StandardEntity next = world.getEntity(id);
					if(next instanceof Road){
						distance += ((Road) next).getState().getLength(world);
					}
				}
				
//				distance = (int)Router.getASGS(hum.getMotionlessPosition(), Collections.singleton(civ.getRealPosition()), 
//						HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/2000;
			} catch (Exception e) {
				distance = (int)(Math.random()*500);
			}
			
			if(distance < minDist){
				result = civ;
				minDist = distance;
			}
			
			if(!excepts.contains(civ)){
				if(distance < minBest){
					minBest = distance;
					best = civ;
				}
			}
		}
		
		if(minBest < Integer.MAX_VALUE/2000){
			int[] id_dist=new int[2];
			id_dist[0] = best.getID().getValue();
			id_dist[1] = minBest;
			return id_dist;
		} else {
			int[] id_dist=new int[2];
			
			try {
				id_dist[0] = result.getID().getValue();
			} catch (Exception e) {
				id_dist[0] = -1;
			}
			
			id_dist[1] = minDist;
			return id_dist;
		}
		
	}
	
	private class IntValue{
		public int value;
		public IntValue(int v){
			value = v;
		}
	}
	
	private class AssignSimulator {
		private HashMap<AmbulanceTeam,Human> m_assign;
		private HashMap<Human,IntValue> m_beAssigned;
		private HashSet<Human> m_savedCiv;
		
		public AssignSimulator(List<Civilian> civs)
		{
			m_assign = new HashMap<AmbulanceTeam,Human>();
			m_beAssigned = new HashMap<Human,IntValue>();
			m_savedCiv = new HashSet<Human>();
			
		}
		
		public HashMap<AmbulanceTeam,Human> getAssignation(){
			return m_assign;
		}
		
		public HashMap<Human,IntValue> getCivBeAssigned(){
			return m_beAssigned;
		}
		
		public HashSet<Human> getSavedCiv(){
			return m_savedCiv;
		}
		
		public void addAssign(AmbulanceTeam at, Human human)
		{
			if(human == null || human.getPosition(world) == null)
				return;
			
			m_assign.put(at, human);
			
			if(m_beAssigned.containsKey(human))
				m_beAssigned.get(human).value++;
			else
				m_beAssigned.put(human, new IntValue(1));
			
			if(m_savedCiv.contains(human))
				return;
			int a=0;
			int b=0;
			for(Iterator it = m_assign.keySet().iterator();it.hasNext();)
			{
				AmbulanceTeam at1 = (AmbulanceTeam)it.next();
				Human civ = m_assign.get(at1);
				if(civ != human || civ == null || at1 == null || human == null)
					continue;
				if(civ.getPosition(world) == human.getPosition(world))
					a++;
				else
					b++;
			}
			int bury = human.getBuriedness();
			int saveTime = 0;
			if(bury < a*7)
				saveTime = wst.getTime() + bury/a + 1;
			else
				saveTime =  wst.getTime() + 7 + (bury-a*7)/(b+a) + 1;
			
			if(saveTime < ((Civilian)human).getState().getDeadTime()-5 || saveTime > 305)
				m_savedCiv.add(human);
		}
		
		public void remove(AmbulanceTeam at){
			Civilian civ = (Civilian) m_assign.get(at);
			if(civ == null)
				return;
			
			m_assign.remove(at);
			m_beAssigned.get(civ).value --;
			if(m_beAssigned.get(civ).value <= 0)
				m_beAssigned.remove(civ);
			if(m_savedCiv.contains(civ))
				m_savedCiv.remove(civ);
			
			int a=0;
			int b=0;
			for(Iterator it = m_assign.keySet().iterator();it.hasNext();)
			{
				AmbulanceTeam at1 = (AmbulanceTeam)it.next();
				Human civ1 = m_assign.get(at1);
				if(civ != civ1)
					continue;
				if(civ.getPosition(world) == civ1.getPosition(world))
					a++;
				else
					b++;
			}
			if((a+b) == 0)
				return;
			int bury = civ.getBuriedness();
			int saveTime = 0;
			if(bury < a*7)
				saveTime =  wst.getTime() + bury/a + 1;
			else
				saveTime =  wst.getTime() + 7 + (bury-a*7)/(b+a) + 1;
			
			if(saveTime < civ.getState().getDeadTime()-5 || saveTime > 305)
				m_savedCiv.add(civ);
		}
		
		
		public void dealBoundary()
		{
			Civilian bestTarget = null;
			int minMargin = 7;
			for(Iterator it = m_assignBoundary.keySet().iterator();it.hasNext();)
			{
				Civilian civ = (Civilian)it.next();
				int margin = m_assignBoundary.get(civ);
				
				if(m_beAssigned.containsKey(civ))
					continue;
				if(margin < minMargin)
				{
					minMargin = margin;
					bestTarget = civ;
				}
			}
			if(bestTarget == null)
				return;
			ArrayList<AmbulanceTeam> source = new ArrayList<AmbulanceTeam>();
			for(Iterator it = m_assign.keySet().iterator();it.hasNext();)
			{
				AmbulanceTeam at = (AmbulanceTeam)it.next();
				Civilian civ = (Civilian)m_assign.get(at);
				if(civ.getState().getRescuePriority() > bestTarget.getState().getRescuePriority())
					source.add(at);
			}
			if(source.isEmpty())
				return;
			
			AmbulanceTeam bestAT = null;
			double bestScore = Double.MIN_VALUE;
			for(Iterator it = source.iterator();it.hasNext();)
			{
				AmbulanceTeam at = (AmbulanceTeam)it.next();
				Human orgCiv = m_assign.get(at);
				
				if(orgCiv == null)
					continue;
				
				if(m_rescueBoundary.get(orgCiv) != null && m_rescueBoundary.get(orgCiv) < 15)
					continue;
				
				if(orgCiv.getBuriedness()==0)
					continue;
				
				int dist = 0;
				
				try {
					List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
							orgCiv.getPosition());
					for(EntityID id : path){
						StandardEntity next = world.getEntity(id);
						if(next instanceof Road){
							dist += ((Road) next).getState().getLength(world);
						}
					}
					
//					dist = (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(bestTarget.getRealPosition()), 
//							HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/2000;
				} catch (Exception e) {
					dist = (int)(Math.random()*500);
					continue;
				}
				
//				int distCost = dist - world.pathPlanner.getNearestDistance(at.id, orgCiv.id);
				
				int distCost = 0;
				
				try {
					List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
							orgCiv.getPosition());
					for(EntityID id : path){
						StandardEntity next = world.getEntity(id);
						if(next instanceof Road){
							distCost += ((Road) next).getState().getLength(world);
						}
					}
//					
//					distCost = dist - (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(orgCiv.getRealPosition()), 
//							HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/2000;
				} catch (Exception e) {
					distCost = dist - (int)(Math.random()*500);
					continue;
				}
			
				double score = (double)orgCiv.getBuriedness()/getCivBeAssigned().get(orgCiv).value;
				score = score/distCost;
				if(score < 0)
					score = 1000;
				if(score > bestScore)
				{
					bestScore = score;
					bestAT = at;
				}
			}
			
			if(bestAT == null)
				return;
			remove(bestAT);
			addAssign(bestAT,bestTarget);
		}
		
		public void adjust()
		{
			ArrayList<AmbulanceTeam> removeList = new ArrayList<AmbulanceTeam>();
			for(Iterator it = m_assign.keySet().iterator();it.hasNext();)
			{
				AmbulanceTeam at = (AmbulanceTeam)it.next();
				Human civ = m_assign.get(at);
				
				try {
					List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
							civ.getPosition());
					int distance = 0;
					for(EntityID id : path){
						StandardEntity next = world.getEntity(id);
						if(next instanceof Road){
							distance += ((Road) next).getState().getLength(world);
						}
					}
					
//					int distance = (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(civ.getRealPosition()), 
//							HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/2000;
				
					if(distance < Integer.MAX_VALUE/2000)
						continue;
					
				} catch (Exception e) {
					// TODO
					if(Math.random() > 0.7)
						continue;
				}
				
				
				removeList.add(at);
			}
			
			for(int i=0;i<removeList.size();i++)
			{
				AmbulanceTeam at = removeList.get(i);
				remove(at);
			}
			for(int i=0;i<removeList.size();i++)
			{
				AmbulanceTeam at = removeList.get(i);
				addAssign(at,m_greedyAssign.get(at));
			}
			
			// adjust boundary assign 
			ArrayList<AmbulanceTeam> adjustAT = new ArrayList<AmbulanceTeam>();
			for(Iterator it = getCivBeAssigned().keySet().iterator();it.hasNext();)
			{
				Civilian civ = (Civilian)it.next();
				int power = getCivBeAssigned().get(civ).value;
				if(civ.getBuriedness()==0 && power > 1)
				{
					int minDist = Integer.MAX_VALUE;
					AmbulanceTeam bestAT = null;
					for(Iterator it2 = m_assign.keySet().iterator();it2.hasNext();)
					{
						AmbulanceTeam at = (AmbulanceTeam)it2.next();
						if(m_assign.get(at)!=civ)
							continue;
//						int dist = world.pathPlanner.getNearestDistance(at.id, civ.id);
						int dist = 0;
						
						try {
							List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
									civ.getPosition());
							for(EntityID id : path){
								StandardEntity next = world.getEntity(id);
								if(next instanceof Road){
									dist += ((Road) next).getState().getLength(world);
								}
							}
							
//							dist = (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(civ.getRealPosition()), 
//									HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/2000;
						} catch (Exception e) {
							dist = (int)(Math.random()*500);
						}
					
						if(dist < minDist)
						{
							minDist = dist;
							bestAT = at;
						}
					}
					//
					for(Iterator it2 = m_assign.keySet().iterator();it2.hasNext();)
					{
						AmbulanceTeam at = (AmbulanceTeam)it2.next();
						if(m_assign.get(at)==civ && at != bestAT)
							adjustAT.add(at);
					}
				}
			}
			
			for(AmbulanceTeam at : adjustAT)
			{
				remove(at);
			}
			
			for(AmbulanceTeam at : adjustAT)
			{
				addAssign(at,m_greedyAssign.get(at));
			}
		}
		

		public void dealWithSpecial()
		{
			ArrayList<AmbulanceTeam> adjustAT = new ArrayList<AmbulanceTeam>();
			for(Iterator it = getCivBeAssigned().keySet().iterator();it.hasNext();)
			{
				Civilian c = (Civilian)it.next();
				int ATs = getCivBeAssigned().get(c).value;
				if(c.getBuriedness()==0 && ATs > 1)
				{
					int minDist = Integer.MAX_VALUE;
					AmbulanceTeam bestAT = null;
					for(Iterator it2 = m_assign.keySet().iterator();it2.hasNext();)
					{
						AmbulanceTeam at = (AmbulanceTeam)it2.next();
						if(m_assign.get(at) != c)
							continue;
//						int dist = world.pathPlanner.getNearestDistance(at.id, c.id);
						int dist = 0;
						
						try {
							List<EntityID> path = search.breadthFirstSearch(at.getPosition(),
									c.getPosition());
							for(EntityID id : path){
								StandardEntity next = world.getEntity(id);
								if(next instanceof Road){
									dist += ((Road) next).getState().getLength(world);
								}
							}
							
//							dist = (int)Router.getASGS(at.getMotionlessPosition(), Collections.singleton(c.getRealPosition()), 
//									HumanoidAgent.SIMPLE_COST_FUNC, HumanoidAgent.LINE_DIST_HFUNC).cost/3000;
						} catch (Exception e) {
							dist = (int)(Math.random()*500);
							continue;
						}
					
						if(dist < minDist)
						{
							minDist = dist;
							bestAT = at;
						}
					}
					//
					for(Iterator it2 = m_assign.keySet().iterator();it2.hasNext();)
					{
						AmbulanceTeam at = (AmbulanceTeam)it2.next();
						if(m_assign.get(at) == c && at != bestAT)
							adjustAT.add(at);
					}
				}
			}
			//
			for(int i=0;i<adjustAT.size();i++)
			{
				AmbulanceTeam at = adjustAT.get(i);
				remove(at);
			}
			//
			for(int i=0;i<adjustAT.size();i++)
			{
				AmbulanceTeam at = adjustAT.get(i);
				addAssign(at,m_greedyAssign.get(at));
				
			}
		}
		
		
	}
	
	private class CivilianSimulator implements Comparable<CivilianSimulator>, Prioritized{
		public int m_time;
		public Civilian m_civ;
		public double m_workLeft;
		public int m_numAT;
		public int m_savedTime;
		public int m_deadTime;
		
		public double getPriority(){
			return m_savedTime;
		}
		public void setPriority(double key){
			m_savedTime = (int)key;
		}
		public int getHashKey(){
			return m_civ.getID().getValue();
		}
		
		public CivilianSimulator(Civilian civ, int deadTime){
			m_time = wst.getTime();
			m_civ = civ;
			m_workLeft = civ.getBuriedness();
			m_numAT = 0;
			m_savedTime = 10000;
			m_deadTime = deadTime;
		}
		
		public int getNumATNeeded(int givenTime)
		{
			if(m_savedTime + TIME_COST_TO_REFUGE < m_deadTime)
				return 0;
			if(givenTime >= m_deadTime - TIME_COST_TO_REFUGE)
				return 100;
			double workLeft = m_workLeft - m_numAT * (givenTime-m_time);
			double tmp = workLeft/(m_deadTime - TIME_COST_TO_REFUGE-givenTime) - m_numAT;
			if(tmp < 0)
				return 0;
			int result = ((int)tmp) + 1;
			return result;
		}
		
		public void addATSkillful(int numAT , int timeReady)
		{
			if(m_civ.getState().m_rescueATs <= 0)
			{
				addAT(numAT,timeReady);
				return;
			}
			
			if(numAT <= m_civ.getState().m_rescueATs)
			{
				addAT(numAT,wst.getTime());
				return;
			}
			
			addAT(m_civ.getState().m_rescueATs, wst.getTime());
			addAT(numAT - m_civ.getState().m_rescueATs, timeReady);
		}
		
		public void addAT(int numAT , int givenTime)
		{
			if(m_numAT == 0)
			{
				m_time = givenTime;
				m_numAT = numAT;
				m_savedTime = m_time + (int)(m_workLeft/numAT);
			}
			else
			{
				if(givenTime < m_time)
				{
//					System.err.println(" givenTime < m_time in addAT() ... ");
					m_workLeft -= (m_time-givenTime)*numAT;
					if(m_workLeft < 0)
						m_workLeft = 0;
					m_numAT += numAT;
					m_savedTime = m_time + (int)(m_workLeft/m_numAT);
				}
				
				int t = (int)(m_workLeft/m_numAT) + 1;
				m_workLeft -= (givenTime-m_time)*m_numAT;
				if(m_workLeft<0)
				{
					m_workLeft = 0;
					m_time += t;
				}
				else
					m_time = givenTime;
				m_numAT += numAT;
				m_savedTime = m_time + (int)(m_workLeft/m_numAT);
			}
		}
		
		public void removeAT(int numAT)
		{
			if(m_numAT <=0)
				return;
			else if(m_numAT==numAT)
			{
				m_numAT = 0;
				m_savedTime = 500;
			}
			else
			{
				m_numAT-=numAT;
				m_savedTime = m_time + (int)(m_workLeft/m_numAT);
			}
		}
		
		public int compareTo(CivilianSimulator o) {
			int me = m_savedTime;
			int him = o.m_savedTime;
			if(me < him)
				return -1;
			else
				return 1;
		}
		
	}

	private static class civilianLifeComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			Civilian civ1 = (Civilian) o1;
			Civilian civ2 = (Civilian) o2;
			double value1 = civ1.getState().getDeadTime();
			double value2 = civ2.getState().getDeadTime();
			if (value1 < value2)
				return -1;
			else
				return 1;
		}
	}
	
//	private static class civilian4KLifeComparator implements Comparator {
//		public int compare(Object o1, Object o2) {
//			Civilian civ1 = (Civilian) o1;
//			Civilian civ2 = (Civilian) o2;
//			double value1 = civ1.m_4kHpDeadTime;
//			double value2 = civ2.m_4kHpDeadTime;
//			if (value1 < value2)
//				return -1;
//			else
//				return 1;
//		}
//	}
	
}


