package seu.utilities;

import java.util.*;

public class BinaryMinHeap {

	private int m_size = 0;

	private Prioritized[] m_array;

	private HashMap<Integer, PositionProperty> m_positionMap;

	public BinaryMinHeap(int capacity) {
		m_array = new Prioritized[capacity];
		m_positionMap = new HashMap<Integer, PositionProperty>();
	}

	public int size() {
		return m_size;
	}

	public boolean isEmpty() {
		return (m_size == 0);
	}

	private void doubleArray() {
		Prioritized[] newArray = new Prioritized[m_array.length * 2];
		System.arraycopy(m_array, 0, newArray, 0, m_array.length);
		m_array = newArray;
	}

	public void insert(Prioritized p) {
		if (m_size + 1 >= m_array.length)
			doubleArray();
		int hole = m_size + 1;
		m_array[0] = p;
		for (; p.getPriority() < m_array[hole / 2].getPriority(); hole /= 2) {
			m_array[hole] = m_array[hole / 2];
			m_positionMap.get(m_array[hole].getHashKey()).m_position = hole;
		}
		m_array[hole] = p;
		m_positionMap.put(p.getHashKey(), new PositionProperty(hole));
		m_size++;
	}

	public boolean contains(Prioritized p) {
		return m_positionMap.containsKey(p.getHashKey());
	}

	public boolean containsKey(int key) {
		return m_positionMap.containsKey(key);
	}

	public Prioritized get(int key) {
		return m_array[m_positionMap.get(key).m_position];
	}

	public int findPos(Prioritized p) {
		PositionProperty pos = m_positionMap.get(p.getHashKey());
		if (pos != null)
			return pos.m_position;
		else
			return -1;
	}

	public Prioritized getMin() {
		return m_array[1];
	}

	private void shift_down(int pos) {
		int child;
		Prioritized tmp = m_array[pos];

		for (; (pos * 2) <= m_size; pos = child) {
			child = pos * 2;

			if ((child != m_size)
					&& m_array[child + 1].getPriority() < m_array[child]
							.getPriority()) {
				child++;
			}

			if (m_array[child].getPriority() < tmp.getPriority()) {
				m_array[pos] = m_array[child];
				m_positionMap.get(m_array[pos].getHashKey()).m_position = pos;
			} else {
				break;
			}
		}

		m_array[pos] = tmp;
		m_positionMap.get(tmp.getHashKey()).m_position = pos;
	}

	public Prioritized deleteMin() {
		Prioritized minItem = getMin();
		m_positionMap.remove(minItem.getHashKey());

		m_array[1] = m_array[m_size--];
		if (m_size == 0)
			return minItem;
		m_positionMap.get(m_array[1].getHashKey()).m_position = 1;
		shift_down(1);
		return minItem;
	}

	public void decreaseKey(Prioritized p, double newKey) {
		p.setPriority(newKey);
		int hole = m_positionMap.get(p.getHashKey()).m_position;
		while (hole > 1 && p.getPriority() < m_array[hole / 2].getPriority()) {
			m_array[hole] = m_array[hole / 2];
			m_positionMap.get(m_array[hole].getHashKey()).m_position = hole;
			hole /= 2;
		}

		m_array[hole] = p;
		m_positionMap.get(p.getHashKey()).m_position = hole;
	}

	private class PositionProperty {
		public int m_position = -1;

		public PositionProperty(int pos) {
			m_position = pos;
		}
	}

}
