package seu.utilities;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.worldmodel.EntityID;
import seu.SeuConstants;
import seu.entities.Node;
import seu.entities.Path;

public class ConnectedComponents
{

	private final StandardWorldModel world;

	public List<Road> totalNodes;

	public ArrayList<Graph> mainInfo;

	public HashMap<Path, Integer> m_additionalPathValue;

	public HashMap<Integer, Integer> mapPathNode;

	public HashSet<Integer> connectedBlockedPaths;

	private HashMap<Integer, Integer> mapPath;

	public HashMap<Integer, Integer> mapEdgeNode;

	private ArrayList<Boolean> isPathVisited;
	

	public Graph myGraph = null;

	public ConnectedComponents(StandardWorldModel world,List<Path> pathss)
	{
		this.world = world;
		mainInfo = new ArrayList<Graph>();
		totalNodes = new ArrayList<Road>();
		m_additionalPathValue = new HashMap<Path, Integer>();
		connectedBlockedPaths = new HashSet<Integer>();
		mapPath = new HashMap<Integer, Integer>();
		mapEdgeNode = new HashMap<Integer, Integer>();
		mapPathNode = new HashMap<Integer, Integer>();
		isPathVisited = new ArrayList<Boolean>();

		for (Iterator it = pathss.iterator(); it.hasNext();)
		{
			Path path = (Path) it.next();
			if (!totalNodes.contains(path.getHeadNode()))
				totalNodes.add(path.getHeadNode());
			if (!totalNodes.contains(path.getTailNode()))
				totalNodes.add(path.getTailNode());
		}
		Collections.sort(totalNodes, AgentUtil.idComparator);

		// System.err.println(" total node= "+totalNodes.size());
	}

	public HashSet<Integer> containedMovingAgents = new HashSet<Integer>();

	public void step()
	{
		createConnectedComponents();

//		logConnectedComponents();

		if (!(world.getState().me instanceof PoliceForce))
			return;

		updateGraphForEvaluation();
		m_additionalPathValue = new HashMap<Path, Integer>();
		for (Graph g : mainInfo)
		{
			g.evaluateGraphFlows(world.getState().getTime());

			for (Integer id : connectedBlockedPaths)
			{
				Path path = world.getState().paths.get(id);
				int value = evaluateBlockedPath(path);
				if (value > 0)
					m_additionalPathValue.put(path, value);
			}
			if (g.vertices.size() < 2
					&& (g.m_numOfCivilian + g.m_numOfAmbulanceTeam
							+ g.m_numOfFireBrigade > 0))
				containedMovingAgents.add(g.vertices.iterator().next());
		}

	}

	private void createConnectedComponents()
	{
		mainInfo = new ArrayList<Graph>();
		myGraph = null;
		mapPathNode = new HashMap<Integer, Integer>();
		mapPath = new HashMap<Integer, Integer>();
		mapEdgeNode = new HashMap<Integer, Integer>();
		connectedBlockedPaths = new HashSet<Integer>();
		isPathVisited = new ArrayList<Boolean>();

		HashSet<Road> connectorsLeft = new HashSet<Road>(totalNodes);

		for (int i = 0; i < world.getState().paths.size(); i++)
		{
			isPathVisited.add(false);
		}

//		long t = System.currentTimeMillis();
		int index = 0;
		while (!connectorsLeft.isEmpty())
		{
			createOneConnectedComponents(connectorsLeft, index);
			index++;
		}
		
		if (!(world.getState().me instanceof Human))
		{
			return;
		}

		// get my graph.
		for (Graph g : mainInfo)
		{
			if (g.isContains(getOutsidePosition((Human)world.getState().me),world))
			{
				myGraph = g;
				// myGraph.vertices.addAll(world.passedNodes);
				break;
			}

		}
		
		if (myGraph == null)
		{
			// System.out.println(world.me + "   new  graph ");
			myGraph = new Graph();

			mainInfo.add(myGraph);

			for (Road r : world.getState().passedRoads)
				myGraph.vertices.add(new Integer(r.getID().getValue()));
			// myGraph.vertices.addAll(world.passedNodes);

		}
		
	}

	public Area getOutsidePosition(Human mo)
	{
		StandardEntity mop = mo.getPosition(world);
		if (mop == null)
			return null;
		if (mop instanceof Road)
		{

			Road rd = (Road) mop;
			return rd;
		}
		if (mop instanceof Building)
			return (Building)mop;
		throw new Error(mop.toString());
	}

	private void createOneConnectedComponents(HashSet connectorsLeft, int index)
	{
		if (connectorsLeft.isEmpty())
			return;
		Graph g = new Graph();
		ArrayList<Road> v = new ArrayList<Road>();

		Road n = (Road) connectorsLeft.iterator().next();
		mapPathNode.put(new Integer(n.getID().getValue()), new Integer(index));
		connectorsLeft.remove(n);
		g.vertices.add(new Integer(n.getID().getValue()));
		g.index = index;
		v.add(n);
		while (!v.isEmpty())
		{
			Road nowC = v.get(0);
			v.remove(0);

			for (Iterator it = nowC.getState().getContainedPaths(world).iterator(); it
					.hasNext();)
			{
				Path path = (Path) it.next();
				if (isPathVisited.get(path.id) == true)
					continue;

				if (path.isBlocked)
				{

					// add connected & passable part of path to graph.
					// Route route = path.toRoute(nowNode.id ==
					// path.getTailNode().id);

					if (path.reachableRoadsFromEnds != null && path.reachableRoadsFromEnds
							.get(nowC.getID().getValue()) != null)
					{
						ArrayList<Road> roadss = path.reachableRoadsFromEnds
								.get(nowC.getID().getValue());

						for (Road rd : roadss)
						{
							mapEdgeNode.put(new Integer(rd.getID().getValue()), new Integer(
									index));
							g.vertices.add(new Integer(rd.getID().getValue()));
							g.buildings.addAll(rd.getState().getConnectedBuildings(world));
						}
					}
					// for (Route rt = route; rt.previous() != null; rt = rt
					// .previous()) {
					// MotionlessObject mo = rt.end;
					//
					// if (mo instanceof Node) {
					// // if(!g.vertices.contains(mo))
					// g.vertices.add((Node) mo);
					// //
					// g.buildings.addAll(((Node)mo).getConnectedBuildings());
					// // if(!mapEdgeNode.keySet().contains(mo.id))
					// mapEdgeNode.put(mo.id, g);
					//							
					// continue;
					// }
					// if(((Road) mo).getCondition() == Road.CONDITION_BLOCKED)
					// {
					// break;
					// }
					// //
					// g.buildings.addAll(((Road)mo).getConnectedBuildings());
					// }

					// long t3 = System.currentTimeMillis();
					// if ((t3 - t2) > 10)
					// System.out.println(world.me
					// + "   add buildings  used  "
					// + (t3 - t2)
					// + "  reachable ends:  "
					// + path.reachableNodesFromEnds.get(nowNode.id)
					// .size());

					connectedBlockedPaths.add(new Integer(path.id));

					continue;
				}

				// if(count[1] == path.getContainedRoads().size()){
				// //TODO
				// continue;
				// }

				isPathVisited.set(path.id, true);

				g.edges.add(new Integer(path.id));
				g.buildings.addAll(path.getConnectedBuildings());

				mapPath.put(new Integer(path.id), new Integer(index));

				if (path.getHeadNode().getID().getValue() == path.getTailNode().getID().getValue())
				{
					System.out.println(" Closed Path: " + path.getHeadRoad());
					continue;
				}

				Road end = path.getHeadNode();
				if (end.getID().getValue() == nowC.getID().getValue())
					end = path.getTailNode();

				if (mapPathNode.containsKey(end.getID().getValue()))
					continue;
				connectorsLeft.remove(end);
				g.vertices.add(new Integer(end.getID().getValue()));
				mapPathNode.put(new Integer(end.getID().getValue()), new Integer(index));
				v.add(end);
				
			}
		}
		mainInfo.add(g);

	}

	private void updateGraphForEvaluation()
	{
		for (Iterator it = world.getEntitiesOfType(StandardEntityURN.REFUGE).iterator(); it.hasNext();)
		{
			Refuge r = (Refuge) it.next();
			Graph g=null;
			for(EntityID eId:r.getNeighbours())
				if(world.getEntity(eId) instanceof Road){
					g=getGraph((Road)world.getEntity(eId));
					break;
				}
			if (g != null)
			{
				g.m_numRefuge++;
			}
		}

		//
		for (Iterator it = world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM).iterator(); it.hasNext();)
		{
			AmbulanceTeam at = (AmbulanceTeam) it.next();

			if (at.getPosition(world) == null)
				continue;
			
			Area mlo=null;
			if(at.getPosition(world) instanceof Area){
				mlo = (Area)at.getPosition(world);
			}else{
				System.out.println(at.getPosition(world).toString());
			}
			Road node = null;

			if (mlo instanceof Building)
			{
				for(EntityID eId:((Building)mlo).getNeighbours())
					if(world.getEntity(eId) instanceof Road){
						node=(Road)world.getEntity(eId);
						break;
					}
			} else if (mlo instanceof Road)
			{
				node = (Road) mlo;
			}

			Graph g = getGraph(node);

			if (g != null)
			{
				g.m_numOfAmbulanceTeam++;
			}
		}

		//
		for (Iterator it = world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE).iterator(); it.hasNext();)
		{
			FireBrigade fb = (FireBrigade) it.next();

			if (fb.getPosition(world) == null)
				continue;
			
			Area mlo=null;
			if(fb.getPosition(world) instanceof Area){
				mlo = (Area)fb.getPosition(world);
			}else{
				System.out.println(fb.getPosition(world).toString());
			}
			Road node = null;

			if (mlo instanceof Building)
			{
				for(EntityID eId:((Building)mlo).getNeighbours())
					if(world.getEntity(eId) instanceof Road){
						node=(Road)world.getEntity(eId);
						break;
					}
			} else if (mlo instanceof Road)
			{
				node = (Road) mlo;
			}

			Graph g = getGraph(node);

			if (g != null)
			{
				g.m_numOfAmbulanceTeam++;
			}
		}
		//
		for (Iterator it = world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE).iterator(); it.hasNext();)
		{
			PoliceForce pf = (PoliceForce) it.next();

			if (pf.getPosition(world) == null)
				continue;
			
			Area mlo=null;
			if(pf.getPosition(world) instanceof Area){
				mlo = (Area)pf.getPosition(world);
			}else{
				System.out.println(pf.getPosition(world).toString());
			}
			Road node = null;

			if (mlo instanceof Building)
			{
				for(EntityID eId:((Building)mlo).getNeighbours())
					if(world.getEntity(eId) instanceof Road){
						node=(Road)world.getEntity(eId);
						break;
					}
			} else if (mlo instanceof Road)
			{
				node = (Road) mlo;
			}

			Graph g = getGraph(node);

			if (g != null)
			{
				g.m_numOfAmbulanceTeam++;
			}
		}
		//

		for (Iterator it = world.getEntitiesOfType(StandardEntityURN.CIVILIAN).iterator(); it.hasNext();)
		{
			Civilian c = (Civilian) it.next();
			if (c.getPosition(world) == null
					|| !(c.getState().needLoad(world) || c.getState().needRescue(world)))
				continue;
			Area mlo=null;
			if(c.getPosition(world) instanceof Area){
				mlo = (Area)c.getPosition(world);
			}else{
				System.out.println(c.getPosition(world).toString());
			}
			Road node = null;
			if (mlo instanceof Building)
			{
				for(EntityID eId:((Building)mlo).getNeighbours())
					if(world.getEntity(eId) instanceof Road){
						node=(Road)world.getEntity(eId);
						break;
					}
				Graph g = getGraph(node);

				if (g != null)
				{
					g.m_numOfCivilian++;
				}
			}

		}
		//
		for (Iterator it = world.getEntitiesOfType(StandardEntityURN.BUILDING).iterator(); it.hasNext();)
		{
			Building b = (Building) it.next();

			if (!b.isOnFire() || b == null)
				continue;
			Road en=null;
			for(EntityID eId:b.getNeighbours())
				if(world.getEntity(eId) instanceof Road){
					en=(Road)world.getEntity(eId);
					break;
				}
			Graph g = getGraph(en);

			if (g != null)
			{
				g.m_totalCivValue += b.getState().getCivPossibility(world);
				g.m_numOfBurningBuilding++;
				g.m_totalAreaOfBurningBuilding += b.getTotalArea();
			}
		}
		//

	}

	private Graph getGraph(Road connector)
	{
		System.out.println(connector);
		if (mapPathNode.containsKey(connector.getID().getValue()))
			return mainInfo.get(mapPathNode.get(connector.getID().getValue()));
		if (mapEdgeNode.containsKey(connector.getID().getValue()))
			return mainInfo.get(mapEdgeNode.get(connector.getID().getValue()));
		if(connector.getState().getContainedPaths(world).isEmpty()){
			return null;
		}
		if (mapPath.containsKey(connector.getState().getContainedPaths(world).get(0).id))
			return mainInfo
					.get(mapPath.get(connector.getState().getContainedPaths(world).get(0).id));
		return null;
	}

	private int evaluateBlockedPath(Path path)
	{
		Graph g1 = mainInfo.get(mapPathNode.get(path.getHeadNode().getID().getValue()));
		Graph g2 = mainInfo.get(mapPathNode.get(path.getTailNode().getID().getValue()));
		int value1 = 0;
		int value2 = 0;
		if (g1 != null && g1.m_numOfAmbulanceTeam > g1.m_maxATNeeded)
		{
			value1 += (g1.m_numOfAmbulanceTeam - g1.m_maxATNeeded);
		}
		//
		if (g1 != null && g1.m_numOfFireBrigade > g1.m_maxFBNeeded)
		{
			value1 += (g1.m_numOfFireBrigade - g1.m_maxFBNeeded);
		}
		//
		if (g2 != null && g2.m_numOfAmbulanceTeam > g2.m_maxATNeeded)
		{
			value2 += (g2.m_numOfAmbulanceTeam - g2.m_maxATNeeded);
		}
		//
		if (g2 != null && g2.m_numOfFireBrigade > g2.m_maxFBNeeded)
		{
			value2 += (g2.m_numOfFireBrigade - g2.m_maxFBNeeded);
		}

		int result = Math.abs(value1 - value2);
		if (g1 != null && g2 != null && g1 != g2)
		{
			result = value1 + value2;
		}
		//
		if (world.getState().getTime() >= 30)
		{
			if (g1 != null
					&& (g1.m_numOfAmbulanceTeam > 0 || g1.m_numOfFireBrigade > 0)
					&& g1.m_numRefuge == 0)
				result += 4;
			else if (g2 != null
					&& (g2.m_numOfAmbulanceTeam > 0 || g2.m_numOfFireBrigade > 0)
					&& g2.m_numRefuge == 0)
				result += 4;
		}

		return result;

	}

	public boolean isReachable(StandardEntity real)
	{
		if (myGraph == null || real instanceof Road)
		{
			// TODO
			return Math.random() < 0.3;
		}

		Road node = null;
		if (real instanceof Human)
		{
			StandardEntity mro = ((Human)real).getPosition(world);
			Area mlo=null;
			if(mro instanceof Area)
				mlo=(Area) mro;
			if (mlo instanceof Building)
			{
				for(EntityID eId:((Building)mlo).getNeighbours())
					if(world.getEntity(eId) instanceof Road){
						node=(Road)world.getEntity(eId);
						break;
					}
			} else if (mlo instanceof Road)
			{
				node = (Road) mlo;
			}
		} else if (real instanceof Building)
		{
			for(EntityID eId:((Building)real).getNeighbours())
				if(world.getEntity(eId) instanceof Road){
					node=(Road)world.getEntity(eId);
					break;
				}
		}

		if (node == null)
		{
			System.out.println("reachable : ?  " + real);
			return false;
		}

		if (mapPathNode.containsKey(node.getID().getValue()))
			return true;

		if (mapEdgeNode.containsKey(node.getID().getValue()))
			return true;

		if(node.getState().getContainedPaths(world).size() == 0){
			System.out.println("no path");
			return false;
		}
		else{
			System.out.println("have path");
		}
		
		if (myGraph.edges.contains(node.getState().getContainedPaths(world).get(0).id))
			return true;

		return false;

	}

	public static void minimumSpanningTreePrim(int n, int dist[][], int tree[])
	{
		int i, j, n1, d, mindist, node, k = 0;
		n1 = n - 1;
		for (i = 1; i <= n1; i++)
			tree[i] = -n;
		tree[n] = 0;
		for (i = 1; i <= n1; i++)
		{
			mindist = Integer.MAX_VALUE;
			for (j = 1; j <= n1; j++)
			{
				node = tree[j];
				if (node <= 0)
				{
					d = dist[-node][j];
					if (d < mindist)
					{
						mindist = d;
						k = j;
					}
				}
			}
			tree[k] = -tree[k];
			for (j = 1; j <= n1; j++)
			{
				node = tree[j];
				if (node <= 0)
					if (dist[j][k] < dist[j][-node])
						tree[j] = -k;
			}
		}
	}

//	private void logConnectedComponents()
//	{
//		Color[] colorTable = new Color[10];
//		colorTable[0] = Color.BLUE;
//		colorTable[1] = Color.CYAN;
//		colorTable[2] = Color.GREEN;
//		colorTable[3] = Color.MAGENTA;
//		colorTable[4] = Color.ORANGE;
//		colorTable[5] = Color.PINK;
//		colorTable[6] = Color.RED;
//		colorTable[7] = Color.WHITE;
//		colorTable[8] = Color.YELLOW;
//		colorTable[9] = new Color(224, 159, 120);
//
//		if (!SeuConstants.IS_COMPETETION_MODE)
//		{
//			for (int i = 0; i < mainInfo.size(); i++)
//			{
//				Color color = colorTable[i % 10];
//				Graph g = mainInfo.get(i);
//				// for(Node node : g.vertices)
//				// world.logFile.logIDObject(node.id, color, "CG");
//				for (Integer id : g.edges)
//				{
//					Path p = world.getState().paths.get(id);
//				}
//			}
//
//			Color color = Color.RED;
//		}
//
//	}

}
