/*
 * @author JiangYh
 * 2010.5.10
 */
package seu.utilities;

import java.util.Comparator;

import rescuecore2.standard.entities.AmbulanceCentre;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.FireStation;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.PoliceOffice;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import seu.agent.PlatoonAgent;
import seu.entities.ExtraProperty;

//saeed 85-2-3
public abstract interface ConstantConditions {
	// -------------------------------------------------------------------
	// Object
	static final Condition TRUE_C = new Condition() {
		public boolean eval(Object obj) {
			return true;
		}
	};

	static final Condition FALSE_C = new Condition() {
		public boolean eval(Object obj) {
			return false;
		}
	};

	static final Condition CIVILIAN_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof Civilian);
		}
	};

	public static final Condition FIRE_BRIGADE_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof FireBrigade);
		}
	};

	public static final Condition AMBULANCE_TEAM_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof AmbulanceTeam);
		}
	};

	public static final Condition POLICE_FORCE_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof PoliceForce);
		}
	};

	static final Condition PLATOON_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof PlatoonAgent);
		}
	};

	public static final Condition AMBULANCE_CENTER_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof AmbulanceCentre);
		}
	};

	public static final Condition FIRE_STATION_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof FireStation);
		}
	};

	public static final Condition POLICE_OFFICE_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof PoliceOffice);
		}
	};

	static final Condition ROAD_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof Road);
		}
	};

	// public static final Condition NODE_C = new Condition() {
	// public boolean eval(Object obj) {
	// return (obj instanceof Node);
	// }
	// };

	public static final Condition Condition_Area = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof Area);
		}
	};

	static final Condition BUILDING_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof Building);
		}
	};

	static final Condition REFUGE_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof Refuge);
		}
	};

	static final Condition CENTER_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof PoliceOffice
					|| obj instanceof AmbulanceCentre || obj instanceof FireStation);
		}
	};

	static final Condition SPECIAL_BUILDING_C = REFUGE_C.or(CENTER_C);

//	public static final Condition VISIBLE_C = new Condition() {
//		public boolean eval(Object obj) {
//			return ((StandardEntity) obj).isVisible();
//		}
//	};

	// public static final Condition MO_POS_SET_C = new Condition() {
	// public boolean eval(Object obj) {
	// return ((MovingObject) obj).position() != null;
	// }
	// };

	// public static final Condition HAS_BEEN_SEEN_CND = Property.get(
	// "RealObject", "hasBeenSeen").equal(Boolean.TRUE);
	//
	// public static final Condition HAS_NOT_BEEN_SEEN_CND = Property.get(
	// "RealObject", "hasBeenSeen").equal(Boolean.FALSE);
	//	
	// public static final Condition HAS_BEEN_HEARED_CND = Property.get(
	// "RealObject", "hasBeenHeared").equal(Boolean.TRUE);
	//
	// public static final Condition HAS_NOT_BEEN_HEARED_CND = Property.get(
	// "RealObject", "hasBeenHeared").equal(Boolean.FALSE);

	// -------------------------------------------------------------
	// MovingObject
	/*
	 * static final Condition MISSED_C = new Condition() { public boolean
	 * eval(Object obj) { return ( (MovingObject) obj).isMissed(); } };
	 * 
	 * static final Condition LOST_C = new Condition() { public boolean
	 * eval(Object obj) { return ( (MovingObject) obj).isLost(); } };
	 */

	// -----------------------------------------------------------------
	// Humanoid
	public static final Condition BURIED_C = new Condition() {
		public boolean eval(Object obj) {
			return ((Human) obj).getBuriedness() > 0;
		}
	};

	public static final Condition NEED_LOAD_C = new Condition() {
		public boolean eval(Object obj, StandardWorldModel model) {
			Human hum = (Human) obj;
			StandardEntity pos = hum.getPosition(model);
			return hum.getBuriedness() == 0 && hum.getDamage() >= 0
					&& hum.getDamage() < 999 && pos != null
					&& !(pos instanceof Refuge) && hum instanceof Civilian
					&& pos instanceof Building;
		}

		@Override
		public boolean eval(Object obj) {
			// TODO Auto-generated method stub
			return false;
		}
	};

	public static final Condition ALIVE_C = new Condition() {
		public boolean eval(Object obj) {
			return ((Human) obj).getHP() > 0;
		}
	};

	public static final Condition HUMANOID_C = new Condition() {
		public boolean eval(Object obj) {
			return obj instanceof Human;
		}
	};

	public static final Condition AGENT_C = HUMANOID_C.and(CIVILIAN_C.not());

	// public static final Condition HUM_TO_REP_C =
	// HUMANOID_C.and(VISIBLE_C).and(
	// BURIED_C).and(ALIVE_C);

	public static final Condition NEED_RESCUE_C = new Condition() {
		public boolean eval(Object obj, StandardWorldModel model) {
			Human hum = (Human) obj;
			StandardEntity pos = hum.getPosition(model);
			return (hum.getHP() > 0 && hum.getBuriedness() > 0 && pos != null && !(pos instanceof Refuge));
		}

		@Override
		public boolean eval(Object obj) {
			// TODO Auto-generated method stub
			return false;
		}
	};

	// /**
	// * the position must be Road
	// */
	// public static final Condition PASSABLE_TO_HEAD = new Condition() {
	// public boolean eval(Object obj,StandardWorldModel model) {
	// if(!(obj instanceof Human)) return false;
	// Human hum = (Human)obj;
	// if(!(hum.getPosition(model) instanceof Road)) return false;
	// Road pos = (Road)hum.getPosition(model);
	// 
	// int posExtra = hum.getPositionExtra();
	// if(pos.aliveLinesToHead() == 0) return false;
	// if(posExtra > pos.getLength()/2 && pos.passableLinesToHead() < 1) return
	// false;
	//			
	// return true;
	// }
	// };
	// /**
	// * the position must be Road
	// */
	// public static final Condition PASSABLE_TO_TAIL = new Condition() {
	// public boolean eval(Object obj) {
	// if(!(obj instanceof Humanoid)) return false;
	// Humanoid hum = (Humanoid)obj;
	// if(!(hum.getRealPosition() instanceof Road)) return false;
	// Road pos = (Road)hum.getRealPosition();
	// 
	// int posExtra = hum.getPositionExtra();
	// if(pos.aliveLinesToTail() == 0) return false;
	// if(posExtra <= pos.getLength()/2 && pos.passableLinesToTail() < 1) return
	// false;
	//			
	// return true;
	// }
	// };

	// -----------------------------------------------------------------
	// Building
	static final Condition NOT_BURNED_C = new Condition() {
		public boolean eval(Object obj) {
			return !((Building) obj).isOnFire();
		}
	};

	public static final Condition BURNING_C = new Condition() {
		public boolean eval(Object obj) {
			return ((Building) obj).isOnFire();
		}
	};

	static final Condition PUT_OUT_C = new Condition() {
		public boolean eval(Object obj) {
			Building bld = (Building) obj;
			return 5 <= bld.getFieryness() && bld.getFieryness() <= 7;
		}
	};

	static final Condition BURNED_C = new Condition() {
		public boolean eval(Object obj) {
			return ((Building) obj).getFieryness() > 7;
		}
	};

	// ---------------------------------------------------------------------
	// Road
	// static final Condition MOVABLE_BOTH_DIRECTIONS_C = new Condition() {
	// public boolean eval(Object obj) {
	// return ((Road) obj).isMovableBothDirections();
	// }
	// };

	// static final Condition NOT_MOVABLE_BOTH_DIRECTIONS_C = new Condition() {
	// public boolean eval(Object obj)
	// { return ((Road) obj).isNotMovableBothDirections(); }
	// };

	// static final Condition MOVABLE_AT_LEAST_ONE_DIRECTION_C = new Condition()
	// {
	// public boolean eval(Object obj) {
	// return ((Road) obj).isMovableAtLeastOneDirection();
	// }
	// };

	// static final Condition NOT_MOVABLE_AT_LEAST_ONE_DIRECTION_C=new
	// Condition(){
	// public boolean eval(Object obj)
	// { return ((Road) obj).isNotMovableAtLeastOneDirection(); }
	// };

	public static final Condition PASSABLE_C = new Condition() {
		public boolean eval(Object obj,StandardWorldModel world) {
			if (obj instanceof Road) {
				Road rd = (Road) obj;
				for (EntityID blk : rd.getBlockades()) {
					if (((Blockade)world.getEntity(blk)).getShape().getBounds2D().intersects(
							rd.getShape().getBounds2D()))
						return false;
				}
				return true;
			}
			throw new Error("Invalid Usage");
		}

		@Override
		public boolean eval(Object obj) {
			// TODO Auto-generated method stub
			return false;
		}
	};

	// public static final Condition PASSABLE_SURE_C = new Condition() {
	// public boolean eval(Object obj) {
	// if (obj instanceof Road) {
	// Road rd = (Road) obj;
	// return rd.getCondition() == rd.CONDITION_PASSABLE;
	// }
	// throw new Error("Invalid Usage");
	// }
	// };

	 public static final Condition NOT_PASSABLE_C = PASSABLE_C.not();

	// public static final Condition CLEARABLE_C = new Condition() {
	// public boolean eval(Object obj) {
	// return ((Road) obj).isClearable();
	// }
	// };

	public static final Condition HAS_BEEN_PASSED_C = new Condition() {
		public boolean eval(Object obj) {
			return ((Road) obj).getState().hasBeenPassed();
		}
	};

	public static final Condition HAS_NOT_BEEN_PASSED_C = HAS_BEEN_PASSED_C
			.not();
	// TODO
	// public static final Condition HAS_NOT_BEEN_INFORMED_C =
	// HAS_NOT_BEEN_PASSED_C.and(HAS_NOT_BEEN_SEEN_CND).and(HAS_NOT_BEEN_HEARED_CND);

	// /** This means certain road HAS NOT BEEN PASSED(modified from HAS NOT
	// BEEN SEEN)<br> or NOT PASSABLE at all.*/
	// public static final Condition SHOULD_CHECK_ROAD_CND =
	// HAS_NOT_BEEN_PASSED_C
	// .or(NOT_PASSABLE_C);

	// Use in FireBrigade
	// 2010 10:32
	//TODO
//	public static final ExtraProperty DAMAGE_PRP = ExtraProperty.get("base.BaseHumanoid",
//			"getDamage"), BURIEDNESS_PRP = ExtraProperty.get("base.BaseHumanoid",
//			"getBuriedness"), POSITION_PRP = ExtraProperty.get("MovingObject",
//			"getRealPosition"), FIERYNESS_PRP = ExtraProperty.get(
//			"base.BaseBuilding", "getFieryness");

	// public static final Property BROKENNESS_PRP =
	// Property.get("base.BaseBuilding",
	// "getBrokenness");

	public static final float IN_BLOCKS = (float) Integer.MAX_VALUE
			* (float) Integer.MAX_VALUE;

	public static final float NOT_SEEN_RD = (float) Integer.MAX_VALUE;

	public static final Condition BROKEN_CND = new Condition() {
		public boolean eval(Object obj) {
			return ((Building) obj).getBrokenness() > 0;
		}
	};

	public static final Comparator BROKENNESS_COMP = new Comparator() {
		public int compare(Object obj1, Object obj2) {
			return ((Building) obj1).getBrokenness()
					- ((Building) obj2).getBrokenness();
		}
	};

	public static final Condition MOVING_AGENT_C = new Condition() {
		public boolean eval(Object obj) {
			return (obj instanceof Human) && !(obj instanceof Civilian);
		}
	};
}
