/**************************************************************************
 * Copyright (C) 2006 Nobel Huang <nobel1984@gmail.com>
 * All rights reserved.
 * Created by Southeast University RoboCup.
 **************************************************************************
 * Last modify:
 * $Id: MapPartition.java,v 1.17 2006/12/29 16:52:29 nobel Exp $
 */
package seu.utilities.Partition;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import seu.entities.Node;
import seu.entities.Path;

/**
 * @author Jiang Yinha
 * @since 2010.06.11
 */
public abstract class MapPartition implements PartitionConstants {
	public final int id;//
	protected HashSet<Path> paths;
	protected HashSet<Road> roads;
	protected HashSet<Node> nodes;
	protected List<Entrance> entrances;
	protected final StandardWorldModel world;
	
	protected boolean isBoundary;
	
	public MapPartition(int id, Boolean isBoundary, HashSet<Path> paths, StandardWorldModel world){
		this.id = id;
		this.world = world;
		this.paths = paths;
		this.isBoundary = isBoundary;
		initialRoads();
//		initialEntrances();
	}
	
	/**
	 * 
	 */
	public abstract boolean isInPartition(StandardEntity ro);
	
	/**
	 * 
	 */
	public HashSet<Path> getConnectedPaths(Node checkNode)
	{
		HashSet<Path> connectedPaths = new HashSet<Path>();
		for(Path path : paths)
		{
			HashSet<Node> nodes = path.getContainedNodes();
			boolean isConnected = false;
			for(Node node : nodes){
				if(node.me().getLine() == checkNode.me().getLine()){isConnected = true;break;}
			}
			if(isConnected == true) connectedPaths.add(path);
		}
		if(connectedPaths.size() == 0) return null;
		return connectedPaths;
	}
	
	/**
	 * 
	 */
	public Road getConnectedRoads(Node checkNode)
	{
		if(world.getEntity(checkNode.me().getNeighbour()) instanceof Road)
			return (Road)world.getEntity(checkNode.me().getNeighbour());
		return null;
	}
	
	/**
	 * roads
	 */
	private void initialRoads(){
		roads = new HashSet<Road>();
		nodes = new HashSet<Node>();
		if(paths == null) return;
		for(Path path : paths){
			roads.addAll(path.getContainedRoads());
			nodes.addAll(path.getContainedNodes());
		}
	}
	
	/**
	 * entrances
	 */
	private void initialEntrances(){
		entrances = new ArrayList<Entrance>();
		if(paths == null) return;
		ArrayList<Node> nodesInPartition = new ArrayList<Node>();
		for(Path path : paths){
			nodesInPartition.addAll(path.getContainedNodes());
		}
		for(Node node : nodesInPartition){
			if(world.getEntity(node.me().getNeighbour()) instanceof Building){
				Entrance et=new Entrance(entrances.size(),node,this);
				entrances.add(et);
			}
		}
		Collections.sort(entrances, seu.utilities.AgentUtil.idComparator2);
	}
	
	public HashSet<Path> getContainedPaths(){
		return paths;
	}
	public HashSet<Road> getContainedRoads(){
		return roads;
	}
	public List<Entrance> getEntrances(){
		return entrances;
	}
	
	public String toString(){
		String result = "partition id:" + this.id +"\t  ";
		
		result += roads.iterator().next();
		return result;
	}
}
