package seu.utilities.Partition;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import seu.entities.Node;
import seu.entities.Path;
import seu.utilities.ConstantConditions;

public class RadialPartition {

	private final StandardWorldModel world;
	public final int id;
	public final double startDir;
	public final double endDir;
	public List<Node> nodes;
	private List<Road> roads;
	private List<Building> buildings;
	private List<Path> paths;
	public HashSet<RadialPartition> neighbourPartitions = new HashSet<RadialPartition>();
	
	
	public RadialPartition(int id, double startDir, double endDir, StandardWorldModel world) {
		this.world = world;
		this.id = id;
		this.startDir = startDir;
		this.endDir = endDir;
		
		nodes = new ArrayList<Node>();
		roads = new ArrayList<Road>();
		buildings = new ArrayList<Building>();
		paths = new ArrayList<Path>();
	}

	public void setNeighbourPartition(RadialPartition rp){
		neighbourPartitions.add(rp);
	}
	
	public void fillPartitions(){
		for(Node node : nodes){
			for(Road r : node.getConnectedRoads(world)){
				if(!roads.contains(r))
					roads.add(r);
			}
			for(Path p : node.getContainedPaths(world)){
				if(!paths.contains(p))
					paths.add(p);
			}
		}
		for(Path path : paths){
			for(Building b : path.getConnectedBuildings()){
				if(!buildings.contains(b))
					buildings.add(b);
			}
		}
		
	}
	
	public List<Building> getContainedBuildings(){
		return new ArrayList<Building>(buildings);
	}
	
	private List<Building> m_unSeenBuildings = null;
	public List<Building> getUnSeenBuildings(){
		if(m_unSeenBuildings == null){
			m_unSeenBuildings = new ArrayList<Building>(buildings);
			m_unSeenBuildings.removeAll(world.getState().specialBuildings);
		}
		return m_unSeenBuildings;
	}
	
	
	private List<Path> m_clearPaths = null;
	public List<Path> getClearingPaths(){
		if(m_clearPaths == null){
			m_clearPaths = new ArrayList<Path>(paths);
		}
		return m_clearPaths;
	}
	
	public List<Path> getContainedPaths(){
		return new ArrayList<Path>(paths);
	}
	
	public List<Road> getContainedRoads(){
		return new ArrayList<Road>(roads);
	}
	
	public boolean isInPartition(StandardEntity ro) {
		Area mlo = null;
		if(ro instanceof Human)
			if(((Human)ro).getPosition(world) instanceof Area)
				mlo=(Area)((Human)ro).getPosition(world);
		else 
			if(ro instanceof Area)
				mlo = (Area)ro;
		
		if(mlo instanceof Road){
			return roads.contains(mlo);
		} else if(mlo instanceof Building){
			return buildings.contains(mlo);
		}
		return false;
	}
	
	public boolean isInNeighbourPartition(StandardEntity ro) {
		for(RadialPartition rp : neighbourPartitions){
			if(rp.isInPartition(ro))
				return true;
		}
		return false;
	}
	
	public String toString(){
		return "RadialPartition["+id + "]";
	}

}
