/**************************************************************************
 * Copyright (C) 2010 Yinhao Jiang <Jiang.Yinhao@gmail.com>
 * All rights reserved.
 * Created by Southeast University RoboCup.
 **************************************************************************
 */
package seu.utilities;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import rescuecore2.config.Config;
import rescuecore2.misc.Pair;
import rescuecore2.misc.EntityTools.IDComparator;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.AbstractEntity;
import rescuecore2.worldmodel.WorldModel;
import seu.agent.PoliceForceAgent;
import seu.states.WorldState;

/**
 * @author JiangYh
 */
public class TrafficEvaluation {

	public final StandardWorldModel world;
	
	public final WorldState worldState;

	private final int BASE_RATION;

	private final int MAP_RATE = 500;// (m)

	public HashMap<StandardEntity, PoliceForce> m_assign =
	        new HashMap<StandardEntity, PoliceForce>();

	public HashMap<StandardEntity, PoliceForce> m_secondAssign =
	        new HashMap<StandardEntity, PoliceForce>();

	public HashMap<StandardEntity, PoliceForce> m_lastAssigned =
	        new HashMap<StandardEntity, PoliceForce>();

	public List<PoliceForce> m_freePFs = new ArrayList<PoliceForce>();

	protected HashMap<StandardEntity, Integer> blockedTargetMap =
	        new HashMap<StandardEntity, Integer>();

	protected List<StandardEntity> blockedTargets =
	        new ArrayList<StandardEntity>();

	// JiangYh
	private final int time;

//	private final Config config;
	final int MAXN = 100;

	//

	public TrafficEvaluation(StandardWorldModel world) {
		this.worldState=world.getState();
		this.world = world;
		this.time = worldState.getTime();
//		this.config = config;
		
		Pair<Integer,Integer> mapTopLeft=world.getWorldBounds().first();
		Pair<Integer,Integer> mapBottomRight=world.getWorldBounds().second();
		if ((mapBottomRight.first()-mapTopLeft.first()+mapTopLeft.second()-mapBottomRight.second()) / 2000 > MAP_RATE) {
			BASE_RATION =
			        35000 + (mapBottomRight.first()-mapTopLeft.first()+mapTopLeft.second()-mapBottomRight.second()) / 1000 / 100;
		} else {
			BASE_RATION = 35000;
		}
	}

	public void evaluate() {
//		if (this.time< config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY))
//			return;
		updateFreePFAndFinishedTargets();
		
	}

	private void updateFreePFAndFinishedTargets() {
		m_freePFs = new ArrayList<PoliceForce>();

		for (StandardEntity se : world
		        .getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
			PoliceForce pf;
			if (se instanceof PoliceForce)
				pf = (PoliceForce) se;
			else
				continue;
			if (pf.getPosition() == null)
				continue;
			if (pf.getPosition(world) instanceof Area) {
				if(pf.getPosition(world) instanceof Building){
					Building pos = (Building) (pf.getPosition(world));
					pos.getState().setAreaChecked(true);
				}
				else if(pf.getPosition(world) instanceof Road){
					Road pos = (Road) (pf.getPosition(world));
					pos.getState().setAreaChecked(true);
				}
			}
			if (pf.isBuriednessDefined() && pf.getBuriedness() == 0)
				m_freePFs.add(pf);
			
		}

		Collections.sort(m_freePFs, new IDComparator());

	}

	public void assignTargets() {
		m_assign = new HashMap<StandardEntity, PoliceForce>();
		if (m_freePFs.isEmpty())
			return;
		/*
		 * Debug Mode if (!SeuConstants.IS_COMPETETION_MODE)
		 * world.logFile.writeString("#########################     " +
		 * world.getTime(), 15, "pfEva");
		 */
		ArrayList<ArrayDistProperty> mainInfo =
		        new ArrayList<ArrayDistProperty>();

		for (Iterator itp = m_freePFs.iterator(); itp.hasNext();) {
			PoliceForce pf = (PoliceForce) itp.next();

			ArrayDistProperty arrayDist = new ArrayDistProperty();
			arrayDist.m_pf = pf;

			
			for (Iterator it =
			        worldState.lockedByBlockadeAgentMap.values().iterator(); it
			        .hasNext();) {
				Human blocked = (Human) it.next();

				if (blocked.getPosition(world) == null)
					continue;


				int dist = world.getDistance(blocked.getID(), pf.getID());
				if (m_lastAssigned.containsKey(blocked)
				        && m_lastAssigned.get(blocked) == pf)
					dist -= BASE_RATION;
				if (dist <= 1)
					dist = 1;
				arrayDist.m_distArray.add(new DistProperty(blocked, dist));
			}

			for (Human bury : worldState.buriedAgents) {
				if (bury.getPosition(world) == null)
					continue;

				if (bury.getPosition(world) instanceof Building){
					Building bldg = (Building)  bury.getPosition(world);
					if(bldg.getState().getAreaChecked()){
						continue;
					}
				}
					


				int dist = world.getDistance(bury.getID(), pf.getID());
				if (m_lastAssigned.containsKey(bury)
				        && m_lastAssigned.get(bury) == pf)
					dist -= BASE_RATION;
				if (dist <= 1)
					dist = 1;
				arrayDist.m_distArray.add(new DistProperty(bury, dist));

			}
			
			for (Iterator it=world.getEntitiesOfType(StandardEntityURN.REFUGE).iterator();it.hasNext();) {
				Refuge rfg=(Refuge)it.next();
				if (rfg.getState().getAreaChecked())
					continue;

				int dist = world.getDistance(rfg.getID(), pf.getID());
				if (m_lastAssigned.containsKey(rfg)
				        && m_lastAssigned.get(rfg) == pf)
					dist -= BASE_RATION;
				if (dist <= 1)
					dist = 1;
				arrayDist.m_distArray.add(new DistProperty(rfg,
				        (int) (0.8 * dist)));

			}

			if (arrayDist.m_distArray.isEmpty())
				continue;
			Collections.sort(arrayDist.m_distArray);
			arrayDist.m_dist = arrayDist.m_distArray.get(0).m_dist;

			mainInfo.add(arrayDist);

		}

		Collections.sort(mainInfo);

		while (mainInfo.size() > 0 && m_freePFs.size() > 0) {
			ArrayDistProperty arrayDist = mainInfo.get(0);

			for (Iterator it = arrayDist.m_distArray.iterator(); it.hasNext();) {
				DistProperty pro = (DistProperty) it.next();

				if (!m_assign.containsKey(pro.m_obj)) {

					m_freePFs.remove(arrayDist.m_pf);
					m_assign.put(pro.m_obj, arrayDist.m_pf);
					break;
				}
			}
			mainInfo.remove(0);

		}

		// JiangYh

		// if(mainInfo.isEmpty())
		// return;

		// int[] s, t, l1, l2, match1, match2;
		// int m, n, p, q, ret = 0, i, j, k, inf = Integer.MAX_VALUE;
		// l2 = new int[MAXN];
		// l1 = new int[MAXN];
		// t = new int[MAXN];
		// s = new int[MAXN];
		// match1 = new int[MAXN];
		// match2 = new int[MAXN];
		//		
		//			
		// m = m_freePFs.size();
		// n = mainInfo.get(0).m_distArray.size();
		// if (m > n) {
		// m = n;
		// }
		//
		// for (i = 0; i < m; i++) {
		// l1[i] = -inf;
		// for (j = 0; j < n; j++)
		// l1[i] = (-mainInfo.get(i).m_distArray.get(j).m_dist) > l1[i] ?
		// (-mainInfo
		// .get(i).m_distArray.get(j).m_dist)
		// : l1[i];
		// }
		//
		// for (i = 0; i < n; l2[i++] = 0)
		// ;
		//
		// for (i = 0; i < MAXN; match1[i++] = -0xff, match2[i - 1] = -0xff)
		// ;
		//
		// for (i = 0; i < m; i++) {
		// for (int iii = 0; iii < MAXN; t[iii++] = -0xff)
		// ;
		// q = 0;
		// p = 0;
		// for (s[0] = i; p <= q && match1[i] < 0; p++)
		// for (k = s[p], j = 0; j < n && match1[i] < 0; j++)
		// if (l1[k] + l2[j] == (-mainInfo.get(k).m_distArray.get(j).m_dist)
		// && t[j] < 0) {
		// s[++q] = match2[j];
		// t[j] = k;
		// if (s[q] < 0)
		// for (p = j; p >= 0; j = p) {
		// match2[j] = k = t[j];
		// p = match1[k];
		// match1[k] = j;
		// }
		// }
		// if (match1[i] < 0) {
		// for (i--, p = inf, k = 0; k <= q; k++)
		// for (j = 0; j < n; j++)
		// if (t[j] < 0
		// && l1[s[k]]
		// + l2[j]
		// - (-mainInfo.get(s[k]).m_distArray
		// .get(j).m_dist) < p)
		// p = l1[s[k]]
		// + l2[j]
		// - (-mainInfo.get(s[k]).m_distArray.get(j).m_dist);
		// for (j = 0; j < n; l2[j] += t[j] < 0 ? 0 : p, j++)
		// ;
		// for (k = 0; k <= q; l1[s[k++]] -= p)
		// ;
		// }
		// }

		// world.logFile.writeString("#########################
		// "+world.getTime(), 10, "assignResult");

		// for (i = 0; i < m; i++) {
		//
		// if (!m_assign.containsKey(mainInfo.get(i).m_distArray
		// .get(match1[i]).m_obj)
		// && !m_freePFs.isEmpty()) {
		// m_freePFs.remove(mainInfo.get(i).m_pf);
		// m_assign.put(mainInfo.get(i).m_distArray.get(match1[i]).m_obj,
		// mainInfo.get(i).m_pf);
		// }
		// }

		// JiangYh

		m_lastAssigned = new HashMap<StandardEntity, PoliceForce>();
		for (Iterator it = m_assign.keySet().iterator(); it.hasNext();) {
			StandardEntity key = (StandardEntity) it.next();
			PoliceForce value = m_assign.get(key);
			m_lastAssigned.put(key, value);
		}
	}

	private class DistProperty implements Comparable {
		public StandardEntity m_obj;

		public int m_dist;

		public DistProperty(StandardEntity obj, int dist) {
			m_obj = obj;
			m_dist = dist;
		}

		public int compareTo(Object o) {
			if (m_dist < ((DistProperty) o).m_dist)
				return -1;
			else
				return 1;
		}
	}

	private class ArrayDistProperty implements Comparable {
		PoliceForce m_pf;

		int m_dist;

		ArrayList<DistProperty> m_distArray;

		public ArrayDistProperty() {
			m_distArray = new ArrayList<DistProperty>();
		}

		public int compareTo(Object o) {
			if (m_dist < ((ArrayDistProperty) o).m_dist)
				return -1;
			else
				return 1;
		}
	}

}
